
function MenuScene(){

  var royalBlue = 0x0076f3;
  var shaders;
  var textures;

  Array.prototype.filterMap = function(f){
    var r = new Array();
    for(var i=0;i<this.length;i++){
      var x = f(this[i]);
      if(x!==undefined){
        r.push(x);
      }
    }
    return r;
  }

  function StaticObj3d(mesh){
    return function(addToScope,lookup,removeFromScope){
      return {
        obj:[mesh],
        handle:function(){},
        update:function(){}
      }
    }
  }

  function texPlane(w,h,tex){

    var g = new THREE.PlaneBufferGeometry(w,h);
    var m = new THREE.Mesh(g,
      new THREE.MeshBasicMaterial({
        map:tex
      , transparent:true
      }) );

    m.renderOrder = 1;

    return m;
  }

  var focusFireNearbyState = {callback:function(){}}; // called when near a FF
  function focusFire(duration,target,focusingModifier,returnValue){
    return function(addToScope,lookup,removeFromScope){

      function findRad(c,r){
        function distance(p1,p2){
          return Math.sqrt((p1.x - p2.x)*(p1.x - p2.x) + (p1.y - p2.y)*(p1.y - p2.y) + (p1.z - p2.z)*(p1.z - p2.z));
        }
        var p = distance(c,{x:0,y:0,z:0}) / r;
        return p * r * 0.5;
      }

      var geometry = new THREE.SphereGeometry( 1.0, 6, 6 );
      var material = new THREE.MeshNormalMaterial();
      var sphere = new THREE.Mesh( geometry, material );

      material.visible = false;

      function updateBoundPosition(){

        target.geometry.computeBoundingSphere();

        var x = target.geometry.boundingSphere.center.x + target.position.x;
        var y = target.geometry.boundingSphere.center.y + target.position.y;
        var z = target.geometry.boundingSphere.center.z + target.position.z;

        var r = findRad({
          x:x,
          y:y,
          z:z
        },target.geometry.boundingSphere.radius);

        sphere.scale.set(r,r,r);

        sphere.position.x = x;
        sphere.position.y = y;
        sphere.position.z = z;
      }

      addToScope( StaticObj3d(sphere) );

      var raycaster = new THREE.Raycaster();

      var gaze = new THREE.Vector2();

      gaze.x = 0;
      gaze.y = 0;

      raycaster.near = 0.001;
      raycaster.far = 10;

      var focusingState = {
        ignition:0.0 / 100
      };

      return {
        handle:function(){},
        update:function(time,timeDelta,camera){

          raycaster.setFromCamera( gaze, camera );

          if(raycaster.intersectObject(target,true).length === 1){

            var prev = focusingState.ignition;

            focusingState.ignition += timeDelta / duration;

            if(focusingState.ignition > 1.0){
              focusingState.ignition = 1.0;
            }

            if(prev < 1.0 && focusingState.ignition === 1.0){
              return returnValue === undefined ? true : returnValue;
            }

            focusingModifier(focusingState.ignition,true);

          }else{

            updateBoundPosition();

            if(raycaster.intersectObject(sphere,true).length === 1){
              focusFireNearbyState.callback();
            }

            if(focusingState.ignition > 0.0){

              focusingState.ignition -= timeDelta / duration;

              if(focusingState.ignition < 0.0){
                focusingState.ignition = 0.0;
              }

              focusingModifier(focusingState.ignition,false);

            }
          }

        }
      }
    }
  }

  function groundFloor(scene){

    var geometry = new THREE.BufferGeometry();

    var vertices = new Float32Array(
      [12.0,-0.0,12.0,12.0,-0.0,-12.0,-12.0,-0.0,12.0,-12.0,0.0,-12.0,-12.0,0.0,12.0,12.0,0.0,-12.0]
    );

    geometry.addAttribute( 'position', new THREE.BufferAttribute( vertices, 3 ) );
    geometry.computeVertexNormals();

    var material =
      new THREE.MeshPhongMaterial({
         color:0xfcfcfc
      });

    var mesh = new THREE.Mesh( geometry, material );

    mesh.position.y = -2.0;

    return mesh;
  }

  function addPanel(verts,tex,v){
    return function(addToScope,lookup,removeFromScope){

      var geometry = new THREE.BufferGeometry();
      var vertices = new Float32Array(verts);

      geometry.addAttribute( 'position', new THREE.BufferAttribute( vertices, 3 ) );
      geometry.computeFaceNormals();
      geometry.computeVertexNormals();

      var material =
        new THREE.MeshBasicMaterial( {
          color: interpColor(0.0,royalBlue,0xffffff)
        , transparent:true
        , opacity:0.0
        } );


      var mesh = new THREE.Mesh( geometry, material );
      mesh.renderOrder = 2;

      geometry.computeBoundingSphere();

      var m = texPlane(0.6,0.6,tex);
      m.material.opacity = 0.0;

      m.position.x = geometry.boundingSphere.center.x;
      m.position.y = geometry.boundingSphere.center.y;
      m.position.z = geometry.boundingSphere.center.z;
      m.renderOrder = 3;

      var zdiff = m.position.z - mesh.position.z;

      m.lookAt({x:0,y:0,z:0});

      addToScope(StaticObj3d(m));

      addToScope(animate(350,function(p){
        material.opacity = interp(p,0.0,0.8);
        m.material.opacity = interp(p,0.0,0.8);
      },0));

      var result;

      return {
        obj:[mesh],
        setOpacity:function(x){
          material.opacity = x;
          m.material.opacity = x;
        },handle:function(_){

          addToScope(focusFire(800,mesh,function(p,i){
            material.opacity = i ? 0.3 : 0.8;
            var s = -0.25 * p + 1;
            m.scale.set(s,s,s);
          },v));

          return function(v){

            material.opacity = 0.8;

            var start = Date.now();

            addToScope(animate(350,function(p){
              var s = interp(p,0.75,1.0,Bezier(.33,1.78,.79,1.33));
              m.scale.set(s,s,s);
            },0));

            return function(_){

              addToScope(animate(150,function(p){
                var s = interp(p,1.0,0.2);
                m.scale.set(s,s,s);
                m.material.opacity = interp(p,1.0,0.0);
              },0));

              return function(_){
                result = v;
              }
            };
          }
        },
        update:function(){
          m.position.z = zdiff + mesh.position.z;
          m.material.opacity = mesh.material.opacity;
          if(result !== undefined)return result
        }
      }
    }
  }

  function addVirtualScreen(choices){
    return function(addToScope,lookup,removeFromScope){

      var zRestingOffset = -0.8;
      var zAnimOffset = -0.25;

      var models = Pregenerated(choices.length);

      // screen

      var screenOpacity = 0.5;
      var skeletonOpacity = 0.94;

      var screen = (function(){
        var geometry = new THREE.BufferGeometry();
        var vertices = new Float32Array(
          models[1]
        );

        geometry.addAttribute( 'position', new THREE.BufferAttribute( vertices, 3 ) );
        geometry.computeVertexNormals();

        var material =
        new THREE.MeshBasicMaterial( {
            color: interpColor(0.1,royalBlue,0xffffff)
          , opacity:0.0
          } );

        material.transparent = true;

        var mesh = new THREE.Mesh( geometry, material );
        mesh.position.z = zAnimOffset;

        return mesh;
      })();

      // skeleton

      var skeleton = (function(){
        var geometry = new THREE.BufferGeometry();
        var vertices = new Float32Array(
          models[0]
        );

        geometry.addAttribute( 'position', new THREE.BufferAttribute( vertices, 3 ) );
        geometry.computeVertexNormals();

        var material = new THREE.MeshBasicMaterial( { color: interpColor(0.15,royalBlue,0x0) } );

        material.transparent = true;
        material.opacity = 0.0;
        material.side = THREE.BackSide;

        var mesh = new THREE.Mesh( geometry, material );
        mesh.position.z = zAnimOffset;

        return mesh;
      })();

      // panels

      var panels = models[2].map(function(x,i){return addToScope(addPanel(x,choices[i],i))});
      panels.map(lookup).forEach(function(x){
        for(var i=0;i<x.object.obj.length;i++){
          x.object.obj[i].position.z = zAnimOffset;
        }
      });

      addToScope(animate(350,function(p){

        screen.material.opacity = interp(p,0.0,screenOpacity);
        skeleton.material.opacity = interp(p,0.0,skeletonOpacity);

        screen.position.z = interp(p,zAnimOffset,zRestingOffset);
        skeleton.position.z = interp(p,zAnimOffset,zRestingOffset);

        panels.map(lookup).forEach(function(x){
          for(var i=0;i<x.object.obj.length;i++){
            x.object.obj[i].position.z = interp(p,zAnimOffset,zRestingOffset);
          }
        });
      }));

      var result;

      return {
        obj:[screen,skeleton],
        handle:function(v){

          addToScope(animate(300,function(p){
            function mods(o,m){
              m.material.opacity = interp(p,o,0.0);
              m.position.z = interp(p,zRestingOffset,zAnimOffset);
            }

            mods(screenOpacity,screen);
            mods(skeletonOpacity,skeleton);

            panels.filterMap(lookup).forEach(function($0){mods(0.5,$0.object.obj[0])});
          },0));

          return function(_){
            result = v;
          }
        },
        update:function(time,timeDelta,scene){
          if(result !== undefined)return result
        }
      }
    }
  }

  function animate(duration,iteration,value){
    return function(addToScope,lookup,removeFromScope){

      var s = {
        i:0,
        done:false
      };

      return {
        update:function(_,delta){
          if(s.done){
            return
          }
          s.i = clamp(s.i + (delta / duration),0,1);
          try{iteration(s.i);}catch(e){}finally{

            if(s.i === 1.0){
              if(value !== undefined){
                return value;
              }else{
                s.done = true;
              }
            }
          }
        }
      }
    }
  }

  function pingPongAnimate(duration,iteration){
    return function(addToScope,lookup,removeFromScope){

      var s = {
        i:0,
        forward:true
      };

      return {
        update:function(_,delta){
          if(s.done){
            return
          }
          s.i = clamp(s.forward ? s.i + (delta / duration) : s.i - (delta / duration),0,1);
          try{iteration(s.i);}catch(e){}finally{

            if(s.i === 1.0 && s.forward){
              s.forward = false;
            }else if(s.i === 0.0 && !s.forward){
              s.forward = true;
            }
          }
        }
      }
    }
  }

  function lookDown(addToScope,lookup,removeFromScope){

    var size = 0.6;

    var background = new THREE.Mesh(new THREE.PlaneBufferGeometry(5.6 * size, 1.2 * size),new THREE.MeshBasicMaterial({
        color:interpColor(0.3,royalBlue,0x0)
      , transparent:true
      , opacity:0.0
      }));

    function makeFace(w,h,tex){
      var geometry = new THREE.PlaneBufferGeometry(w,h);
      var material = new THREE.MeshBasicMaterial({
            map:tex
          , color:0xffffff
          , transparent:true
          , opacity:0.0
        });

      var mesh = new THREE.Mesh(geometry,material);

      mesh.renderOrder = 1;
      mesh.position.z = 0.001;

      background.add(mesh);

      return mesh;
    }

    var text = makeFace(4*size,size,textures['lookdown.png']);
    var icon = makeFace(size,size,textures['lookdown_icon.png']);

    background.position.y = 0.4;
    background.position.z = -3;

    background.lookAt({x:0,y:0.5,z:0});

    text.position.x = 0.35;
    icon.position.x = -1.3;

    addToScope(animate(400,function(p){},0));

    return {
        obj:[background]
      , handle:function(){

        addToScope(animate(500,function(p){
          var s = interp(p,0.5,1.0,Bezier(.33,1.28,.79,1.23));
          background.scale.set(s,s,s);
          background.material.opacity = interp(p,0.0,1.0);
          text.material.opacity = interp(p,0.0,1.0);
          icon.material.opacity = interp(p,0.0,1.0);
        }));

        return function(){}
      }
      , update:function(){}
    }
  }

  function MagicBall(coords,light,choices){
    return function(addToScope,lookup,removeFromScope){

      var circleVerts = new Float32Array(
        [-0.0,0.0,0.0,0.3,0.0,-5.510910596163089e-17,0.2977261529015807,0.0,-3.686648719941418e-2,-0.0,0.0,0.0,0.2977261529015807,0.0,-3.686648719941418e-2,0.2909390808105028,0.0,-7.317411603251325e-2,-0.0,0.0,0.0,0.2909390808105028,0.0,-7.317411603251325e-2,0.27974166882130674,0.0,-0.10837249985614587,-0.0,0.0,0.0,0.27974166882130674,0.0,-0.10837249985614587,0.2643036582857354,0.0,-0.1419280670508029,-0.0,0.0,0.0,0.2643036582857354,0.0,-0.1419280670508029,0.24485907370686646,0.0,-0.1733321494224754,-0.0,0.0,0.0,0.24485907370686646,0.0,-0.1733321494224754,0.22170267516619765,0.0,-0.20210869309396723,-0.0,0.0,0.0,0.22170267516619765,0.0,-0.20210869309396723,0.19518549006127264,0.0,-0.22782147499641212,-0.0,0.0,0.0,0.19518549006127264,0.0,-0.22782147499641212,0.16570949188815173,0.0,-0.2500807155663358,-0.0,0.0,0.0,0.16570949188815173,0.0,-0.2500807155663358,0.13372150673296132,0.0,-0.2685489874065188,-0.0,0.0,0.0,0.13372150673296132,0.0,-0.2685489874065188,9.970643984389797e-2,0.0,-0.28294633034138317,-0.0,0.0,0.0,9.970643984389797e-2,0.0,-0.28294633034138317,6.417992496194914e-2,0.0,-0.29305449532788025,-0.0,0.0,0.0,6.417992496194914e-2,0.0,-0.29305449532788025,2.7680507838990485e-2,0.0,-0.29872025288851034,-0.0,0.0,0.0,2.7680507838990485e-2,0.0,-0.29872025288851034,-9.238517566851116e-3,0.0,-0.2998577159140098,-0.0,0.0,0.0,-9.238517566851116e-3,0.0,-0.2998577159140098,-4.601749646360564e-2,0.0,-0.2964496416243778,-0.0,0.0,0.0,-4.601749646360564e-2,0.0,-0.2964496416243778,-8.209889702162491e-2,0.0,-0.2885476929518457,-0.0,0.0,0.0,-8.209889702162491e-2,0.0,-0.2885476929518457,-0.11693576198780375,0.0,-0.2762716553834861,-0.0,0.0,0.0,-0.11693576198780375,0.0,-0.2762716553834861,-0.14999999999999997,0.0,-0.2598076211353316,-0.0,0.0,0.0,-0.14999999999999997,0.0,-0.2598076211353316,-0.1807903909137769,0.0,-0.23940516818407184,-0.0,0.0,0.0,-0.1807903909137769,0.0,-0.23940516818407184,-0.208840183788878,0.0,-0.21537357691933254,-0.0,0.0,0.0,-0.208840183788878,0.0,-0.21537357691933254,-0.23372417235770115,0.0,-0.18807714176823187,-0.0,0.0,0.0,-0.23372417235770115,0.0,-0.18807714176823187,-0.2550651407188842,0.0,-0.15792964886320673,-0.0,0.0,0.0,-0.2550651407188842,0.0,-0.15792964886320673,-0.27253958154585717,0.0,-0.1253881034660349,-0.0,0.0,0.0,-0.27253958154585717,0.0,-0.1253881034660349,-0.285882600128147,0.0,-9.094580223391303e-2,-0.0,0.0,0.0,-0.285882600128147,0.0,-9.094580223391303e-2,-0.29489192990517055,0.0,-5.512485534497113e-2,-0.0,0.0,0.0,-0.29489192990517055,0.0,-5.512485534497113e-2,-0.2994309986211132,0.0,-1.8468271840182845e-2,-0.0,0.0,0.0,-0.2994309986211132,0.0,-1.8468271840182845e-2,-0.2994309986211132,0.0,1.8468271840182817e-2,-0.0,0.0,0.0,-0.2994309986211132,0.0,1.8468271840182817e-2,-0.29489192990517055,0.0,5.5124855344971156e-2,-0.0,0.0,0.0,-0.29489192990517055,0.0,5.5124855344971156e-2,-0.2858826001281469,0.0,9.094580223391312e-2,-0.0,0.0,0.0,-0.2858826001281469,0.0,9.094580223391312e-2,-0.2725395815458571,0.0,0.125388103466035,-0.0,0.0,0.0,-0.2725395815458571,0.0,0.125388103466035,-0.2550651407188842,0.0,0.15792964886320682,-0.0,0.0,0.0,-0.2550651407188842,0.0,0.15792964886320682,-0.23372417235770115,0.0,0.18807714176823187,-0.0,0.0,0.0,-0.23372417235770115,0.0,0.18807714176823187,-0.208840183788878,0.0,0.2153735769193326,-0.0,0.0,0.0,-0.208840183788878,0.0,0.2153735769193326,-0.18079039091377688,0.0,0.23940516818407187,-0.0,0.0,0.0,-0.18079039091377688,0.0,0.23940516818407187,-0.15000000000000002,0.0,0.25980762113533157,-0.0,0.0,0.0,-0.15000000000000002,0.0,0.25980762113533157,-0.1169357619878038,0.0,0.2762716553834861,-0.0,0.0,0.0,-0.1169357619878038,0.0,0.2762716553834861,-8.209889702162498e-2,0.0,0.2885476929518457,-0.0,0.0,0.0,-8.209889702162498e-2,0.0,0.2885476929518457,-4.601749646360565e-2,0.0,0.2964496416243778,-0.0,0.0,0.0,-4.601749646360565e-2,0.0,0.2964496416243778,-9.238517566851073e-3,0.0,0.2998577159140098,-0.0,0.0,0.0,-9.238517566851073e-3,0.0,0.2998577159140098,2.76805078389905e-2,0.0,0.29872025288851034,-0.0,0.0,0.0,2.76805078389905e-2,0.0,0.29872025288851034,6.41799249619492e-2,0.0,0.2930544953278802,-0.0,0.0,0.0,6.41799249619492e-2,0.0,0.2930544953278802,9.970643984389795e-2,0.0,0.2829463303413832,-0.0,0.0,0.0,9.970643984389795e-2,0.0,0.2829463303413832,0.1337215067329614,0.0,0.2685489874065187,-0.0,0.0,0.0,0.1337215067329614,0.0,0.2685489874065187,0.16570949188815173,0.0,0.2500807155663358,-0.0,0.0,0.0,0.16570949188815173,0.0,0.2500807155663358,0.1951854900612727,0.0,0.2278214749964121,-0.0,0.0,0.0,0.1951854900612727,0.0,0.2278214749964121,0.22170267516619768,0.0,0.2021086930939672,-0.0,0.0,0.0,0.22170267516619768,0.0,0.2021086930939672,0.2448590737068665,0.0,0.17333214942247532,-0.0,0.0,0.0,0.2448590737068665,0.0,0.17333214942247532,0.26430365828573527,0.0,0.14192806705080313,-0.0,0.0,0.0,0.26430365828573527,0.0,0.14192806705080313,0.27974166882130674,0.0,0.1083724998561459,-0.0,0.0,0.0,0.27974166882130674,0.0,0.1083724998561459,0.29093908081050285,0.0,7.317411603251309e-2,-0.0,0.0,0.0,0.29093908081050285,0.0,7.317411603251309e-2,0.2977261529015807,0.0,3.686648719941434e-2,-0.0,0.0,0.0,0.2977261529015807,0.0,3.686648719941434e-2,0.3,0.0,1.8369701987210297e-17]
      );

      var geometry = new THREE.BufferGeometry();
      geometry.addAttribute( 'position', new THREE.BufferAttribute( circleVerts, 3 ) );
      geometry.computeVertexNormals();

      var material = new THREE.MeshPhongMaterial( {
          color: interpColor(0.5,0x0,royalBlue)
        , emissiveIntensity:0.5
        , emissive: royalBlue
        , transparent:true
      } );

      // split visible and target sphere so size doesn't scale animation doesn't
      // interfere with selection semantics
      var sphereVisible = new THREE.Mesh(geometry, material);
      var sphereTarget = new THREE.Mesh(geometry, new THREE.MeshBasicMaterial({visible:false}));

      var icon = texPlane(0.60,0.60,textures['menu.png']);
      icon.renderOrder = 4;
      var shadow = texPlane(0.65,0.65,textures['shadow.png']);
      var shadowOpacity = 0.65;

      shadow.material.opacity = shadowOpacity;

      function setPosition(s,x,y,z){
        s.position.x = coords[0] + x;
        s.position.y = coords[2] + y;
        s.position.z = coords[1] + z;
      }

      setPosition(sphereVisible,0,0.013,0);
      setPosition(sphereTarget,0,0,0);
      setPosition(icon,0,0.014,0);
      setPosition(shadow,0,0.01,0.004);

      var shadowz = shadow.position.z;

      icon.lookAt({x:0,y:999999,z:0});
      shadow.lookAt({x:0,y:999999,z:0});

      var lightExplodeColor1 = interpColor(0.8,royalBlue,0xffffff);
      var lightExplodeColor2 = interpColor(0.5,royalBlue,0xffffff);
      var lightExplodeColor3 = interpColor(0.6,royalBlue,0xffffff);

      var baseSVY = sphereVisible.position.y;
      var baseICY = icon.position.y;
      var baseSHZ = shadow.position.z;

      function activateBall(){

        (function(){
          // hover animation to accent depth

          var magnitude = 0.03;

          addToScope(pingPongAnimate(2000,function(p){
            sphereVisible.position.y = baseSVY + magnitude * p;
            icon.position.y = baseICY + magnitude * p;
            shadow.material.opacity = shadowOpacity - 0.12 * p;
            shadow.position.z = baseSHZ + magnitude * 0.02 * (p * p);

            var s = 0.98 + 0.012 * (p * p);

            shadow.scale.set(s,s,s);
          }));

        })();

        var curColor = light.color;

        addToScope(animate(250,function(p){

          var s = interp(p,0.8,1.0,Bezier(.33,1.78,.79,1.93));
          material.visible = true;
          sphereVisible.scale.set(s,s,s);
          icon.scale.set(s,s,s);
          material.opacity = interp(p,0.0,1.0);
          shadow.material.opacity = interp(p,0.0,shadowOpacity);
          icon.material.opacity = interp(p,0.0,1.0);

          light.color = interpColor(p,curColor,lightExplodeColor1);

        }));

        addToScope(focusFire(600,sphereTarget,function(p,i){
          var s = interp(p,1.0,0.85);
          material.opacity = i ? 0.7 : 1.0;
          icon.scale.set(s,s,s);
          light.color = interpColor(p,lightExplodeColor1,lightExplodeColor2);
        }));
      }

      activateBall();

      var lookDownRef = addToScope(lookDown);

      function ballHandler(v){

        removeFromScope(lookDownRef);

        material.opacity = 1.0;

        addToScope(animate(400,function(p){
          var s = interp(p,0.85,1.0,Bezier(.33,1.78,.79,1.23));
          icon.scale.set(s,s,s);
        },0));

        return function(){

          addToScope(animate(150,function(p){

            var s = interp(p,1.0,0.5);

            sphereVisible.scale.set(s,s,s);
            icon.scale.set(s,s,s);
            shadow.scale.set(s,s,s);
            material.opacity = interp(p,1.0,0.0);
            shadow.material.opacity = interp(p,shadowOpacity,0.0);
            icon.material.opacity = interp(p,1.0,0.0);

            if(material.opacity < 0.05){
              material.visible = false;
            }

            light.color = interpColor(p,lightExplodeColor2,lightExplodeColor3,Bezier(.33,2.78,.79,1.93));
          },0));

          return function(){

            addToScope(addVirtualScreen(choices));

            return function(){
              activateBall();
              return ballHandler
            }
          }
        }
      }

      return {
          obj:[sphereVisible,sphereTarget,icon,shadow]
        , handle:ballHandler
        , update:function(){
          // shadow.scale.x = clamp(1.0 - (1.0 -sphereVisible.scale.x)*1.00,0.0,1.02);
          // shadow.scale.y = sphereVisible.scale.y <= 1.0 ? 1.0 - (1.0 - sphereVisible.scale.y)*1.4 : 1.0 - (1.0 - sphereVisible.scale.y)*1.1;
          // shadow.position.z = shadowz + 0.05 * (1.0 - clamp(sphereVisible.scale.y,0.0,1.0));
        }
      }
    }
  }

  var choiceTextures = ['circle.png','triangle.png','square.png','x.png'];

  return {
    shaders:[],
    textures:choiceTextures.concat(['lookdown.png','lookdown_icon.png','menu.png','shadow.png']),
    scene:function(shaders_,textures_){

      shaders = shaders_;
      textures = textures_;

      return function(addToScope,lookup,removeFromScope){

        var light = new THREE.PointLight( interpColor(.80,royalBlue,0xffffff), 1, 12 );

        light.position.set( 0.0, 0.4, -2.0);
        addToScope(StaticObj3d(light));

        addToScope(StaticObj3d(groundFloor()));
        addToScope(MagicBall(
            [0,-1.5,-1.6]
          , light
          , choiceTextures.map(function($0){return textures[$0]})
          ));

        var geometry = new THREE.SphereGeometry( 0.01, 32, 32 );
        var material = new THREE.MeshBasicMaterial({
          color:interpColor(0.5,0x0,royalBlue)
        , transparent:true
        , opacity:0.5
        });

        var aimer = new THREE.Mesh( geometry, material );

        aimer.position.z -= 1.0;
        aimer.visible = false;

        addToScope(StaticObj3d(aimer));

        // callback when near a focused object; we activate our aimer here
        focusFireNearbyState.callback = function(){
          aimer.visible = true;
          clearInterval(focusFireNearbyState.interval);
          focusFireNearbyState.interval = setInterval(function(){
            aimer.visible = false;
            clearInterval(focusFireNearbyState.interval);
          },500);
        };

        return {
          handle:function(){},
          update:function(){},
          realTimeUpdate:function(_,_,camera){
            var d = camera.getWorldDirection();
            aimer.position.x = d.x;
            aimer.position.y = d.y;
            aimer.position.z = d.z;
          }
        };
      };
    }
  };

  function Pregenerated(n){
    switch(n){
      case 1:
        return [[0.5842865730243545,0.6784329353268667,-1.2022225055484645,0.7150866609909294,0.6826289037478287,-1.1267962757485852,0.7162322333317126,0.6784329353268667,-1.1286014089130136,0.583352040192686,0.6826289037478287,-1.2002996196663018,0.7150866609909294,0.6826289037478287,-1.1267962757485852,0.5842865730243545,0.6784329353268667,-1.2022225055484645,0.44487526175223463,0.6784329353268667,-1.260482349636302,0.583352040192686,0.6826289037478287,-1.2002996196663018,0.5842865730243545,0.6784329353268667,-1.2022225055484645,0.4441637093098217,0.6826289037478287,-1.2584662804780182,0.583352040192686,0.6826289037478287,-1.2002996196663018,0.44487526175223463,0.6784329353268667,-1.260482349636302,0.2997796106707845,0.6784329353268667,-1.3026365330704666,0.4441637093098217,0.6826289037478287,-1.2584662804780182,0.44487526175223463,0.6784329353268667,-1.260482349636302,0.2993001303928337,0.6826289037478287,-1.3005530407157067,0.4441637093098217,0.6826289037478287,-1.2584662804780182,0.2997796106707845,0.6784329353268667,-1.3026365330704666,0.1508535618993176,0.6784329353268667,-1.3281464358744712,0.2993001303928337,0.6826289037478287,-1.3005530407157067,0.2997796106707845,0.6784329353268667,-1.3026365330704666,0.15061228028704413,0.6826289037478287,-1.326022141894612,0.2993001303928337,0.6826289037478287,-1.3005530407157067,0.1508535618993176,0.6784329353268667,-1.3281464358744712,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,0.15061228028704413,0.6826289037478287,-1.326022141894612,0.1508535618993176,0.6784329353268667,-1.3281464358744712,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,0.15061228028704413,0.6826289037478287,-1.326022141894612,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,-0.15085356189931773,0.6784329353268667,-1.3281464358744712,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,-0.15061228028704426,0.6826289037478287,-1.326022141894612,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,-0.15085356189931773,0.6784329353268667,-1.3281464358744712,-0.2997796106707846,0.6784329353268667,-1.3026365330704666,-0.15061228028704426,0.6826289037478287,-1.326022141894612,-0.15085356189931773,0.6784329353268667,-1.3281464358744712,-0.2993001303928339,0.6826289037478287,-1.3005530407157067,-0.15061228028704426,0.6826289037478287,-1.326022141894612,-0.2997796106707846,0.6784329353268667,-1.3026365330704666,-0.44487526175223446,0.6784329353268667,-1.260482349636302,-0.2993001303928339,0.6826289037478287,-1.3005530407157067,-0.2997796106707846,0.6784329353268667,-1.3026365330704666,-0.4441637093098215,0.6826289037478287,-1.2584662804780182,-0.2993001303928339,0.6826289037478287,-1.3005530407157067,-0.44487526175223446,0.6784329353268667,-1.260482349636302,-0.5842865730243543,0.6784329353268667,-1.2022225055484645,-0.4441637093098215,0.6826289037478287,-1.2584662804780182,-0.44487526175223446,0.6784329353268667,-1.260482349636302,-0.5833520401926859,0.6826289037478287,-1.2002996196663018,-0.4441637093098215,0.6826289037478287,-1.2584662804780182,-0.5842865730243543,0.6784329353268667,-1.2022225055484645,-0.7162322333317123,0.6784329353268667,-1.1286014089130139,-0.5833520401926859,0.6826289037478287,-1.2002996196663018,-0.5842865730243543,0.6784329353268667,-1.2022225055484645,-0.7150866609909291,0.6826289037478287,-1.1267962757485854,-0.5833520401926859,0.6826289037478287,-1.2002996196663018,-0.7162322333317123,0.6784329353268667,-1.1286014089130139,0.6295058797224125,0.41594602267436875,-1.2952653217068253,0.7709591344021218,0.42046828134539266,-1.214837206157427,0.7716631237908568,0.41594602267436875,-1.2159465159301666,0.6289315805939357,0.42046828134539266,-1.294083649272359,0.7709591344021218,0.42046828134539266,-1.214837206157427,0.6295058797224125,0.41594602267436875,-1.2952653217068253,0.47930520047121755,0.41594602267436875,-1.3580340316143111,0.6289315805939357,0.42046828134539266,-1.294083649272359,0.6295058797224125,0.41594602267436875,-1.2952653217068253,0.47886792964059977,0.42046828134539266,-1.3567950951946197,0.6289315805939357,0.42046828134539266,-1.294083649272359,0.47930520047121755,0.41594602267436875,-1.3580340316143111,0.3229802570361111,0.41594602267436875,-1.4034506260593071,0.47886792964059977,0.42046828134539266,-1.3567950951946197,0.47930520047121755,0.41594602267436875,-1.3580340316143111,0.3226856016784632,0.42046828134539266,-1.4021702560145326,0.47886792964059977,0.42046828134539266,-1.3567950951946197,0.3229802570361111,0.41594602267436875,-1.4034506260593071,0.1625284724602614,0.41594602267436875,-1.430934800003518,0.3226856016784632,0.42046828134539266,-1.4021702560145326,0.3229802570361111,0.41594602267436875,-1.4034506260593071,0.1623801974987507,0.42046828134539266,-1.429629356106931,0.3226856016784632,0.42046828134539266,-1.4021702560145326,0.1625284724602614,0.41594602267436875,-1.430934800003518,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,0.1623801974987507,0.42046828134539266,-1.429629356106931,0.1625284724602614,0.41594602267436875,-1.430934800003518,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,0.1623801974987507,0.42046828134539266,-1.429629356106931,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,-0.16252847246026156,0.41594602267436875,-1.430934800003518,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,-0.16238019749875085,0.42046828134539266,-1.429629356106931,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,-0.16252847246026156,0.41594602267436875,-1.430934800003518,-0.3229802570361113,0.41594602267436875,-1.4034506260593071,-0.16238019749875085,0.42046828134539266,-1.429629356106931,-0.16252847246026156,0.41594602267436875,-1.430934800003518,-0.3226856016784634,0.42046828134539266,-1.4021702560145326,-0.16238019749875085,0.42046828134539266,-1.429629356106931,-0.3229802570361113,0.41594602267436875,-1.4034506260593071,-0.4793052004712175,0.41594602267436875,-1.3580340316143114,-0.3226856016784634,0.42046828134539266,-1.4021702560145326,-0.3229802570361113,0.41594602267436875,-1.4034506260593071,-0.4788679296405997,0.42046828134539266,-1.35679509519462,-0.3226856016784634,0.42046828134539266,-1.4021702560145326,-0.4793052004712175,0.41594602267436875,-1.3580340316143114,-0.6295058797224122,0.41594602267436875,-1.2952653217068253,-0.4788679296405997,0.42046828134539266,-1.35679509519462,-0.4793052004712175,0.41594602267436875,-1.3580340316143114,-0.6289315805939355,0.42046828134539266,-1.294083649272359,-0.4788679296405997,0.42046828134539266,-1.35679509519462,-0.6295058797224122,0.41594602267436875,-1.2952653217068253,-0.7716631237908564,0.41594602267436875,-1.2159465159301668,-0.6289315805939355,0.42046828134539266,-1.294083649272359,-0.6295058797224122,0.41594602267436875,-1.2952653217068253,-0.7709591344021216,0.42046828134539266,-1.2148372061574273,-0.6289315805939355,0.42046828134539266,-1.294083649272359,-0.7716631237908564,0.41594602267436875,-1.2159465159301668,0.6524246267957916,0.13872401480181032,-1.3424227180987511,0.7995199961211304,0.14341236045493638,-1.2598419228900153,0.7997574633827521,0.13872401480181032,-1.2602161114193284,0.6522309062034455,0.14341236045493638,-1.3420241204470866,0.7995199961211304,0.14341236045493638,-1.2598419228900153,0.6524246267957916,0.13872401480181032,-1.3424227180987511,0.496755513509436,0.13872401480181032,-1.4074766809845354,0.6522309062034455,0.14341236045493638,-1.3420241204470866,0.6524246267957916,0.13872401480181032,-1.3424227180987511,0.49660801482778616,0.14341236045493638,-1.4070587672437669,0.6522309062034455,0.14341236045493638,-1.3420241204470866,0.496755513509436,0.13872401480181032,-1.4074766809845354,0.334739166776509,0.13872401480181032,-1.454546780940041,0.49660801482778616,0.14341236045493638,-1.4070587672437669,0.496755513509436,0.13872401480181032,-1.4074766809845354,0.33463977465210704,0.14341236045493638,-1.454114890952407,0.49660801482778616,0.14341236045493638,-1.4070587672437669,0.334739166776509,0.13872401480181032,-1.454546780940041,0.16844573085692952,0.13872401480181032,-1.483031585460452,0.33463977465210704,0.14341236045493638,-1.454114890952407,0.334739166776509,0.13872401480181032,-1.454546780940041,0.16839571526061453,0.14341236045493638,-1.4825912376479937,0.33463977465210704,0.14341236045493638,-1.454114890952407,0.16844573085692952,0.13872401480181032,-1.483031585460452,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,0.16839571526061453,0.14341236045493638,-1.4825912376479937,0.16844573085692952,0.13872401480181032,-1.483031585460452,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,0.16839571526061453,0.14341236045493638,-1.4825912376479937,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,-0.16844573085692968,0.13872401480181032,-1.483031585460452,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,-0.1683957152606147,0.14341236045493638,-1.4825912376479937,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,-0.16844573085692968,0.13872401480181032,-1.483031585460452,-0.33473916677650917,0.13872401480181032,-1.454546780940041,-0.1683957152606147,0.14341236045493638,-1.4825912376479937,-0.16844573085692968,0.13872401480181032,-1.483031585460452,-0.3346397746521072,0.14341236045493638,-1.454114890952407,-0.1683957152606147,0.14341236045493638,-1.4825912376479937,-0.33473916677650917,0.13872401480181032,-1.454546780940041,-0.4967555135094358,0.13872401480181032,-1.4074766809845356,-0.3346397746521072,0.14341236045493638,-1.454114890952407,-0.33473916677650917,0.13872401480181032,-1.454546780940041,-0.496608014827786,0.14341236045493638,-1.407058767243767,-0.3346397746521072,0.14341236045493638,-1.454114890952407,-0.4967555135094358,0.13872401480181032,-1.4074766809845356,-0.6524246267957915,0.13872401480181032,-1.3424227180987511,-0.496608014827786,0.14341236045493638,-1.407058767243767,-0.4967555135094358,0.13872401480181032,-1.4074766809845356,-0.6522309062034454,0.14341236045493638,-1.3420241204470866,-0.496608014827786,0.14341236045493638,-1.407058767243767,-0.6524246267957915,0.13872401480181032,-1.3424227180987511,-0.7997574633827518,0.13872401480181032,-1.2602161114193287,-0.6522309062034454,0.14341236045493638,-1.3420241204470866,-0.6524246267957915,0.13872401480181032,-1.3424227180987511,-0.7995199961211301,0.14341236045493638,-1.2598419228900155,-0.6522309062034454,0.14341236045493638,-1.3420241204470866,-0.7997574633827518,0.13872401480181032,-1.2602161114193287,0.6522309062034455,-0.14341236045493638,-1.3420241204470866,0.7997574633827521,-0.13872401480181032,-1.2602161114193284,0.7995199961211304,-0.14341236045493638,-1.2598419228900153,0.6524246267957916,-0.13872401480181032,-1.3424227180987511,0.7997574633827521,-0.13872401480181032,-1.2602161114193284,0.6522309062034455,-0.14341236045493638,-1.3420241204470866,0.49660801482778616,-0.14341236045493638,-1.4070587672437669,0.6524246267957916,-0.13872401480181032,-1.3424227180987511,0.6522309062034455,-0.14341236045493638,-1.3420241204470866,0.496755513509436,-0.13872401480181032,-1.4074766809845354,0.6524246267957916,-0.13872401480181032,-1.3424227180987511,0.49660801482778616,-0.14341236045493638,-1.4070587672437669,0.33463977465210704,-0.14341236045493638,-1.454114890952407,0.496755513509436,-0.13872401480181032,-1.4074766809845354,0.49660801482778616,-0.14341236045493638,-1.4070587672437669,0.334739166776509,-0.13872401480181032,-1.454546780940041,0.496755513509436,-0.13872401480181032,-1.4074766809845354,0.33463977465210704,-0.14341236045493638,-1.454114890952407,0.16839571526061453,-0.14341236045493638,-1.4825912376479937,0.334739166776509,-0.13872401480181032,-1.454546780940041,0.33463977465210704,-0.14341236045493638,-1.454114890952407,0.16844573085692952,-0.13872401480181032,-1.483031585460452,0.334739166776509,-0.13872401480181032,-1.454546780940041,0.16839571526061453,-0.14341236045493638,-1.4825912376479937,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,0.16844573085692952,-0.13872401480181032,-1.483031585460452,0.16839571526061453,-0.14341236045493638,-1.4825912376479937,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,0.16844573085692952,-0.13872401480181032,-1.483031585460452,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,-0.1683957152606147,-0.14341236045493638,-1.4825912376479937,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,-0.16844573085692968,-0.13872401480181032,-1.483031585460452,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,-0.1683957152606147,-0.14341236045493638,-1.4825912376479937,-0.3346397746521072,-0.14341236045493638,-1.454114890952407,-0.16844573085692968,-0.13872401480181032,-1.483031585460452,-0.1683957152606147,-0.14341236045493638,-1.4825912376479937,-0.33473916677650917,-0.13872401480181032,-1.454546780940041,-0.16844573085692968,-0.13872401480181032,-1.483031585460452,-0.3346397746521072,-0.14341236045493638,-1.454114890952407,-0.496608014827786,-0.14341236045493638,-1.407058767243767,-0.33473916677650917,-0.13872401480181032,-1.454546780940041,-0.3346397746521072,-0.14341236045493638,-1.454114890952407,-0.4967555135094358,-0.13872401480181032,-1.4074766809845356,-0.33473916677650917,-0.13872401480181032,-1.454546780940041,-0.496608014827786,-0.14341236045493638,-1.407058767243767,-0.6522309062034454,-0.14341236045493638,-1.3420241204470866,-0.4967555135094358,-0.13872401480181032,-1.4074766809845356,-0.496608014827786,-0.14341236045493638,-1.407058767243767,-0.6524246267957915,-0.13872401480181032,-1.3424227180987511,-0.4967555135094358,-0.13872401480181032,-1.4074766809845356,-0.6522309062034454,-0.14341236045493638,-1.3420241204470866,-0.7995199961211301,-0.14341236045493638,-1.2598419228900155,-0.6524246267957915,-0.13872401480181032,-1.3424227180987511,-0.6522309062034454,-0.14341236045493638,-1.3420241204470866,-0.7997574633827518,-0.13872401480181032,-1.2602161114193287,-0.6524246267957915,-0.13872401480181032,-1.3424227180987511,-0.7995199961211301,-0.14341236045493638,-1.2598419228900155,0.6289315805939357,-0.42046828134539266,-1.294083649272359,0.7716631237908568,-0.41594602267436875,-1.2159465159301666,0.7709591344021218,-0.42046828134539266,-1.214837206157427,0.6295058797224125,-0.41594602267436875,-1.2952653217068253,0.7716631237908568,-0.41594602267436875,-1.2159465159301666,0.6289315805939357,-0.42046828134539266,-1.294083649272359,0.47886792964059977,-0.42046828134539266,-1.3567950951946197,0.6295058797224125,-0.41594602267436875,-1.2952653217068253,0.6289315805939357,-0.42046828134539266,-1.294083649272359,0.47930520047121755,-0.41594602267436875,-1.3580340316143111,0.6295058797224125,-0.41594602267436875,-1.2952653217068253,0.47886792964059977,-0.42046828134539266,-1.3567950951946197,0.3226856016784632,-0.42046828134539266,-1.4021702560145326,0.47930520047121755,-0.41594602267436875,-1.3580340316143111,0.47886792964059977,-0.42046828134539266,-1.3567950951946197,0.3229802570361111,-0.41594602267436875,-1.4034506260593071,0.47930520047121755,-0.41594602267436875,-1.3580340316143111,0.3226856016784632,-0.42046828134539266,-1.4021702560145326,0.1623801974987507,-0.42046828134539266,-1.429629356106931,0.3229802570361111,-0.41594602267436875,-1.4034506260593071,0.3226856016784632,-0.42046828134539266,-1.4021702560145326,0.1625284724602614,-0.41594602267436875,-1.430934800003518,0.3229802570361111,-0.41594602267436875,-1.4034506260593071,0.1623801974987507,-0.42046828134539266,-1.429629356106931,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,0.1625284724602614,-0.41594602267436875,-1.430934800003518,0.1623801974987507,-0.42046828134539266,-1.429629356106931,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,0.1625284724602614,-0.41594602267436875,-1.430934800003518,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,-0.16238019749875085,-0.42046828134539266,-1.429629356106931,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,-0.16252847246026156,-0.41594602267436875,-1.430934800003518,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,-0.16238019749875085,-0.42046828134539266,-1.429629356106931,-0.3226856016784634,-0.42046828134539266,-1.4021702560145326,-0.16252847246026156,-0.41594602267436875,-1.430934800003518,-0.16238019749875085,-0.42046828134539266,-1.429629356106931,-0.3229802570361113,-0.41594602267436875,-1.4034506260593071,-0.16252847246026156,-0.41594602267436875,-1.430934800003518,-0.3226856016784634,-0.42046828134539266,-1.4021702560145326,-0.4788679296405997,-0.42046828134539266,-1.35679509519462,-0.3229802570361113,-0.41594602267436875,-1.4034506260593071,-0.3226856016784634,-0.42046828134539266,-1.4021702560145326,-0.4793052004712175,-0.41594602267436875,-1.3580340316143114,-0.3229802570361113,-0.41594602267436875,-1.4034506260593071,-0.4788679296405997,-0.42046828134539266,-1.35679509519462,-0.6289315805939355,-0.42046828134539266,-1.294083649272359,-0.4793052004712175,-0.41594602267436875,-1.3580340316143114,-0.4788679296405997,-0.42046828134539266,-1.35679509519462,-0.6295058797224122,-0.41594602267436875,-1.2952653217068253,-0.4793052004712175,-0.41594602267436875,-1.3580340316143114,-0.6289315805939355,-0.42046828134539266,-1.294083649272359,-0.7709591344021216,-0.42046828134539266,-1.2148372061574273,-0.6295058797224122,-0.41594602267436875,-1.2952653217068253,-0.6289315805939355,-0.42046828134539266,-1.294083649272359,-0.7716631237908564,-0.41594602267436875,-1.2159465159301668,-0.6295058797224122,-0.41594602267436875,-1.2952653217068253,-0.7709591344021216,-0.42046828134539266,-1.2148372061574273,0.583352040192686,-0.6826289037478287,-1.2002996196663018,0.7162322333317126,-0.6784329353268667,-1.1286014089130136,0.7150866609909294,-0.6826289037478287,-1.1267962757485852,0.5842865730243545,-0.6784329353268667,-1.2022225055484645,0.7162322333317126,-0.6784329353268667,-1.1286014089130136,0.583352040192686,-0.6826289037478287,-1.2002996196663018,0.4441637093098217,-0.6826289037478287,-1.2584662804780182,0.5842865730243545,-0.6784329353268667,-1.2022225055484645,0.583352040192686,-0.6826289037478287,-1.2002996196663018,0.44487526175223463,-0.6784329353268667,-1.260482349636302,0.5842865730243545,-0.6784329353268667,-1.2022225055484645,0.4441637093098217,-0.6826289037478287,-1.2584662804780182,0.2993001303928337,-0.6826289037478287,-1.3005530407157067,0.44487526175223463,-0.6784329353268667,-1.260482349636302,0.4441637093098217,-0.6826289037478287,-1.2584662804780182,0.2997796106707845,-0.6784329353268667,-1.3026365330704666,0.44487526175223463,-0.6784329353268667,-1.260482349636302,0.2993001303928337,-0.6826289037478287,-1.3005530407157067,0.15061228028704413,-0.6826289037478287,-1.326022141894612,0.2997796106707845,-0.6784329353268667,-1.3026365330704666,0.2993001303928337,-0.6826289037478287,-1.3005530407157067,0.1508535618993176,-0.6784329353268667,-1.3281464358744712,0.2997796106707845,-0.6784329353268667,-1.3026365330704666,0.15061228028704413,-0.6826289037478287,-1.326022141894612,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,0.1508535618993176,-0.6784329353268667,-1.3281464358744712,0.15061228028704413,-0.6826289037478287,-1.326022141894612,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,0.1508535618993176,-0.6784329353268667,-1.3281464358744712,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,-0.15061228028704426,-0.6826289037478287,-1.326022141894612,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,-0.15085356189931773,-0.6784329353268667,-1.3281464358744712,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,-0.15061228028704426,-0.6826289037478287,-1.326022141894612,-0.2993001303928339,-0.6826289037478287,-1.3005530407157067,-0.15085356189931773,-0.6784329353268667,-1.3281464358744712,-0.15061228028704426,-0.6826289037478287,-1.326022141894612,-0.2997796106707846,-0.6784329353268667,-1.3026365330704666,-0.15085356189931773,-0.6784329353268667,-1.3281464358744712,-0.2993001303928339,-0.6826289037478287,-1.3005530407157067,-0.4441637093098215,-0.6826289037478287,-1.2584662804780182,-0.2997796106707846,-0.6784329353268667,-1.3026365330704666,-0.2993001303928339,-0.6826289037478287,-1.3005530407157067,-0.44487526175223446,-0.6784329353268667,-1.260482349636302,-0.2997796106707846,-0.6784329353268667,-1.3026365330704666,-0.4441637093098215,-0.6826289037478287,-1.2584662804780182,-0.5833520401926859,-0.6826289037478287,-1.2002996196663018,-0.44487526175223446,-0.6784329353268667,-1.260482349636302,-0.4441637093098215,-0.6826289037478287,-1.2584662804780182,-0.5842865730243543,-0.6784329353268667,-1.2022225055484645,-0.44487526175223446,-0.6784329353268667,-1.260482349636302,-0.5833520401926859,-0.6826289037478287,-1.2002996196663018,-0.7150866609909291,-0.6826289037478287,-1.1267962757485854,-0.5842865730243543,-0.6784329353268667,-1.2022225055484645,-0.5833520401926859,-0.6826289037478287,-1.2002996196663018,-0.7162322333317123,-0.6784329353268667,-1.1286014089130139,-0.5842865730243543,-0.6784329353268667,-1.2022225055484645,-0.7150866609909291,-0.6826289037478287,-1.1267962757485854,0.7174308338174925,0.6805317591095806,-1.1265746861753807,0.7694019944553447,0.41820766795280473,-1.2166034347286772,0.7138880605051496,0.6805317591095806,-1.1288229984862177,0.7732202637376339,0.41820766795280473,-1.2141802873589163,0.7694019944553447,0.41820766795280473,-1.2166034347286772,0.7174308338174925,0.6805317591095806,-1.1265746861753807,0.7732202637376339,0.41820766795280473,-1.2141802873589163,0.7976594791722711,0.14106836166445297,-1.261285087767206,0.7694019944553447,0.41820766795280473,-1.2166034347286772,0.8016179803316116,0.14106836166445297,-1.2587729465421376,0.7976594791722711,0.14106836166445297,-1.261285087767206,0.7732202637376339,0.41820766795280473,-1.2141802873589163,0.8016179803316116,0.14106836166445297,-1.2587729465421376,0.7976594791722711,-0.14106836166445297,-1.261285087767206,0.7976594791722711,0.14106836166445297,-1.261285087767206,0.8016179803316116,-0.14106836166445297,-1.2587729465421376,0.7976594791722711,-0.14106836166445297,-1.261285087767206,0.8016179803316116,0.14106836166445297,-1.2587729465421376,0.8016179803316116,-0.14106836166445297,-1.2587729465421376,0.7694019944553447,-0.41820766795280473,-1.2166034347286772,0.7976594791722711,-0.14106836166445297,-1.261285087767206,0.7732202637376339,-0.41820766795280473,-1.2141802873589163,0.7694019944553447,-0.41820766795280473,-1.2166034347286772,0.8016179803316116,-0.14106836166445297,-1.2587729465421376,0.7732202637376339,-0.41820766795280473,-1.2141802873589163,0.7138880605051496,-0.6805317591095806,-1.1288229984862177,0.7694019944553447,-0.41820766795280473,-1.2166034347286772,0.7174308338174925,-0.6805317591095806,-1.1265746861753807,0.7138880605051496,-0.6805317591095806,-1.1288229984862177,0.7732202637376339,-0.41820766795280473,-1.2141802873589163,0.5857062446251304,0.6805317591095806,-1.200344000630797,0.6271850585593439,0.41820766795280473,-1.2956628607725813,0.5819323685919104,0.6805317591095806,-1.202178124583969,0.6312524017570044,0.41820766795280473,-1.2936861102066028,0.6271850585593439,0.41820766795280473,-1.2956628607725813,0.5857062446251304,0.6805317591095806,-1.200344000630797,0.6312524017570044,0.41820766795280473,-1.2936861102066028,0.6502194051488286,0.14106836166445297,-1.343248094175164,0.6271850585593439,0.41820766795280473,-1.2956628607725813,0.6544361278504087,0.14106836166445297,-1.3411987443706734,0.6502194051488286,0.14106836166445297,-1.343248094175164,0.6312524017570044,0.41820766795280473,-1.2936861102066028,0.6544361278504087,0.14106836166445297,-1.3411987443706734,0.6502194051488286,-0.14106836166445297,-1.343248094175164,0.6502194051488286,0.14106836166445297,-1.343248094175164,0.6544361278504087,-0.14106836166445297,-1.3411987443706734,0.6502194051488286,-0.14106836166445297,-1.343248094175164,0.6544361278504087,0.14106836166445297,-1.3411987443706734,0.6544361278504087,-0.14106836166445297,-1.3411987443706734,0.6271850585593439,-0.41820766795280473,-1.2956628607725813,0.6502194051488286,-0.14106836166445297,-1.343248094175164,0.6312524017570044,-0.41820766795280473,-1.2936861102066028,0.6271850585593439,-0.41820766795280473,-1.2956628607725813,0.6544361278504087,-0.14106836166445297,-1.3411987443706734,0.6312524017570044,-0.41820766795280473,-1.2936861102066028,0.5819323685919104,-0.6805317591095806,-1.202178124583969,0.6271850585593439,-0.41820766795280473,-1.2956628607725813,0.5857062446251304,-0.6805317591095806,-1.200344000630797,0.5819323685919104,-0.6805317591095806,-1.202178124583969,0.6312524017570044,-0.41820766795280473,-1.2936861102066028,0.44649786478596426,0.6805317591095806,-1.2587760649078115,0.476954341492091,0.41820766795280473,-1.3581671114400182,0.4425411062760917,0.6805317591095806,-1.260172565206509,0.481218788619726,0.41820766795280473,-1.3566620153689128,0.476954341492091,0.41820766795280473,-1.3581671114400182,0.44649786478596426,0.6805317591095806,-1.2587760649078115,0.481218788619726,0.41820766795280473,-1.3566620153689128,0.4944712313786647,0.14106836166445297,-1.408047910646571,0.476954341492091,0.41820766795280473,-1.3581671114400182,0.4988922969585571,0.14106836166445297,-1.4064875375817314,0.4944712313786647,0.14106836166445297,-1.408047910646571,0.481218788619726,0.41820766795280473,-1.3566620153689128,0.4988922969585571,0.14106836166445297,-1.4064875375817314,0.4944712313786647,-0.14106836166445297,-1.408047910646571,0.4944712313786647,0.14106836166445297,-1.408047910646571,0.4988922969585571,-0.14106836166445297,-1.4064875375817314,0.4944712313786647,-0.14106836166445297,-1.408047910646571,0.4988922969585571,0.14106836166445297,-1.4064875375817314,0.4988922969585571,-0.14106836166445297,-1.4064875375817314,0.476954341492091,-0.41820766795280473,-1.3581671114400182,0.4944712313786647,-0.14106836166445297,-1.408047910646571,0.481218788619726,-0.41820766795280473,-1.3566620153689128,0.476954341492091,-0.41820766795280473,-1.3581671114400182,0.4988922969585571,-0.14106836166445297,-1.4064875375817314,0.481218788619726,-0.41820766795280473,-1.3566620153689128,0.4425411062760917,-0.6805317591095806,-1.260172565206509,0.476954341492091,-0.41820766795280473,-1.3581671114400182,0.44649786478596426,-0.6805317591095806,-1.2587760649078115,0.4425411062760917,-0.6805317591095806,-1.260172565206509,0.481218788619726,-0.41820766795280473,-1.3566620153689128,0.301584412523605,0.6805317591095806,-1.3011242703777424,0.3206293980569848,0.41820766795280473,-1.4033175462336003,0.29749532854001354,0.6805317591095806,-1.302065303408431,0.32503646065758984,0.41820766795280473,-1.4023033358402397,0.3206293980569848,0.41820766795280473,-1.4033175462336003,0.301584412523605,0.6805317591095806,-1.3011242703777424,0.32503646065758984,0.41820766795280473,-1.4023033358402397,0.3324050113003664,0.14106836166445297,-1.4548565653698342,0.3206293980569848,0.41820766795280473,-1.4033175462336003,0.33697393012824994,0.14106836166445297,-1.4538051065226139,0.3324050113003664,0.14106836166445297,-1.4548565653698342,0.32503646065758984,0.41820766795280473,-1.4023033358402397,0.33697393012824994,0.14106836166445297,-1.4538051065226139,0.3324050113003664,-0.14106836166445297,-1.4548565653698342,0.3324050113003664,0.14106836166445297,-1.4548565653698342,0.33697393012824994,-0.14106836166445297,-1.4538051065226139,0.3324050113003664,-0.14106836166445297,-1.4548565653698342,0.33697393012824994,0.14106836166445297,-1.4538051065226139,0.33697393012824994,-0.14106836166445297,-1.4538051065226139,0.3206293980569848,-0.41820766795280473,-1.4033175462336003,0.3324050113003664,-0.14106836166445297,-1.4548565653698342,0.32503646065758984,-0.41820766795280473,-1.4023033358402397,0.3206293980569848,-0.41820766795280473,-1.4033175462336003,0.33697393012824994,-0.14106836166445297,-1.4538051065226139,0.32503646065758984,-0.41820766795280473,-1.4023033358402397,0.29749532854001354,-0.6805317591095806,-1.302065303408431,0.3206293980569848,-0.41820766795280473,-1.4033175462336003,0.301584412523605,-0.6805317591095806,-1.3011242703777424,0.29749532854001354,-0.6805317591095806,-1.302065303408431,0.32503646065758984,-0.41820766795280473,-1.4023033358402397,0.15281750193400737,0.6805317591095806,-1.326847517971025,0.16020765129719292,0.41820766795280473,-1.4305372609377618,0.14864834025235465,0.6805317591095806,-1.3273210597980583,0.16470101866181952,0.41820766795280473,-1.4300268951726873,0.16020765129719292,0.41820766795280473,-1.4305372609377618,0.15281750193400737,0.6805317591095806,-1.326847517971025,0.16470101866181952,0.41820766795280473,-1.4300268951726873,0.1660915264244855,0.14106836166445297,-1.4830759664249473,0.16020765129719292,0.41820766795280473,-1.4305372609377618,0.17074991969305892,0.14106836166445297,-1.4825468566834983,0.1660915264244855,0.14106836166445297,-1.4830759664249473,0.16470101866181952,0.41820766795280473,-1.4300268951726873,0.17074991969305892,0.14106836166445297,-1.4825468566834983,0.1660915264244855,-0.14106836166445297,-1.4830759664249473,0.1660915264244855,0.14106836166445297,-1.4830759664249473,0.17074991969305892,-0.14106836166445297,-1.4825468566834983,0.1660915264244855,-0.14106836166445297,-1.4830759664249473,0.17074991969305892,0.14106836166445297,-1.4825468566834983,0.17074991969305892,-0.14106836166445297,-1.4825468566834983,0.16020765129719292,-0.41820766795280473,-1.4305372609377618,0.1660915264244855,-0.14106836166445297,-1.4830759664249473,0.16470101866181952,-0.41820766795280473,-1.4300268951726873,0.16020765129719292,-0.41820766795280473,-1.4305372609377618,0.17074991969305892,-0.14106836166445297,-1.4825468566834983,0.16470101866181952,-0.41820766795280473,-1.4300268951726873,0.14864834025235465,-0.6805317591095806,-1.3273210597980583,0.16020765129719292,-0.41820766795280473,-1.4305372609377618,0.15281750193400737,-0.6805317591095806,-1.326847517971025,0.14864834025235465,-0.6805317591095806,-1.3273210597980583,0.16470101866181952,-0.41820766795280473,-1.4300268951726873,2.0979842104809974e-3,0.6805317591095806,-1.3356171320050791,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,-2.097984210480834e-3,0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,2.0979842104809974e-3,0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-2.097984210480834e-3,-0.6805317591095806,-1.3356171320050791,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,2.0979842104809974e-3,-0.6805317591095806,-1.3356171320050791,-2.097984210480834e-3,-0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-0.1486483402523545,0.6805317591095806,-1.3273210597980583,-0.16470101866181935,0.41820766795280473,-1.4300268951726873,-0.15281750193400725,0.6805317591095806,-1.326847517971025,-0.16020765129719275,0.41820766795280473,-1.4305372609377618,-0.16470101866181935,0.41820766795280473,-1.4300268951726873,-0.1486483402523545,0.6805317591095806,-1.3273210597980583,-0.16020765129719275,0.41820766795280473,-1.4305372609377618,-0.17074991969305875,0.14106836166445297,-1.4825468566834983,-0.16470101866181935,0.41820766795280473,-1.4300268951726873,-0.16609152642448533,0.14106836166445297,-1.4830759664249473,-0.17074991969305875,0.14106836166445297,-1.4825468566834983,-0.16020765129719275,0.41820766795280473,-1.4305372609377618,-0.16609152642448533,0.14106836166445297,-1.4830759664249473,-0.17074991969305875,-0.14106836166445297,-1.4825468566834983,-0.17074991969305875,0.14106836166445297,-1.4825468566834983,-0.16609152642448533,-0.14106836166445297,-1.4830759664249473,-0.17074991969305875,-0.14106836166445297,-1.4825468566834983,-0.16609152642448533,0.14106836166445297,-1.4830759664249473,-0.16609152642448533,-0.14106836166445297,-1.4830759664249473,-0.16470101866181935,-0.41820766795280473,-1.4300268951726873,-0.17074991969305875,-0.14106836166445297,-1.4825468566834983,-0.16020765129719275,-0.41820766795280473,-1.4305372609377618,-0.16470101866181935,-0.41820766795280473,-1.4300268951726873,-0.16609152642448533,-0.14106836166445297,-1.4830759664249473,-0.16020765129719275,-0.41820766795280473,-1.4305372609377618,-0.15281750193400725,-0.6805317591095806,-1.326847517971025,-0.16470101866181935,-0.41820766795280473,-1.4300268951726873,-0.1486483402523545,-0.6805317591095806,-1.3273210597980583,-0.15281750193400725,-0.6805317591095806,-1.326847517971025,-0.16020765129719275,-0.41820766795280473,-1.4305372609377618,-0.29749532854001337,0.6805317591095806,-1.302065303408431,-0.3250364606575897,0.41820766795280473,-1.4023033358402397,-0.30158441252360485,0.6805317591095806,-1.3011242703777424,-0.32062939805698465,0.41820766795280473,-1.4033175462336003,-0.3250364606575897,0.41820766795280473,-1.4023033358402397,-0.29749532854001337,0.6805317591095806,-1.302065303408431,-0.32062939805698465,0.41820766795280473,-1.4033175462336003,-0.3369739301282498,0.14106836166445297,-1.4538051065226139,-0.3250364606575897,0.41820766795280473,-1.4023033358402397,-0.33240501130036626,0.14106836166445297,-1.4548565653698342,-0.3369739301282498,0.14106836166445297,-1.4538051065226139,-0.32062939805698465,0.41820766795280473,-1.4033175462336003,-0.33240501130036626,0.14106836166445297,-1.4548565653698342,-0.3369739301282498,-0.14106836166445297,-1.4538051065226139,-0.3369739301282498,0.14106836166445297,-1.4538051065226139,-0.33240501130036626,-0.14106836166445297,-1.4548565653698342,-0.3369739301282498,-0.14106836166445297,-1.4538051065226139,-0.33240501130036626,0.14106836166445297,-1.4548565653698342,-0.33240501130036626,-0.14106836166445297,-1.4548565653698342,-0.3250364606575897,-0.41820766795280473,-1.4023033358402397,-0.3369739301282498,-0.14106836166445297,-1.4538051065226139,-0.32062939805698465,-0.41820766795280473,-1.4033175462336003,-0.3250364606575897,-0.41820766795280473,-1.4023033358402397,-0.33240501130036626,-0.14106836166445297,-1.4548565653698342,-0.32062939805698465,-0.41820766795280473,-1.4033175462336003,-0.30158441252360485,-0.6805317591095806,-1.3011242703777424,-0.3250364606575897,-0.41820766795280473,-1.4023033358402397,-0.29749532854001337,-0.6805317591095806,-1.302065303408431,-0.30158441252360485,-0.6805317591095806,-1.3011242703777424,-0.32062939805698465,-0.41820766795280473,-1.4033175462336003,-0.4425411062760919,0.6805317591095806,-1.260172565206509,-0.48121878861972617,0.41820766795280473,-1.3566620153689128,-0.4464978647859644,0.6805317591095806,-1.2587760649078115,-0.47695434149209115,0.41820766795280473,-1.3581671114400182,-0.48121878861972617,0.41820766795280473,-1.3566620153689128,-0.4425411062760919,0.6805317591095806,-1.260172565206509,-0.47695434149209115,0.41820766795280473,-1.3581671114400182,-0.49889229695855725,0.14106836166445297,-1.4064875375817314,-0.48121878861972617,0.41820766795280473,-1.3566620153689128,-0.4944712313786649,0.14106836166445297,-1.408047910646571,-0.49889229695855725,0.14106836166445297,-1.4064875375817314,-0.47695434149209115,0.41820766795280473,-1.3581671114400182,-0.4944712313786649,0.14106836166445297,-1.408047910646571,-0.49889229695855725,-0.14106836166445297,-1.4064875375817314,-0.49889229695855725,0.14106836166445297,-1.4064875375817314,-0.4944712313786649,-0.14106836166445297,-1.408047910646571,-0.49889229695855725,-0.14106836166445297,-1.4064875375817314,-0.4944712313786649,0.14106836166445297,-1.408047910646571,-0.4944712313786649,-0.14106836166445297,-1.408047910646571,-0.48121878861972617,-0.41820766795280473,-1.3566620153689128,-0.49889229695855725,-0.14106836166445297,-1.4064875375817314,-0.47695434149209115,-0.41820766795280473,-1.3581671114400182,-0.48121878861972617,-0.41820766795280473,-1.3566620153689128,-0.4944712313786649,-0.14106836166445297,-1.408047910646571,-0.47695434149209115,-0.41820766795280473,-1.3581671114400182,-0.4464978647859644,-0.6805317591095806,-1.2587760649078115,-0.48121878861972617,-0.41820766795280473,-1.3566620153689128,-0.4425411062760919,-0.6805317591095806,-1.260172565206509,-0.4464978647859644,-0.6805317591095806,-1.2587760649078115,-0.47695434149209115,-0.41820766795280473,-1.3581671114400182,-0.5819323685919102,0.6805317591095806,-1.202178124583969,-0.6312524017570039,0.41820766795280473,-1.293686110206603,-0.58570624462513,0.6805317591095806,-1.2003440006307973,-0.6271850585593437,0.41820766795280473,-1.2956628607725813,-0.6312524017570039,0.41820766795280473,-1.293686110206603,-0.5819323685919102,0.6805317591095806,-1.202178124583969,-0.6271850585593437,0.41820766795280473,-1.2956628607725813,-0.6544361278504083,0.14106836166445297,-1.3411987443706737,-0.6312524017570039,0.41820766795280473,-1.293686110206603,-0.6502194051488285,0.14106836166445297,-1.343248094175164,-0.6544361278504083,0.14106836166445297,-1.3411987443706737,-0.6271850585593437,0.41820766795280473,-1.2956628607725813,-0.6502194051488285,0.14106836166445297,-1.343248094175164,-0.6544361278504083,-0.14106836166445297,-1.3411987443706737,-0.6544361278504083,0.14106836166445297,-1.3411987443706737,-0.6502194051488285,-0.14106836166445297,-1.343248094175164,-0.6544361278504083,-0.14106836166445297,-1.3411987443706737,-0.6502194051488285,0.14106836166445297,-1.343248094175164,-0.6502194051488285,-0.14106836166445297,-1.343248094175164,-0.6312524017570039,-0.41820766795280473,-1.293686110206603,-0.6544361278504083,-0.14106836166445297,-1.3411987443706737,-0.6271850585593437,-0.41820766795280473,-1.2956628607725813,-0.6312524017570039,-0.41820766795280473,-1.293686110206603,-0.6502194051488285,-0.14106836166445297,-1.343248094175164,-0.6271850585593437,-0.41820766795280473,-1.2956628607725813,-0.58570624462513,-0.6805317591095806,-1.2003440006307973,-0.6312524017570039,-0.41820766795280473,-1.293686110206603,-0.5819323685919102,-0.6805317591095806,-1.202178124583969,-0.58570624462513,-0.6805317591095806,-1.2003440006307973,-0.6271850585593437,-0.41820766795280473,-1.2956628607725813,-0.7138880605051494,0.6805317591095806,-1.128822998486218,-0.7732202637376335,0.41820766795280473,-1.2141802873589167,-0.7174308338174921,0.6805317591095806,-1.126574686175381,-0.7694019944553445,0.41820766795280473,-1.2166034347286772,-0.7732202637376335,0.41820766795280473,-1.2141802873589167,-0.7138880605051494,0.6805317591095806,-1.128822998486218,-0.7694019944553445,0.41820766795280473,-1.2166034347286772,-0.801617980331611,0.14106836166445297,-1.258772946542138,-0.7732202637376335,0.41820766795280473,-1.2141802873589167,-0.797659479172271,0.14106836166445297,-1.2612850877672062,-0.801617980331611,0.14106836166445297,-1.258772946542138,-0.7694019944553445,0.41820766795280473,-1.2166034347286772,-0.797659479172271,0.14106836166445297,-1.2612850877672062,-0.801617980331611,-0.14106836166445297,-1.258772946542138,-0.801617980331611,0.14106836166445297,-1.258772946542138,-0.797659479172271,-0.14106836166445297,-1.2612850877672062,-0.801617980331611,-0.14106836166445297,-1.258772946542138,-0.797659479172271,0.14106836166445297,-1.2612850877672062,-0.797659479172271,-0.14106836166445297,-1.2612850877672062,-0.7732202637376335,-0.41820766795280473,-1.2141802873589167,-0.801617980331611,-0.14106836166445297,-1.258772946542138,-0.7694019944553445,-0.41820766795280473,-1.2166034347286772,-0.7732202637376335,-0.41820766795280473,-1.2141802873589167,-0.797659479172271,-0.14106836166445297,-1.2612850877672062,-0.7694019944553445,-0.41820766795280473,-1.2166034347286772,-0.7174308338174921,-0.6805317591095806,-1.126574686175381,-0.7732202637376335,-0.41820766795280473,-1.2141802873589167,-0.7138880605051494,-0.6805317591095806,-1.128822998486218,-0.7174308338174921,-0.6805317591095806,-1.126574686175381,-0.7694019944553445,-0.41820766795280473,-1.2166034347286772],[0.584209499867372,0.6809857496093201,-1.2020639205523753,0.7718266317522666,0.4184866590588439,-1.2162041632504461,0.7161377552418435,0.6809857496093201,-1.1284525352650885,0.6296392659370903,0.4184866590588439,-1.295539775915803,0.7718266317522666,0.4184866590588439,-1.2162041632504461,0.584209499867372,0.6809857496093201,-1.2020639205523753,0.6296392659370903,0.4184866590588439,-1.295539775915803,0.8001731650443925,0.14116246997777146,-1.2608711524230436,0.7718266317522666,0.4184866590588439,-1.2162041632504461,0.6527637471089762,0.14116246997777146,-1.3431204888356159,0.8001731650443925,0.14116246997777146,-1.2608711524230436,0.6296392659370903,0.4184866590588439,-1.295539775915803,0.6527637471089762,0.14116246997777146,-1.3431204888356159,0.8001731650443925,-0.14116246997777146,-1.2608711524230436,0.8001731650443925,0.14116246997777146,-1.2608711524230436,0.6527637471089762,-0.14116246997777146,-1.3431204888356159,0.8001731650443925,-0.14116246997777146,-1.2608711524230436,0.6527637471089762,0.14116246997777146,-1.3431204888356159,0.6527637471089762,-0.14116246997777146,-1.3431204888356159,0.7718266317522666,-0.4184866590588439,-1.2162041632504461,0.8001731650443925,-0.14116246997777146,-1.2608711524230436,0.6296392659370903,-0.4184866590588439,-1.295539775915803,0.7718266317522666,-0.4184866590588439,-1.2162041632504461,0.6527637471089762,-0.14116246997777146,-1.3431204888356159,0.6296392659370903,-0.4184866590588439,-1.295539775915803,0.7161377552418435,-0.6809857496093201,-1.1284525352650885,0.7718266317522666,-0.4184866590588439,-1.2162041632504461,0.584209499867372,-0.6809857496093201,-1.2020639205523753,0.7161377552418435,-0.6809857496093201,-1.1284525352650885,0.6296392659370903,-0.4184866590588439,-1.295539775915803,0.44481657832107313,0.6809857496093201,-1.2603160795926414,0.6296392659370903,0.4184866590588439,-1.295539775915803,0.584209499867372,0.6809857496093201,-1.2020639205523753,0.47940676061295046,0.4184866590588439,-1.3583217859065517,0.6296392659370903,0.4184866590588439,-1.295539775915803,0.44481657832107313,0.6809857496093201,-1.2603160795926414,0.47940676061295046,0.4184866590588439,-1.3583217859065517,0.6527637471089762,0.14116246997777146,-1.3431204888356159,0.6296392659370903,0.4184866590588439,-1.295539775915803,0.4970137194054103,0.14116246997777146,-1.408208265773417,0.6527637471089762,0.14116246997777146,-1.3431204888356159,0.47940676061295046,0.4184866590588439,-1.3583217859065517,0.4970137194054103,0.14116246997777146,-1.408208265773417,0.6527637471089762,-0.14116246997777146,-1.3431204888356159,0.6527637471089762,0.14116246997777146,-1.3431204888356159,0.4970137194054103,-0.14116246997777146,-1.408208265773417,0.6527637471089762,-0.14116246997777146,-1.3431204888356159,0.4970137194054103,0.14116246997777146,-1.408208265773417,0.4970137194054103,-0.14116246997777146,-1.408208265773417,0.6296392659370903,-0.4184866590588439,-1.295539775915803,0.6527637471089762,-0.14116246997777146,-1.3431204888356159,0.47940676061295046,-0.4184866590588439,-1.3583217859065517,0.6296392659370903,-0.4184866590588439,-1.295539775915803,0.4970137194054103,-0.14116246997777146,-1.408208265773417,0.47940676061295046,-0.4184866590588439,-1.3583217859065517,0.584209499867372,-0.6809857496093201,-1.2020639205523753,0.6296392659370903,-0.4184866590588439,-1.295539775915803,0.44481657832107313,-0.6809857496093201,-1.2603160795926414,0.584209499867372,-0.6809857496093201,-1.2020639205523753,0.47940676061295046,-0.4184866590588439,-1.3583217859065517,0.29974006678588266,0.6809857496093201,-1.3024647024745917,0.47940676061295046,0.4184866590588439,-1.3583217859065517,0.44481657832107313,0.6809857496093201,-1.2603160795926414,0.32304869343248077,0.4184866590588439,-1.4037480036891714,0.47940676061295046,0.4184866590588439,-1.3583217859065517,0.29974006678588266,0.6809857496093201,-1.3024647024745917,0.32304869343248077,0.4184866590588439,-1.4037480036891714,0.4970137194054103,0.14116246997777146,-1.408208265773417,0.47940676061295046,0.4184866590588439,-1.3583217859065517,0.3349131590606904,0.14116246997777146,-1.4553028320448507,0.4970137194054103,0.14116246997777146,-1.408208265773417,0.32304869343248077,0.4184866590588439,-1.4037480036891714,0.3349131590606904,0.14116246997777146,-1.4553028320448507,0.4970137194054103,-0.14116246997777146,-1.408208265773417,0.4970137194054103,0.14116246997777146,-1.408208265773417,0.3349131590606904,-0.14116246997777146,-1.4553028320448507,0.4970137194054103,-0.14116246997777146,-1.408208265773417,0.3349131590606904,0.14116246997777146,-1.4553028320448507,0.3349131590606904,-0.14116246997777146,-1.4553028320448507,0.47940676061295046,-0.4184866590588439,-1.3583217859065517,0.4970137194054103,-0.14116246997777146,-1.408208265773417,0.32304869343248077,-0.4184866590588439,-1.4037480036891714,0.47940676061295046,-0.4184866590588439,-1.3583217859065517,0.3349131590606904,-0.14116246997777146,-1.4553028320448507,0.32304869343248077,-0.4184866590588439,-1.4037480036891714,0.44481657832107313,-0.6809857496093201,-1.2603160795926414,0.47940676061295046,-0.4184866590588439,-1.3583217859065517,0.29974006678588266,-0.6809857496093201,-1.3024647024745917,0.44481657832107313,-0.6809857496093201,-1.2603160795926414,0.32304869343248077,-0.4184866590588439,-1.4037480036891714,0.1508336628278783,0.6809857496093201,-1.3279712402710226,0.32304869343248077,0.4184866590588439,-1.4037480036891714,0.29974006678588266,0.6809857496093201,-1.3024647024745917,0.16256291067349685,0.4184866590588439,-1.4312380012643346,0.32304869343248077,0.4184866590588439,-1.4037480036891714,0.1508336628278783,0.6809857496093201,-1.3279712402710226,0.16256291067349685,0.4184866590588439,-1.4312380012643346,0.3349131590606904,0.14116246997777146,-1.4553028320448507,0.32304869343248077,0.4184866590588439,-1.4037480036891714,0.16853328636396697,0.14116246997777146,-1.4838024425297107,0.3349131590606904,0.14116246997777146,-1.4553028320448507,0.16256291067349685,0.4184866590588439,-1.4312380012643346,0.16853328636396697,0.14116246997777146,-1.4838024425297107,0.3349131590606904,-0.14116246997777146,-1.4553028320448507,0.3349131590606904,0.14116246997777146,-1.4553028320448507,0.16853328636396697,-0.14116246997777146,-1.4838024425297107,0.3349131590606904,-0.14116246997777146,-1.4553028320448507,0.16853328636396697,0.14116246997777146,-1.4838024425297107,0.16853328636396697,-0.14116246997777146,-1.4838024425297107,0.32304869343248077,-0.4184866590588439,-1.4037480036891714,0.3349131590606904,-0.14116246997777146,-1.4553028320448507,0.16256291067349685,-0.4184866590588439,-1.4312380012643346,0.32304869343248077,-0.4184866590588439,-1.4037480036891714,0.16853328636396697,-0.14116246997777146,-1.4838024425297107,0.16256291067349685,-0.4184866590588439,-1.4312380012643346,0.29974006678588266,-0.6809857496093201,-1.3024647024745917,0.32304869343248077,-0.4184866590588439,-1.4037480036891714,0.1508336628278783,-0.6809857496093201,-1.3279712402710226,0.29974006678588266,-0.6809857496093201,-1.3024647024745917,0.16256291067349685,-0.4184866590588439,-1.4312380012643346,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,0.16256291067349685,0.4184866590588439,-1.4312380012643346,0.1508336628278783,0.6809857496093201,-1.3279712402710226,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,0.16256291067349685,0.4184866590588439,-1.4312380012643346,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,0.16853328636396697,0.14116246997777146,-1.4838024425297107,0.16256291067349685,0.4184866590588439,-1.4312380012643346,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,0.16853328636396697,0.14116246997777146,-1.4838024425297107,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,0.16853328636396697,-0.14116246997777146,-1.4838024425297107,0.16853328636396697,0.14116246997777146,-1.4838024425297107,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,0.16853328636396697,-0.14116246997777146,-1.4838024425297107,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,0.16256291067349685,-0.4184866590588439,-1.4312380012643346,0.16853328636396697,-0.14116246997777146,-1.4838024425297107,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,0.16256291067349685,-0.4184866590588439,-1.4312380012643346,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,0.1508336628278783,-0.6809857496093201,-1.3279712402710226,0.16256291067349685,-0.4184866590588439,-1.4312380012643346,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,0.1508336628278783,-0.6809857496093201,-1.3279712402710226,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.15083366282787844,0.6809857496093201,-1.3279712402710226,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,-0.162562910673497,0.4184866590588439,-1.4312380012643346,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,-0.15083366282787844,0.6809857496093201,-1.3279712402710226,-0.162562910673497,0.4184866590588439,-1.4312380012643346,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,-0.16853328636396714,0.14116246997777146,-1.4838024425297107,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,-0.162562910673497,0.4184866590588439,-1.4312380012643346,-0.16853328636396714,0.14116246997777146,-1.4838024425297107,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,-0.16853328636396714,-0.14116246997777146,-1.4838024425297107,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,-0.16853328636396714,0.14116246997777146,-1.4838024425297107,-0.16853328636396714,-0.14116246997777146,-1.4838024425297107,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,-0.162562910673497,-0.4184866590588439,-1.4312380012643346,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.16853328636396714,-0.14116246997777146,-1.4838024425297107,-0.162562910673497,-0.4184866590588439,-1.4312380012643346,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.15083366282787844,-0.6809857496093201,-1.3279712402710226,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,-0.162562910673497,-0.4184866590588439,-1.4312380012643346,-0.29974006678588283,0.6809857496093201,-1.3024647024745917,-0.162562910673497,0.4184866590588439,-1.4312380012643346,-0.15083366282787844,0.6809857496093201,-1.3279712402710226,-0.32304869343248094,0.4184866590588439,-1.4037480036891714,-0.162562910673497,0.4184866590588439,-1.4312380012643346,-0.29974006678588283,0.6809857496093201,-1.3024647024745917,-0.32304869343248094,0.4184866590588439,-1.4037480036891714,-0.16853328636396714,0.14116246997777146,-1.4838024425297107,-0.162562910673497,0.4184866590588439,-1.4312380012643346,-0.3349131590606906,0.14116246997777146,-1.4553028320448507,-0.16853328636396714,0.14116246997777146,-1.4838024425297107,-0.32304869343248094,0.4184866590588439,-1.4037480036891714,-0.3349131590606906,0.14116246997777146,-1.4553028320448507,-0.16853328636396714,-0.14116246997777146,-1.4838024425297107,-0.16853328636396714,0.14116246997777146,-1.4838024425297107,-0.3349131590606906,-0.14116246997777146,-1.4553028320448507,-0.16853328636396714,-0.14116246997777146,-1.4838024425297107,-0.3349131590606906,0.14116246997777146,-1.4553028320448507,-0.3349131590606906,-0.14116246997777146,-1.4553028320448507,-0.162562910673497,-0.4184866590588439,-1.4312380012643346,-0.16853328636396714,-0.14116246997777146,-1.4838024425297107,-0.32304869343248094,-0.4184866590588439,-1.4037480036891714,-0.162562910673497,-0.4184866590588439,-1.4312380012643346,-0.3349131590606906,-0.14116246997777146,-1.4553028320448507,-0.32304869343248094,-0.4184866590588439,-1.4037480036891714,-0.15083366282787844,-0.6809857496093201,-1.3279712402710226,-0.162562910673497,-0.4184866590588439,-1.4312380012643346,-0.29974006678588283,-0.6809857496093201,-1.3024647024745917,-0.15083366282787844,-0.6809857496093201,-1.3279712402710226,-0.32304869343248094,-0.4184866590588439,-1.4037480036891714,-0.444816578321073,0.6809857496093201,-1.2603160795926414,-0.32304869343248094,0.4184866590588439,-1.4037480036891714,-0.29974006678588283,0.6809857496093201,-1.3024647024745917,-0.4794067606129503,0.4184866590588439,-1.358321785906552,-0.32304869343248094,0.4184866590588439,-1.4037480036891714,-0.444816578321073,0.6809857496093201,-1.2603160795926414,-0.4794067606129503,0.4184866590588439,-1.358321785906552,-0.3349131590606906,0.14116246997777146,-1.4553028320448507,-0.32304869343248094,0.4184866590588439,-1.4037480036891714,-0.4970137194054101,0.14116246997777146,-1.4082082657734172,-0.3349131590606906,0.14116246997777146,-1.4553028320448507,-0.4794067606129503,0.4184866590588439,-1.358321785906552,-0.4970137194054101,0.14116246997777146,-1.4082082657734172,-0.3349131590606906,-0.14116246997777146,-1.4553028320448507,-0.3349131590606906,0.14116246997777146,-1.4553028320448507,-0.4970137194054101,-0.14116246997777146,-1.4082082657734172,-0.3349131590606906,-0.14116246997777146,-1.4553028320448507,-0.4970137194054101,0.14116246997777146,-1.4082082657734172,-0.4970137194054101,-0.14116246997777146,-1.4082082657734172,-0.32304869343248094,-0.4184866590588439,-1.4037480036891714,-0.3349131590606906,-0.14116246997777146,-1.4553028320448507,-0.4794067606129503,-0.4184866590588439,-1.358321785906552,-0.32304869343248094,-0.4184866590588439,-1.4037480036891714,-0.4970137194054101,-0.14116246997777146,-1.4082082657734172,-0.4794067606129503,-0.4184866590588439,-1.358321785906552,-0.29974006678588283,-0.6809857496093201,-1.3024647024745917,-0.32304869343248094,-0.4184866590588439,-1.4037480036891714,-0.444816578321073,-0.6809857496093201,-1.2603160795926414,-0.29974006678588283,-0.6809857496093201,-1.3024647024745917,-0.4794067606129503,-0.4184866590588439,-1.358321785906552,-0.5842094998673719,0.6809857496093201,-1.2020639205523753,-0.4794067606129503,0.4184866590588439,-1.358321785906552,-0.444816578321073,0.6809857496093201,-1.2603160795926414,-0.6296392659370902,0.4184866590588439,-1.295539775915803,-0.4794067606129503,0.4184866590588439,-1.358321785906552,-0.5842094998673719,0.6809857496093201,-1.2020639205523753,-0.6296392659370902,0.4184866590588439,-1.295539775915803,-0.4970137194054101,0.14116246997777146,-1.4082082657734172,-0.4794067606129503,0.4184866590588439,-1.358321785906552,-0.652763747108976,0.14116246997777146,-1.3431204888356159,-0.4970137194054101,0.14116246997777146,-1.4082082657734172,-0.6296392659370902,0.4184866590588439,-1.295539775915803,-0.652763747108976,0.14116246997777146,-1.3431204888356159,-0.4970137194054101,-0.14116246997777146,-1.4082082657734172,-0.4970137194054101,0.14116246997777146,-1.4082082657734172,-0.652763747108976,-0.14116246997777146,-1.3431204888356159,-0.4970137194054101,-0.14116246997777146,-1.4082082657734172,-0.652763747108976,0.14116246997777146,-1.3431204888356159,-0.652763747108976,-0.14116246997777146,-1.3431204888356159,-0.4794067606129503,-0.4184866590588439,-1.358321785906552,-0.4970137194054101,-0.14116246997777146,-1.4082082657734172,-0.6296392659370902,-0.4184866590588439,-1.295539775915803,-0.4794067606129503,-0.4184866590588439,-1.358321785906552,-0.652763747108976,-0.14116246997777146,-1.3431204888356159,-0.6296392659370902,-0.4184866590588439,-1.295539775915803,-0.444816578321073,-0.6809857496093201,-1.2603160795926414,-0.4794067606129503,-0.4184866590588439,-1.358321785906552,-0.5842094998673719,-0.6809857496093201,-1.2020639205523753,-0.444816578321073,-0.6809857496093201,-1.2603160795926414,-0.6296392659370902,-0.4184866590588439,-1.295539775915803,-0.7161377552418432,0.6809857496093201,-1.1284525352650887,-0.6296392659370902,0.4184866590588439,-1.295539775915803,-0.5842094998673719,0.6809857496093201,-1.2020639205523753,-0.7718266317522664,0.4184866590588439,-1.2162041632504463,-0.6296392659370902,0.4184866590588439,-1.295539775915803,-0.7161377552418432,0.6809857496093201,-1.1284525352650887,-0.7718266317522664,0.4184866590588439,-1.2162041632504463,-0.652763747108976,0.14116246997777146,-1.3431204888356159,-0.6296392659370902,0.4184866590588439,-1.295539775915803,-0.8001731650443922,0.14116246997777146,-1.2608711524230438,-0.652763747108976,0.14116246997777146,-1.3431204888356159,-0.7718266317522664,0.4184866590588439,-1.2162041632504463,-0.8001731650443922,0.14116246997777146,-1.2608711524230438,-0.652763747108976,-0.14116246997777146,-1.3431204888356159,-0.652763747108976,0.14116246997777146,-1.3431204888356159,-0.8001731650443922,-0.14116246997777146,-1.2608711524230438,-0.652763747108976,-0.14116246997777146,-1.3431204888356159,-0.8001731650443922,0.14116246997777146,-1.2608711524230438,-0.8001731650443922,-0.14116246997777146,-1.2608711524230438,-0.6296392659370902,-0.4184866590588439,-1.295539775915803,-0.652763747108976,-0.14116246997777146,-1.3431204888356159,-0.7718266317522664,-0.4184866590588439,-1.2162041632504463,-0.6296392659370902,-0.4184866590588439,-1.295539775915803,-0.8001731650443922,-0.14116246997777146,-1.2608711524230438,-0.7718266317522664,-0.4184866590588439,-1.2162041632504463,-0.5842094998673719,-0.6809857496093201,-1.2020639205523753,-0.6296392659370902,-0.4184866590588439,-1.295539775915803,-0.7161377552418432,-0.6809857496093201,-1.1284525352650887,-0.5842094998673719,-0.6809857496093201,-1.2020639205523753,-0.7718266317522664,-0.4184866590588439,-1.2162041632504463],[[0.17811055558800623,0.303478973012677,-1.251469992819365,0.300451362902075,0.1831716015188576,-1.2514699928193653,0.29509382483070545,0.30347897301267707,-1.2291542407224392,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.300451362902075,0.1831716015188576,-1.2514699928193653,0.17811055558800623,0.303478973012677,-1.251469992819365,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.3031420734062201,6.123838592253551e-2,-1.2626776086636606,0.300451362902075,0.1831716015188576,-1.2514699928193653,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.3031420734062201,6.123838592253551e-2,-1.2626776086636606,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.3031420734062201,-6.123838592253551e-2,-1.2626776086636606,0.3031420734062201,6.123838592253551e-2,-1.2626776086636606,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.3031420734062201,-6.123838592253551e-2,-1.2626776086636606,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.300451362902075,-0.1831716015188576,-1.2514699928193653,0.3031420734062201,-6.123838592253551e-2,-1.2626776086636606,0.18134421892549893,-0.1831716015188576,-1.274190895690013,0.300451362902075,-0.1831716015188576,-1.2514699928193653,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.18134421892549893,-0.1831716015188576,-1.274190895690013,0.2950938248307055,-0.3034789730126769,-1.2291542407224392,0.300451362902075,-0.1831716015188576,-1.2514699928193653,0.17811055558800623,-0.30347897301267684,-1.2514699928193653,0.2950938248307055,-0.3034789730126769,-1.2291542407224392,0.18134421892549893,-0.1831716015188576,-1.274190895690013,5.9546364444778024e-2,0.30347897301267707,-1.2626776086636606,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.17811055558800623,0.303478973012677,-1.251469992819365,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,0.18134421892549893,0.1831716015188576,-1.274190895690013,5.9546364444778024e-2,0.30347897301267707,-1.2626776086636606,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.18134421892549893,0.1831716015188576,-1.274190895690013,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,0.18134421892549893,-0.1831716015188576,-1.274190895690013,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,0.18134421892549893,-0.1831716015188576,-1.274190895690013,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,0.17811055558800623,-0.30347897301267684,-1.2514699928193653,0.18134421892549893,-0.1831716015188576,-1.274190895690013,5.954636444477803e-2,-0.3034789730126769,-1.2626776086636609,0.17811055558800623,-0.30347897301267684,-1.2514699928193653,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,-5.954636444477759e-2,0.30347897301267707,-1.2626776086636606,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,5.9546364444778024e-2,0.30347897301267707,-1.2626776086636606,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,-5.954636444477759e-2,0.30347897301267707,-1.2626776086636606,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,5.954636444477803e-2,-0.3034789730126769,-1.2626776086636609,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,-5.95463644447776e-2,-0.3034789730126769,-1.2626776086636609,5.954636444477803e-2,-0.3034789730126769,-1.2626776086636609,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-0.17811055558800643,0.30347897301267707,-1.2514699928193653,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-5.954636444477759e-2,0.30347897301267707,-1.2626776086636606,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-0.17811055558800643,0.30347897301267707,-1.2514699928193653,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-5.95463644447776e-2,-0.3034789730126769,-1.2626776086636609,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-0.17811055558800643,-0.3034789730126769,-1.2514699928193656,-5.95463644447776e-2,-0.3034789730126769,-1.2626776086636609,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.2950938248307048,0.30347897301267707,-1.2291542407224392,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.17811055558800643,0.30347897301267707,-1.2514699928193653,-0.30045136290207436,0.1831716015188576,-1.2514699928193656,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.2950938248307048,0.30347897301267707,-1.2291542407224392,-0.30045136290207436,0.1831716015188576,-1.2514699928193656,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.3031420734062194,6.123838592253551e-2,-1.2626776086636609,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.30045136290207436,0.1831716015188576,-1.2514699928193656,-0.3031420734062194,6.123838592253551e-2,-1.2626776086636609,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.3031420734062194,-6.123838592253551e-2,-1.2626776086636609,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.3031420734062194,6.123838592253551e-2,-1.2626776086636609,-0.3031420734062194,-6.123838592253551e-2,-1.2626776086636609,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.30045136290207436,-0.1831716015188576,-1.2514699928193656,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.3031420734062194,-6.123838592253551e-2,-1.2626776086636609,-0.30045136290207436,-0.1831716015188576,-1.2514699928193656,-0.17811055558800643,-0.3034789730126769,-1.2514699928193656,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.29509382483070484,-0.3034789730126769,-1.2291542407224392,-0.17811055558800643,-0.3034789730126769,-1.2514699928193656,-0.30045136290207436,-0.1831716015188576,-1.2514699928193656]]];
      case 2:
        return [[0.8258843617301825,0.6784329353268667,-1.051020824395645,0.9870733136704201,0.6826289037478287,-0.8981676086386854,0.988654609963794,0.6784329353268667,-0.8996064775562205,0.8245634071047778,0.6826289037478287,-1.0493397769225172,0.9870733136704201,0.6826289037478287,-0.8981676086386854,0.8258843617301825,0.6784329353268667,-1.051020824395645,0.6402703770698974,0.6784329353268667,-1.1733642215912938,0.8245634071047778,0.6826289037478287,-1.0493397769225172,0.8258843617301825,0.6784329353268667,-1.051020824395645,0.6392463013574958,0.6826289037478287,-1.171487492877666,0.8245634071047778,0.6826289037478287,-1.0493397769225172,0.6402703770698974,0.6784329353268667,-1.1733642215912938,0.4369466882266312,0.6784329353268667,-1.263252684110171,0.6392463013574958,0.6826289037478287,-1.171487492877666,0.6402703770698974,0.6784329353268667,-1.1733642215912938,0.43624781708242033,0.6826289037478287,-1.2612321839609317,0.6392463013574958,0.6826289037478287,-1.171487492877666,0.4369466882266312,0.6784329353268667,-1.263252684110171,0.22153717303662712,0.6784329353268667,-1.3181999215698321,0.43624781708242033,0.6826289037478287,-1.2612321839609317,0.4369466882266312,0.6784329353268667,-1.263252684110171,0.22118283704604272,0.6826289037478287,-1.316091536468608,0.43624781708242033,0.6826289037478287,-1.2612321839609317,0.22153717303662712,0.6784329353268667,-1.3181999215698321,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,0.22118283704604272,0.6826289037478287,-1.316091536468608,0.22153717303662712,0.6784329353268667,-1.3181999215698321,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,0.22118283704604272,0.6826289037478287,-1.316091536468608,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,-0.22153717303662665,0.6784329353268667,-1.3181999215698321,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,-0.2211828370460423,0.6826289037478287,-1.316091536468608,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,-0.22153717303662665,0.6784329353268667,-1.3181999215698321,-0.4369466882266308,0.6784329353268667,-1.2632526841101712,-0.2211828370460423,0.6826289037478287,-1.316091536468608,-0.22153717303662665,0.6784329353268667,-1.3181999215698321,-0.43624781708241983,0.6826289037478287,-1.2612321839609317,-0.2211828370460423,0.6826289037478287,-1.316091536468608,-0.4369466882266308,0.6784329353268667,-1.2632526841101712,-0.6402703770698972,0.6784329353268667,-1.173364221591294,-0.43624781708241983,0.6826289037478287,-1.2612321839609317,-0.4369466882266308,0.6784329353268667,-1.2632526841101712,-0.6392463013574957,0.6826289037478287,-1.171487492877666,-0.43624781708241983,0.6826289037478287,-1.2612321839609317,-0.6402703770698972,0.6784329353268667,-1.173364221591294,-0.8258843617301822,0.6784329353268667,-1.051020824395645,-0.6392463013574957,0.6826289037478287,-1.171487492877666,-0.6402703770698972,0.6784329353268667,-1.173364221591294,-0.8245634071047775,0.6826289037478287,-1.0493397769225175,-0.6392463013574957,0.6826289037478287,-1.171487492877666,-0.8258843617301822,0.6784329353268667,-1.051020824395645,-0.988654609963794,0.6784329353268667,-0.8996064775562205,-0.8245634071047775,0.6826289037478287,-1.0493397769225175,-0.8258843617301822,0.6784329353268667,-1.051020824395645,-0.9870733136704201,0.6826289037478287,-0.8981676086386854,-0.8245634071047775,0.6826289037478287,-1.0493397769225175,-0.988654609963794,0.6784329353268667,-0.8996064775562205,0.8898014872887915,0.41594602267436875,-1.1323617882284922,1.0641971512155421,0.42046828134539266,-0.96834490122434,1.065168906356051,0.41594602267436875,-0.969229130367855,0.8889897200994323,0.42046828134539266,-1.131328732924248,1.0641971512155421,0.42046828134539266,-0.96834490122434,0.8898014872887915,0.41594602267436875,-1.1323617882284922,0.6898224015166375,0.41594602267436875,-1.2641736275477322,0.8889897200994323,0.42046828134539266,-1.131328732924248,0.8898014872887915,0.41594602267436875,-1.1323617882284922,0.6891930755377136,0.42046828134539266,-1.2630203201110104,0.8889897200994323,0.42046828134539266,-1.131328732924248,0.6898224015166375,0.41594602267436875,-1.2641736275477322,0.4707630160661504,0.41594602267436875,-1.3610187687631923,0.6891930755377136,0.42046828134539266,-1.2630203201110104,0.6898224015166375,0.41594602267436875,-1.2641736275477322,0.4703335382827735,0.42046828134539266,-1.3597771093634643,0.6891930755377136,0.42046828134539266,-1.2630203201110104,0.4707630160661504,0.41594602267436875,-1.3610187687631923,0.2386824538544124,0.41594602267436875,-1.4202185016550835,0.4703335382827735,0.42046828134539266,-1.3597771093634643,0.4707630160661504,0.41594602267436875,-1.3610187687631923,0.2384647035050563,0.42046828134539266,-1.4189228342530462,0.4703335382827735,0.42046828134539266,-1.3597771093634643,0.2386824538544124,0.41594602267436875,-1.4202185016550835,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,0.2384647035050563,0.42046828134539266,-1.4189228342530462,0.2386824538544124,0.41594602267436875,-1.4202185016550835,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,0.2384647035050563,0.42046828134539266,-1.4189228342530462,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,-0.23868245385441195,0.41594602267436875,-1.4202185016550835,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,-0.2384647035050558,0.42046828134539266,-1.4189228342530462,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,-0.23868245385441195,0.41594602267436875,-1.4202185016550835,-0.4707630160661499,0.41594602267436875,-1.3610187687631923,-0.2384647035050558,0.42046828134539266,-1.4189228342530462,-0.23868245385441195,0.41594602267436875,-1.4202185016550835,-0.470333538282773,0.42046828134539266,-1.3597771093634645,-0.2384647035050558,0.42046828134539266,-1.4189228342530462,-0.4707630160661499,0.41594602267436875,-1.3610187687631923,-0.6898224015166374,0.41594602267436875,-1.2641736275477324,-0.470333538282773,0.42046828134539266,-1.3597771093634645,-0.4707630160661499,0.41594602267436875,-1.3610187687631923,-0.6891930755377134,0.42046828134539266,-1.2630203201110106,-0.470333538282773,0.42046828134539266,-1.3597771093634645,-0.6898224015166374,0.41594602267436875,-1.2641736275477324,-0.889801487288791,0.41594602267436875,-1.1323617882284924,-0.6891930755377134,0.42046828134539266,-1.2630203201110106,-0.6898224015166374,0.41594602267436875,-1.2641736275477324,-0.8889897200994319,0.42046828134539266,-1.1313287329242485,-0.6891930755377134,0.42046828134539266,-1.2630203201110106,-0.889801487288791,0.41594602267436875,-1.1323617882284924,-1.065168906356051,0.41594602267436875,-0.969229130367855,-0.8889897200994319,0.42046828134539266,-1.1313287329242485,-0.889801487288791,0.41594602267436875,-1.1323617882284924,-1.0641971512155421,0.42046828134539266,-0.96834490122434,-0.8889897200994319,0.42046828134539266,-1.1313287329242485,-1.065168906356051,0.41594602267436875,-0.969229130367855,0.9221969515562278,0.13872401480181032,-1.1735882711827297,1.1036212741312148,0.14341236045493638,-1.0042180929229159,1.103949063208545,0.13872401480181032,-1.0045163580341756,0.9219231290602979,0.14341236045493638,-1.1732398045465442,1.1036212741312148,0.14341236045493638,-1.0042180929229159,0.9221969515562278,0.13872401480181032,-1.1735882711827297,0.7149371234837818,0.13872401480181032,-1.3101990525038563,0.9219231290602979,0.14341236045493638,-1.1732398045465442,0.9221969515562278,0.13872401480181032,-1.1735882711827297,0.7147248414248843,0.14341236045493638,-1.3098100228349596,0.9219231290602979,0.14341236045493638,-1.1732398045465442,0.7149371234837818,0.13872401480181032,-1.3101990525038563,0.4879023293660975,0.13872401480181032,-1.410570085006911,0.7147248414248843,0.14341236045493638,-1.3098100228349596,0.7149371234837818,0.13872401480181032,-1.3101990525038563,0.4877574594081438,0.14341236045493638,-1.4101512527599502,0.7147248414248843,0.14341236045493638,-1.3098100228349596,0.4879023293660975,0.13872401480181032,-1.410570085006911,0.24737228975102862,0.13872401480181032,-1.471925133279747,0.4877574594081438,0.14341236045493638,-1.4101512527599502,0.4879023293660975,0.13872401480181032,-1.410570085006911,0.24729883895758453,0.14341236045493638,-1.471488083240559,0.4877574594081438,0.14341236045493638,-1.4101512527599502,0.24737228975102862,0.13872401480181032,-1.471925133279747,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,0.24729883895758453,0.14341236045493638,-1.471488083240559,0.24737228975102862,0.13872401480181032,-1.471925133279747,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,0.24729883895758453,0.14341236045493638,-1.471488083240559,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,-0.24737228975102812,0.13872401480181032,-1.471925133279747,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,-0.24729883895758406,0.14341236045493638,-1.471488083240559,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,-0.24737228975102812,0.13872401480181032,-1.471925133279747,-0.487902329366097,0.13872401480181032,-1.410570085006911,-0.24729883895758406,0.14341236045493638,-1.471488083240559,-0.24737228975102812,0.13872401480181032,-1.471925133279747,-0.4877574594081433,0.14341236045493638,-1.4101512527599505,-0.24729883895758406,0.14341236045493638,-1.471488083240559,-0.487902329366097,0.13872401480181032,-1.410570085006911,-0.7149371234837816,0.13872401480181032,-1.3101990525038565,-0.4877574594081433,0.14341236045493638,-1.4101512527599505,-0.487902329366097,0.13872401480181032,-1.410570085006911,-0.7147248414248842,0.14341236045493638,-1.3098100228349598,-0.4877574594081433,0.14341236045493638,-1.4101512527599505,-0.7149371234837816,0.13872401480181032,-1.3101990525038565,-0.9221969515562274,0.13872401480181032,-1.17358827118273,-0.7147248414248842,0.14341236045493638,-1.3098100228349598,-0.7149371234837816,0.13872401480181032,-1.3101990525038565,-0.9219231290602975,0.14341236045493638,-1.1732398045465444,-0.7147248414248842,0.14341236045493638,-1.3098100228349598,-0.9221969515562274,0.13872401480181032,-1.17358827118273,-1.103949063208545,0.13872401480181032,-1.0045163580341756,-0.9219231290602975,0.14341236045493638,-1.1732398045465444,-0.9221969515562274,0.13872401480181032,-1.17358827118273,-1.1036212741312148,0.14341236045493638,-1.0042180929229159,-0.9219231290602975,0.14341236045493638,-1.1732398045465444,-1.103949063208545,0.13872401480181032,-1.0045163580341756,0.9219231290602979,-0.14341236045493638,-1.1732398045465442,1.103949063208545,-0.13872401480181032,-1.0045163580341756,1.1036212741312148,-0.14341236045493638,-1.0042180929229159,0.9221969515562278,-0.13872401480181032,-1.1735882711827297,1.103949063208545,-0.13872401480181032,-1.0045163580341756,0.9219231290602979,-0.14341236045493638,-1.1732398045465442,0.7147248414248843,-0.14341236045493638,-1.3098100228349596,0.9221969515562278,-0.13872401480181032,-1.1735882711827297,0.9219231290602979,-0.14341236045493638,-1.1732398045465442,0.7149371234837818,-0.13872401480181032,-1.3101990525038563,0.9221969515562278,-0.13872401480181032,-1.1735882711827297,0.7147248414248843,-0.14341236045493638,-1.3098100228349596,0.4877574594081438,-0.14341236045493638,-1.4101512527599502,0.7149371234837818,-0.13872401480181032,-1.3101990525038563,0.7147248414248843,-0.14341236045493638,-1.3098100228349596,0.4879023293660975,-0.13872401480181032,-1.410570085006911,0.7149371234837818,-0.13872401480181032,-1.3101990525038563,0.4877574594081438,-0.14341236045493638,-1.4101512527599502,0.24729883895758453,-0.14341236045493638,-1.471488083240559,0.4879023293660975,-0.13872401480181032,-1.410570085006911,0.4877574594081438,-0.14341236045493638,-1.4101512527599502,0.24737228975102862,-0.13872401480181032,-1.471925133279747,0.4879023293660975,-0.13872401480181032,-1.410570085006911,0.24729883895758453,-0.14341236045493638,-1.471488083240559,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,0.24737228975102862,-0.13872401480181032,-1.471925133279747,0.24729883895758453,-0.14341236045493638,-1.471488083240559,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,0.24737228975102862,-0.13872401480181032,-1.471925133279747,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,-0.24729883895758406,-0.14341236045493638,-1.471488083240559,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,-0.24737228975102812,-0.13872401480181032,-1.471925133279747,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,-0.24729883895758406,-0.14341236045493638,-1.471488083240559,-0.4877574594081433,-0.14341236045493638,-1.4101512527599505,-0.24737228975102812,-0.13872401480181032,-1.471925133279747,-0.24729883895758406,-0.14341236045493638,-1.471488083240559,-0.487902329366097,-0.13872401480181032,-1.410570085006911,-0.24737228975102812,-0.13872401480181032,-1.471925133279747,-0.4877574594081433,-0.14341236045493638,-1.4101512527599505,-0.7147248414248842,-0.14341236045493638,-1.3098100228349598,-0.487902329366097,-0.13872401480181032,-1.410570085006911,-0.4877574594081433,-0.14341236045493638,-1.4101512527599505,-0.7149371234837816,-0.13872401480181032,-1.3101990525038565,-0.487902329366097,-0.13872401480181032,-1.410570085006911,-0.7147248414248842,-0.14341236045493638,-1.3098100228349598,-0.9219231290602975,-0.14341236045493638,-1.1732398045465444,-0.7149371234837816,-0.13872401480181032,-1.3101990525038565,-0.7147248414248842,-0.14341236045493638,-1.3098100228349598,-0.9221969515562274,-0.13872401480181032,-1.17358827118273,-0.7149371234837816,-0.13872401480181032,-1.3101990525038565,-0.9219231290602975,-0.14341236045493638,-1.1732398045465444,-1.1036212741312148,-0.14341236045493638,-1.0042180929229159,-0.9221969515562274,-0.13872401480181032,-1.17358827118273,-0.9219231290602975,-0.14341236045493638,-1.1732398045465444,-1.103949063208545,-0.13872401480181032,-1.0045163580341756,-0.9221969515562274,-0.13872401480181032,-1.17358827118273,-1.1036212741312148,-0.14341236045493638,-1.0042180929229159,0.8889897200994323,-0.42046828134539266,-1.131328732924248,1.065168906356051,-0.41594602267436875,-0.969229130367855,1.0641971512155421,-0.42046828134539266,-0.96834490122434,0.8898014872887915,-0.41594602267436875,-1.1323617882284922,1.065168906356051,-0.41594602267436875,-0.969229130367855,0.8889897200994323,-0.42046828134539266,-1.131328732924248,0.6891930755377136,-0.42046828134539266,-1.2630203201110104,0.8898014872887915,-0.41594602267436875,-1.1323617882284922,0.8889897200994323,-0.42046828134539266,-1.131328732924248,0.6898224015166375,-0.41594602267436875,-1.2641736275477322,0.8898014872887915,-0.41594602267436875,-1.1323617882284922,0.6891930755377136,-0.42046828134539266,-1.2630203201110104,0.4703335382827735,-0.42046828134539266,-1.3597771093634643,0.6898224015166375,-0.41594602267436875,-1.2641736275477322,0.6891930755377136,-0.42046828134539266,-1.2630203201110104,0.4707630160661504,-0.41594602267436875,-1.3610187687631923,0.6898224015166375,-0.41594602267436875,-1.2641736275477322,0.4703335382827735,-0.42046828134539266,-1.3597771093634643,0.2384647035050563,-0.42046828134539266,-1.4189228342530462,0.4707630160661504,-0.41594602267436875,-1.3610187687631923,0.4703335382827735,-0.42046828134539266,-1.3597771093634643,0.2386824538544124,-0.41594602267436875,-1.4202185016550835,0.4707630160661504,-0.41594602267436875,-1.3610187687631923,0.2384647035050563,-0.42046828134539266,-1.4189228342530462,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,0.2386824538544124,-0.41594602267436875,-1.4202185016550835,0.2384647035050563,-0.42046828134539266,-1.4189228342530462,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,0.2386824538544124,-0.41594602267436875,-1.4202185016550835,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,-0.2384647035050558,-0.42046828134539266,-1.4189228342530462,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,-0.23868245385441195,-0.41594602267436875,-1.4202185016550835,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,-0.2384647035050558,-0.42046828134539266,-1.4189228342530462,-0.470333538282773,-0.42046828134539266,-1.3597771093634645,-0.23868245385441195,-0.41594602267436875,-1.4202185016550835,-0.2384647035050558,-0.42046828134539266,-1.4189228342530462,-0.4707630160661499,-0.41594602267436875,-1.3610187687631923,-0.23868245385441195,-0.41594602267436875,-1.4202185016550835,-0.470333538282773,-0.42046828134539266,-1.3597771093634645,-0.6891930755377134,-0.42046828134539266,-1.2630203201110106,-0.4707630160661499,-0.41594602267436875,-1.3610187687631923,-0.470333538282773,-0.42046828134539266,-1.3597771093634645,-0.6898224015166374,-0.41594602267436875,-1.2641736275477324,-0.4707630160661499,-0.41594602267436875,-1.3610187687631923,-0.6891930755377134,-0.42046828134539266,-1.2630203201110106,-0.8889897200994319,-0.42046828134539266,-1.1313287329242485,-0.6898224015166374,-0.41594602267436875,-1.2641736275477324,-0.6891930755377134,-0.42046828134539266,-1.2630203201110106,-0.889801487288791,-0.41594602267436875,-1.1323617882284924,-0.6898224015166374,-0.41594602267436875,-1.2641736275477324,-0.8889897200994319,-0.42046828134539266,-1.1313287329242485,-1.0641971512155421,-0.42046828134539266,-0.96834490122434,-0.889801487288791,-0.41594602267436875,-1.1323617882284924,-0.8889897200994319,-0.42046828134539266,-1.1313287329242485,-1.065168906356051,-0.41594602267436875,-0.969229130367855,-0.889801487288791,-0.41594602267436875,-1.1323617882284924,-1.0641971512155421,-0.42046828134539266,-0.96834490122434,0.8245634071047778,-0.6826289037478287,-1.0493397769225172,0.988654609963794,-0.6784329353268667,-0.8996064775562205,0.9870733136704201,-0.6826289037478287,-0.8981676086386854,0.8258843617301825,-0.6784329353268667,-1.051020824395645,0.988654609963794,-0.6784329353268667,-0.8996064775562205,0.8245634071047778,-0.6826289037478287,-1.0493397769225172,0.6392463013574958,-0.6826289037478287,-1.171487492877666,0.8258843617301825,-0.6784329353268667,-1.051020824395645,0.8245634071047778,-0.6826289037478287,-1.0493397769225172,0.6402703770698974,-0.6784329353268667,-1.1733642215912938,0.8258843617301825,-0.6784329353268667,-1.051020824395645,0.6392463013574958,-0.6826289037478287,-1.171487492877666,0.43624781708242033,-0.6826289037478287,-1.2612321839609317,0.6402703770698974,-0.6784329353268667,-1.1733642215912938,0.6392463013574958,-0.6826289037478287,-1.171487492877666,0.4369466882266312,-0.6784329353268667,-1.263252684110171,0.6402703770698974,-0.6784329353268667,-1.1733642215912938,0.43624781708242033,-0.6826289037478287,-1.2612321839609317,0.22118283704604272,-0.6826289037478287,-1.316091536468608,0.4369466882266312,-0.6784329353268667,-1.263252684110171,0.43624781708242033,-0.6826289037478287,-1.2612321839609317,0.22153717303662712,-0.6784329353268667,-1.3181999215698321,0.4369466882266312,-0.6784329353268667,-1.263252684110171,0.22118283704604272,-0.6826289037478287,-1.316091536468608,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,0.22153717303662712,-0.6784329353268667,-1.3181999215698321,0.22118283704604272,-0.6826289037478287,-1.316091536468608,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,0.22153717303662712,-0.6784329353268667,-1.3181999215698321,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,-0.2211828370460423,-0.6826289037478287,-1.316091536468608,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,-0.22153717303662665,-0.6784329353268667,-1.3181999215698321,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,-0.2211828370460423,-0.6826289037478287,-1.316091536468608,-0.43624781708241983,-0.6826289037478287,-1.2612321839609317,-0.22153717303662665,-0.6784329353268667,-1.3181999215698321,-0.2211828370460423,-0.6826289037478287,-1.316091536468608,-0.4369466882266308,-0.6784329353268667,-1.2632526841101712,-0.22153717303662665,-0.6784329353268667,-1.3181999215698321,-0.43624781708241983,-0.6826289037478287,-1.2612321839609317,-0.6392463013574957,-0.6826289037478287,-1.171487492877666,-0.4369466882266308,-0.6784329353268667,-1.2632526841101712,-0.43624781708241983,-0.6826289037478287,-1.2612321839609317,-0.6402703770698972,-0.6784329353268667,-1.173364221591294,-0.4369466882266308,-0.6784329353268667,-1.2632526841101712,-0.6392463013574957,-0.6826289037478287,-1.171487492877666,-0.8245634071047775,-0.6826289037478287,-1.0493397769225175,-0.6402703770698972,-0.6784329353268667,-1.173364221591294,-0.6392463013574957,-0.6826289037478287,-1.171487492877666,-0.8258843617301822,-0.6784329353268667,-1.051020824395645,-0.6402703770698972,-0.6784329353268667,-1.173364221591294,-0.8245634071047775,-0.6826289037478287,-1.0493397769225175,-0.9870733136704201,-0.6826289037478287,-0.8981676086386854,-0.8258843617301822,-0.6784329353268667,-1.051020824395645,-0.8245634071047775,-0.6826289037478287,-1.0493397769225175,-0.988654609963794,-0.6784329353268667,-0.8996064775562205,-0.8258843617301822,-0.6784329353268667,-1.051020824395645,-0.9870733136704201,-0.6826289037478287,-0.8981676086386854,0.9892759314439068,0.6805317591095806,-0.897335308738607,1.063161260448333,0.41820766795280473,-0.9704594173624104,0.9864519921903073,0.6805317591095806,-0.9004387774562989,1.0662047971232602,0.41820766795280473,-0.9671146142297845,1.063161260448333,0.41820766795280473,-0.9704594173624104,0.9892759314439068,0.6805317591095806,-0.897335308738607,1.0662047971232602,0.41820766795280473,-0.9671146142297845,1.1022075110237735,0.14106836166445297,-1.0061010485930755,1.063161260448333,0.41820766795280473,-0.9704594173624104,1.1053628263159865,0.14106836166445297,-1.0026334023640155,1.1022075110237735,0.14106836166445297,-1.0061010485930755,1.0662047971232602,0.41820766795280473,-0.9671146142297845,1.1053628263159865,0.14106836166445297,-1.0026334023640155,1.1022075110237735,-0.14106836166445297,-1.0061010485930755,1.1022075110237735,0.14106836166445297,-1.0061010485930755,1.1053628263159865,-0.14106836166445297,-1.0026334023640155,1.1022075110237735,-0.14106836166445297,-1.0061010485930755,1.1053628263159865,0.14106836166445297,-1.0026334023640155,1.1053628263159865,-0.14106836166445297,-1.0026334023640155,1.063161260448333,-0.41820766795280473,-0.9704594173624104,1.1022075110237735,-0.14106836166445297,-1.0061010485930755,1.0662047971232602,-0.41820766795280473,-0.9671146142297845,1.063161260448333,-0.41820766795280473,-0.9704594173624104,1.1053628263159865,-0.14106836166445297,-1.0026334023640155,1.0662047971232602,-0.41820766795280473,-0.9671146142297845,0.9864519921903073,-0.6805317591095806,-0.9004387774562989,1.063161260448333,-0.41820766795280473,-0.9704594173624104,0.9892759314439068,-0.6805317591095806,-0.897335308738607,0.9864519921903073,-0.6805317591095806,-0.9004387774562989,1.0662047971232602,-0.41820766795280473,-0.9671146142297845,0.826873505132987,0.6805317591095806,-1.0488840409465234,0.8876177038540342,0.41820766795280473,-1.133242321072757,0.8235742637019738,0.6805317591095806,-1.0514765603716383,0.8911735035341899,0.41820766795280473,-1.1304482000799834,0.8876177038540342,0.41820766795280473,-1.133242321072757,0.826873505132987,0.6805317591095806,-1.0488840409465234,0.8911735035341899,0.41820766795280473,-1.1304482000799834,0.9202168443318074,0.14106836166445297,-1.1748624075802738,0.8876177038540342,0.41820766795280473,-1.133242321072757,0.9239032362847184,0.14106836166445297,-1.1719656681489996,0.9202168443318074,0.14106836166445297,-1.1748624075802738,0.8911735035341899,0.41820766795280473,-1.1304482000799834,0.9239032362847184,0.14106836166445297,-1.1719656681489996,0.9202168443318074,-0.14106836166445297,-1.1748624075802738,0.9202168443318074,0.14106836166445297,-1.1748624075802738,0.9239032362847184,-0.14106836166445297,-1.1719656681489996,0.9202168443318074,-0.14106836166445297,-1.1748624075802738,0.9239032362847184,0.14106836166445297,-1.1719656681489996,0.9239032362847184,-0.14106836166445297,-1.1719656681489996,0.8876177038540342,-0.41820766795280473,-1.133242321072757,0.9202168443318074,-0.14106836166445297,-1.1748624075802738,0.8911735035341899,-0.41820766795280473,-1.1304482000799834,0.8876177038540342,-0.41820766795280473,-1.133242321072757,0.9239032362847184,-0.14106836166445297,-1.1719656681489996,0.8911735035341899,-0.41820766795280473,-1.1304482000799834,0.8235742637019738,-0.6805317591095806,-1.0514765603716383,0.8876177038540342,-0.41820766795280473,-1.133242321072757,0.826873505132987,-0.6805317591095806,-1.0488840409465234,0.8235742637019738,-0.6805317591095806,-1.0514765603716383,0.8911735035341899,-0.41820766795280473,-1.1304482000799834,0.6415999829583717,0.6805317591095806,-1.1714209263586841,0.6875228834096573,0.41820766795280473,-1.2646800509431417,0.637916695469022,0.6805317591095806,-1.173430788110276,0.6914925936446944,0.41820766795280473,-1.2625138967156009,0.6875228834096573,0.41820766795280473,-1.2646800509431417,0.6415999829583717,0.6805317591095806,-1.1714209263586841,0.6914925936446944,0.41820766795280473,-1.2625138967156009,0.712773230446044,0.14106836166445297,-1.3111273924744375,0.6875228834096573,0.41820766795280473,-1.2646800509431417,0.7168887344626226,0.14106836166445297,-1.3088816828643783,0.712773230446044,0.14106836166445297,-1.3111273924744375,0.6914925936446944,0.41820766795280473,-1.2625138967156009,0.7168887344626226,0.14106836166445297,-1.3088816828643783,0.712773230446044,-0.14106836166445297,-1.3111273924744375,0.712773230446044,0.14106836166445297,-1.3111273924744375,0.7168887344626226,-0.14106836166445297,-1.3088816828643783,0.712773230446044,-0.14106836166445297,-1.3111273924744375,0.7168887344626226,0.14106836166445297,-1.3088816828643783,0.7168887344626226,-0.14106836166445297,-1.3088816828643783,0.6875228834096573,-0.41820766795280473,-1.2646800509431417,0.712773230446044,-0.14106836166445297,-1.3111273924744375,0.6914925936446944,-0.41820766795280473,-1.2625138967156009,0.6875228834096573,-0.41820766795280473,-1.2646800509431417,0.7168887344626226,-0.14106836166445297,-1.3088816828643783,0.6914925936446944,-0.41820766795280473,-1.2625138967156009,0.637916695469022,-0.6805317591095806,-1.173430788110276,0.6875228834096573,-0.41820766795280473,-1.2646800509431417,0.6415999829583717,-0.6805317591095806,-1.1714209263586841,0.637916695469022,-0.6805317591095806,-1.173430788110276,0.6914925936446944,-0.41820766795280473,-1.2625138967156009,0.43857998006416155,0.6805317591095806,-1.26155662811074,0.46841136733126354,0.41820766795280473,-1.361137075176609,0.43461452524488964,0.6805317591095806,-1.2629282399603627,0.47268518701766,0.41820766795280473,-1.3596588029500476,0.46841136733126354,0.41820766795280473,-1.361137075176609,0.43857998006416155,0.6805317591095806,-1.26155662811074,0.47268518701766,0.41820766795280473,-1.3596588029500476,0.48561450320689586,0.14106836166445297,-1.4111269509198763,0.46841136733126354,0.41820766795280473,-1.361137075176609,0.4900452855673451,0.14106836166445297,-1.4095943868469845,0.48561450320689586,0.14106836166445297,-1.4111269509198763,0.47268518701766,0.41820766795280473,-1.3596588029500476,0.4900452855673451,0.14106836166445297,-1.4095943868469845,0.48561450320689586,-0.14106836166445297,-1.4111269509198763,0.48561450320689586,0.14106836166445297,-1.4111269509198763,0.4900452855673451,-0.14106836166445297,-1.4095943868469845,0.48561450320689586,-0.14106836166445297,-1.4111269509198763,0.4900452855673451,0.14106836166445297,-1.4095943868469845,0.4900452855673451,-0.14106836166445297,-1.4095943868469845,0.46841136733126354,-0.41820766795280473,-1.361137075176609,0.48561450320689586,-0.14106836166445297,-1.4111269509198763,0.47268518701766,-0.41820766795280473,-1.3596588029500476,0.46841136733126354,-0.41820766795280473,-1.361137075176609,0.4900452855673451,-0.14106836166445297,-1.4095943868469845,0.47268518701766,-0.41820766795280473,-1.3596588029500476,0.43461452524488964,-0.6805317591095806,-1.2629282399603627,0.46841136733126354,-0.41820766795280473,-1.361137075176609,0.43857998006416155,-0.6805317591095806,-1.26155662811074,0.43461452524488964,-0.6805317591095806,-1.2629282399603627,0.47268518701766,-0.41820766795280473,-1.3596588029500476,0.22342897441599144,0.6805317591095806,-1.3167980172504203,0.23634372045490132,0.41820766795280473,-1.4199454187633456,0.21929103566667868,0.6805317591095806,-1.3174934407880199,0.2408034369045677,0.41820766795280473,-1.419195917144784,0.23634372045490132,0.41820766795280473,-1.4199454187633456,0.22342897441599144,0.6805317591095806,-1.3167980172504203,0.2408034369045677,0.41820766795280473,-1.419195917144784,0.24502381111859076,0.14106836166445297,-1.4720951223756664,0.23634372045490132,0.41820766795280473,-1.4199454187633456,0.2496473175900227,0.14106836166445297,-1.4713180941446393,0.24502381111859076,0.14106836166445297,-1.4720951223756664,0.2408034369045677,0.41820766795280473,-1.419195917144784,0.2496473175900227,0.14106836166445297,-1.4713180941446393,0.24502381111859076,-0.14106836166445297,-1.4720951223756664,0.24502381111859076,0.14106836166445297,-1.4720951223756664,0.2496473175900227,-0.14106836166445297,-1.4713180941446393,0.24502381111859076,-0.14106836166445297,-1.4720951223756664,0.2496473175900227,0.14106836166445297,-1.4713180941446393,0.2496473175900227,-0.14106836166445297,-1.4713180941446393,0.23634372045490132,-0.41820766795280473,-1.4199454187633456,0.24502381111859076,-0.14106836166445297,-1.4720951223756664,0.2408034369045677,-0.41820766795280473,-1.419195917144784,0.23634372045490132,-0.41820766795280473,-1.4199454187633456,0.2496473175900227,-0.14106836166445297,-1.4713180941446393,0.2408034369045677,-0.41820766795280473,-1.419195917144784,0.21929103566667868,-0.6805317591095806,-1.3174934407880199,0.23634372045490132,-0.41820766795280473,-1.4199454187633456,0.22342897441599144,-0.6805317591095806,-1.3167980172504203,0.21929103566667868,-0.6805317591095806,-1.3174934407880199,0.2408034369045677,-0.41820766795280473,-1.419195917144784,2.0979842104809974e-3,0.6805317591095806,-1.3356171320050791,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,-2.097984210480834e-3,0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,2.0979842104809974e-3,0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-2.097984210480834e-3,-0.6805317591095806,-1.3356171320050791,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,2.0979842104809974e-3,-0.6805317591095806,-1.3356171320050791,-2.097984210480834e-3,-0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-0.21929103566667826,0.6805317591095806,-1.3174934407880199,-0.2408034369045672,0.41820766795280473,-1.419195917144784,-0.22342897441599097,0.6805317591095806,-1.3167980172504203,-0.23634372045490082,0.41820766795280473,-1.4199454187633456,-0.2408034369045672,0.41820766795280473,-1.419195917144784,-0.21929103566667826,0.6805317591095806,-1.3174934407880199,-0.23634372045490082,0.41820766795280473,-1.4199454187633456,-0.2496473175900222,0.14106836166445297,-1.4713180941446393,-0.2408034369045672,0.41820766795280473,-1.419195917144784,-0.24502381111859026,0.14106836166445297,-1.4720951223756666,-0.2496473175900222,0.14106836166445297,-1.4713180941446393,-0.23634372045490082,0.41820766795280473,-1.4199454187633456,-0.24502381111859026,0.14106836166445297,-1.4720951223756666,-0.2496473175900222,-0.14106836166445297,-1.4713180941446393,-0.2496473175900222,0.14106836166445297,-1.4713180941446393,-0.24502381111859026,-0.14106836166445297,-1.4720951223756666,-0.2496473175900222,-0.14106836166445297,-1.4713180941446393,-0.24502381111859026,0.14106836166445297,-1.4720951223756666,-0.24502381111859026,-0.14106836166445297,-1.4720951223756666,-0.2408034369045672,-0.41820766795280473,-1.419195917144784,-0.2496473175900222,-0.14106836166445297,-1.4713180941446393,-0.23634372045490082,-0.41820766795280473,-1.4199454187633456,-0.2408034369045672,-0.41820766795280473,-1.419195917144784,-0.24502381111859026,-0.14106836166445297,-1.4720951223756666,-0.23634372045490082,-0.41820766795280473,-1.4199454187633456,-0.22342897441599097,-0.6805317591095806,-1.3167980172504203,-0.2408034369045672,-0.41820766795280473,-1.419195917144784,-0.21929103566667826,-0.6805317591095806,-1.3174934407880199,-0.22342897441599097,-0.6805317591095806,-1.3167980172504203,-0.23634372045490082,-0.41820766795280473,-1.4199454187633456,-0.4346145252448893,0.6805317591095806,-1.262928239960363,-0.4726851870176598,0.41820766795280473,-1.3596588029500478,-0.4385799800641614,0.6805317591095806,-1.26155662811074,-0.4684113673312631,0.41820766795280473,-1.361137075176609,-0.4726851870176598,0.41820766795280473,-1.3596588029500478,-0.4346145252448893,0.6805317591095806,-1.262928239960363,-0.4684113673312631,0.41820766795280473,-1.361137075176609,-0.49004528556734495,0.14106836166445297,-1.4095943868469847,-0.4726851870176598,0.41820766795280473,-1.3596588029500478,-0.4856145032068954,0.14106836166445297,-1.4111269509198765,-0.49004528556734495,0.14106836166445297,-1.4095943868469847,-0.4684113673312631,0.41820766795280473,-1.361137075176609,-0.4856145032068954,0.14106836166445297,-1.4111269509198765,-0.49004528556734495,-0.14106836166445297,-1.4095943868469847,-0.49004528556734495,0.14106836166445297,-1.4095943868469847,-0.4856145032068954,-0.14106836166445297,-1.4111269509198765,-0.49004528556734495,-0.14106836166445297,-1.4095943868469847,-0.4856145032068954,0.14106836166445297,-1.4111269509198765,-0.4856145032068954,-0.14106836166445297,-1.4111269509198765,-0.4726851870176598,-0.41820766795280473,-1.3596588029500478,-0.49004528556734495,-0.14106836166445297,-1.4095943868469847,-0.4684113673312631,-0.41820766795280473,-1.361137075176609,-0.4726851870176598,-0.41820766795280473,-1.3596588029500478,-0.4856145032068954,-0.14106836166445297,-1.4111269509198765,-0.4684113673312631,-0.41820766795280473,-1.361137075176609,-0.4385799800641614,-0.6805317591095806,-1.26155662811074,-0.4726851870176598,-0.41820766795280473,-1.3596588029500478,-0.4346145252448893,-0.6805317591095806,-1.262928239960363,-0.4385799800641614,-0.6805317591095806,-1.26155662811074,-0.4684113673312631,-0.41820766795280473,-1.361137075176609,-0.6379166954690219,0.6805317591095806,-1.173430788110276,-0.6914925936446938,0.41820766795280473,-1.262513896715601,-0.6415999829583712,0.6805317591095806,-1.1714209263586843,-0.6875228834096571,0.41820766795280473,-1.264680050943142,-0.6914925936446938,0.41820766795280473,-1.262513896715601,-0.6379166954690219,0.6805317591095806,-1.173430788110276,-0.6875228834096571,0.41820766795280473,-1.264680050943142,-0.716888734462622,0.14106836166445297,-1.3088816828643786,-0.6914925936446938,0.41820766795280473,-1.262513896715601,-0.7127732304460439,0.14106836166445297,-1.3111273924744378,-0.716888734462622,0.14106836166445297,-1.3088816828643786,-0.6875228834096571,0.41820766795280473,-1.264680050943142,-0.7127732304460439,0.14106836166445297,-1.3111273924744378,-0.716888734462622,-0.14106836166445297,-1.3088816828643786,-0.716888734462622,0.14106836166445297,-1.3088816828643786,-0.7127732304460439,-0.14106836166445297,-1.3111273924744378,-0.716888734462622,-0.14106836166445297,-1.3088816828643786,-0.7127732304460439,0.14106836166445297,-1.3111273924744378,-0.7127732304460439,-0.14106836166445297,-1.3111273924744378,-0.6914925936446938,-0.41820766795280473,-1.262513896715601,-0.716888734462622,-0.14106836166445297,-1.3088816828643786,-0.6875228834096571,-0.41820766795280473,-1.264680050943142,-0.6914925936446938,-0.41820766795280473,-1.262513896715601,-0.7127732304460439,-0.14106836166445297,-1.3111273924744378,-0.6875228834096571,-0.41820766795280473,-1.264680050943142,-0.6415999829583712,-0.6805317591095806,-1.1714209263586843,-0.6914925936446938,-0.41820766795280473,-1.262513896715601,-0.6379166954690219,-0.6805317591095806,-1.173430788110276,-0.6415999829583712,-0.6805317591095806,-1.1714209263586843,-0.6875228834096571,-0.41820766795280473,-1.264680050943142,-0.8235742637019734,0.6805317591095806,-1.0514765603716387,-0.8911735035341898,0.41820766795280473,-1.1304482000799834,-0.8268735051329867,0.6805317591095806,-1.0488840409465234,-0.8876177038540336,0.41820766795280473,-1.133242321072757,-0.8911735035341898,0.41820766795280473,-1.1304482000799834,-0.8235742637019734,0.6805317591095806,-1.0514765603716387,-0.8876177038540336,0.41820766795280473,-1.133242321072757,-0.9239032362847183,0.14106836166445297,-1.1719656681489996,-0.8911735035341898,0.41820766795280473,-1.1304482000799834,-0.9202168443318068,0.14106836166445297,-1.1748624075802743,-0.9239032362847183,0.14106836166445297,-1.1719656681489996,-0.8876177038540336,0.41820766795280473,-1.133242321072757,-0.9202168443318068,0.14106836166445297,-1.1748624075802743,-0.9239032362847183,-0.14106836166445297,-1.1719656681489996,-0.9239032362847183,0.14106836166445297,-1.1719656681489996,-0.9202168443318068,-0.14106836166445297,-1.1748624075802743,-0.9239032362847183,-0.14106836166445297,-1.1719656681489996,-0.9202168443318068,0.14106836166445297,-1.1748624075802743,-0.9202168443318068,-0.14106836166445297,-1.1748624075802743,-0.8911735035341898,-0.41820766795280473,-1.1304482000799834,-0.9239032362847183,-0.14106836166445297,-1.1719656681489996,-0.8876177038540336,-0.41820766795280473,-1.133242321072757,-0.8911735035341898,-0.41820766795280473,-1.1304482000799834,-0.9202168443318068,-0.14106836166445297,-1.1748624075802743,-0.8876177038540336,-0.41820766795280473,-1.133242321072757,-0.8268735051329867,-0.6805317591095806,-1.0488840409465234,-0.8911735035341898,-0.41820766795280473,-1.1304482000799834,-0.8235742637019734,-0.6805317591095806,-1.0514765603716387,-0.8268735051329867,-0.6805317591095806,-1.0488840409465234,-0.8876177038540336,-0.41820766795280473,-1.133242321072757,-0.9864519921903068,0.6805317591095806,-0.9004387774562992,-1.06620479712326,0.41820766795280473,-0.9671146142297845,-0.9892759314439067,0.6805317591095806,-0.897335308738607,-1.0631612604483325,0.41820766795280473,-0.9704594173624107,-1.06620479712326,0.41820766795280473,-0.9671146142297845,-0.9864519921903068,0.6805317591095806,-0.9004387774562992,-1.0631612604483325,0.41820766795280473,-0.9704594173624107,-1.1053628263159865,0.14106836166445297,-1.0026334023640155,-1.06620479712326,0.41820766795280473,-0.9671146142297845,-1.1022075110237728,0.14106836166445297,-1.006101048593076,-1.1053628263159865,0.14106836166445297,-1.0026334023640155,-1.0631612604483325,0.41820766795280473,-0.9704594173624107,-1.1022075110237728,0.14106836166445297,-1.006101048593076,-1.1053628263159865,-0.14106836166445297,-1.0026334023640155,-1.1053628263159865,0.14106836166445297,-1.0026334023640155,-1.1022075110237728,-0.14106836166445297,-1.006101048593076,-1.1053628263159865,-0.14106836166445297,-1.0026334023640155,-1.1022075110237728,0.14106836166445297,-1.006101048593076,-1.1022075110237728,-0.14106836166445297,-1.006101048593076,-1.06620479712326,-0.41820766795280473,-0.9671146142297845,-1.1053628263159865,-0.14106836166445297,-1.0026334023640155,-1.0631612604483325,-0.41820766795280473,-0.9704594173624107,-1.06620479712326,-0.41820766795280473,-0.9671146142297845,-1.1022075110237728,-0.14106836166445297,-1.006101048593076,-1.0631612604483325,-0.41820766795280473,-0.9704594173624107,-0.9892759314439067,-0.6805317591095806,-0.897335308738607,-1.06620479712326,-0.41820766795280473,-0.9671146142297845,-0.9864519921903068,-0.6805317591095806,-0.9004387774562992,-0.9892759314439067,-0.6805317591095806,-0.897335308738607,-1.0631612604483325,-0.41820766795280473,-0.9704594173624107],[0.825775419443983,0.6809857496093201,-1.0508821843913552,1.0653946053574237,0.4184866590588439,-0.9694345006575056,0.9885241966777916,0.6809857496093201,-0.8994878105964301,0.8899900276281709,0.4184866590588439,-1.1326017247524278,1.0653946053574237,0.4184866590588439,-0.9694345006575056,0.825775419443983,0.6809857496093201,-1.0508821843913552,0.8899900276281709,0.4184866590588439,-1.1326017247524278,1.104522878997648,0.14116246997777146,-1.0050384902283702,1.0653946053574237,0.4184866590588439,-0.9694345006575056,0.9226762953856684,0.14116246997777146,-1.1741982843639116,1.104522878997648,0.14116246997777146,-1.0050384902283702,0.8899900276281709,0.4184866590588439,-1.1326017247524278,0.9226762953856684,0.14116246997777146,-1.1741982843639116,1.104522878997648,-0.14116246997777146,-1.0050384902283702,1.104522878997648,0.14116246997777146,-1.0050384902283702,0.9226762953856684,-0.14116246997777146,-1.1741982843639116,1.104522878997648,-0.14116246997777146,-1.0050384902283702,0.9226762953856684,0.14116246997777146,-1.1741982843639116,0.9226762953856684,-0.14116246997777146,-1.1741982843639116,1.0653946053574237,-0.4184866590588439,-0.9694345006575056,1.104522878997648,-0.14116246997777146,-1.0050384902283702,0.8899900276281709,-0.4184866590588439,-1.1326017247524278,1.0653946053574237,-0.4184866590588439,-0.9694345006575056,0.9226762953856684,-0.14116246997777146,-1.1741982843639116,0.8899900276281709,-0.4184866590588439,-1.1326017247524278,0.9885241966777916,-0.6809857496093201,-0.8994878105964301,1.0653946053574237,-0.4184866590588439,-0.9694345006575056,0.825775419443983,-0.6809857496093201,-1.0508821843913552,0.9885241966777916,-0.6809857496093201,-0.8994878105964301,0.8899900276281709,-0.4184866590588439,-1.1326017247524278,0.6401859190974547,0.6809857496093201,-1.1732094432872502,0.8899900276281709,0.4184866590588439,-1.1326017247524278,0.825775419443983,0.6809857496093201,-1.0508821843913552,0.6899685682196062,0.4184866590588439,-1.2644414937270738,0.8899900276281709,0.4184866590588439,-1.1326017247524278,0.6401859190974547,0.6809857496093201,-1.1732094432872502,0.6899685682196062,0.4184866590588439,-1.2644414937270738,0.9226762953856684,0.14116246997777146,-1.1741982843639116,0.8899900276281709,0.4184866590588439,-1.1326017247524278,0.7153087368338386,0.14116246997777146,-1.3108800738736373,0.9226762953856684,0.14116246997777146,-1.1741982843639116,0.6899685682196062,0.4184866590588439,-1.2644414937270738,0.7153087368338386,0.14116246997777146,-1.3108800738736373,0.9226762953856684,-0.14116246997777146,-1.1741982843639116,0.9226762953856684,0.14116246997777146,-1.1741982843639116,0.7153087368338386,-0.14116246997777146,-1.3108800738736373,0.9226762953856684,-0.14116246997777146,-1.1741982843639116,0.7153087368338386,0.14116246997777146,-1.3108800738736373,0.7153087368338386,-0.14116246997777146,-1.3108800738736373,0.8899900276281709,-0.4184866590588439,-1.1326017247524278,0.9226762953856684,-0.14116246997777146,-1.1741982843639116,0.6899685682196062,-0.4184866590588439,-1.2644414937270738,0.8899900276281709,-0.4184866590588439,-1.1326017247524278,0.7153087368338386,-0.14116246997777146,-1.3108800738736373,0.6899685682196062,-0.4184866590588439,-1.2644414937270738,0.825775419443983,-0.6809857496093201,-1.0508821843913552,0.8899900276281709,-0.4184866590588439,-1.1326017247524278,0.6401859190974547,-0.6809857496093201,-1.1732094432872502,0.825775419443983,-0.6809857496093201,-1.0508821843913552,0.6899685682196062,-0.4184866590588439,-1.2644414937270738,0.43688905065245187,0.6809857496093201,-1.263086048632092,0.6899685682196062,0.4184866590588439,-1.2644414937270738,0.6401859190974547,0.6809857496093201,-1.1732094432872502,0.47086276620152817,0.4184866590588439,-1.361307155492401,0.6899685682196062,0.4184866590588439,-1.2644414937270738,0.43688905065245187,0.6809857496093201,-1.263086048632092,0.47086276620152817,0.4184866590588439,-1.361307155492401,0.7153087368338386,0.14116246997777146,-1.3108800738736373,0.6899685682196062,0.4184866590588439,-1.2644414937270738,0.4881559335127569,0.14116246997777146,-1.4113032776997523,0.7153087368338386,0.14116246997777146,-1.3108800738736373,0.47086276620152817,0.4184866590588439,-1.361307155492401,0.4881559335127569,0.14116246997777146,-1.4113032776997523,0.7153087368338386,-0.14116246997777146,-1.3108800738736373,0.7153087368338386,0.14116246997777146,-1.3108800738736373,0.4881559335127569,-0.14116246997777146,-1.4113032776997523,0.7153087368338386,-0.14116246997777146,-1.3108800738736373,0.4881559335127569,0.14116246997777146,-1.4113032776997523,0.4881559335127569,-0.14116246997777146,-1.4113032776997523,0.6899685682196062,-0.4184866590588439,-1.2644414937270738,0.7153087368338386,-0.14116246997777146,-1.3108800738736373,0.47086276620152817,-0.4184866590588439,-1.361307155492401,0.6899685682196062,-0.4184866590588439,-1.2644414937270738,0.4881559335127569,-0.14116246997777146,-1.4113032776997523,0.47086276620152817,-0.4184866590588439,-1.361307155492401,0.6401859190974547,-0.6809857496093201,-1.1732094432872502,0.6899685682196062,-0.4184866590588439,-1.2644414937270738,0.43688905065245187,-0.6809857496093201,-1.263086048632092,0.6401859190974547,-0.6809857496093201,-1.1732094432872502,0.47086276620152817,-0.4184866590588439,-1.361307155492401,0.22150795010030921,0.6809857496093201,-1.318026038009641,0.47086276620152817,0.4184866590588439,-1.361307155492401,0.43688905065245187,0.6809857496093201,-1.263086048632092,0.2387330283606326,0.4184866590588439,-1.420519432235803,0.47086276620152817,0.4184866590588439,-1.361307155492401,0.22150795010030921,0.6809857496093201,-1.318026038009641,0.2387330283606326,0.4184866590588439,-1.420519432235803,0.4881559335127569,0.14116246997777146,-1.4113032776997523,0.47086276620152817,0.4184866590588439,-1.361307155492401,0.24750087007269053,0.14116246997777146,-1.4726902173855287,0.4881559335127569,0.14116246997777146,-1.4113032776997523,0.2387330283606326,0.4184866590588439,-1.420519432235803,0.24750087007269053,0.14116246997777146,-1.4726902173855287,0.4881559335127569,-0.14116246997777146,-1.4113032776997523,0.4881559335127569,0.14116246997777146,-1.4113032776997523,0.24750087007269053,-0.14116246997777146,-1.4726902173855287,0.4881559335127569,-0.14116246997777146,-1.4113032776997523,0.24750087007269053,0.14116246997777146,-1.4726902173855287,0.24750087007269053,-0.14116246997777146,-1.4726902173855287,0.47086276620152817,-0.4184866590588439,-1.361307155492401,0.4881559335127569,-0.14116246997777146,-1.4113032776997523,0.2387330283606326,-0.4184866590588439,-1.420519432235803,0.47086276620152817,-0.4184866590588439,-1.361307155492401,0.24750087007269053,-0.14116246997777146,-1.4726902173855287,0.2387330283606326,-0.4184866590588439,-1.420519432235803,0.43688905065245187,-0.6809857496093201,-1.263086048632092,0.47086276620152817,-0.4184866590588439,-1.361307155492401,0.22150795010030921,-0.6809857496093201,-1.318026038009641,0.43688905065245187,-0.6809857496093201,-1.263086048632092,0.2387330283606326,-0.4184866590588439,-1.420519432235803,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,0.2387330283606326,0.4184866590588439,-1.420519432235803,0.22150795010030921,0.6809857496093201,-1.318026038009641,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,0.2387330283606326,0.4184866590588439,-1.420519432235803,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,0.24750087007269053,0.14116246997777146,-1.4726902173855287,0.2387330283606326,0.4184866590588439,-1.420519432235803,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,0.24750087007269053,0.14116246997777146,-1.4726902173855287,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,0.24750087007269053,-0.14116246997777146,-1.4726902173855287,0.24750087007269053,0.14116246997777146,-1.4726902173855287,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,0.24750087007269053,-0.14116246997777146,-1.4726902173855287,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,0.2387330283606326,-0.4184866590588439,-1.420519432235803,0.24750087007269053,-0.14116246997777146,-1.4726902173855287,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,0.2387330283606326,-0.4184866590588439,-1.420519432235803,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,0.22150795010030921,-0.6809857496093201,-1.318026038009641,0.2387330283606326,-0.4184866590588439,-1.420519432235803,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,0.22150795010030921,-0.6809857496093201,-1.318026038009641,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.22150795010030874,0.6809857496093201,-1.318026038009641,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,-0.2387330283606321,0.4184866590588439,-1.420519432235803,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,-0.22150795010030874,0.6809857496093201,-1.318026038009641,-0.2387330283606321,0.4184866590588439,-1.420519432235803,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,-0.24750087007269003,0.14116246997777146,-1.4726902173855287,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,-0.2387330283606321,0.4184866590588439,-1.420519432235803,-0.24750087007269003,0.14116246997777146,-1.4726902173855287,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,-0.24750087007269003,-0.14116246997777146,-1.4726902173855287,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,-0.24750087007269003,0.14116246997777146,-1.4726902173855287,-0.24750087007269003,-0.14116246997777146,-1.4726902173855287,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,-0.2387330283606321,-0.4184866590588439,-1.420519432235803,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.24750087007269003,-0.14116246997777146,-1.4726902173855287,-0.2387330283606321,-0.4184866590588439,-1.420519432235803,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.22150795010030874,-0.6809857496093201,-1.318026038009641,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,-0.2387330283606321,-0.4184866590588439,-1.420519432235803,-0.4368890506524514,0.6809857496093201,-1.2630860486320923,-0.2387330283606321,0.4184866590588439,-1.420519432235803,-0.22150795010030874,0.6809857496093201,-1.318026038009641,-0.47086276620152767,0.4184866590588439,-1.3613071554924012,-0.2387330283606321,0.4184866590588439,-1.420519432235803,-0.4368890506524514,0.6809857496093201,-1.2630860486320923,-0.47086276620152767,0.4184866590588439,-1.3613071554924012,-0.24750087007269003,0.14116246997777146,-1.4726902173855287,-0.2387330283606321,0.4184866590588439,-1.420519432235803,-0.48815593351275643,0.14116246997777146,-1.4113032776997523,-0.24750087007269003,0.14116246997777146,-1.4726902173855287,-0.47086276620152767,0.4184866590588439,-1.3613071554924012,-0.48815593351275643,0.14116246997777146,-1.4113032776997523,-0.24750087007269003,-0.14116246997777146,-1.4726902173855287,-0.24750087007269003,0.14116246997777146,-1.4726902173855287,-0.48815593351275643,-0.14116246997777146,-1.4113032776997523,-0.24750087007269003,-0.14116246997777146,-1.4726902173855287,-0.48815593351275643,0.14116246997777146,-1.4113032776997523,-0.48815593351275643,-0.14116246997777146,-1.4113032776997523,-0.2387330283606321,-0.4184866590588439,-1.420519432235803,-0.24750087007269003,-0.14116246997777146,-1.4726902173855287,-0.47086276620152767,-0.4184866590588439,-1.3613071554924012,-0.2387330283606321,-0.4184866590588439,-1.420519432235803,-0.48815593351275643,-0.14116246997777146,-1.4113032776997523,-0.47086276620152767,-0.4184866590588439,-1.3613071554924012,-0.22150795010030874,-0.6809857496093201,-1.318026038009641,-0.2387330283606321,-0.4184866590588439,-1.420519432235803,-0.4368890506524514,-0.6809857496093201,-1.2630860486320923,-0.22150795010030874,-0.6809857496093201,-1.318026038009641,-0.47086276620152767,-0.4184866590588439,-1.3613071554924012,-0.6401859190974546,0.6809857496093201,-1.1732094432872504,-0.47086276620152767,0.4184866590588439,-1.3613071554924012,-0.4368890506524514,0.6809857496093201,-1.2630860486320923,-0.689968568219606,0.4184866590588439,-1.264441493727074,-0.47086276620152767,0.4184866590588439,-1.3613071554924012,-0.6401859190974546,0.6809857496093201,-1.1732094432872504,-0.689968568219606,0.4184866590588439,-1.264441493727074,-0.48815593351275643,0.14116246997777146,-1.4113032776997523,-0.47086276620152767,0.4184866590588439,-1.3613071554924012,-0.7153087368338384,0.14116246997777146,-1.3108800738736375,-0.48815593351275643,0.14116246997777146,-1.4113032776997523,-0.689968568219606,0.4184866590588439,-1.264441493727074,-0.7153087368338384,0.14116246997777146,-1.3108800738736375,-0.48815593351275643,-0.14116246997777146,-1.4113032776997523,-0.48815593351275643,0.14116246997777146,-1.4113032776997523,-0.7153087368338384,-0.14116246997777146,-1.3108800738736375,-0.48815593351275643,-0.14116246997777146,-1.4113032776997523,-0.7153087368338384,0.14116246997777146,-1.3108800738736375,-0.7153087368338384,-0.14116246997777146,-1.3108800738736375,-0.47086276620152767,-0.4184866590588439,-1.3613071554924012,-0.48815593351275643,-0.14116246997777146,-1.4113032776997523,-0.689968568219606,-0.4184866590588439,-1.264441493727074,-0.47086276620152767,-0.4184866590588439,-1.3613071554924012,-0.7153087368338384,-0.14116246997777146,-1.3108800738736375,-0.689968568219606,-0.4184866590588439,-1.264441493727074,-0.4368890506524514,-0.6809857496093201,-1.2630860486320923,-0.47086276620152767,-0.4184866590588439,-1.3613071554924012,-0.6401859190974546,-0.6809857496093201,-1.1732094432872504,-0.4368890506524514,-0.6809857496093201,-1.2630860486320923,-0.689968568219606,-0.4184866590588439,-1.264441493727074,-0.8257754194439826,0.6809857496093201,-1.0508821843913556,-0.689968568219606,0.4184866590588439,-1.264441493727074,-0.6401859190974546,0.6809857496093201,-1.1732094432872504,-0.8899900276281705,0.4184866590588439,-1.1326017247524283,-0.689968568219606,0.4184866590588439,-1.264441493727074,-0.8257754194439826,0.6809857496093201,-1.0508821843913556,-0.8899900276281705,0.4184866590588439,-1.1326017247524283,-0.7153087368338384,0.14116246997777146,-1.3108800738736375,-0.689968568219606,0.4184866590588439,-1.264441493727074,-0.9226762953856679,0.14116246997777146,-1.174198284363912,-0.7153087368338384,0.14116246997777146,-1.3108800738736375,-0.8899900276281705,0.4184866590588439,-1.1326017247524283,-0.9226762953856679,0.14116246997777146,-1.174198284363912,-0.7153087368338384,-0.14116246997777146,-1.3108800738736375,-0.7153087368338384,0.14116246997777146,-1.3108800738736375,-0.9226762953856679,-0.14116246997777146,-1.174198284363912,-0.7153087368338384,-0.14116246997777146,-1.3108800738736375,-0.9226762953856679,0.14116246997777146,-1.174198284363912,-0.9226762953856679,-0.14116246997777146,-1.174198284363912,-0.689968568219606,-0.4184866590588439,-1.264441493727074,-0.7153087368338384,-0.14116246997777146,-1.3108800738736375,-0.8899900276281705,-0.4184866590588439,-1.1326017247524283,-0.689968568219606,-0.4184866590588439,-1.264441493727074,-0.9226762953856679,-0.14116246997777146,-1.174198284363912,-0.8899900276281705,-0.4184866590588439,-1.1326017247524283,-0.6401859190974546,-0.6809857496093201,-1.1732094432872504,-0.689968568219606,-0.4184866590588439,-1.264441493727074,-0.8257754194439826,-0.6809857496093201,-1.0508821843913556,-0.6401859190974546,-0.6809857496093201,-1.1732094432872504,-0.8899900276281705,-0.4184866590588439,-1.1326017247524283,-0.9885241966777916,0.6809857496093201,-0.8994878105964301,-0.8899900276281705,0.4184866590588439,-1.1326017247524283,-0.8257754194439826,0.6809857496093201,-1.0508821843913556,-1.0653946053574237,0.4184866590588439,-0.9694345006575056,-0.8899900276281705,0.4184866590588439,-1.1326017247524283,-0.9885241966777916,0.6809857496093201,-0.8994878105964301,-1.0653946053574237,0.4184866590588439,-0.9694345006575056,-0.9226762953856679,0.14116246997777146,-1.174198284363912,-0.8899900276281705,0.4184866590588439,-1.1326017247524283,-1.104522878997648,0.14116246997777146,-1.0050384902283702,-0.9226762953856679,0.14116246997777146,-1.174198284363912,-1.0653946053574237,0.4184866590588439,-0.9694345006575056,-1.104522878997648,0.14116246997777146,-1.0050384902283702,-0.9226762953856679,-0.14116246997777146,-1.174198284363912,-0.9226762953856679,0.14116246997777146,-1.174198284363912,-1.104522878997648,-0.14116246997777146,-1.0050384902283702,-0.9226762953856679,-0.14116246997777146,-1.174198284363912,-1.104522878997648,0.14116246997777146,-1.0050384902283702,-1.104522878997648,-0.14116246997777146,-1.0050384902283702,-0.8899900276281705,-0.4184866590588439,-1.1326017247524283,-0.9226762953856679,-0.14116246997777146,-1.174198284363912,-1.0653946053574237,-0.4184866590588439,-0.9694345006575056,-0.8899900276281705,-0.4184866590588439,-1.1326017247524283,-1.104522878997648,-0.14116246997777146,-1.0050384902283702,-1.0653946053574237,-0.4184866590588439,-0.9694345006575056,-0.8257754194439826,-0.6809857496093201,-1.0508821843913556,-0.8899900276281705,-0.4184866590588439,-1.1326017247524283,-0.9885241966777916,-0.6809857496093201,-0.8994878105964301,-0.8257754194439826,-0.6809857496093201,-1.0508821843913556,-1.0653946053574237,-0.4184866590588439,-0.9694345006575056],[[0.5020270616434407,0.303478973012677,-1.1601160900172116,0.6200317949183417,0.1831716015188576,-1.1278336480560227,0.608975616258845,0.30347897301267707,-1.1077225337336651,0.5111415495428591,0.1831716015188576,-1.1811784288277287,0.6200317949183417,0.1831716015188576,-1.1278336480560227,0.5020270616434407,0.303478973012677,-1.1601160900172116,0.5111415495428591,0.1831716015188576,-1.1811784288277287,0.6255845274717116,6.123838592253551e-2,-1.1379340310745603,0.6200317949183417,0.1831716015188576,-1.1278336480560227,0.5157191088628619,6.123838592253551e-2,-1.191756544283814,0.6255845274717116,6.123838592253551e-2,-1.1379340310745603,0.5111415495428591,0.1831716015188576,-1.1811784288277287,0.5157191088628619,6.123838592253551e-2,-1.191756544283814,0.6255845274717116,-6.123838592253551e-2,-1.1379340310745603,0.6255845274717116,6.123838592253551e-2,-1.1379340310745603,0.5157191088628619,-6.123838592253551e-2,-1.191756544283814,0.6255845274717116,-6.123838592253551e-2,-1.1379340310745603,0.5157191088628619,6.123838592253551e-2,-1.191756544283814,0.5157191088628619,-6.123838592253551e-2,-1.191756544283814,0.6200317949183417,-0.1831716015188576,-1.1278336480560227,0.6255845274717116,-6.123838592253551e-2,-1.1379340310745603,0.5111415495428591,-0.1831716015188576,-1.1811784288277287,0.6200317949183417,-0.1831716015188576,-1.1278336480560227,0.5157191088628619,-6.123838592253551e-2,-1.191756544283814,0.5111415495428591,-0.1831716015188576,-1.1811784288277287,0.608975616258845,-0.3034789730126769,-1.1077225337336654,0.6200317949183417,-0.1831716015188576,-1.1278336480560227,0.5020270616434408,-0.30347897301267684,-1.1601160900172118,0.608975616258845,-0.3034789730126769,-1.1077225337336654,0.5111415495428591,-0.1831716015188576,-1.1811784288277287,0.390622479288466,0.30347897301267707,-1.202212373756693,0.5111415495428591,0.1831716015188576,-1.1811784288277287,0.5020270616434407,0.303478973012677,-1.1601160900172116,0.3977143755879611,0.1831716015188576,-1.2240389862450032,0.5111415495428591,0.1831716015188576,-1.1811784288277287,0.390622479288466,0.30347897301267707,-1.202212373756693,0.3977143755879611,0.1831716015188576,-1.2240389862450032,0.5157191088628619,6.123838592253551e-2,-1.191756544283814,0.5111415495428591,0.1831716015188576,-1.1811784288277287,0.4012761309340096,6.123838592253552e-2,-1.2350009420369827,0.5157191088628619,6.123838592253551e-2,-1.191756544283814,0.3977143755879611,0.1831716015188576,-1.2240389862450032,0.4012761309340096,6.123838592253552e-2,-1.2350009420369827,0.5157191088628619,-6.123838592253551e-2,-1.191756544283814,0.5157191088628619,6.123838592253551e-2,-1.191756544283814,0.4012761309340096,-6.123838592253552e-2,-1.2350009420369827,0.5157191088628619,-6.123838592253551e-2,-1.191756544283814,0.4012761309340096,6.123838592253552e-2,-1.2350009420369827,0.4012761309340096,-6.123838592253552e-2,-1.2350009420369827,0.5111415495428591,-0.1831716015188576,-1.1811784288277287,0.5157191088628619,-6.123838592253551e-2,-1.191756544283814,0.3977143755879611,-0.1831716015188576,-1.2240389862450032,0.5111415495428591,-0.1831716015188576,-1.1811784288277287,0.4012761309340096,-6.123838592253552e-2,-1.2350009420369827,0.3977143755879611,-0.1831716015188576,-1.2240389862450032,0.5020270616434408,-0.30347897301267684,-1.1601160900172118,0.5111415495428591,-0.1831716015188576,-1.1811784288277287,0.3906224792884661,-0.3034789730126769,-1.2022123737566932,0.5020270616434408,-0.30347897301267684,-1.1601160900172118,0.3977143755879611,-0.1831716015188576,-1.2240389862450032,0.27575070415366154,0.30347897301267707,-1.2336377353574794,0.3977143755879611,0.1831716015188576,-1.2240389862450032,0.390622479288466,0.30347897301267707,-1.202212373756693,0.2807570606796165,0.1831716015188576,-1.2560348869659468,0.3977143755879611,0.1831716015188576,-1.2240389862450032,0.27575070415366154,0.30347897301267707,-1.2336377353574794,0.2807570606796165,0.1831716015188576,-1.2560348869659468,0.4012761309340096,6.123838592253552e-2,-1.2350009420369827,0.3977143755879611,0.1831716015188576,-1.2240389862450032,0.2832713976591088,6.123838592253552e-2,-1.267283383998172,0.4012761309340096,6.123838592253552e-2,-1.2350009420369827,0.2807570606796165,0.1831716015188576,-1.2560348869659468,0.2832713976591088,6.123838592253552e-2,-1.267283383998172,0.4012761309340096,-6.123838592253552e-2,-1.2350009420369827,0.4012761309340096,6.123838592253552e-2,-1.2350009420369827,0.2832713976591088,-6.123838592253552e-2,-1.267283383998172,0.4012761309340096,-6.123838592253552e-2,-1.2350009420369827,0.2832713976591088,6.123838592253552e-2,-1.267283383998172,0.2832713976591088,-6.123838592253552e-2,-1.267283383998172,0.3977143755879611,-0.1831716015188576,-1.2240389862450032,0.4012761309340096,-6.123838592253552e-2,-1.2350009420369827,0.2807570606796165,-0.1831716015188576,-1.2560348869659468,0.3977143755879611,-0.1831716015188576,-1.2240389862450032,0.2832713976591088,-6.123838592253552e-2,-1.267283383998172,0.2807570606796165,-0.1831716015188576,-1.2560348869659468,0.3906224792884661,-0.3034789730126769,-1.2022123737566932,0.3977143755879611,-0.1831716015188576,-1.2240389862450032,0.2757507041536616,-0.3034789730126769,-1.2336377353574794,0.3906224792884661,-0.3034789730126769,-1.2022123737566932,0.2807570606796165,-0.1831716015188576,-1.2560348869659468,0.15843134624733743,0.30347897301267707,-1.2541132410852807,0.2807570606796165,0.1831716015188576,-1.2560348869659468,0.27575070415366154,0.30347897301267707,-1.2336377353574794,0.16130772622480924,0.1831716015188576,-1.2768821331106481,0.2807570606796165,0.1831716015188576,-1.2560348869659468,0.15843134624733743,0.30347897301267707,-1.2541132410852807,0.16130772622480924,0.1831716015188576,-1.2768821331106481,0.2832713976591088,6.123838592253552e-2,-1.267283383998172,0.2807570606796165,0.1831716015188576,-1.2560348869659468,0.16275232740471518,6.123838592253551e-2,-1.2883173289271364,0.2832713976591088,6.123838592253552e-2,-1.267283383998172,0.16130772622480924,0.1831716015188576,-1.2768821331106481,0.16275232740471518,6.123838592253551e-2,-1.2883173289271364,0.2832713976591088,-6.123838592253552e-2,-1.267283383998172,0.2832713976591088,6.123838592253552e-2,-1.267283383998172,0.16275232740471518,-6.123838592253551e-2,-1.2883173289271364,0.2832713976591088,-6.123838592253552e-2,-1.267283383998172,0.16275232740471518,6.123838592253551e-2,-1.2883173289271364,0.16275232740471518,-6.123838592253551e-2,-1.2883173289271364,0.2807570606796165,-0.1831716015188576,-1.2560348869659468,0.2832713976591088,-6.123838592253552e-2,-1.267283383998172,0.16130772622480924,-0.1831716015188576,-1.2768821331106481,0.2807570606796165,-0.1831716015188576,-1.2560348869659468,0.16275232740471518,-6.123838592253551e-2,-1.2883173289271364,0.16130772622480924,-0.1831716015188576,-1.2768821331106481,0.2757507041536616,-0.3034789730126769,-1.2336377353574794,0.2807570606796165,-0.1831716015188576,-1.2560348869659468,0.15843134624733746,-0.3034789730126769,-1.2541132410852809,0.2757507041536616,-0.3034789730126769,-1.2336377353574794,0.16130772622480924,-0.1831716015188576,-1.2768821331106481,3.970574049575968e-2,0.30347897301267707,-1.2634571489017166,0.16130772622480924,0.1831716015188576,-1.2768821331106481,0.15843134624733743,0.30347897301267707,-1.2541132410852807,4.04266129724372e-2,0.1831716015188576,-1.28639568304647,0.16130772622480924,0.1831716015188576,-1.2768821331106481,3.970574049575968e-2,0.30347897301267707,-1.2634571489017166,4.04266129724372e-2,0.1831716015188576,-1.28639568304647,0.16275232740471518,6.123838592253551e-2,-1.2883173289271364,0.16130772622480924,0.1831716015188576,-1.2768821331106481,4.07886559704162e-2,6.123838592253551e-2,-1.2979160780396126,0.16275232740471518,6.123838592253551e-2,-1.2883173289271364,4.04266129724372e-2,0.1831716015188576,-1.28639568304647,4.07886559704162e-2,6.123838592253551e-2,-1.2979160780396126,0.16275232740471518,-6.123838592253551e-2,-1.2883173289271364,0.16275232740471518,6.123838592253551e-2,-1.2883173289271364,4.07886559704162e-2,-6.123838592253551e-2,-1.2979160780396126,0.16275232740471518,-6.123838592253551e-2,-1.2883173289271364,4.07886559704162e-2,6.123838592253551e-2,-1.2979160780396126,4.07886559704162e-2,-6.123838592253551e-2,-1.2979160780396126,0.16130772622480924,-0.1831716015188576,-1.2768821331106481,0.16275232740471518,-6.123838592253551e-2,-1.2883173289271364,4.04266129724372e-2,-0.1831716015188576,-1.28639568304647,0.16130772622480924,-0.1831716015188576,-1.2768821331106481,4.07886559704162e-2,-6.123838592253551e-2,-1.2979160780396126,4.04266129724372e-2,-0.1831716015188576,-1.28639568304647,0.15843134624733746,-0.3034789730126769,-1.2541132410852809,0.16130772622480924,-0.1831716015188576,-1.2768821331106481,3.970574049575969e-2,-0.3034789730126769,-1.2634571489017166,0.15843134624733746,-0.3034789730126769,-1.2541132410852809,4.04266129724372e-2,-0.1831716015188576,-1.28639568304647],[-0.15843134624733782,0.303478973012677,-1.2541132410852802,-4.0426612972436764e-2,0.1831716015188576,-1.28639568304647,-3.970574049575925e-2,0.30347897301267707,-1.2634571489017166,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-4.0426612972436764e-2,0.1831716015188576,-1.28639568304647,-0.15843134624733782,0.303478973012677,-1.2541132410852802,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-4.0788655970415755e-2,6.123838592253551e-2,-1.2979160780396126,-4.0426612972436764e-2,0.1831716015188576,-1.28639568304647,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-4.0788655970415755e-2,6.123838592253551e-2,-1.2979160780396126,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-4.0788655970415755e-2,-6.123838592253551e-2,-1.2979160780396126,-4.0788655970415755e-2,6.123838592253551e-2,-1.2979160780396126,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-4.0788655970415755e-2,-6.123838592253551e-2,-1.2979160780396126,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-4.0426612972436764e-2,-0.1831716015188576,-1.28639568304647,-4.0788655970415755e-2,-6.123838592253551e-2,-1.2979160780396126,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-4.0426612972436764e-2,-0.1831716015188576,-1.28639568304647,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-3.970574049575925e-2,-0.3034789730126769,-1.2634571489017166,-4.0426612972436764e-2,-0.1831716015188576,-1.28639568304647,-0.15843134624733785,-0.30347897301267684,-1.2541132410852804,-3.970574049575925e-2,-0.3034789730126769,-1.2634571489017166,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-0.27575070415366143,0.30347897301267707,-1.2336377353574794,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-0.15843134624733782,0.303478973012677,-1.2541132410852802,-0.28075706067961637,0.1831716015188576,-1.256034886965947,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-0.27575070415366143,0.30347897301267707,-1.2336377353574794,-0.28075706067961637,0.1831716015188576,-1.256034886965947,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-0.28327139765910864,6.123838592253552e-2,-1.267283383998172,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-0.28075706067961637,0.1831716015188576,-1.256034886965947,-0.28327139765910864,6.123838592253552e-2,-1.267283383998172,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-0.28327139765910864,-6.123838592253552e-2,-1.267283383998172,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-0.28327139765910864,6.123838592253552e-2,-1.267283383998172,-0.28327139765910864,-6.123838592253552e-2,-1.267283383998172,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-0.28075706067961637,-0.1831716015188576,-1.256034886965947,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-0.28327139765910864,-6.123838592253552e-2,-1.267283383998172,-0.28075706067961637,-0.1831716015188576,-1.256034886965947,-0.15843134624733785,-0.30347897301267684,-1.2541132410852804,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-0.27575070415366143,-0.3034789730126769,-1.2336377353574797,-0.15843134624733785,-0.30347897301267684,-1.2541132410852804,-0.28075706067961637,-0.1831716015188576,-1.256034886965947,-0.3906224792884654,0.30347897301267707,-1.2022123737566934,-0.28075706067961637,0.1831716015188576,-1.256034886965947,-0.27575070415366143,0.30347897301267707,-1.2336377353574794,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.28075706067961637,0.1831716015188576,-1.256034886965947,-0.3906224792884654,0.30347897301267707,-1.2022123737566934,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.28327139765910864,6.123838592253552e-2,-1.267283383998172,-0.28075706067961637,0.1831716015188576,-1.256034886965947,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.28327139765910864,6.123838592253552e-2,-1.267283383998172,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.28327139765910864,-6.123838592253552e-2,-1.267283383998172,-0.28327139765910864,6.123838592253552e-2,-1.267283383998172,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.28327139765910864,-6.123838592253552e-2,-1.267283383998172,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.28075706067961637,-0.1831716015188576,-1.256034886965947,-0.28327139765910864,-6.123838592253552e-2,-1.267283383998172,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.28075706067961637,-0.1831716015188576,-1.256034886965947,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.27575070415366143,-0.3034789730126769,-1.2336377353574797,-0.28075706067961637,-0.1831716015188576,-1.256034886965947,-0.39062247928846544,-0.3034789730126769,-1.2022123737566934,-0.27575070415366143,-0.3034789730126769,-1.2336377353574797,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.5020270616434412,0.30347897301267707,-1.1601160900172116,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.3906224792884654,0.30347897301267707,-1.2022123737566934,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.5020270616434412,0.30347897301267707,-1.1601160900172116,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.5157191088628623,6.123838592253551e-2,-1.1917565442838138,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.5157191088628623,6.123838592253551e-2,-1.1917565442838138,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.5157191088628623,-6.123838592253551e-2,-1.1917565442838138,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.5157191088628623,6.123838592253551e-2,-1.1917565442838138,-0.5157191088628623,-6.123838592253551e-2,-1.1917565442838138,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.5157191088628623,-6.123838592253551e-2,-1.1917565442838138,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.39062247928846544,-0.3034789730126769,-1.2022123737566934,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.5020270616434412,-0.3034789730126769,-1.1601160900172118,-0.39062247928846544,-0.3034789730126769,-1.2022123737566934,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.6089756162588447,0.30347897301267707,-1.1077225337336651,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.5020270616434412,0.30347897301267707,-1.1601160900172116,-0.6200317949183414,0.1831716015188576,-1.1278336480560227,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.6089756162588447,0.30347897301267707,-1.1077225337336651,-0.6200317949183414,0.1831716015188576,-1.1278336480560227,-0.5157191088628623,6.123838592253551e-2,-1.1917565442838138,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.6255845274717112,6.123838592253551e-2,-1.1379340310745603,-0.5157191088628623,6.123838592253551e-2,-1.1917565442838138,-0.6200317949183414,0.1831716015188576,-1.1278336480560227,-0.6255845274717112,6.123838592253551e-2,-1.1379340310745603,-0.5157191088628623,-6.123838592253551e-2,-1.1917565442838138,-0.5157191088628623,6.123838592253551e-2,-1.1917565442838138,-0.6255845274717112,-6.123838592253551e-2,-1.1379340310745603,-0.5157191088628623,-6.123838592253551e-2,-1.1917565442838138,-0.6255845274717112,6.123838592253551e-2,-1.1379340310745603,-0.6255845274717112,-6.123838592253551e-2,-1.1379340310745603,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.5157191088628623,-6.123838592253551e-2,-1.1917565442838138,-0.6200317949183414,-0.1831716015188576,-1.1278336480560227,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.6255845274717112,-6.123838592253551e-2,-1.1379340310745603,-0.6200317949183414,-0.1831716015188576,-1.1278336480560227,-0.5020270616434412,-0.3034789730126769,-1.1601160900172118,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.6089756162588448,-0.3034789730126769,-1.1077225337336654,-0.5020270616434412,-0.3034789730126769,-1.1601160900172118,-0.6200317949183414,-0.1831716015188576,-1.1278336480560227]]];
      case 3:
        return [[1.0299343479203862,0.6784329353268667,-0.8520357922279913,1.1890911135341198,0.6826289037478287,-0.6058721841133033,1.1909960433344438,0.6784329353268667,-0.6068427943260688,1.0282870270558786,0.6826289037478287,-0.850673009890769,1.1890911135341198,0.6826289037478287,-0.6058721841133033,1.0299343479203862,0.6784329353268667,-0.8520357922279913,0.8192643443294739,0.6784329353268667,-1.0561892284880439,1.0282870270558786,0.6826289037478287,-0.850673009890769,1.0299343479203862,0.6784329353268667,-0.8520357922279913,0.8179539780418692,0.6826289037478287,-1.0544999144492708,1.0282870270558786,0.6826289037478287,-0.850673009890769,0.8192643443294739,0.6784329353268667,-1.0561892284880439,0.5691332642593958,0.6784329353268667,-1.2094697514932802,0.8179539780418692,0.6826289037478287,-1.0544999144492708,0.8192643443294739,0.6784329353268667,-1.0561892284880439,0.5682229682751976,0.6826289037478287,-1.2075352740572682,0.8179539780418692,0.6826289037478287,-1.0544999144492708,0.5691332642593958,0.6784329353268667,-1.2094697514932802,0.29158904040584394,0.6784329353268667,-1.304494378592323,0.5682229682751976,0.6826289037478287,-1.2075352740572682,0.5691332642593958,0.6784329353268667,-1.2094697514932802,0.2911226604748394,0.6826289037478287,-1.3024079147203036,0.5682229682751976,0.6826289037478287,-1.2075352740572682,0.29158904040584394,0.6784329353268667,-1.304494378592323,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,0.2911226604748394,0.6826289037478287,-1.3024079147203036,0.29158904040584394,0.6784329353268667,-1.304494378592323,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,0.2911226604748394,0.6826289037478287,-1.3024079147203036,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,-0.29158904040584405,0.6784329353268667,-1.304494378592323,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,-0.2911226604748396,0.6826289037478287,-1.3024079147203036,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,-0.29158904040584405,0.6784329353268667,-1.304494378592323,-0.5691332642593959,0.6784329353268667,-1.20946975149328,-0.2911226604748396,0.6826289037478287,-1.3024079147203036,-0.29158904040584405,0.6784329353268667,-1.304494378592323,-0.5682229682751977,0.6826289037478287,-1.207535274057268,-0.2911226604748396,0.6826289037478287,-1.3024079147203036,-0.5691332642593959,0.6784329353268667,-1.20946975149328,-0.8192643443294736,0.6784329353268667,-1.0561892284880439,-0.5682229682751977,0.6826289037478287,-1.207535274057268,-0.5691332642593959,0.6784329353268667,-1.20946975149328,-0.8179539780418689,0.6826289037478287,-1.0544999144492708,-0.5682229682751977,0.6826289037478287,-1.207535274057268,-0.8192643443294736,0.6784329353268667,-1.0561892284880439,-1.0299343479203864,0.6784329353268667,-0.8520357922279911,-0.8179539780418689,0.6826289037478287,-1.0544999144492708,-0.8192643443294736,0.6784329353268667,-1.0561892284880439,-1.0282870270558788,0.6826289037478287,-0.850673009890769,-0.8179539780418689,0.6826289037478287,-1.0544999144492708,-1.0299343479203864,0.6784329353268667,-0.8520357922279911,-1.1909960433344438,0.6784329353268667,-0.6068427943260688,-1.0282870270558788,0.6826289037478287,-0.850673009890769,-1.0299343479203864,0.6784329353268667,-0.8520357922279911,-1.1890911135341198,0.6826289037478287,-0.6058721841133035,-1.0282870270558788,0.6826289037478287,-0.850673009890769,-1.1909960433344438,0.6784329353268667,-0.6068427943260688,1.1096433799393968,0.41594602267436875,-0.9179768382579405,1.2819993794110904,0.42046828134539266,-0.6532113100460147,1.2831700172817604,0.41594602267436875,-0.653807779833263,1.1086310506720354,0.42046828134539266,-0.9171393666549654,1.2819993794110904,0.42046828134539266,-0.6532113100460147,1.1096433799393968,0.41594602267436875,-0.9179768382579405,0.8826691312326869,0.41594602267436875,-1.1379301872216536,1.1086310506720354,0.42046828134539266,-0.9171393666549654,1.1096433799393968,0.41594602267436875,-0.9179768382579405,0.8818638709021182,0.42046828134539266,-1.1368920518589196,1.1086310506720354,0.42046828134539266,-0.9171393666549654,0.8826691312326869,0.41594602267436875,-1.1379301872216536,0.6131798208923851,0.41594602267436875,-1.3030734489934774,0.8818638709021182,0.42046828134539266,-1.1368920518589196,0.8826691312326869,0.41594602267436875,-1.1379301872216536,0.6126204160510939,0.42046828134539266,-1.3018846531931454,0.8818638709021182,0.42046828134539266,-1.1368920518589196,0.6131798208923851,0.41594602267436875,-1.3030734489934774,0.31415579935026805,0.41594602267436875,-1.4054522545984862,0.6126204160510939,0.42046828134539266,-1.3018846531931454,0.6131798208923851,0.41594602267436875,-1.3030734489934774,0.3138691945581852,0.42046828134539266,-1.4041700584650878,0.6126204160510939,0.42046828134539266,-1.3018846531931454,0.31415579935026805,0.41594602267436875,-1.4054522545984862,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,0.3138691945581852,0.42046828134539266,-1.4041700584650878,0.31415579935026805,0.41594602267436875,-1.4054522545984862,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,0.3138691945581852,0.42046828134539266,-1.4041700584650878,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,-0.31415579935026816,0.41594602267436875,-1.4054522545984862,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,-0.3138691945581853,0.42046828134539266,-1.4041700584650878,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,-0.31415579935026816,0.41594602267436875,-1.4054522545984862,-0.6131798208923852,0.41594602267436875,-1.3030734489934772,-0.3138691945581853,0.42046828134539266,-1.4041700584650878,-0.31415579935026816,0.41594602267436875,-1.4054522545984862,-0.6126204160510941,0.42046828134539266,-1.3018846531931454,-0.3138691945581853,0.42046828134539266,-1.4041700584650878,-0.6131798208923852,0.41594602267436875,-1.3030734489934772,-0.8826691312326865,0.41594602267436875,-1.1379301872216538,-0.6126204160510941,0.42046828134539266,-1.3018846531931454,-0.6131798208923852,0.41594602267436875,-1.3030734489934772,-0.8818638709021178,0.42046828134539266,-1.1368920518589198,-0.6126204160510941,0.42046828134539266,-1.3018846531931454,-0.8826691312326865,0.41594602267436875,-1.1379301872216538,-1.109643379939397,0.41594602267436875,-0.9179768382579404,-0.8818638709021178,0.42046828134539266,-1.1368920518589198,-0.8826691312326865,0.41594602267436875,-1.1379301872216538,-1.1086310506720354,0.42046828134539266,-0.9171393666549652,-0.8818638709021178,0.42046828134539266,-1.1368920518589198,-1.109643379939397,0.41594602267436875,-0.9179768382579404,-1.2831700172817604,0.41594602267436875,-0.6538077798332631,-1.1086310506720354,0.42046828134539266,-0.9171393666549652,-1.109643379939397,0.41594602267436875,-0.9179768382579404,-1.2819993794110904,0.42046828134539266,-0.6532113100460148,-1.1086310506720354,0.42046828134539266,-0.9171393666549652,-1.2831700172817604,0.41594602267436875,-0.6538077798332631,1.1500427420195336,0.13872401480181032,-0.9513980971420218,1.3294921781410904,0.14341236045493638,-0.6774100996666663,1.329887053651925,0.13872401480181032,-0.6776112987888185,1.1497012666183035,0.14341236045493638,-0.9511156041223446,1.3294921781410904,0.14341236045493638,-0.6774100996666663,1.1500427420195336,0.13872401480181032,-0.9513980971420218,0.9148049241137984,0.13872401480181032,-1.1793594017662363,1.1497012666183035,0.14341236045493638,-0.9511156041223446,1.1500427420195336,0.13872401480181032,-0.9513980971420218,0.9145332964889327,0.14341236045493638,-1.1790092215423211,1.1497012666183035,0.14341236045493638,-0.9511156041223446,0.9148049241137984,0.13872401480181032,-1.1793594017662363,0.6355041766739856,0.13872401480181032,-1.350515119925426,0.9145332964889327,0.14341236045493638,-1.1790092215423211,0.9148049241137984,0.13872401480181032,-1.1793594017662363,0.6353154801709913,0.14341236045493638,-1.3501141194446664,0.9145332964889327,0.14341236045493638,-1.1790092215423211,0.6355041766739856,0.13872401480181032,-1.350515119925426,0.32559343248265277,0.13872401480181032,-1.4566212838076449,0.6353154801709913,0.14341236045493638,-1.3501141194446664,0.6355041766739856,0.13872401480181032,-1.350515119925426,0.32549675594713584,0.14341236045493638,-1.456188777850123,0.6353154801709913,0.14341236045493638,-1.3501141194446664,0.32559343248265277,0.13872401480181032,-1.4566212838076449,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,0.32549675594713584,0.14341236045493638,-1.456188777850123,0.32559343248265277,0.13872401480181032,-1.4566212838076449,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,0.32549675594713584,0.14341236045493638,-1.456188777850123,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,-0.3255934324826529,0.13872401480181032,-1.4566212838076449,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,-0.32549675594713595,0.14341236045493638,-1.456188777850123,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,-0.3255934324826529,0.13872401480181032,-1.4566212838076449,-0.6355041766739858,0.13872401480181032,-1.3505151199254257,-0.32549675594713595,0.14341236045493638,-1.456188777850123,-0.3255934324826529,0.13872401480181032,-1.4566212838076449,-0.6353154801709915,0.14341236045493638,-1.3501141194446662,-0.32549675594713595,0.14341236045493638,-1.456188777850123,-0.6355041766739858,0.13872401480181032,-1.3505151199254257,-0.9148049241137981,0.13872401480181032,-1.1793594017662365,-0.6353154801709915,0.14341236045493638,-1.3501141194446662,-0.6355041766739858,0.13872401480181032,-1.3505151199254257,-0.9145332964889323,0.14341236045493638,-1.1790092215423214,-0.6353154801709915,0.14341236045493638,-1.3501141194446662,-0.9148049241137981,0.13872401480181032,-1.1793594017662365,-1.1500427420195336,0.13872401480181032,-0.9513980971420216,-0.9145332964889323,0.14341236045493638,-1.1790092215423214,-0.9148049241137981,0.13872401480181032,-1.1793594017662365,-1.1497012666183037,0.14341236045493638,-0.9511156041223445,-0.9145332964889323,0.14341236045493638,-1.1790092215423214,-1.1500427420195336,0.13872401480181032,-0.9513980971420216,-1.329887053651925,0.13872401480181032,-0.6776112987888187,-1.1497012666183037,0.14341236045493638,-0.9511156041223445,-1.1500427420195336,0.13872401480181032,-0.9513980971420216,-1.3294921781410904,0.14341236045493638,-0.6774100996666664,-1.1497012666183037,0.14341236045493638,-0.9511156041223445,-1.329887053651925,0.13872401480181032,-0.6776112987888187,1.1497012666183035,-0.14341236045493638,-0.9511156041223446,1.329887053651925,-0.13872401480181032,-0.6776112987888185,1.3294921781410904,-0.14341236045493638,-0.6774100996666663,1.1500427420195336,-0.13872401480181032,-0.9513980971420218,1.329887053651925,-0.13872401480181032,-0.6776112987888185,1.1497012666183035,-0.14341236045493638,-0.9511156041223446,0.9145332964889327,-0.14341236045493638,-1.1790092215423211,1.1500427420195336,-0.13872401480181032,-0.9513980971420218,1.1497012666183035,-0.14341236045493638,-0.9511156041223446,0.9148049241137984,-0.13872401480181032,-1.1793594017662363,1.1500427420195336,-0.13872401480181032,-0.9513980971420218,0.9145332964889327,-0.14341236045493638,-1.1790092215423211,0.6353154801709913,-0.14341236045493638,-1.3501141194446664,0.9148049241137984,-0.13872401480181032,-1.1793594017662363,0.9145332964889327,-0.14341236045493638,-1.1790092215423211,0.6355041766739856,-0.13872401480181032,-1.350515119925426,0.9148049241137984,-0.13872401480181032,-1.1793594017662363,0.6353154801709913,-0.14341236045493638,-1.3501141194446664,0.32549675594713584,-0.14341236045493638,-1.456188777850123,0.6355041766739856,-0.13872401480181032,-1.350515119925426,0.6353154801709913,-0.14341236045493638,-1.3501141194446664,0.32559343248265277,-0.13872401480181032,-1.4566212838076449,0.6355041766739856,-0.13872401480181032,-1.350515119925426,0.32549675594713584,-0.14341236045493638,-1.456188777850123,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,0.32559343248265277,-0.13872401480181032,-1.4566212838076449,0.32549675594713584,-0.14341236045493638,-1.456188777850123,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,0.32559343248265277,-0.13872401480181032,-1.4566212838076449,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,-0.32549675594713595,-0.14341236045493638,-1.456188777850123,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,-0.3255934324826529,-0.13872401480181032,-1.4566212838076449,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,-0.32549675594713595,-0.14341236045493638,-1.456188777850123,-0.6353154801709915,-0.14341236045493638,-1.3501141194446662,-0.3255934324826529,-0.13872401480181032,-1.4566212838076449,-0.32549675594713595,-0.14341236045493638,-1.456188777850123,-0.6355041766739858,-0.13872401480181032,-1.3505151199254257,-0.3255934324826529,-0.13872401480181032,-1.4566212838076449,-0.6353154801709915,-0.14341236045493638,-1.3501141194446662,-0.9145332964889323,-0.14341236045493638,-1.1790092215423214,-0.6355041766739858,-0.13872401480181032,-1.3505151199254257,-0.6353154801709915,-0.14341236045493638,-1.3501141194446662,-0.9148049241137981,-0.13872401480181032,-1.1793594017662365,-0.6355041766739858,-0.13872401480181032,-1.3505151199254257,-0.9145332964889323,-0.14341236045493638,-1.1790092215423214,-1.1497012666183037,-0.14341236045493638,-0.9511156041223445,-0.9148049241137981,-0.13872401480181032,-1.1793594017662365,-0.9145332964889323,-0.14341236045493638,-1.1790092215423214,-1.1500427420195336,-0.13872401480181032,-0.9513980971420216,-0.9148049241137981,-0.13872401480181032,-1.1793594017662365,-1.1497012666183037,-0.14341236045493638,-0.9511156041223445,-1.3294921781410904,-0.14341236045493638,-0.6774100996666664,-1.1500427420195336,-0.13872401480181032,-0.9513980971420216,-1.1497012666183037,-0.14341236045493638,-0.9511156041223445,-1.329887053651925,-0.13872401480181032,-0.6776112987888187,-1.1500427420195336,-0.13872401480181032,-0.9513980971420216,-1.3294921781410904,-0.14341236045493638,-0.6774100996666664,1.1086310506720354,-0.42046828134539266,-0.9171393666549654,1.2831700172817604,-0.41594602267436875,-0.653807779833263,1.2819993794110904,-0.42046828134539266,-0.6532113100460147,1.1096433799393968,-0.41594602267436875,-0.9179768382579405,1.2831700172817604,-0.41594602267436875,-0.653807779833263,1.1086310506720354,-0.42046828134539266,-0.9171393666549654,0.8818638709021182,-0.42046828134539266,-1.1368920518589196,1.1096433799393968,-0.41594602267436875,-0.9179768382579405,1.1086310506720354,-0.42046828134539266,-0.9171393666549654,0.8826691312326869,-0.41594602267436875,-1.1379301872216536,1.1096433799393968,-0.41594602267436875,-0.9179768382579405,0.8818638709021182,-0.42046828134539266,-1.1368920518589196,0.6126204160510939,-0.42046828134539266,-1.3018846531931454,0.8826691312326869,-0.41594602267436875,-1.1379301872216536,0.8818638709021182,-0.42046828134539266,-1.1368920518589196,0.6131798208923851,-0.41594602267436875,-1.3030734489934774,0.8826691312326869,-0.41594602267436875,-1.1379301872216536,0.6126204160510939,-0.42046828134539266,-1.3018846531931454,0.3138691945581852,-0.42046828134539266,-1.4041700584650878,0.6131798208923851,-0.41594602267436875,-1.3030734489934774,0.6126204160510939,-0.42046828134539266,-1.3018846531931454,0.31415579935026805,-0.41594602267436875,-1.4054522545984862,0.6131798208923851,-0.41594602267436875,-1.3030734489934774,0.3138691945581852,-0.42046828134539266,-1.4041700584650878,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,0.31415579935026805,-0.41594602267436875,-1.4054522545984862,0.3138691945581852,-0.42046828134539266,-1.4041700584650878,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,0.31415579935026805,-0.41594602267436875,-1.4054522545984862,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,-0.3138691945581853,-0.42046828134539266,-1.4041700584650878,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,-0.31415579935026816,-0.41594602267436875,-1.4054522545984862,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,-0.3138691945581853,-0.42046828134539266,-1.4041700584650878,-0.6126204160510941,-0.42046828134539266,-1.3018846531931454,-0.31415579935026816,-0.41594602267436875,-1.4054522545984862,-0.3138691945581853,-0.42046828134539266,-1.4041700584650878,-0.6131798208923852,-0.41594602267436875,-1.3030734489934772,-0.31415579935026816,-0.41594602267436875,-1.4054522545984862,-0.6126204160510941,-0.42046828134539266,-1.3018846531931454,-0.8818638709021178,-0.42046828134539266,-1.1368920518589198,-0.6131798208923852,-0.41594602267436875,-1.3030734489934772,-0.6126204160510941,-0.42046828134539266,-1.3018846531931454,-0.8826691312326865,-0.41594602267436875,-1.1379301872216538,-0.6131798208923852,-0.41594602267436875,-1.3030734489934772,-0.8818638709021178,-0.42046828134539266,-1.1368920518589198,-1.1086310506720354,-0.42046828134539266,-0.9171393666549652,-0.8826691312326865,-0.41594602267436875,-1.1379301872216538,-0.8818638709021178,-0.42046828134539266,-1.1368920518589198,-1.109643379939397,-0.41594602267436875,-0.9179768382579404,-0.8826691312326865,-0.41594602267436875,-1.1379301872216538,-1.1086310506720354,-0.42046828134539266,-0.9171393666549652,-1.2819993794110904,-0.42046828134539266,-0.6532113100460148,-1.109643379939397,-0.41594602267436875,-0.9179768382579404,-1.1086310506720354,-0.42046828134539266,-0.9171393666549652,-1.2831700172817604,-0.41594602267436875,-0.6538077798332631,-1.109643379939397,-0.41594602267436875,-0.9179768382579404,-1.2819993794110904,-0.42046828134539266,-0.6532113100460148,1.0282870270558786,-0.6826289037478287,-0.850673009890769,1.1909960433344438,-0.6784329353268667,-0.6068427943260688,1.1890911135341198,-0.6826289037478287,-0.6058721841133033,1.0299343479203862,-0.6784329353268667,-0.8520357922279913,1.1909960433344438,-0.6784329353268667,-0.6068427943260688,1.0282870270558786,-0.6826289037478287,-0.850673009890769,0.8179539780418692,-0.6826289037478287,-1.0544999144492708,1.0299343479203862,-0.6784329353268667,-0.8520357922279913,1.0282870270558786,-0.6826289037478287,-0.850673009890769,0.8192643443294739,-0.6784329353268667,-1.0561892284880439,1.0299343479203862,-0.6784329353268667,-0.8520357922279913,0.8179539780418692,-0.6826289037478287,-1.0544999144492708,0.5682229682751976,-0.6826289037478287,-1.2075352740572682,0.8192643443294739,-0.6784329353268667,-1.0561892284880439,0.8179539780418692,-0.6826289037478287,-1.0544999144492708,0.5691332642593958,-0.6784329353268667,-1.2094697514932802,0.8192643443294739,-0.6784329353268667,-1.0561892284880439,0.5682229682751976,-0.6826289037478287,-1.2075352740572682,0.2911226604748394,-0.6826289037478287,-1.3024079147203036,0.5691332642593958,-0.6784329353268667,-1.2094697514932802,0.5682229682751976,-0.6826289037478287,-1.2075352740572682,0.29158904040584394,-0.6784329353268667,-1.304494378592323,0.5691332642593958,-0.6784329353268667,-1.2094697514932802,0.2911226604748394,-0.6826289037478287,-1.3024079147203036,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,0.29158904040584394,-0.6784329353268667,-1.304494378592323,0.2911226604748394,-0.6826289037478287,-1.3024079147203036,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,0.29158904040584394,-0.6784329353268667,-1.304494378592323,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,-0.2911226604748396,-0.6826289037478287,-1.3024079147203036,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,-0.29158904040584405,-0.6784329353268667,-1.304494378592323,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,-0.2911226604748396,-0.6826289037478287,-1.3024079147203036,-0.5682229682751977,-0.6826289037478287,-1.207535274057268,-0.29158904040584405,-0.6784329353268667,-1.304494378592323,-0.2911226604748396,-0.6826289037478287,-1.3024079147203036,-0.5691332642593959,-0.6784329353268667,-1.20946975149328,-0.29158904040584405,-0.6784329353268667,-1.304494378592323,-0.5682229682751977,-0.6826289037478287,-1.207535274057268,-0.8179539780418689,-0.6826289037478287,-1.0544999144492708,-0.5691332642593959,-0.6784329353268667,-1.20946975149328,-0.5682229682751977,-0.6826289037478287,-1.207535274057268,-0.8192643443294736,-0.6784329353268667,-1.0561892284880439,-0.5691332642593959,-0.6784329353268667,-1.20946975149328,-0.8179539780418689,-0.6826289037478287,-1.0544999144492708,-1.0282870270558788,-0.6826289037478287,-0.850673009890769,-0.8192643443294736,-0.6784329353268667,-1.0561892284880439,-0.8179539780418689,-0.6826289037478287,-1.0544999144492708,-1.0299343479203864,-0.6784329353268667,-0.8520357922279911,-0.8192643443294736,-0.6784329353268667,-1.0561892284880439,-1.0282870270558788,-0.6826289037478287,-0.850673009890769,-1.1890911135341198,-0.6826289037478287,-0.6058721841133035,-1.0299343479203864,-0.6784329353268667,-0.8520357922279911,-1.0282870270558788,-0.6826289037478287,-0.850673009890769,-1.1909960433344438,-0.6784329353268667,-0.6068427943260688,-1.0299343479203864,-0.6784329353268667,-0.8520357922279911,-1.1890911135341198,-0.6826289037478287,-0.6058721841133035,1.190996043334444,0.6805317591095806,-0.6044881716005033,1.2815581671094207,0.41820766795280473,-0.6555242259296138,1.18909111353412,0.6805317591095806,-0.608226806838869,1.2836112295834303,0.41820766795280473,-0.651494863949664,1.2815581671094207,0.41820766795280473,-0.6555242259296138,1.190996043334444,0.6805317591095806,-0.6044881716005033,1.2836112295834303,0.41820766795280473,-0.651494863949664,1.3286253837035005,0.14106836166445297,-0.6795993725100353,1.2815581671094207,0.41820766795280473,-0.6555242259296138,1.330753848089515,0.14106836166445297,-0.6754220259454498,1.3286253837035005,0.14106836166445297,-0.6795993725100353,1.2836112295834303,0.41820766795280473,-0.651494863949664,1.330753848089515,0.14106836166445297,-0.6754220259454498,1.3286253837035005,-0.14106836166445297,-0.6795993725100353,1.3286253837035005,0.14106836166445297,-0.6795993725100353,1.330753848089515,-0.14106836166445297,-0.6754220259454498,1.3286253837035005,-0.14106836166445297,-0.6795993725100353,1.330753848089515,0.14106836166445297,-0.6754220259454498,1.330753848089515,-0.14106836166445297,-0.6754220259454498,1.2815581671094207,-0.41820766795280473,-0.6555242259296138,1.3286253837035005,-0.14106836166445297,-0.6795993725100353,1.2836112295834303,-0.41820766795280473,-0.651494863949664,1.2815581671094207,-0.41820766795280473,-0.6555242259296138,1.330753848089515,-0.14106836166445297,-0.6754220259454498,1.2836112295834303,-0.41820766795280473,-0.651494863949664,1.18909111353412,-0.6805317591095806,-0.608226806838869,1.2815581671094207,-0.41820766795280473,-0.6555242259296138,1.190996043334444,-0.6805317591095806,-0.6044881716005033,1.18909111353412,-0.6805317591095806,-0.608226806838869,1.2836112295834303,-0.41820766795280473,-0.651494863949664,1.0304479929540074,0.6805317591095806,-0.8497378764420699,1.1076959172233365,0.41820766795280473,-0.9193003325530936,1.027773382022258,0.6805317591095806,-0.8529709256766902,1.110578513388096,0.41820766795280473,-0.9158158723598121,1.1076959172233365,0.41820766795280473,-0.9193003325530936,1.0304479929540074,0.6805317591095806,-0.8497378764420699,1.110578513388096,0.41820766795280473,-0.9158158723598121,1.1483777723231503,0.14106836166445297,-0.9530630668384051,1.1076959172233365,0.41820766795280473,-0.9193003325530936,1.151366236314687,0.14106836166445297,-0.9494506344259612,1.1483777723231503,0.14106836166445297,-0.9530630668384051,1.110578513388096,0.41820766795280473,-0.9158158723598121,1.151366236314687,0.14106836166445297,-0.9494506344259612,1.1483777723231503,-0.14106836166445297,-0.9530630668384051,1.1483777723231503,0.14106836166445297,-0.9530630668384051,1.151366236314687,-0.14106836166445297,-0.9494506344259612,1.1483777723231503,-0.14106836166445297,-0.9530630668384051,1.151366236314687,0.14106836166445297,-0.9494506344259612,1.151366236314687,-0.14106836166445297,-0.9494506344259612,1.1076959172233365,-0.41820766795280473,-0.9193003325530936,1.1483777723231503,-0.14106836166445297,-0.9530630668384051,1.110578513388096,-0.41820766795280473,-0.9158158723598121,1.1076959172233365,-0.41820766795280473,-0.9193003325530936,1.151366236314687,-0.14106836166445297,-0.9494506344259612,1.110578513388096,-0.41820766795280473,-0.9158158723598121,1.027773382022258,-0.6805317591095806,-0.8529709256766902,1.1076959172233365,-0.41820766795280473,-0.9193003325530936,1.0304479929540074,-0.6805317591095806,-0.8497378764420699,1.027773382022258,-0.6805317591095806,-0.8529709256766902,1.110578513388096,-0.41820766795280473,-0.9158158723598121,0.8202668939254681,0.6805317591095806,-1.0540587021476011,0.880479858389318,0.41820766795280473,-1.1387969816592436,0.8169514284458749,0.6805317591095806,-1.0566304407897136,0.8840531437454869,0.41820766795280473,-1.1360252574213296,0.880479858389318,0.41820766795280473,-1.1387969816592436,0.8202668939254681,0.6805317591095806,-1.0540587021476011,0.8840531437454869,0.41820766795280473,-1.1360252574213296,0.9128168503925818,0.14106836166445297,-1.1806210717146608,0.880479858389318,0.41820766795280473,-1.1387969816592436,0.916521370210149,0.14106836166445297,-1.1777475515938967,0.9128168503925818,0.14106836166445297,-1.1806210717146608,0.8840531437454869,0.41820766795280473,-1.1360252574213296,0.916521370210149,0.14106836166445297,-1.1777475515938967,0.9128168503925818,-0.14106836166445297,-1.1806210717146608,0.9128168503925818,0.14106836166445297,-1.1806210717146608,0.916521370210149,-0.14106836166445297,-1.1777475515938967,0.9128168503925818,-0.14106836166445297,-1.1806210717146608,0.916521370210149,0.14106836166445297,-1.1777475515938967,0.916521370210149,-0.14106836166445297,-1.1777475515938967,0.880479858389318,-0.41820766795280473,-1.1387969816592436,0.9128168503925818,-0.14106836166445297,-1.1806210717146608,0.8840531437454869,-0.41820766795280473,-1.1360252574213296,0.880479858389318,-0.41820766795280473,-1.1387969816592436,0.916521370210149,-0.14106836166445297,-1.1777475515938967,0.8840531437454869,-0.41820766795280473,-1.1360252574213296,0.8169514284458749,-0.6805317591095806,-1.0566304407897136,0.880479858389318,-0.41820766795280473,-1.1387969816592436,0.8202668939254681,-0.6805317591095806,-1.0540587021476011,0.8169514284458749,-0.6805317591095806,-1.0566304407897136,0.8840531437454869,-0.41820766795280473,-1.1360252574213296,0.5705764291365863,0.6805317591095806,-1.207609234544421,0.6108541874798435,0.41820766795280473,-1.3034417931399227,0.5667798033980067,0.6805317591095806,-1.2093957910061275,0.6149460494636351,0.41820766795280473,-1.3015163090467001,0.6108541874798435,0.41820766795280473,-1.3034417931399227,0.5705764291365863,0.6805317591095806,-1.207609234544421,0.6149460494636351,0.41820766795280473,-1.3015163090467001,0.6332887574333583,0.14106836166445297,-1.351312719930446,0.6108541874798435,0.41820766795280473,-1.3034417931399227,0.6375308994116182,0.14106836166445297,-1.3493165194396461,0.6332887574333583,0.14106836166445297,-1.351312719930446,0.6149460494636351,0.41820766795280473,-1.3015163090467001,0.6375308994116182,0.14106836166445297,-1.3493165194396461,0.6332887574333583,-0.14106836166445297,-1.351312719930446,0.6332887574333583,0.14106836166445297,-1.351312719930446,0.6375308994116182,-0.14106836166445297,-1.3493165194396461,0.6332887574333583,-0.14106836166445297,-1.351312719930446,0.6375308994116182,0.14106836166445297,-1.3493165194396461,0.6375308994116182,-0.14106836166445297,-1.3493165194396461,0.6108541874798435,-0.41820766795280473,-1.3034417931399227,0.6332887574333583,-0.14106836166445297,-1.351312719930446,0.6149460494636351,-0.41820766795280473,-1.3015163090467001,0.6108541874798435,-0.41820766795280473,-1.3034417931399227,0.6375308994116182,-0.14106836166445297,-1.3493165194396461,0.6149460494636351,-0.41820766795280473,-1.3015163090467001,0.5667798033980067,-0.6805317591095806,-1.2093957910061275,0.6108541874798435,-0.41820766795280473,-1.3034417931399227,0.5705764291365863,-0.6805317591095806,-1.207609234544421,0.5667798033980067,-0.6805317591095806,-1.2093957910061275,0.6149460494636351,-0.41820766795280473,-1.3015163090467001,0.29340330839763284,0.6805317591095806,-1.30299348558024,0.31180582293478887,0.41820766795280473,-1.4053044066142524,0.2893083924830502,0.6805317591095806,-1.3039088077323864,0.31621917097366403,0.41820766795280473,-1.4043179064493214,0.31180582293478887,0.41820766795280473,-1.4053044066142524,0.29340330839763284,0.6805317591095806,-1.30299348558024,0.31621917097366403,0.41820766795280473,-1.4043179064493214,0.32325737666056986,0.14106836166445297,-1.4569163962876641,0.31180582293478887,0.41820766795280473,-1.4053044066142524,0.32783281176921836,0.14106836166445297,-1.4558936653701038,0.32325737666056986,0.14106836166445297,-1.4569163962876641,0.31621917097366403,0.41820766795280473,-1.4043179064493214,0.32783281176921836,0.14106836166445297,-1.4558936653701038,0.32325737666056986,-0.14106836166445297,-1.4569163962876641,0.32325737666056986,0.14106836166445297,-1.4569163962876641,0.32783281176921836,-0.14106836166445297,-1.4558936653701038,0.32325737666056986,-0.14106836166445297,-1.4569163962876641,0.32783281176921836,0.14106836166445297,-1.4558936653701038,0.32783281176921836,-0.14106836166445297,-1.4558936653701038,0.31180582293478887,-0.41820766795280473,-1.4053044066142524,0.32325737666056986,-0.14106836166445297,-1.4569163962876641,0.31621917097366403,-0.41820766795280473,-1.4043179064493214,0.31180582293478887,-0.41820766795280473,-1.4053044066142524,0.32783281176921836,-0.14106836166445297,-1.4558936653701038,0.31621917097366403,-0.41820766795280473,-1.4043179064493214,0.2893083924830502,-0.6805317591095806,-1.3039088077323864,0.31180582293478887,-0.41820766795280473,-1.4053044066142524,0.29340330839763284,-0.6805317591095806,-1.30299348558024,0.2893083924830502,-0.6805317591095806,-1.3039088077323864,0.31621917097366403,-0.41820766795280473,-1.4043179064493214,2.0979842104809974e-3,0.6805317591095806,-1.3356171320050791,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,-2.097984210480834e-3,0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,2.0979842104809974e-3,0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-2.097984210480834e-3,-0.6805317591095806,-1.3356171320050791,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,2.0979842104809974e-3,-0.6805317591095806,-1.3356171320050791,-2.097984210480834e-3,-0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-0.28930839248305035,0.6805317591095806,-1.3039088077323864,-0.3162191709736642,0.41820766795280473,-1.4043179064493214,-0.29340330839763296,0.6805317591095806,-1.30299348558024,-0.31180582293478903,0.41820766795280473,-1.4053044066142524,-0.3162191709736642,0.41820766795280473,-1.4043179064493214,-0.28930839248305035,0.6805317591095806,-1.3039088077323864,-0.31180582293478903,0.41820766795280473,-1.4053044066142524,-0.3278328117692185,0.14106836166445297,-1.4558936653701038,-0.3162191709736642,0.41820766795280473,-1.4043179064493214,-0.32325737666057003,0.14106836166445297,-1.4569163962876641,-0.3278328117692185,0.14106836166445297,-1.4558936653701038,-0.31180582293478903,0.41820766795280473,-1.4053044066142524,-0.32325737666057003,0.14106836166445297,-1.4569163962876641,-0.3278328117692185,-0.14106836166445297,-1.4558936653701038,-0.3278328117692185,0.14106836166445297,-1.4558936653701038,-0.32325737666057003,-0.14106836166445297,-1.4569163962876641,-0.3278328117692185,-0.14106836166445297,-1.4558936653701038,-0.32325737666057003,0.14106836166445297,-1.4569163962876641,-0.32325737666057003,-0.14106836166445297,-1.4569163962876641,-0.3162191709736642,-0.41820766795280473,-1.4043179064493214,-0.3278328117692185,-0.14106836166445297,-1.4558936653701038,-0.31180582293478903,-0.41820766795280473,-1.4053044066142524,-0.3162191709736642,-0.41820766795280473,-1.4043179064493214,-0.32325737666057003,-0.14106836166445297,-1.4569163962876641,-0.31180582293478903,-0.41820766795280473,-1.4053044066142524,-0.29340330839763296,-0.6805317591095806,-1.30299348558024,-0.3162191709736642,-0.41820766795280473,-1.4043179064493214,-0.28930839248305035,-0.6805317591095806,-1.3039088077323864,-0.29340330839763296,-0.6805317591095806,-1.30299348558024,-0.31180582293478903,-0.41820766795280473,-1.4053044066142524,-0.5667798033980066,0.6805317591095806,-1.2093957910061277,-0.6149460494636353,0.41820766795280473,-1.3015163090467001,-0.5705764291365865,0.6805317591095806,-1.207609234544421,-0.6108541874798434,0.41820766795280473,-1.303441793139923,-0.6149460494636353,0.41820766795280473,-1.3015163090467001,-0.5667798033980066,0.6805317591095806,-1.2093957910061277,-0.6108541874798434,0.41820766795280473,-1.303441793139923,-0.6375308994116184,0.14106836166445297,-1.3493165194396461,-0.6149460494636353,0.41820766795280473,-1.3015163090467001,-0.6332887574333581,0.14106836166445297,-1.3513127199304462,-0.6375308994116184,0.14106836166445297,-1.3493165194396461,-0.6108541874798434,0.41820766795280473,-1.303441793139923,-0.6332887574333581,0.14106836166445297,-1.3513127199304462,-0.6375308994116184,-0.14106836166445297,-1.3493165194396461,-0.6375308994116184,0.14106836166445297,-1.3493165194396461,-0.6332887574333581,-0.14106836166445297,-1.3513127199304462,-0.6375308994116184,-0.14106836166445297,-1.3493165194396461,-0.6332887574333581,0.14106836166445297,-1.3513127199304462,-0.6332887574333581,-0.14106836166445297,-1.3513127199304462,-0.6149460494636353,-0.41820766795280473,-1.3015163090467001,-0.6375308994116184,-0.14106836166445297,-1.3493165194396461,-0.6108541874798434,-0.41820766795280473,-1.303441793139923,-0.6149460494636353,-0.41820766795280473,-1.3015163090467001,-0.6332887574333581,-0.14106836166445297,-1.3513127199304462,-0.6108541874798434,-0.41820766795280473,-1.303441793139923,-0.5705764291365865,-0.6805317591095806,-1.207609234544421,-0.6149460494636353,-0.41820766795280473,-1.3015163090467001,-0.5667798033980066,-0.6805317591095806,-1.2093957910061277,-0.5705764291365865,-0.6805317591095806,-1.207609234544421,-0.6108541874798434,-0.41820766795280473,-1.303441793139923,-0.8169514284458748,0.6805317591095806,-1.0566304407897138,-0.8840531437454872,0.41820766795280473,-1.1360252574213296,-0.8202668939254684,0.6805317591095806,-1.054058702147601,-0.8804798583893177,0.41820766795280473,-1.1387969816592438,-0.8840531437454872,0.41820766795280473,-1.1360252574213296,-0.8169514284458748,0.6805317591095806,-1.0566304407897138,-0.8804798583893177,0.41820766795280473,-1.1387969816592438,-0.9165213702101491,0.14106836166445297,-1.1777475515938964,-0.8840531437454872,0.41820766795280473,-1.1360252574213296,-0.9128168503925816,0.14106836166445297,-1.180621071714661,-0.9165213702101491,0.14106836166445297,-1.1777475515938964,-0.8804798583893177,0.41820766795280473,-1.1387969816592438,-0.9128168503925816,0.14106836166445297,-1.180621071714661,-0.9165213702101491,-0.14106836166445297,-1.1777475515938964,-0.9165213702101491,0.14106836166445297,-1.1777475515938964,-0.9128168503925816,-0.14106836166445297,-1.180621071714661,-0.9165213702101491,-0.14106836166445297,-1.1777475515938964,-0.9128168503925816,0.14106836166445297,-1.180621071714661,-0.9128168503925816,-0.14106836166445297,-1.180621071714661,-0.8840531437454872,-0.41820766795280473,-1.1360252574213296,-0.9165213702101491,-0.14106836166445297,-1.1777475515938964,-0.8804798583893177,-0.41820766795280473,-1.1387969816592438,-0.8840531437454872,-0.41820766795280473,-1.1360252574213296,-0.9128168503925816,-0.14106836166445297,-1.180621071714661,-0.8804798583893177,-0.41820766795280473,-1.1387969816592438,-0.8202668939254684,-0.6805317591095806,-1.054058702147601,-0.8840531437454872,-0.41820766795280473,-1.1360252574213296,-0.8169514284458748,-0.6805317591095806,-1.0566304407897138,-0.8202668939254684,-0.6805317591095806,-1.054058702147601,-0.8804798583893177,-0.41820766795280473,-1.1387969816592438,-1.027773382022258,0.6805317591095806,-0.85297092567669,-1.110578513388096,0.41820766795280473,-0.9158158723598121,-1.0304479929540074,0.6805317591095806,-0.8497378764420699,-1.1076959172233365,0.41820766795280473,-0.9193003325530935,-1.110578513388096,0.41820766795280473,-0.9158158723598121,-1.027773382022258,0.6805317591095806,-0.85297092567669,-1.1076959172233365,0.41820766795280473,-0.9193003325530935,-1.151366236314687,0.14106836166445297,-0.9494506344259612,-1.110578513388096,0.41820766795280473,-0.9158158723598121,-1.1483777723231505,0.14106836166445297,-0.9530630668384049,-1.151366236314687,0.14106836166445297,-0.9494506344259612,-1.1076959172233365,0.41820766795280473,-0.9193003325530935,-1.1483777723231505,0.14106836166445297,-0.9530630668384049,-1.151366236314687,-0.14106836166445297,-0.9494506344259612,-1.151366236314687,0.14106836166445297,-0.9494506344259612,-1.1483777723231505,-0.14106836166445297,-0.9530630668384049,-1.151366236314687,-0.14106836166445297,-0.9494506344259612,-1.1483777723231505,0.14106836166445297,-0.9530630668384049,-1.1483777723231505,-0.14106836166445297,-0.9530630668384049,-1.110578513388096,-0.41820766795280473,-0.9158158723598121,-1.151366236314687,-0.14106836166445297,-0.9494506344259612,-1.1076959172233365,-0.41820766795280473,-0.9193003325530935,-1.110578513388096,-0.41820766795280473,-0.9158158723598121,-1.1483777723231505,-0.14106836166445297,-0.9530630668384049,-1.1076959172233365,-0.41820766795280473,-0.9193003325530935,-1.0304479929540074,-0.6805317591095806,-0.8497378764420699,-1.110578513388096,-0.41820766795280473,-0.9158158723598121,-1.027773382022258,-0.6805317591095806,-0.85297092567669,-1.0304479929540074,-0.6805317591095806,-0.8497378764420699,-1.1076959172233365,-0.41820766795280473,-0.9193003325530935,-1.1890911135341198,0.6805317591095806,-0.6082268068388693,-1.2836112295834303,0.41820766795280473,-0.6514948639496642,-1.1909960433344438,0.6805317591095806,-0.6044881716005035,-1.2815581671094203,0.41820766795280473,-0.6555242259296142,-1.2836112295834303,0.41820766795280473,-0.6514948639496642,-1.1890911135341198,0.6805317591095806,-0.6082268068388693,-1.2815581671094203,0.41820766795280473,-0.6555242259296142,-1.3307538480895147,0.14106836166445297,-0.67542202594545,-1.2836112295834303,0.41820766795280473,-0.6514948639496642,-1.3286253837035,0.14106836166445297,-0.6795993725100358,-1.3307538480895147,0.14106836166445297,-0.67542202594545,-1.2815581671094203,0.41820766795280473,-0.6555242259296142,-1.3286253837035,0.14106836166445297,-0.6795993725100358,-1.3307538480895147,-0.14106836166445297,-0.67542202594545,-1.3307538480895147,0.14106836166445297,-0.67542202594545,-1.3286253837035,-0.14106836166445297,-0.6795993725100358,-1.3307538480895147,-0.14106836166445297,-0.67542202594545,-1.3286253837035,0.14106836166445297,-0.6795993725100358,-1.3286253837035,-0.14106836166445297,-0.6795993725100358,-1.2836112295834303,-0.41820766795280473,-0.6514948639496642,-1.3307538480895147,-0.14106836166445297,-0.67542202594545,-1.2815581671094203,-0.41820766795280473,-0.6555242259296142,-1.2836112295834303,-0.41820766795280473,-0.6514948639496642,-1.3286253837035,-0.14106836166445297,-0.6795993725100358,-1.2815581671094203,-0.41820766795280473,-0.6555242259296142,-1.1909960433344438,-0.6805317591095806,-0.6044881716005035,-1.2836112295834303,-0.41820766795280473,-0.6514948639496642,-1.1890911135341198,-0.6805317591095806,-0.6082268068388693,-1.1909960433344438,-0.6805317591095806,-0.6044881716005035,-1.2815581671094203,-0.41820766795280473,-0.6555242259296142],[1.0297984894301457,0.6809857496093201,-0.8519234003103932,1.283441908612575,0.4184866590588439,-0.65394631538581,1.1908389392193548,0.6809857496093201,-0.6067627457812106,1.1098785026520837,0.4184866590588439,-0.9181713486820071,1.283441908612575,0.4184866590588439,-0.65394631538581,1.0297984894301457,0.6809857496093201,-0.8519234003103932,1.1098785026520837,0.4184866590588439,-0.9181713486820071,1.3305783085426997,0.14116246997777146,-0.677963510747756,1.283441908612575,0.4184866590588439,-0.65394631538581,1.1506405165958318,0.14116246997777146,-0.951892619279009,1.3305783085426997,0.14116246997777146,-0.677963510747756,1.1098785026520837,0.4184866590588439,-0.9181713486820071,1.1506405165958318,0.14116246997777146,-0.951892619279009,1.3305783085426997,-0.14116246997777146,-0.677963510747756,1.3305783085426997,0.14116246997777146,-0.677963510747756,1.1506405165958318,-0.14116246997777146,-0.951892619279009,1.3305783085426997,-0.14116246997777146,-0.677963510747756,1.1506405165958318,0.14116246997777146,-0.951892619279009,1.1506405165958318,-0.14116246997777146,-0.951892619279009,1.283441908612575,-0.4184866590588439,-0.65394631538581,1.3305783085426997,-0.14116246997777146,-0.677963510747756,1.1098785026520837,-0.4184866590588439,-0.9181713486820071,1.283441908612575,-0.4184866590588439,-0.65394631538581,1.1506405165958318,-0.14116246997777146,-0.951892619279009,1.1098785026520837,-0.4184866590588439,-0.9181713486820071,1.1908389392193548,-0.6809857496093201,-0.6067627457812106,1.283441908612575,-0.4184866590588439,-0.65394631538581,1.0297984894301457,-0.6809857496093201,-0.8519234003103932,1.1908389392193548,-0.6809857496093201,-0.6067627457812106,1.1098785026520837,-0.4184866590588439,-0.9181713486820071,0.8191562752888082,0.6809857496093201,-1.056049906720321,1.1098785026520837,0.4184866590588439,-0.9181713486820071,1.0297984894301457,0.6809857496093201,-0.8519234003103932,0.882856160294719,0.4184866590588439,-1.1381713036355436,1.1098785026520837,0.4184866590588439,-0.9181713486820071,0.8191562752888082,0.6809857496093201,-1.056049906720321,0.882856160294719,0.4184866590588439,-1.1381713036355436,1.1506405165958318,0.14116246997777146,-0.951892619279009,1.1098785026520837,0.4184866590588439,-0.9181713486820071,0.915280425680764,0.14116246997777146,-1.1799724146925699,1.1506405165958318,0.14116246997777146,-0.951892619279009,0.882856160294719,0.4184866590588439,-1.1381713036355436,0.915280425680764,0.14116246997777146,-1.1799724146925699,1.1506405165958318,-0.14116246997777146,-0.951892619279009,1.1506405165958318,0.14116246997777146,-0.951892619279009,0.915280425680764,-0.14116246997777146,-1.1799724146925699,1.1506405165958318,-0.14116246997777146,-0.951892619279009,0.915280425680764,0.14116246997777146,-1.1799724146925699,0.915280425680764,-0.14116246997777146,-1.1799724146925699,1.1098785026520837,-0.4184866590588439,-0.9181713486820071,1.1506405165958318,-0.14116246997777146,-0.951892619279009,0.882856160294719,-0.4184866590588439,-1.1381713036355436,1.1098785026520837,-0.4184866590588439,-0.9181713486820071,0.915280425680764,-0.14116246997777146,-1.1799724146925699,0.882856160294719,-0.4184866590588439,-1.1381713036355436,1.0297984894301457,-0.6809857496093201,-0.8519234003103932,1.1098785026520837,-0.4184866590588439,-0.9181713486820071,0.8191562752888082,-0.6809857496093201,-1.056049906720321,1.0297984894301457,-0.6809857496093201,-0.8519234003103932,0.882856160294719,-0.4184866590588439,-1.1381713036355436,0.5690581899731716,0.6809857496093201,-1.2093102105140312,0.882856160294719,0.4184866590588439,-1.1381713036355436,0.8191562752888082,0.6809857496093201,-1.056049906720321,0.6133097477729121,0.4184866590588439,-1.303349557669198,0.882856160294719,0.4184866590588439,-1.1381713036355436,0.5690581899731716,0.6809857496093201,-1.2093102105140312,0.6133097477729121,0.4184866590588439,-1.303349557669198,0.915280425680764,0.14116246997777146,-1.1799724146925699,0.882856160294719,0.4184866590588439,-1.1381713036355436,0.6358345019967471,0.14116246997777146,-1.351217096968627,0.915280425680764,0.14116246997777146,-1.1799724146925699,0.6133097477729121,0.4184866590588439,-1.303349557669198,0.6358345019967471,0.14116246997777146,-1.351217096968627,0.915280425680764,-0.14116246997777146,-1.1799724146925699,0.915280425680764,0.14116246997777146,-1.1799724146925699,0.6358345019967471,-0.14116246997777146,-1.351217096968627,0.915280425680764,-0.14116246997777146,-1.1799724146925699,0.6358345019967471,0.14116246997777146,-1.351217096968627,0.6358345019967471,-0.14116246997777146,-1.351217096968627,0.882856160294719,-0.4184866590588439,-1.1381713036355436,0.915280425680764,-0.14116246997777146,-1.1799724146925699,0.6133097477729121,-0.4184866590588439,-1.303349557669198,0.882856160294719,-0.4184866590588439,-1.1381713036355436,0.6358345019967471,-0.14116246997777146,-1.351217096968627,0.6133097477729121,-0.4184866590588439,-1.303349557669198,0.8191562752888082,-0.6809857496093201,-1.056049906720321,0.882856160294719,-0.4184866590588439,-1.1381713036355436,0.5690581899731716,-0.6809857496093201,-1.2093102105140312,0.8191562752888082,-0.6809857496093201,-1.056049906720321,0.6133097477729121,-0.4184866590588439,-1.303349557669198,0.2915505769378761,0.6809857496093201,-1.3043223029283153,0.6133097477729121,0.4184866590588439,-1.303349557669198,0.5690581899731716,0.6809857496093201,-1.2093102105140312,0.3142223659293013,0.4184866590588439,-1.4057500563541014,0.6133097477729121,0.4184866590588439,-1.303349557669198,0.2915505769378761,0.6809857496093201,-1.3043223029283153,0.3142223659293013,0.4184866590588439,-1.4057500563541014,0.6358345019967471,0.14116246997777146,-1.351217096968627,0.6133097477729121,0.4184866590588439,-1.303349557669198,0.32576267095443867,0.14116246997777146,-1.4573784132072236,0.6358345019967471,0.14116246997777146,-1.351217096968627,0.3142223659293013,0.4184866590588439,-1.4057500563541014,0.32576267095443867,0.14116246997777146,-1.4573784132072236,0.6358345019967471,-0.14116246997777146,-1.351217096968627,0.6358345019967471,0.14116246997777146,-1.351217096968627,0.32576267095443867,-0.14116246997777146,-1.4573784132072236,0.6358345019967471,-0.14116246997777146,-1.351217096968627,0.32576267095443867,0.14116246997777146,-1.4573784132072236,0.32576267095443867,-0.14116246997777146,-1.4573784132072236,0.6133097477729121,-0.4184866590588439,-1.303349557669198,0.6358345019967471,-0.14116246997777146,-1.351217096968627,0.3142223659293013,-0.4184866590588439,-1.4057500563541014,0.6133097477729121,-0.4184866590588439,-1.303349557669198,0.32576267095443867,-0.14116246997777146,-1.4573784132072236,0.3142223659293013,-0.4184866590588439,-1.4057500563541014,0.5690581899731716,-0.6809857496093201,-1.2093102105140312,0.6133097477729121,-0.4184866590588439,-1.303349557669198,0.2915505769378761,-0.6809857496093201,-1.3043223029283153,0.5690581899731716,-0.6809857496093201,-1.2093102105140312,0.3142223659293013,-0.4184866590588439,-1.4057500563541014,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,0.3142223659293013,0.4184866590588439,-1.4057500563541014,0.2915505769378761,0.6809857496093201,-1.3043223029283153,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,0.3142223659293013,0.4184866590588439,-1.4057500563541014,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,0.32576267095443867,0.14116246997777146,-1.4573784132072236,0.3142223659293013,0.4184866590588439,-1.4057500563541014,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,0.32576267095443867,0.14116246997777146,-1.4573784132072236,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,0.32576267095443867,-0.14116246997777146,-1.4573784132072236,0.32576267095443867,0.14116246997777146,-1.4573784132072236,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,0.32576267095443867,-0.14116246997777146,-1.4573784132072236,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,0.3142223659293013,-0.4184866590588439,-1.4057500563541014,0.32576267095443867,-0.14116246997777146,-1.4573784132072236,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,0.3142223659293013,-0.4184866590588439,-1.4057500563541014,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,0.2915505769378761,-0.6809857496093201,-1.3043223029283153,0.3142223659293013,-0.4184866590588439,-1.4057500563541014,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,0.2915505769378761,-0.6809857496093201,-1.3043223029283153,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.2915505769378762,0.6809857496093201,-1.3043223029283153,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,-0.3142223659293014,0.4184866590588439,-1.4057500563541014,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,-0.2915505769378762,0.6809857496093201,-1.3043223029283153,-0.3142223659293014,0.4184866590588439,-1.4057500563541014,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,-0.3257626709544388,0.14116246997777146,-1.4573784132072236,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,-0.3142223659293014,0.4184866590588439,-1.4057500563541014,-0.3257626709544388,0.14116246997777146,-1.4573784132072236,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,-0.3257626709544388,-0.14116246997777146,-1.4573784132072236,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,-0.3257626709544388,0.14116246997777146,-1.4573784132072236,-0.3257626709544388,-0.14116246997777146,-1.4573784132072236,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,-0.3142223659293014,-0.4184866590588439,-1.4057500563541014,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.3257626709544388,-0.14116246997777146,-1.4573784132072236,-0.3142223659293014,-0.4184866590588439,-1.4057500563541014,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.2915505769378762,-0.6809857496093201,-1.3043223029283153,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,-0.3142223659293014,-0.4184866590588439,-1.4057500563541014,-0.5690581899731717,0.6809857496093201,-1.209310210514031,-0.3142223659293014,0.4184866590588439,-1.4057500563541014,-0.2915505769378762,0.6809857496093201,-1.3043223029283153,-0.6133097477729123,0.4184866590588439,-1.3033495576691978,-0.3142223659293014,0.4184866590588439,-1.4057500563541014,-0.5690581899731717,0.6809857496093201,-1.209310210514031,-0.6133097477729123,0.4184866590588439,-1.3033495576691978,-0.3257626709544388,0.14116246997777146,-1.4573784132072236,-0.3142223659293014,0.4184866590588439,-1.4057500563541014,-0.6358345019967472,0.14116246997777146,-1.3512170969686268,-0.3257626709544388,0.14116246997777146,-1.4573784132072236,-0.6133097477729123,0.4184866590588439,-1.3033495576691978,-0.6358345019967472,0.14116246997777146,-1.3512170969686268,-0.3257626709544388,-0.14116246997777146,-1.4573784132072236,-0.3257626709544388,0.14116246997777146,-1.4573784132072236,-0.6358345019967472,-0.14116246997777146,-1.3512170969686268,-0.3257626709544388,-0.14116246997777146,-1.4573784132072236,-0.6358345019967472,0.14116246997777146,-1.3512170969686268,-0.6358345019967472,-0.14116246997777146,-1.3512170969686268,-0.3142223659293014,-0.4184866590588439,-1.4057500563541014,-0.3257626709544388,-0.14116246997777146,-1.4573784132072236,-0.6133097477729123,-0.4184866590588439,-1.3033495576691978,-0.3142223659293014,-0.4184866590588439,-1.4057500563541014,-0.6358345019967472,-0.14116246997777146,-1.3512170969686268,-0.6133097477729123,-0.4184866590588439,-1.3033495576691978,-0.2915505769378762,-0.6809857496093201,-1.3043223029283153,-0.3142223659293014,-0.4184866590588439,-1.4057500563541014,-0.5690581899731717,-0.6809857496093201,-1.209310210514031,-0.2915505769378762,-0.6809857496093201,-1.3043223029283153,-0.6133097477729123,-0.4184866590588439,-1.3033495576691978,-0.8191562752888077,0.6809857496093201,-1.0560499067203213,-0.6133097477729123,0.4184866590588439,-1.3033495576691978,-0.5690581899731717,0.6809857496093201,-1.209310210514031,-0.8828561602947187,0.4184866590588439,-1.1381713036355436,-0.6133097477729123,0.4184866590588439,-1.3033495576691978,-0.8191562752888077,0.6809857496093201,-1.0560499067203213,-0.8828561602947187,0.4184866590588439,-1.1381713036355436,-0.6358345019967472,0.14116246997777146,-1.3512170969686268,-0.6133097477729123,0.4184866590588439,-1.3033495576691978,-0.9152804256807636,0.14116246997777146,-1.17997241469257,-0.6358345019967472,0.14116246997777146,-1.3512170969686268,-0.8828561602947187,0.4184866590588439,-1.1381713036355436,-0.9152804256807636,0.14116246997777146,-1.17997241469257,-0.6358345019967472,-0.14116246997777146,-1.3512170969686268,-0.6358345019967472,0.14116246997777146,-1.3512170969686268,-0.9152804256807636,-0.14116246997777146,-1.17997241469257,-0.6358345019967472,-0.14116246997777146,-1.3512170969686268,-0.9152804256807636,0.14116246997777146,-1.17997241469257,-0.9152804256807636,-0.14116246997777146,-1.17997241469257,-0.6133097477729123,-0.4184866590588439,-1.3033495576691978,-0.6358345019967472,-0.14116246997777146,-1.3512170969686268,-0.8828561602947187,-0.4184866590588439,-1.1381713036355436,-0.6133097477729123,-0.4184866590588439,-1.3033495576691978,-0.9152804256807636,-0.14116246997777146,-1.17997241469257,-0.8828561602947187,-0.4184866590588439,-1.1381713036355436,-0.5690581899731717,-0.6809857496093201,-1.209310210514031,-0.6133097477729123,-0.4184866590588439,-1.3033495576691978,-0.8191562752888077,-0.6809857496093201,-1.0560499067203213,-0.5690581899731717,-0.6809857496093201,-1.209310210514031,-0.8828561602947187,-0.4184866590588439,-1.1381713036355436,-1.029798489430146,0.6809857496093201,-0.851923400310393,-0.8828561602947187,0.4184866590588439,-1.1381713036355436,-0.8191562752888077,0.6809857496093201,-1.0560499067203213,-1.1098785026520837,0.4184866590588439,-0.918171348682007,-0.8828561602947187,0.4184866590588439,-1.1381713036355436,-1.029798489430146,0.6809857496093201,-0.851923400310393,-1.1098785026520837,0.4184866590588439,-0.918171348682007,-0.9152804256807636,0.14116246997777146,-1.17997241469257,-0.8828561602947187,0.4184866590588439,-1.1381713036355436,-1.150640516595832,0.14116246997777146,-0.9518926192790089,-0.9152804256807636,0.14116246997777146,-1.17997241469257,-1.1098785026520837,0.4184866590588439,-0.918171348682007,-1.150640516595832,0.14116246997777146,-0.9518926192790089,-0.9152804256807636,-0.14116246997777146,-1.17997241469257,-0.9152804256807636,0.14116246997777146,-1.17997241469257,-1.150640516595832,-0.14116246997777146,-0.9518926192790089,-0.9152804256807636,-0.14116246997777146,-1.17997241469257,-1.150640516595832,0.14116246997777146,-0.9518926192790089,-1.150640516595832,-0.14116246997777146,-0.9518926192790089,-0.8828561602947187,-0.4184866590588439,-1.1381713036355436,-0.9152804256807636,-0.14116246997777146,-1.17997241469257,-1.1098785026520837,-0.4184866590588439,-0.918171348682007,-0.8828561602947187,-0.4184866590588439,-1.1381713036355436,-1.150640516595832,-0.14116246997777146,-0.9518926192790089,-1.1098785026520837,-0.4184866590588439,-0.918171348682007,-0.8191562752888077,-0.6809857496093201,-1.0560499067203213,-0.8828561602947187,-0.4184866590588439,-1.1381713036355436,-1.029798489430146,-0.6809857496093201,-0.851923400310393,-0.8191562752888077,-0.6809857496093201,-1.0560499067203213,-1.1098785026520837,-0.4184866590588439,-0.918171348682007,-1.1908389392193548,0.6809857496093201,-0.6067627457812107,-1.1098785026520837,0.4184866590588439,-0.918171348682007,-1.029798489430146,0.6809857496093201,-0.851923400310393,-1.283441908612575,0.4184866590588439,-0.65394631538581,-1.1098785026520837,0.4184866590588439,-0.918171348682007,-1.1908389392193548,0.6809857496093201,-0.6067627457812107,-1.283441908612575,0.4184866590588439,-0.65394631538581,-1.150640516595832,0.14116246997777146,-0.9518926192790089,-1.1098785026520837,0.4184866590588439,-0.918171348682007,-1.3305783085426997,0.14116246997777146,-0.677963510747756,-1.150640516595832,0.14116246997777146,-0.9518926192790089,-1.283441908612575,0.4184866590588439,-0.65394631538581,-1.3305783085426997,0.14116246997777146,-0.677963510747756,-1.150640516595832,-0.14116246997777146,-0.9518926192790089,-1.150640516595832,0.14116246997777146,-0.9518926192790089,-1.3305783085426997,-0.14116246997777146,-0.677963510747756,-1.150640516595832,-0.14116246997777146,-0.9518926192790089,-1.3305783085426997,0.14116246997777146,-0.677963510747756,-1.3305783085426997,-0.14116246997777146,-0.677963510747756,-1.1098785026520837,-0.4184866590588439,-0.918171348682007,-1.150640516595832,-0.14116246997777146,-0.9518926192790089,-1.283441908612575,-0.4184866590588439,-0.65394631538581,-1.1098785026520837,-0.4184866590588439,-0.918171348682007,-1.3305783085426997,-0.14116246997777146,-0.677963510747756,-1.283441908612575,-0.4184866590588439,-0.65394631538581,-1.029798489430146,-0.6809857496093201,-0.851923400310393,-1.1098785026520837,-0.4184866590588439,-0.918171348682007,-1.1908389392193548,-0.6809857496093201,-0.6067627457812107,-1.029798489430146,-0.6809857496093201,-0.851923400310393,-1.283441908612575,-0.4184866590588439,-0.65394631538581],[[0.7903572975614961,0.303478973012677,-0.9865271689773473,0.8956611720343065,0.1831716015188576,-0.9242506312181502,0.8796900718140513,0.30347897301267707,-0.9077697342889108,0.804706528061659,0.1831716015188576,-1.004437936406211,0.8956611720343065,0.1831716015188576,-0.9242506312181502,0.7903572975614961,0.303478973012677,-0.9865271689773473,0.804706528061659,0.1831716015188576,-1.004437936406211,0.9036823202842915,6.123838592253551e-2,-0.9325278141134461,0.8956611720343065,0.1831716015188576,-0.9242506312181502,0.8119131264504815,6.123838592253551e-2,-1.0134332416035146,0.9036823202842915,6.123838592253551e-2,-0.9325278141134461,0.804706528061659,0.1831716015188576,-1.004437936406211,0.8119131264504815,6.123838592253551e-2,-1.0134332416035146,0.9036823202842915,-6.123838592253551e-2,-0.9325278141134461,0.9036823202842915,6.123838592253551e-2,-0.9325278141134461,0.8119131264504815,-6.123838592253551e-2,-1.0134332416035146,0.9036823202842915,-6.123838592253551e-2,-0.9325278141134461,0.8119131264504815,6.123838592253551e-2,-1.0134332416035146,0.8119131264504815,-6.123838592253551e-2,-1.0134332416035146,0.8956611720343065,-0.1831716015188576,-0.9242506312181502,0.9036823202842915,-6.123838592253551e-2,-0.9325278141134461,0.804706528061659,-0.1831716015188576,-1.004437936406211,0.8956611720343065,-0.1831716015188576,-0.9242506312181502,0.8119131264504815,-6.123838592253551e-2,-1.0134332416035146,0.804706528061659,-0.1831716015188576,-1.004437936406211,0.8796900718140515,-0.3034789730126769,-0.9077697342889111,0.8956611720343065,-0.1831716015188576,-0.9242506312181502,0.7903572975614961,-0.30347897301267684,-0.9865271689773473,0.8796900718140515,-0.3034789730126769,-0.9077697342889111,0.804706528061659,-0.1831716015188576,-1.004437936406211,0.694009255983357,0.30347897301267707,-1.056528118673894,0.804706528061659,0.1831716015188576,-1.004437936406211,0.7903572975614961,0.303478973012677,-0.9865271689773473,0.7066092519776711,0.1831716015188576,-1.0757097793627113,0.804706528061659,0.1831716015188576,-1.004437936406211,0.694009255983357,0.30347897301267707,-1.056528118673894,0.7066092519776711,0.1831716015188576,-1.0757097793627113,0.8119131264504815,6.123838592253551e-2,-1.0134332416035146,0.804706528061659,0.1831716015188576,-1.004437936406211,0.7129373342278493,6.123838592253552e-2,-1.0853433638962793,0.8119131264504815,6.123838592253551e-2,-1.0134332416035146,0.7066092519776711,0.1831716015188576,-1.0757097793627113,0.7129373342278493,6.123838592253552e-2,-1.0853433638962793,0.8119131264504815,-6.123838592253551e-2,-1.0134332416035146,0.8119131264504815,6.123838592253551e-2,-1.0134332416035146,0.7129373342278493,-6.123838592253552e-2,-1.0853433638962793,0.8119131264504815,-6.123838592253551e-2,-1.0134332416035146,0.7129373342278493,6.123838592253552e-2,-1.0853433638962793,0.7129373342278493,-6.123838592253552e-2,-1.0853433638962793,0.804706528061659,-0.1831716015188576,-1.004437936406211,0.8119131264504815,-6.123838592253551e-2,-1.0134332416035146,0.7066092519776711,-0.1831716015188576,-1.0757097793627113,0.804706528061659,-0.1831716015188576,-1.004437936406211,0.7129373342278493,-6.123838592253552e-2,-1.0853433638962793,0.7066092519776711,-0.1831716015188576,-1.0757097793627113,0.7903572975614961,-0.30347897301267684,-0.9865271689773473,0.804706528061659,-0.1831716015188576,-1.004437936406211,0.6940092559833572,-0.3034789730126769,-1.0565281186738942,0.7903572975614961,-0.30347897301267684,-0.9865271689773473,0.7066092519776711,-0.1831716015188576,-1.0757097793627113,0.5915011391175298,0.30347897301267707,-1.1171512499934086,0.7066092519776711,0.1831716015188576,-1.0757097793627113,0.694009255983357,0.30347897301267707,-1.056528118673894,0.6022400621495474,0.1831716015188576,-1.1374335461639622,0.7066092519776711,0.1831716015188576,-1.0757097793627113,0.5915011391175298,0.30347897301267707,-1.1171512499934086,0.6022400621495474,0.1831716015188576,-1.1374335461639622,0.7129373342278493,6.123838592253552e-2,-1.0853433638962793,0.7066092519776711,0.1831716015188576,-1.0757097793627113,0.6076334597550391,6.123838592253552e-2,-1.147619901655476,0.7129373342278493,6.123838592253552e-2,-1.0853433638962793,0.6022400621495474,0.1831716015188576,-1.1374335461639622,0.6076334597550391,6.123838592253552e-2,-1.147619901655476,0.7129373342278493,-6.123838592253552e-2,-1.0853433638962793,0.7129373342278493,6.123838592253552e-2,-1.0853433638962793,0.6076334597550391,-6.123838592253552e-2,-1.147619901655476,0.7129373342278493,-6.123838592253552e-2,-1.0853433638962793,0.6076334597550391,6.123838592253552e-2,-1.147619901655476,0.6076334597550391,-6.123838592253552e-2,-1.147619901655476,0.7066092519776711,-0.1831716015188576,-1.0757097793627113,0.7129373342278493,-6.123838592253552e-2,-1.0853433638962793,0.6022400621495474,-0.1831716015188576,-1.1374335461639622,0.7066092519776711,-0.1831716015188576,-1.0757097793627113,0.6076334597550391,-6.123838592253552e-2,-1.147619901655476,0.6022400621495474,-0.1831716015188576,-1.1374335461639622,0.6940092559833572,-0.3034789730126769,-1.0565281186738942,0.7066092519776711,-0.1831716015188576,-1.0757097793627113,0.5915011391175299,-0.3034789730126769,-1.1171512499934089,0.6940092559833572,-0.3034789730126769,-1.0565281186738942,0.6022400621495474,-0.1831716015188576,-1.1374335461639622,0.48374281626625765,0.30347897301267707,-1.1678584677305552,0.6022400621495474,0.1831716015188576,-1.1374335461639622,0.5915011391175298,0.30347897301267707,-1.1171512499934086,0.4925253468948975,0.1831716015188576,-1.1890613722861734,0.6022400621495474,0.1831716015188576,-1.1374335461639622,0.48374281626625765,0.30347897301267707,-1.1678584677305552,0.4925253468948975,0.1831716015188576,-1.1890613722861734,0.6076334597550391,6.123838592253552e-2,-1.147619901655476,0.6022400621495474,0.1831716015188576,-1.1374335461639622,0.4969361876767367,6.123838592253551e-2,-1.1997100839231594,0.6076334597550391,6.123838592253552e-2,-1.147619901655476,0.4925253468948975,0.1831716015188576,-1.1890613722861734,0.4969361876767367,6.123838592253551e-2,-1.1997100839231594,0.6076334597550391,-6.123838592253552e-2,-1.147619901655476,0.6076334597550391,6.123838592253552e-2,-1.147619901655476,0.4969361876767367,-6.123838592253551e-2,-1.1997100839231594,0.6076334597550391,-6.123838592253552e-2,-1.147619901655476,0.4969361876767367,6.123838592253551e-2,-1.1997100839231594,0.4969361876767367,-6.123838592253551e-2,-1.1997100839231594,0.6022400621495474,-0.1831716015188576,-1.1374335461639622,0.6076334597550391,-6.123838592253552e-2,-1.147619901655476,0.4925253468948975,-0.1831716015188576,-1.1890613722861734,0.6022400621495474,-0.1831716015188576,-1.1374335461639622,0.4969361876767367,-6.123838592253551e-2,-1.1997100839231594,0.4925253468948975,-0.1831716015188576,-1.1890613722861734,0.5915011391175299,-0.3034789730126769,-1.1171512499934089,0.6022400621495474,-0.1831716015188576,-1.1374335461639622,0.48374281626625776,-0.3034789730126769,-1.1678584677305555,0.5915011391175299,-0.3034789730126769,-1.1171512499934089,0.4925253468948975,-0.1831716015188576,-1.1890613722861734,0.3716907579317957,0.30347897301267707,-1.2081996910309398,0.4925253468948975,0.1831716015188576,-1.1890613722861734,0.48374281626625765,0.30347897301267707,-1.1678584677305552,0.378438941793448,0.1831716015188576,-1.230135005489752,0.4925253468948975,0.1831716015188576,-1.1890613722861734,0.3716907579317957,0.30347897301267707,-1.2081996910309398,0.378438941793448,0.1831716015188576,-1.230135005489752,0.4969361876767367,6.123838592253551e-2,-1.1997100839231594,0.4925253468948975,0.1831716015188576,-1.1890613722861734,0.38182807481659575,6.123838592253551e-2,-1.2411515545538567,0.4969361876767367,6.123838592253551e-2,-1.1997100839231594,0.378438941793448,0.1831716015188576,-1.230135005489752,0.38182807481659575,6.123838592253551e-2,-1.2411515545538567,0.4969361876767367,-6.123838592253551e-2,-1.1997100839231594,0.4969361876767367,6.123838592253551e-2,-1.1997100839231594,0.38182807481659575,-6.123838592253551e-2,-1.2411515545538567,0.4969361876767367,-6.123838592253551e-2,-1.1997100839231594,0.38182807481659575,6.123838592253551e-2,-1.2411515545538567,0.38182807481659575,-6.123838592253551e-2,-1.2411515545538567,0.4925253468948975,-0.1831716015188576,-1.1890613722861734,0.4969361876767367,-6.123838592253551e-2,-1.1997100839231594,0.378438941793448,-0.1831716015188576,-1.230135005489752,0.4925253468948975,-0.1831716015188576,-1.1890613722861734,0.38182807481659575,-6.123838592253551e-2,-1.2411515545538567,0.378438941793448,-0.1831716015188576,-1.230135005489752,0.48374281626625776,-0.3034789730126769,-1.1678584677305555,0.4925253468948975,-0.1831716015188576,-1.1890613722861734,0.3716907579317958,-0.3034789730126769,-1.2081996910309398,0.48374281626625776,-0.3034789730126769,-1.1678584677305555,0.378438941793448,-0.1831716015188576,-1.230135005489752],[0.17811055558800623,0.303478973012677,-1.251469992819365,0.300451362902075,0.1831716015188576,-1.2514699928193653,0.29509382483070545,0.30347897301267707,-1.2291542407224392,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.300451362902075,0.1831716015188576,-1.2514699928193653,0.17811055558800623,0.303478973012677,-1.251469992819365,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.3031420734062201,6.123838592253551e-2,-1.2626776086636606,0.300451362902075,0.1831716015188576,-1.2514699928193653,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.3031420734062201,6.123838592253551e-2,-1.2626776086636606,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.3031420734062201,-6.123838592253551e-2,-1.2626776086636606,0.3031420734062201,6.123838592253551e-2,-1.2626776086636606,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.3031420734062201,-6.123838592253551e-2,-1.2626776086636606,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.300451362902075,-0.1831716015188576,-1.2514699928193653,0.3031420734062201,-6.123838592253551e-2,-1.2626776086636606,0.18134421892549893,-0.1831716015188576,-1.274190895690013,0.300451362902075,-0.1831716015188576,-1.2514699928193653,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.18134421892549893,-0.1831716015188576,-1.274190895690013,0.2950938248307055,-0.3034789730126769,-1.2291542407224392,0.300451362902075,-0.1831716015188576,-1.2514699928193653,0.17811055558800623,-0.30347897301267684,-1.2514699928193653,0.2950938248307055,-0.3034789730126769,-1.2291542407224392,0.18134421892549893,-0.1831716015188576,-1.274190895690013,5.9546364444778024e-2,0.30347897301267707,-1.2626776086636606,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.17811055558800623,0.303478973012677,-1.251469992819365,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,0.18134421892549893,0.1831716015188576,-1.274190895690013,5.9546364444778024e-2,0.30347897301267707,-1.2626776086636606,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.18134421892549893,0.1831716015188576,-1.274190895690013,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,0.18134421892549893,-0.1831716015188576,-1.274190895690013,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,0.18134421892549893,-0.1831716015188576,-1.274190895690013,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,0.17811055558800623,-0.30347897301267684,-1.2514699928193653,0.18134421892549893,-0.1831716015188576,-1.274190895690013,5.954636444477803e-2,-0.3034789730126769,-1.2626776086636609,0.17811055558800623,-0.30347897301267684,-1.2514699928193653,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,-5.954636444477759e-2,0.30347897301267707,-1.2626776086636606,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,5.9546364444778024e-2,0.30347897301267707,-1.2626776086636606,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,-5.954636444477759e-2,0.30347897301267707,-1.2626776086636606,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,5.954636444477803e-2,-0.3034789730126769,-1.2626776086636609,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,-5.95463644447776e-2,-0.3034789730126769,-1.2626776086636609,5.954636444477803e-2,-0.3034789730126769,-1.2626776086636609,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-0.17811055558800643,0.30347897301267707,-1.2514699928193653,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-5.954636444477759e-2,0.30347897301267707,-1.2626776086636606,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-0.17811055558800643,0.30347897301267707,-1.2514699928193653,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-5.95463644447776e-2,-0.3034789730126769,-1.2626776086636609,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-0.17811055558800643,-0.3034789730126769,-1.2514699928193656,-5.95463644447776e-2,-0.3034789730126769,-1.2626776086636609,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.2950938248307048,0.30347897301267707,-1.2291542407224392,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.17811055558800643,0.30347897301267707,-1.2514699928193653,-0.30045136290207436,0.1831716015188576,-1.2514699928193656,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.2950938248307048,0.30347897301267707,-1.2291542407224392,-0.30045136290207436,0.1831716015188576,-1.2514699928193656,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.3031420734062194,6.123838592253551e-2,-1.2626776086636609,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.30045136290207436,0.1831716015188576,-1.2514699928193656,-0.3031420734062194,6.123838592253551e-2,-1.2626776086636609,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.3031420734062194,-6.123838592253551e-2,-1.2626776086636609,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.3031420734062194,6.123838592253551e-2,-1.2626776086636609,-0.3031420734062194,-6.123838592253551e-2,-1.2626776086636609,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.30045136290207436,-0.1831716015188576,-1.2514699928193656,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.3031420734062194,-6.123838592253551e-2,-1.2626776086636609,-0.30045136290207436,-0.1831716015188576,-1.2514699928193656,-0.17811055558800643,-0.3034789730126769,-1.2514699928193656,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.29509382483070484,-0.3034789730126769,-1.2291542407224392,-0.17811055558800643,-0.3034789730126769,-1.2514699928193656,-0.30045136290207436,-0.1831716015188576,-1.2514699928193656],[-0.4837428162662574,0.303478973012677,-1.167858467730555,-0.37843894179344706,0.1831716015188576,-1.2301350054897524,-0.37169075793179485,0.30347897301267707,-1.20819969103094,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.37843894179344706,0.1831716015188576,-1.2301350054897524,-0.4837428162662574,0.303478973012677,-1.167858467730555,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.38182807481659486,6.123838592253551e-2,-1.241151554553857,-0.37843894179344706,0.1831716015188576,-1.2301350054897524,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.38182807481659486,6.123838592253551e-2,-1.241151554553857,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.38182807481659486,-6.123838592253551e-2,-1.241151554553857,-0.38182807481659486,6.123838592253551e-2,-1.241151554553857,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.38182807481659486,-6.123838592253551e-2,-1.241151554553857,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.37843894179344706,-0.1831716015188576,-1.2301350054897524,-0.38182807481659486,-6.123838592253551e-2,-1.241151554553857,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.37843894179344706,-0.1831716015188576,-1.2301350054897524,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.37169075793179485,-0.3034789730126769,-1.20819969103094,-0.37843894179344706,-0.1831716015188576,-1.2301350054897524,-0.48374281626625754,-0.30347897301267684,-1.1678584677305552,-0.37169075793179485,-0.3034789730126769,-1.20819969103094,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.5915011391175291,0.30347897301267707,-1.1171512499934089,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.4837428162662574,0.303478973012677,-1.167858467730555,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.5915011391175291,0.30347897301267707,-1.1171512499934089,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.48374281626625754,-0.30347897301267684,-1.1678584677305552,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.5915011391175292,-0.3034789730126769,-1.117151249993409,-0.48374281626625754,-0.30347897301267684,-1.1678584677305552,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.694009255983356,0.30347897301267707,-1.0565281186738948,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.5915011391175291,0.30347897301267707,-1.1171512499934089,-0.7066092519776702,0.1831716015188576,-1.075709779362712,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.694009255983356,0.30347897301267707,-1.0565281186738948,-0.7066092519776702,0.1831716015188576,-1.075709779362712,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.7129373342278483,6.123838592253552e-2,-1.08534336389628,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.7066092519776702,0.1831716015188576,-1.075709779362712,-0.7129373342278483,6.123838592253552e-2,-1.08534336389628,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.7129373342278483,-6.123838592253552e-2,-1.08534336389628,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.7129373342278483,6.123838592253552e-2,-1.08534336389628,-0.7129373342278483,-6.123838592253552e-2,-1.08534336389628,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.7066092519776702,-0.1831716015188576,-1.075709779362712,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.7129373342278483,-6.123838592253552e-2,-1.08534336389628,-0.7066092519776702,-0.1831716015188576,-1.075709779362712,-0.5915011391175292,-0.3034789730126769,-1.117151249993409,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.6940092559833562,-0.3034789730126769,-1.0565281186738948,-0.5915011391175292,-0.3034789730126769,-1.117151249993409,-0.7066092519776702,-0.1831716015188576,-1.075709779362712,-0.790357297561496,0.30347897301267707,-0.9865271689773474,-0.7066092519776702,0.1831716015188576,-1.075709779362712,-0.694009255983356,0.30347897301267707,-1.0565281186738948,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.7066092519776702,0.1831716015188576,-1.075709779362712,-0.790357297561496,0.30347897301267707,-0.9865271689773474,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.7129373342278483,6.123838592253552e-2,-1.08534336389628,-0.7066092519776702,0.1831716015188576,-1.075709779362712,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.7129373342278483,6.123838592253552e-2,-1.08534336389628,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.7129373342278483,-6.123838592253552e-2,-1.08534336389628,-0.7129373342278483,6.123838592253552e-2,-1.08534336389628,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.7129373342278483,-6.123838592253552e-2,-1.08534336389628,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.7066092519776702,-0.1831716015188576,-1.075709779362712,-0.7129373342278483,-6.123838592253552e-2,-1.08534336389628,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.7066092519776702,-0.1831716015188576,-1.075709779362712,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.6940092559833562,-0.3034789730126769,-1.0565281186738948,-0.7066092519776702,-0.1831716015188576,-1.075709779362712,-0.790357297561496,-0.3034789730126769,-0.9865271689773474,-0.6940092559833562,-0.3034789730126769,-1.0565281186738948,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.8796900718140508,0.30347897301267707,-0.9077697342889117,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.790357297561496,0.30347897301267707,-0.9865271689773474,-0.8956611720343057,0.1831716015188576,-0.9242506312181511,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.8796900718140508,0.30347897301267707,-0.9077697342889117,-0.8956611720343057,0.1831716015188576,-0.9242506312181511,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.9036823202842907,6.123838592253551e-2,-0.932527814113447,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.8956611720343057,0.1831716015188576,-0.9242506312181511,-0.9036823202842907,6.123838592253551e-2,-0.932527814113447,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.9036823202842907,-6.123838592253551e-2,-0.932527814113447,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.9036823202842907,6.123838592253551e-2,-0.932527814113447,-0.9036823202842907,-6.123838592253551e-2,-0.932527814113447,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.8956611720343057,-0.1831716015188576,-0.9242506312181511,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.9036823202842907,-6.123838592253551e-2,-0.932527814113447,-0.8956611720343057,-0.1831716015188576,-0.9242506312181511,-0.790357297561496,-0.3034789730126769,-0.9865271689773474,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.8796900718140508,-0.3034789730126769,-0.9077697342889117,-0.790357297561496,-0.3034789730126769,-0.9865271689773474,-0.8956611720343057,-0.1831716015188576,-0.9242506312181511]]];
      case 4:
        return [[1.2565353837765347,0.6784329353268667,-0.4559041363941858,1.3068199958927385,0.6826289037478287,-0.2706294110087461,1.3089135279404522,0.6784329353268667,-0.2710629606535185,1.2545256275635621,0.6826289037478287,-0.4551749438999128,1.3068199958927385,0.6826289037478287,-0.2706294110087461,1.2565353837765347,0.6784329353268667,-0.4559041363941858,1.1782001064829568,0.6784329353268667,-0.631327380482837,1.2545256275635621,0.6826289037478287,-0.4551749438999128,1.2565353837765347,0.6784329353268667,-0.4559041363941858,1.1763156430490562,0.6826289037478287,-0.6303176085800898,1.2545256275635621,0.6826289037478287,-0.4551749438999128,1.1782001064829568,0.6784329353268667,-0.631327380482837,1.0755259228503937,0.6784329353268667,-0.7937088518723858,1.1763156430490562,0.6826289037478287,-0.6303176085800898,1.1782001064829568,0.6784329353268667,-0.631327380482837,1.0738056808790413,0.6826289037478287,-0.7924393601279133,1.1763156430490562,0.6826289037478287,-0.6303176085800898,1.0755259228503937,0.6784329353268667,-0.7937088518723858,0.9506338455362997,0.6784329353268667,-0.9396941225657628,1.0738056808790413,0.6826289037478287,-0.7924393601279133,1.0755259228503937,0.6784329353268667,-0.7937088518723858,0.9491133612729862,0.6826289037478287,-0.9381911357613287,1.0738056808790413,0.6826289037478287,-0.7924393601279133,0.9506338455362997,0.6784329353268667,-0.9396941225657628,0.806103857818136,0.6784329353268667,-1.0662674723888421,0.9491133612729862,0.6826289037478287,-0.9381911357613287,0.9506338455362997,0.6784329353268667,-0.9396941225657628,0.8048145409732078,0.6826289037478287,-1.0645620387775072,0.9491133612729862,0.6826289037478287,-0.9381911357613287,0.806103857818136,0.6784329353268667,-1.0662674723888421,0.6449216170684412,0.6784329353268667,-1.1708141868382014,0.8048145409732078,0.6826289037478287,-1.0645620387775072,0.806103857818136,0.6784329353268667,-1.0662674723888421,0.6438901019646738,0.6826289037478287,-1.168941536758772,0.8048145409732078,0.6826289037478287,-1.0645620387775072,0.6449216170684412,0.6784329353268667,-1.1708141868382014,0.4704167779354621,0.6784329353268667,-1.2511745710733535,0.6438901019646738,0.6826289037478287,-1.168941536758772,0.6449216170684412,0.6784329353268667,-1.1708141868382014,0.4696643733041648,0.6826289037478287,-1.2491733891725538,0.6438901019646738,0.6826289037478287,-1.168941536758772,0.4704167779354621,0.6784329353268667,-1.2511745710733535,0.2861942093306735,0.6784329353268667,-1.3056885642485203,0.4696643733041648,0.6826289037478287,-1.2491733891725538,0.4704167779354621,0.6784329353268667,-1.2511745710733535,0.28573645812227505,0.6826289037478287,-1.3036001903450978,0.4696643733041648,0.6826289037478287,-1.2491733891725538,0.2861942093306735,0.6784329353268667,-1.3056885642485203,9.605952612215037e-2,0.6784329353268667,-1.3332300325543824,0.28573645812227505,0.6826289037478287,-1.3036001903450978,0.2861942093306735,0.6784329353268667,-1.3056885642485203,9.590588442456523e-2,0.6826289037478287,-1.3310976076534657,0.28573645812227505,0.6826289037478287,-1.3036001903450978,9.605952612215037e-2,0.6784329353268667,-1.3332300325543824,-9.605952612215021e-2,0.6784329353268667,-1.3332300325543827,9.590588442456523e-2,0.6826289037478287,-1.3310976076534657,9.605952612215037e-2,0.6784329353268667,-1.3332300325543824,-9.590588442456507e-2,0.6826289037478287,-1.331097607653466,9.590588442456523e-2,0.6826289037478287,-1.3310976076534657,-9.605952612215021e-2,0.6784329353268667,-1.3332300325543827,-0.2861942093306734,0.6784329353268667,-1.3056885642485203,-9.590588442456507e-2,0.6826289037478287,-1.331097607653466,-9.605952612215021e-2,0.6784329353268667,-1.3332300325543827,-0.28573645812227494,0.6826289037478287,-1.3036001903450978,-9.590588442456507e-2,0.6826289037478287,-1.331097607653466,-0.2861942093306734,0.6784329353268667,-1.3056885642485203,-0.47041677793546194,0.6784329353268667,-1.2511745710733535,-0.28573645812227494,0.6826289037478287,-1.3036001903450978,-0.2861942093306734,0.6784329353268667,-1.3056885642485203,-0.4696643733041646,0.6826289037478287,-1.2491733891725538,-0.28573645812227494,0.6826289037478287,-1.3036001903450978,-0.47041677793546194,0.6784329353268667,-1.2511745710733535,-0.6449216170684412,0.6784329353268667,-1.1708141868382014,-0.4696643733041646,0.6826289037478287,-1.2491733891725538,-0.47041677793546194,0.6784329353268667,-1.2511745710733535,-0.6438901019646738,0.6826289037478287,-1.168941536758772,-0.4696643733041646,0.6826289037478287,-1.2491733891725538,-0.6449216170684412,0.6784329353268667,-1.1708141868382014,-0.8061038578181356,0.6784329353268667,-1.0662674723888423,-0.6438901019646738,0.6826289037478287,-1.168941536758772,-0.6449216170684412,0.6784329353268667,-1.1708141868382014,-0.8048145409732076,0.6826289037478287,-1.0645620387775074,-0.6438901019646738,0.6826289037478287,-1.168941536758772,-0.8061038578181356,0.6784329353268667,-1.0662674723888423,-0.9506338455362997,0.6784329353268667,-0.9396941225657628,-0.8048145409732076,0.6826289037478287,-1.0645620387775074,-0.8061038578181356,0.6784329353268667,-1.0662674723888423,-0.9491133612729862,0.6826289037478287,-0.9381911357613287,-0.8048145409732076,0.6826289037478287,-1.0645620387775074,-0.9506338455362997,0.6784329353268667,-0.9396941225657628,-1.0755259228503935,0.6784329353268667,-0.793708851872386,-0.9491133612729862,0.6826289037478287,-0.9381911357613287,-0.9506338455362997,0.6784329353268667,-0.9396941225657628,-1.073805680879041,0.6826289037478287,-0.7924393601279135,-0.9491133612729862,0.6826289037478287,-0.9381911357613287,-1.0755259228503935,0.6784329353268667,-0.793708851872386,-1.1782001064829568,0.6784329353268667,-0.631327380482837,-1.073805680879041,0.6826289037478287,-0.7924393601279135,-1.0755259228503935,0.6784329353268667,-0.793708851872386,-1.1763156430490562,0.6826289037478287,-0.6303176085800899,-1.073805680879041,0.6826289037478287,-0.7924393601279135,-1.1782001064829568,0.6784329353268667,-0.631327380482837,-1.2565353837765347,0.6784329353268667,-0.4559041363941857,-1.1763156430490562,0.6826289037478287,-0.6303176085800899,-1.1782001064829568,0.6784329353268667,-0.631327380482837,-1.2545256275635621,0.6826289037478287,-0.4551749438999127,-1.1763156430490562,0.6826289037478287,-0.6303176085800899,-1.2565353837765347,0.6784329353268667,-0.4559041363941857,-1.3089135279404522,0.6784329353268667,-0.27106296065351865,-1.2545256275635621,0.6826289037478287,-0.4551749438999127,-1.2565353837765347,0.6784329353268667,-0.4559041363941857,-1.3068199958927385,0.6826289037478287,-0.27062941100874627,-1.2545256275635621,0.6826289037478287,-0.4551749438999127,-1.3089135279404522,0.6784329353268667,-0.27106296065351865,1.3163152466094454,0.5349946639441819,-0.4775938453275618,1.369532710817733,0.5393911233209289,-0.28361658992875993,1.3711852889833047,0.5349946639441819,-0.2839588224144888,1.3147287988455776,0.5393911233209289,-0.4770182402892486,1.369532710817733,0.5393911233209289,-0.28361658992875993,1.3163152466094454,0.5349946639441819,-0.4775938453275618,1.2342531565320416,0.5349946639441819,-0.6613628770515804,1.3147287988455776,0.5393911233209289,-0.4770182402892486,1.3163152466094454,0.5349946639441819,-0.4775938453275618,1.2327656115345413,0.5393911233209289,-0.6605657901378595,1.3147287988455776,0.5393911233209289,-0.4770182402892486,1.2342531565320416,0.5349946639441819,-0.6613628770515804,1.1266942329285368,0.5349946639441819,-0.8314696717480611,1.2327656115345413,0.5393911233209289,-0.6605657901378595,1.2342531565320416,0.5349946639441819,-0.6613628770515804,1.1253363199582236,0.5393911233209289,-0.8304675689426229,1.2327656115345413,0.5393911233209289,-0.6605657901378595,1.1266942329285368,0.5349946639441819,-0.8314696717480611,0.9958603959575721,0.5349946639441819,-0.9844002139955476,1.1253363199582236,0.5393911233209289,-0.8304675689426229,1.1266942329285368,0.5349946639441819,-0.8314696717480611,0.9946601663755165,0.5393911233209289,-0.9832137964392162,1.1253363199582236,0.5393911233209289,-0.8304675689426229,0.9958603959575721,0.5349946639441819,-0.9844002139955476,0.8444543719951552,0.5349946639441819,-1.1169953102719459,0.9946601663755165,0.5393911233209289,-0.9832137964392162,0.9958603959575721,0.5349946639441819,-0.9844002139955476,0.8434366197860315,0.5393911233209289,-1.115649086624688,0.9946601663755165,0.5393911233209289,-0.9832137964392162,0.8444543719951552,0.5349946639441819,-1.1169953102719459,0.6756038615194155,0.5349946639441819,-1.2265158506319054,0.8434366197860315,0.5393911233209289,-1.115649086624688,0.8444543719951552,0.5349946639441819,-1.1169953102719459,0.6747896111047609,0.5393911233209289,-1.2250376307802435,0.8434366197860315,0.5393911233209289,-1.115649086624688,0.6756038615194155,0.5349946639441819,-1.2265158506319054,0.4927969280071316,0.5349946639441819,-1.3106993924229864,0.6747896111047609,0.5393911233209289,-1.2250376307802435,0.6756038615194155,0.5349946639441819,-1.2265158506319054,0.49220299992911876,0.5393911233209289,-1.3091197129916587,0.6747896111047609,0.5393911233209289,-1.2250376307802435,0.4927969280071316,0.5349946639441819,-1.3106993924229864,0.2998099425589259,0.5349946639441819,-1.3678068971510802,0.49220299992911876,0.5393911233209289,-1.3091197129916587,0.4927969280071316,0.5349946639441819,-1.3106993924229864,0.2994486060066197,0.5393911233209289,-1.3661583906865555,0.49220299992911876,0.5393911233209289,-1.3091197129916587,0.2998099425589259,0.5349946639441819,-1.3678068971510802,0.10062957275157164,0.5349946639441819,-1.3966586550188589,0.2994486060066197,0.5393911233209289,-1.3661583906865555,0.2998099425589259,0.5349946639441819,-1.3678068971510802,0.1005082921076819,0.5393911233209289,-1.394975375875927,0.2994486060066197,0.5393911233209289,-1.3661583906865555,0.10062957275157164,0.5349946639441819,-1.3966586550188589,-0.1006295727515715,0.5349946639441819,-1.396658655018859,0.1005082921076819,0.5393911233209289,-1.394975375875927,0.10062957275157164,0.5349946639441819,-1.3966586550188589,-0.10050829210768174,0.5393911233209289,-1.3949753758759271,0.1005082921076819,0.5393911233209289,-1.394975375875927,-0.1006295727515715,0.5349946639441819,-1.396658655018859,-0.29980994255892573,0.5349946639441819,-1.3678068971510802,-0.10050829210768174,0.5393911233209289,-1.3949753758759271,-0.1006295727515715,0.5349946639441819,-1.396658655018859,-0.29944860600661954,0.5393911233209289,-1.3661583906865555,-0.10050829210768174,0.5393911233209289,-1.3949753758759271,-0.29980994255892573,0.5349946639441819,-1.3678068971510802,-0.49279692800713143,0.5349946639441819,-1.3106993924229864,-0.29944860600661954,0.5393911233209289,-1.3661583906865555,-0.29980994255892573,0.5349946639441819,-1.3678068971510802,-0.4922029999291186,0.5393911233209289,-1.3091197129916587,-0.29944860600661954,0.5393911233209289,-1.3661583906865555,-0.49279692800713143,0.5349946639441819,-1.3106993924229864,-0.6756038615194155,0.5349946639441819,-1.2265158506319054,-0.4922029999291186,0.5393911233209289,-1.3091197129916587,-0.49279692800713143,0.5349946639441819,-1.3106993924229864,-0.6747896111047609,0.5393911233209289,-1.2250376307802435,-0.4922029999291186,0.5393911233209289,-1.3091197129916587,-0.6756038615194155,0.5349946639441819,-1.2265158506319054,-0.8444543719951548,0.5349946639441819,-1.1169953102719459,-0.6747896111047609,0.5393911233209289,-1.2250376307802435,-0.6756038615194155,0.5349946639441819,-1.2265158506319054,-0.8434366197860311,0.5393911233209289,-1.1156490866246882,-0.6747896111047609,0.5393911233209289,-1.2250376307802435,-0.8444543719951548,0.5349946639441819,-1.1169953102719459,-0.9958603959575721,0.5349946639441819,-0.9844002139955476,-0.8434366197860311,0.5393911233209289,-1.1156490866246882,-0.8444543719951548,0.5349946639441819,-1.1169953102719459,-0.9946601663755165,0.5393911233209289,-0.9832137964392162,-0.8434366197860311,0.5393911233209289,-1.1156490866246882,-0.9958603959575721,0.5349946639441819,-0.9844002139955476,-1.1266942329285365,0.5349946639441819,-0.8314696717480614,-0.9946601663755165,0.5393911233209289,-0.9832137964392162,-0.9958603959575721,0.5349946639441819,-0.9844002139955476,-1.1253363199582231,0.5393911233209289,-0.830467568942623,-0.9946601663755165,0.5393911233209289,-0.9832137964392162,-1.1266942329285365,0.5349946639441819,-0.8314696717480614,-1.2342531565320416,0.5349946639441819,-0.6613628770515805,-1.1253363199582231,0.5393911233209289,-0.830467568942623,-1.1266942329285365,0.5349946639441819,-0.8314696717480614,-1.2327656115345413,0.5393911233209289,-0.6605657901378597,-1.1253363199582231,0.5393911233209289,-0.830467568942623,-1.2342531565320416,0.5349946639441819,-0.6613628770515805,-1.3163152466094454,0.5349946639441819,-0.4775938453275617,-1.2327656115345413,0.5393911233209289,-0.6605657901378597,-1.2342531565320416,0.5349946639441819,-0.6613628770515805,-1.3147287988455776,0.5393911233209289,-0.4770182402892485,-1.2327656115345413,0.5393911233209289,-0.6605657901378597,-1.3163152466094454,0.5349946639441819,-0.4775938453275617,-1.3711852889833047,0.5349946639441819,-0.28395882241448894,-1.3147287988455776,0.5393911233209289,-0.4770182402892485,-1.3163152466094454,0.5349946639441819,-0.4775938453275617,-1.369532710817733,0.5393911233209289,-0.2836165899287601,-1.3147287988455776,0.5393911233209289,-0.4770182402892485,-1.3711852889833047,0.5349946639441819,-0.28395882241448894,1.3616732841438288,0.3856948790684939,-0.49405093614859175,1.417240538769864,0.39024366087207574,-0.2934964061389272,1.4184340570610876,0.3856948790684939,-0.29374357189488226,1.3605275262827525,0.39024366087207574,-0.49363522501549106,1.417240538769864,0.39024366087207574,-0.2934964061389272,1.3616732841438288,0.3856948790684939,-0.49405093614859175,1.2767834707140835,0.3856948790684939,-0.6841523435402671,1.3605275262827525,0.39024366087207574,-0.49363522501549106,1.3616732841438288,0.3856948790684939,-0.49405093614859175,1.2757091420072653,0.39024366087207574,-0.6835766746666005,1.3605275262827525,0.39024366087207574,-0.49363522501549106,1.2767834707140835,0.3856948790684939,-0.6841523435402671,1.1655182452147888,0.3856948790684939,-0.8601207359038496,1.2757091420072653,0.39024366087207574,-0.6835766746666005,1.2767834707140835,0.3856948790684939,-0.6841523435402671,1.1645375388241788,0.39024366087207574,-0.8593970012854856,1.2757091420072653,0.39024366087207574,-0.6835766746666005,1.1655182452147888,0.3856948790684939,-0.8601207359038496,1.0301760914835478,0.3856948790684939,-1.018321010681809,1.1645375388241788,0.39024366087207574,-0.8593970012854856,1.1655182452147888,0.3856948790684939,-0.8601207359038496,1.0293092665492154,0.39024366087207574,-1.0174641610126012,1.1645375388241788,0.39024366087207574,-0.8593970012854856,1.0301760914835478,0.3856948790684939,-1.018321010681809,0.8735528673591574,0.3856948790684939,-1.155485113789414,1.0293092665492154,0.39024366087207574,-1.0174641610126012,1.0301760914835478,0.3856948790684939,-1.018321010681809,0.8728178304920196,0.39024366087207574,-1.1545128496142278,1.0293092665492154,0.39024366087207574,-1.0174641610126012,0.8735528673591574,0.3856948790684939,-1.155485113789414,0.6988840486843858,0.3856948790684939,-1.2687795501011438,0.8728178304920196,0.39024366087207574,-1.1545128496142278,0.8735528673591574,0.3856948790684939,-1.155485113789414,0.69829598405678,0.39024366087207574,-1.2677119561632815,0.8728178304920196,0.39024366087207574,-1.1545128496142278,0.6988840486843858,0.3856948790684939,-1.2687795501011438,0.5097778918111681,0.3856948790684939,-1.3558639169477524,0.69829598405678,0.39024366087207574,-1.2677119561632815,0.6988840486843858,0.3856948790684939,-1.2687795501011438,0.5093489474867497,0.39024366087207574,-1.3547230472843153,0.69829598405678,0.39024366087207574,-1.2677119561632815,0.5097778918111681,0.3856948790684939,-1.3558639169477524,0.3101408953172387,0.3856948790684939,-1.4149392514564583,0.5093489474867497,0.39024366087207574,-1.3547230472843153,0.5097778918111681,0.3856948790684939,-1.3558639169477524,0.3098799322999064,0.39024366087207574,-1.4137486738126286,0.5093489474867497,0.39024366087207574,-1.3547230472843153,0.3101408953172387,0.3856948790684939,-1.4149392514564583,0.10409710072383482,0.3856948790684939,-1.4447851929893356,0.3098799322999064,0.39024366087207574,-1.4137486738126286,0.3101408953172387,0.3856948790684939,-1.4149392514564583,0.10400950990975451,0.39024366087207574,-1.443569501962927,0.3098799322999064,0.39024366087207574,-1.4137486738126286,0.10409710072383482,0.3856948790684939,-1.4447851929893356,-0.10409710072383466,0.3856948790684939,-1.4447851929893358,0.10400950990975451,0.39024366087207574,-1.443569501962927,0.10409710072383482,0.3856948790684939,-1.4447851929893356,-0.10400950990975434,0.39024366087207574,-1.443569501962927,0.10400950990975451,0.39024366087207574,-1.443569501962927,-0.10409710072383466,0.3856948790684939,-1.4447851929893358,-0.31014089531723854,0.3856948790684939,-1.4149392514564583,-0.10400950990975434,0.39024366087207574,-1.443569501962927,-0.10409710072383466,0.3856948790684939,-1.4447851929893358,-0.3098799322999063,0.39024366087207574,-1.4137486738126286,-0.10400950990975434,0.39024366087207574,-1.443569501962927,-0.31014089531723854,0.3856948790684939,-1.4149392514564583,-0.509777891811168,0.3856948790684939,-1.3558639169477524,-0.3098799322999063,0.39024366087207574,-1.4137486738126286,-0.31014089531723854,0.3856948790684939,-1.4149392514564583,-0.5093489474867496,0.39024366087207574,-1.3547230472843153,-0.3098799322999063,0.39024366087207574,-1.4137486738126286,-0.509777891811168,0.3856948790684939,-1.3558639169477524,-0.6988840486843858,0.3856948790684939,-1.2687795501011438,-0.5093489474867496,0.39024366087207574,-1.3547230472843153,-0.509777891811168,0.3856948790684939,-1.3558639169477524,-0.69829598405678,0.39024366087207574,-1.2677119561632815,-0.5093489474867496,0.39024366087207574,-1.3547230472843153,-0.6988840486843858,0.3856948790684939,-1.2687795501011438,-0.873552867359157,0.3856948790684939,-1.155485113789414,-0.69829598405678,0.39024366087207574,-1.2677119561632815,-0.6988840486843858,0.3856948790684939,-1.2687795501011438,-0.8728178304920192,0.39024366087207574,-1.154512849614228,-0.69829598405678,0.39024366087207574,-1.2677119561632815,-0.873552867359157,0.3856948790684939,-1.155485113789414,-1.0301760914835478,0.3856948790684939,-1.018321010681809,-0.8728178304920192,0.39024366087207574,-1.154512849614228,-0.873552867359157,0.3856948790684939,-1.155485113789414,-1.0293092665492154,0.39024366087207574,-1.0174641610126012,-0.8728178304920192,0.39024366087207574,-1.154512849614228,-1.0301760914835478,0.3856948790684939,-1.018321010681809,-1.1655182452147883,0.3856948790684939,-0.86012073590385,-1.0293092665492154,0.39024366087207574,-1.0174641610126012,-1.0301760914835478,0.3856948790684939,-1.018321010681809,-1.1645375388241785,0.39024366087207574,-0.8593970012854859,-1.0293092665492154,0.39024366087207574,-1.0174641610126012,-1.1655182452147883,0.3856948790684939,-0.86012073590385,-1.2767834707140835,0.3856948790684939,-0.6841523435402672,-1.1645375388241785,0.39024366087207574,-0.8593970012854859,-1.1655182452147883,0.3856948790684939,-0.86012073590385,-1.2757091420072653,0.39024366087207574,-0.6835766746666005,-1.1645375388241785,0.39024366087207574,-0.8593970012854859,-1.2767834707140835,0.3856948790684939,-0.6841523435402672,-1.3616732841438288,0.3856948790684939,-0.49405093614859164,-1.2757091420072653,0.39024366087207574,-0.6835766746666005,-1.2767834707140835,0.3856948790684939,-0.6841523435402672,-1.3605275262827525,0.39024366087207574,-0.49363522501549095,-1.2757091420072653,0.39024366087207574,-0.6835766746666005,-1.3616732841438288,0.3856948790684939,-0.49405093614859164,-1.4184340570610876,0.3856948790684939,-0.29374357189488237,-1.3605275262827525,0.39024366087207574,-0.49363522501549095,-1.3616732841438288,0.3856948790684939,-0.49405093614859164,-1.417240538769864,0.39024366087207574,-0.29349640613892736,-1.3605275262827525,0.39024366087207574,-0.49363522501549095,-1.4184340570610876,0.3856948790684939,-0.29374357189488237,1.3921125442286786,0.23216934038588918,-0.5050951015263726,1.4494207828025831,0.23682060721097192,-0.3001606143053649,1.4501421647832997,0.23216934038588918,-0.300310005331801,1.3914200294332852,0.23682060721097192,-0.504843838916652,1.4494207828025831,0.23682060721097192,-0.3001606143053649,1.3921125442286786,0.23216934038588918,-0.5050951015263726,1.3053250780068635,0.23216934038588918,-0.6994460937850446,1.3914200294332852,0.23682060721097192,-0.504843838916652,1.3921125442286786,0.23216934038588918,-0.5050951015263726,1.3046757361608572,0.23682060721097192,-0.6990981501000787,1.3914200294332852,0.23682060721097192,-0.504843838916652,1.3053250780068635,0.23216934038588918,-0.6994460937850446,1.1915725957060945,0.23216934038588918,-0.8793481372852403,1.3046757361608572,0.23682060721097192,-0.6990981501000787,1.3053250780068635,0.23216934038588918,-0.6994460937850446,1.1909798407196297,0.23682060721097192,-0.8789107002418804,1.3046757361608572,0.23682060721097192,-0.6990981501000787,1.1915725957060945,0.23216934038588918,-0.8793481372852403,1.0532049621730233,0.23216934038588918,-1.0410848692776693,1.1909798407196297,0.23682060721097192,-0.8789107002418804,1.1915725957060945,0.23216934038588918,-0.8793481372852403,1.0526810390017896,0.23682060721097192,-1.0405669753198683,1.1909798407196297,0.23682060721097192,-0.8789107002418804,1.0532049621730233,0.23216934038588918,-1.0410848692776693,0.8930805347056829,0.23216934038588918,-1.1813151805994,1.0526810390017896,0.23682060721097192,-1.0405669753198683,1.0532049621730233,0.23216934038588918,-1.0410848692776693,0.8926362663982638,0.23682060721097192,-1.1807275282260488,1.0526810390017896,0.23682060721097192,-1.0405669753198683,0.8930805347056829,0.23216934038588918,-1.1813151805994,0.7145071159610805,0.23216934038588918,-1.297142235310284,0.8926362663982638,0.23682060721097192,-1.1807275282260488,0.8930805347056829,0.23216934038588918,-1.1813151805994,0.7141516800796439,0.23682060721097192,-1.2964969640688129,0.8926362663982638,0.23682060721097192,-1.1807275282260488,0.7145071159610805,0.23216934038588918,-1.297142235310284,0.5211736223546393,0.23216934038588918,-1.386173312665673,0.7141516800796439,0.23682060721097192,-1.2964969640688129,0.7145071159610805,0.23216934038588918,-1.297142235310284,0.5209143613875964,0.23682060721097192,-1.3854837523769017,0.7141516800796439,0.23682060721097192,-1.2964969640688129,0.5211736223546393,0.23216934038588918,-1.386173312665673,0.31707387952530114,0.23216934038588918,-1.446569235227805,0.5209143613875964,0.23682060721097192,-1.3854837523769017,0.5211736223546393,0.23216934038588918,-1.386173312665673,0.3169161492083707,0.23682060721097192,-1.445849630622481,0.5209143613875964,0.23682060721097192,-1.3854837523769017,0.31707387952530114,0.23216934038588918,-1.446569235227805,0.10642411907684889,0.23216934038588918,-1.4770823620446822,0.3169161492083707,0.23682060721097192,-1.445849630622481,0.31707387952530114,0.23216934038588918,-1.446569235227805,0.10637117775586664,0.23682060721097192,-1.4763475785000821,0.3169161492083707,0.23682060721097192,-1.445849630622481,0.10642411907684889,0.23216934038588918,-1.4770823620446822,-0.10642411907684872,0.23216934038588918,-1.4770823620446825,0.10637117775586664,0.23682060721097192,-1.4763475785000821,0.10642411907684889,0.23216934038588918,-1.4770823620446822,-0.10637117775586648,0.23682060721097192,-1.4763475785000824,0.10637117775586664,0.23682060721097192,-1.4763475785000821,-0.10642411907684872,0.23216934038588918,-1.4770823620446825,-0.317073879525301,0.23216934038588918,-1.446569235227805,-0.10637117775586648,0.23682060721097192,-1.4763475785000824,-0.10642411907684872,0.23216934038588918,-1.4770823620446825,-0.3169161492083705,0.23682060721097192,-1.445849630622481,-0.10637117775586648,0.23682060721097192,-1.4763475785000824,-0.317073879525301,0.23216934038588918,-1.446569235227805,-0.5211736223546392,0.23216934038588918,-1.386173312665673,-0.3169161492083705,0.23682060721097192,-1.445849630622481,-0.317073879525301,0.23216934038588918,-1.446569235227805,-0.5209143613875962,0.23682060721097192,-1.3854837523769017,-0.3169161492083705,0.23682060721097192,-1.445849630622481,-0.5211736223546392,0.23216934038588918,-1.386173312665673,-0.7145071159610805,0.23216934038588918,-1.297142235310284,-0.5209143613875962,0.23682060721097192,-1.3854837523769017,-0.5211736223546392,0.23216934038588918,-1.386173312665673,-0.7141516800796439,0.23682060721097192,-1.2964969640688129,-0.5209143613875962,0.23682060721097192,-1.3854837523769017,-0.7145071159610805,0.23216934038588918,-1.297142235310284,-0.8930805347056826,0.23216934038588918,-1.1813151805994,-0.7141516800796439,0.23682060721097192,-1.2964969640688129,-0.7145071159610805,0.23216934038588918,-1.297142235310284,-0.8926362663982634,0.23682060721097192,-1.180727528226049,-0.7141516800796439,0.23682060721097192,-1.2964969640688129,-0.8930805347056826,0.23216934038588918,-1.1813151805994,-1.0532049621730233,0.23216934038588918,-1.0410848692776693,-0.8926362663982634,0.23682060721097192,-1.180727528226049,-0.8930805347056826,0.23216934038588918,-1.1813151805994,-1.0526810390017896,0.23682060721097192,-1.0405669753198683,-0.8926362663982634,0.23682060721097192,-1.180727528226049,-1.0532049621730233,0.23216934038588918,-1.0410848692776693,-1.191572595706094,0.23216934038588918,-0.8793481372852406,-1.0526810390017896,0.23682060721097192,-1.0405669753198683,-1.0532049621730233,0.23216934038588918,-1.0410848692776693,-1.1909798407196293,0.23682060721097192,-0.8789107002418808,-1.0526810390017896,0.23682060721097192,-1.0405669753198683,-1.191572595706094,0.23216934038588918,-0.8793481372852406,-1.3053250780068635,0.23216934038588918,-0.6994460937850447,-1.1909798407196293,0.23682060721097192,-0.8789107002418808,-1.191572595706094,0.23216934038588918,-0.8793481372852406,-1.3046757361608572,0.23682060721097192,-0.6990981501000788,-1.1909798407196293,0.23682060721097192,-0.8789107002418808,-1.3053250780068635,0.23216934038588918,-0.6994460937850447,-1.3921125442286786,0.23216934038588918,-0.5050951015263725,-1.3046757361608572,0.23682060721097192,-0.6990981501000788,-1.3053250780068635,0.23216934038588918,-0.6994460937850447,-1.3914200294332852,0.23682060721097192,-0.504843838916652,-1.3046757361608572,0.23682060721097192,-0.6990981501000788,-1.3921125442286786,0.23216934038588918,-0.5050951015263725,-1.4501421647832997,0.23216934038588918,-0.3003100053318012,-1.3914200294332852,0.23682060721097192,-0.504843838916652,-1.3921125442286786,0.23216934038588918,-0.5050951015263725,-1.4494207828025831,0.23682060721097192,-0.30016061430536506,-1.3914200294332852,0.23682060721097192,-0.504843838916652,-1.4501421647832997,0.23216934038588918,-0.3003100053318012,1.4072995279606808,7.610010582545855e-2,-0.5106053392738854,1.4657208694281187,8.080289741936658e-2,-0.30353619996882636,1.465962211486389,7.610010582545855e-2,-0.3035861795063959,1.4070678435679864,8.080289741936658e-2,-0.510521278073277,1.4657208694281187,8.080289741936658e-2,-0.30353619996882636,1.4072995279606808,7.610010582545855e-2,-0.5106053392738854,1.3195652705881669,7.610010582545855e-2,-0.7070765662578083,1.4070678435679864,8.080289741936658e-2,-0.510521278073277,1.4072995279606808,7.610010582545855e-2,-0.5106053392738854,1.3193480299281206,8.080289741936658e-2,-0.7069601599053673,1.4070678435679864,8.080289741936658e-2,-0.510521278073277,1.3195652705881669,7.610010582545855e-2,-0.7070765662578083,1.204571827486248,7.610010582545855e-2,-0.888941216459106,1.3193480299281206,8.080289741936658e-2,-0.7069601599053673,1.3195652705881669,7.610010582545855e-2,-0.7070765662578083,1.2043735182516018,8.080289741936658e-2,-0.8887948696425368,1.3193480299281206,8.080289741936658e-2,-0.7069601599053673,1.204571827486248,7.610010582545855e-2,-0.888941216459106,1.064694698899624,7.610010582545855e-2,-1.0524423841847086,1.2043735182516018,8.080289741936658e-2,-0.8887948696425368,1.204571827486248,7.610010582545855e-2,-0.888941216459106,1.0645194177033908,8.080289741936658e-2,-1.0522691200928922,1.2043735182516018,8.080289741936658e-2,-0.8887948696425368,1.064694698899624,7.610010582545855e-2,-1.0524423841847086,0.9028234248248564,7.610010582545855e-2,-1.194202511084645,1.0645194177033908,8.080289741936658e-2,-1.0522691200928922,1.064694698899624,7.610010582545855e-2,-1.0524423841847086,0.902674792573701,8.080289741936658e-2,-1.194005908955505,1.0645194177033908,8.080289741936658e-2,-1.0522691200928922,0.9028234248248564,7.610010582545855e-2,-1.194202511084645,0.7223018937550791,7.610010582545855e-2,-1.3112931587449013,0.902674792573701,8.080289741936658e-2,-1.194005908955505,0.9028234248248564,7.610010582545855e-2,-1.194202511084645,0.7221829808497084,8.080289741936658e-2,-1.3110772799265744,0.902674792573701,8.080289741936658e-2,-1.194005908955505,0.7223018937550791,7.610010582545855e-2,-1.3112931587449013,0.5268592656290011,7.610010582545855e-2,-1.4012955034946142,0.7221829808497084,8.080289741936658e-2,-1.3110772799265744,0.7223018937550791,7.610010582545855e-2,-1.3112931587449013,0.5267725285367422,8.080289741936658e-2,-1.401064807547332,0.7221829808497084,8.080289741936658e-2,-1.3110772799265744,0.5268592656290011,7.610010582545855e-2,-1.4012955034946142,0.32053293595730936,7.610010582545855e-2,-1.4623503037439223,0.5267725285367422,8.080289741936658e-2,-1.401064807547332,0.5268592656290011,7.610010582545855e-2,-1.4012955034946142,0.32048016646714017,8.080289741936658e-2,-1.462109556315747,0.5267725285367422,8.080289741936658e-2,-1.401064807547332,0.32053293595730936,7.610010582545855e-2,-1.4623503037439223,0.10758513251057832,7.610010582545855e-2,-1.4931963076421115,0.32048016646714017,8.080289741936658e-2,-1.462109556315747,0.32053293595730936,7.610010582545855e-2,-1.4623503037439223,0.10756742071888549,8.080289741936658e-2,-1.492950482021598,0.32048016646714017,8.080289741936658e-2,-1.462109556315747,0.10758513251057832,7.610010582545855e-2,-1.4931963076421115,-0.10758513251057815,7.610010582545855e-2,-1.4931963076421118,0.10756742071888549,8.080289741936658e-2,-1.492950482021598,0.10758513251057832,7.610010582545855e-2,-1.4931963076421115,-0.10756742071888532,8.080289741936658e-2,-1.4929504820215982,0.10756742071888549,8.080289741936658e-2,-1.492950482021598,-0.10758513251057815,7.610010582545855e-2,-1.4931963076421118,-0.3205329359573092,7.610010582545855e-2,-1.4623503037439223,-0.10756742071888532,8.080289741936658e-2,-1.4929504820215982,-0.10758513251057815,7.610010582545855e-2,-1.4931963076421118,-0.32048016646714,8.080289741936658e-2,-1.462109556315747,-0.10756742071888532,8.080289741936658e-2,-1.4929504820215982,-0.3205329359573092,7.610010582545855e-2,-1.4623503037439223,-0.526859265629001,7.610010582545855e-2,-1.4012955034946142,-0.32048016646714,8.080289741936658e-2,-1.462109556315747,-0.3205329359573092,7.610010582545855e-2,-1.4623503037439223,-0.526772528536742,8.080289741936658e-2,-1.401064807547332,-0.32048016646714,8.080289741936658e-2,-1.462109556315747,-0.526859265629001,7.610010582545855e-2,-1.4012955034946142,-0.7223018937550791,7.610010582545855e-2,-1.3112931587449013,-0.526772528536742,8.080289741936658e-2,-1.401064807547332,-0.526859265629001,7.610010582545855e-2,-1.4012955034946142,-0.7221829808497084,8.080289741936658e-2,-1.3110772799265744,-0.526772528536742,8.080289741936658e-2,-1.401064807547332,-0.7223018937550791,7.610010582545855e-2,-1.3112931587449013,-0.9028234248248561,7.610010582545855e-2,-1.1942025110846453,-0.7221829808497084,8.080289741936658e-2,-1.3110772799265744,-0.7223018937550791,7.610010582545855e-2,-1.3112931587449013,-0.9026747925737006,8.080289741936658e-2,-1.1940059089555053,-0.7221829808497084,8.080289741936658e-2,-1.3110772799265744,-0.9028234248248561,7.610010582545855e-2,-1.1942025110846453,-1.064694698899624,7.610010582545855e-2,-1.0524423841847086,-0.9026747925737006,8.080289741936658e-2,-1.1940059089555053,-0.9028234248248561,7.610010582545855e-2,-1.1942025110846453,-1.0645194177033908,8.080289741936658e-2,-1.0522691200928922,-0.9026747925737006,8.080289741936658e-2,-1.1940059089555053,-1.064694698899624,7.610010582545855e-2,-1.0524423841847086,-1.2045718274862478,7.610010582545855e-2,-0.8889412164591063,-1.0645194177033908,8.080289741936658e-2,-1.0522691200928922,-1.064694698899624,7.610010582545855e-2,-1.0524423841847086,-1.2043735182516015,8.080289741936658e-2,-0.8887948696425372,-1.0645194177033908,8.080289741936658e-2,-1.0522691200928922,-1.2045718274862478,7.610010582545855e-2,-0.8889412164591063,-1.3195652705881669,7.610010582545855e-2,-0.7070765662578083,-1.2043735182516015,8.080289741936658e-2,-0.8887948696425372,-1.2045718274862478,7.610010582545855e-2,-0.8889412164591063,-1.3193480299281206,8.080289741936658e-2,-0.7069601599053674,-1.2043735182516015,8.080289741936658e-2,-0.8887948696425372,-1.3195652705881669,7.610010582545855e-2,-0.7070765662578083,-1.4072995279606808,7.610010582545855e-2,-0.5106053392738854,-1.3193480299281206,8.080289741936658e-2,-0.7069601599053674,-1.3195652705881669,7.610010582545855e-2,-0.7070765662578083,-1.4070678435679864,8.080289741936658e-2,-0.5105212780732769,-1.3193480299281206,8.080289741936658e-2,-0.7069601599053674,-1.4072995279606808,7.610010582545855e-2,-0.5106053392738854,-1.465962211486389,7.610010582545855e-2,-0.30358617950639605,-1.4070678435679864,8.080289741936658e-2,-0.5105212780732769,-1.4072995279606808,7.610010582545855e-2,-0.5106053392738854,-1.4657208694281187,8.080289741936658e-2,-0.3035361999688265,-1.4070678435679864,8.080289741936658e-2,-0.5105212780732769,-1.465962211486389,7.610010582545855e-2,-0.30358617950639605,1.4070678435679864,-8.080289741936658e-2,-0.510521278073277,1.465962211486389,-7.610010582545855e-2,-0.3035861795063959,1.4657208694281187,-8.080289741936658e-2,-0.30353619996882636,1.4072995279606808,-7.610010582545855e-2,-0.5106053392738854,1.465962211486389,-7.610010582545855e-2,-0.3035861795063959,1.4070678435679864,-8.080289741936658e-2,-0.510521278073277,1.3193480299281206,-8.080289741936658e-2,-0.7069601599053673,1.4072995279606808,-7.610010582545855e-2,-0.5106053392738854,1.4070678435679864,-8.080289741936658e-2,-0.510521278073277,1.3195652705881669,-7.610010582545855e-2,-0.7070765662578083,1.4072995279606808,-7.610010582545855e-2,-0.5106053392738854,1.3193480299281206,-8.080289741936658e-2,-0.7069601599053673,1.2043735182516018,-8.080289741936658e-2,-0.8887948696425368,1.3195652705881669,-7.610010582545855e-2,-0.7070765662578083,1.3193480299281206,-8.080289741936658e-2,-0.7069601599053673,1.204571827486248,-7.610010582545855e-2,-0.888941216459106,1.3195652705881669,-7.610010582545855e-2,-0.7070765662578083,1.2043735182516018,-8.080289741936658e-2,-0.8887948696425368,1.0645194177033908,-8.080289741936658e-2,-1.0522691200928922,1.204571827486248,-7.610010582545855e-2,-0.888941216459106,1.2043735182516018,-8.080289741936658e-2,-0.8887948696425368,1.064694698899624,-7.610010582545855e-2,-1.0524423841847086,1.204571827486248,-7.610010582545855e-2,-0.888941216459106,1.0645194177033908,-8.080289741936658e-2,-1.0522691200928922,0.902674792573701,-8.080289741936658e-2,-1.194005908955505,1.064694698899624,-7.610010582545855e-2,-1.0524423841847086,1.0645194177033908,-8.080289741936658e-2,-1.0522691200928922,0.9028234248248564,-7.610010582545855e-2,-1.194202511084645,1.064694698899624,-7.610010582545855e-2,-1.0524423841847086,0.902674792573701,-8.080289741936658e-2,-1.194005908955505,0.7221829808497084,-8.080289741936658e-2,-1.3110772799265744,0.9028234248248564,-7.610010582545855e-2,-1.194202511084645,0.902674792573701,-8.080289741936658e-2,-1.194005908955505,0.7223018937550791,-7.610010582545855e-2,-1.3112931587449013,0.9028234248248564,-7.610010582545855e-2,-1.194202511084645,0.7221829808497084,-8.080289741936658e-2,-1.3110772799265744,0.5267725285367422,-8.080289741936658e-2,-1.401064807547332,0.7223018937550791,-7.610010582545855e-2,-1.3112931587449013,0.7221829808497084,-8.080289741936658e-2,-1.3110772799265744,0.5268592656290011,-7.610010582545855e-2,-1.4012955034946142,0.7223018937550791,-7.610010582545855e-2,-1.3112931587449013,0.5267725285367422,-8.080289741936658e-2,-1.401064807547332,0.32048016646714017,-8.080289741936658e-2,-1.462109556315747,0.5268592656290011,-7.610010582545855e-2,-1.4012955034946142,0.5267725285367422,-8.080289741936658e-2,-1.401064807547332,0.32053293595730936,-7.610010582545855e-2,-1.4623503037439223,0.5268592656290011,-7.610010582545855e-2,-1.4012955034946142,0.32048016646714017,-8.080289741936658e-2,-1.462109556315747,0.10756742071888549,-8.080289741936658e-2,-1.492950482021598,0.32053293595730936,-7.610010582545855e-2,-1.4623503037439223,0.32048016646714017,-8.080289741936658e-2,-1.462109556315747,0.10758513251057832,-7.610010582545855e-2,-1.4931963076421115,0.32053293595730936,-7.610010582545855e-2,-1.4623503037439223,0.10756742071888549,-8.080289741936658e-2,-1.492950482021598,-0.10756742071888532,-8.080289741936658e-2,-1.4929504820215982,0.10758513251057832,-7.610010582545855e-2,-1.4931963076421115,0.10756742071888549,-8.080289741936658e-2,-1.492950482021598,-0.10758513251057815,-7.610010582545855e-2,-1.4931963076421118,0.10758513251057832,-7.610010582545855e-2,-1.4931963076421115,-0.10756742071888532,-8.080289741936658e-2,-1.4929504820215982,-0.32048016646714,-8.080289741936658e-2,-1.462109556315747,-0.10758513251057815,-7.610010582545855e-2,-1.4931963076421118,-0.10756742071888532,-8.080289741936658e-2,-1.4929504820215982,-0.3205329359573092,-7.610010582545855e-2,-1.4623503037439223,-0.10758513251057815,-7.610010582545855e-2,-1.4931963076421118,-0.32048016646714,-8.080289741936658e-2,-1.462109556315747,-0.526772528536742,-8.080289741936658e-2,-1.401064807547332,-0.3205329359573092,-7.610010582545855e-2,-1.4623503037439223,-0.32048016646714,-8.080289741936658e-2,-1.462109556315747,-0.526859265629001,-7.610010582545855e-2,-1.4012955034946142,-0.3205329359573092,-7.610010582545855e-2,-1.4623503037439223,-0.526772528536742,-8.080289741936658e-2,-1.401064807547332,-0.7221829808497084,-8.080289741936658e-2,-1.3110772799265744,-0.526859265629001,-7.610010582545855e-2,-1.4012955034946142,-0.526772528536742,-8.080289741936658e-2,-1.401064807547332,-0.7223018937550791,-7.610010582545855e-2,-1.3112931587449013,-0.526859265629001,-7.610010582545855e-2,-1.4012955034946142,-0.7221829808497084,-8.080289741936658e-2,-1.3110772799265744,-0.9026747925737006,-8.080289741936658e-2,-1.1940059089555053,-0.7223018937550791,-7.610010582545855e-2,-1.3112931587449013,-0.7221829808497084,-8.080289741936658e-2,-1.3110772799265744,-0.9028234248248561,-7.610010582545855e-2,-1.1942025110846453,-0.7223018937550791,-7.610010582545855e-2,-1.3112931587449013,-0.9026747925737006,-8.080289741936658e-2,-1.1940059089555053,-1.0645194177033908,-8.080289741936658e-2,-1.0522691200928922,-0.9028234248248561,-7.610010582545855e-2,-1.1942025110846453,-0.9026747925737006,-8.080289741936658e-2,-1.1940059089555053,-1.064694698899624,-7.610010582545855e-2,-1.0524423841847086,-0.9028234248248561,-7.610010582545855e-2,-1.1942025110846453,-1.0645194177033908,-8.080289741936658e-2,-1.0522691200928922,-1.2043735182516015,-8.080289741936658e-2,-0.8887948696425372,-1.064694698899624,-7.610010582545855e-2,-1.0524423841847086,-1.0645194177033908,-8.080289741936658e-2,-1.0522691200928922,-1.2045718274862478,-7.610010582545855e-2,-0.8889412164591063,-1.064694698899624,-7.610010582545855e-2,-1.0524423841847086,-1.2043735182516015,-8.080289741936658e-2,-0.8887948696425372,-1.3193480299281206,-8.080289741936658e-2,-0.7069601599053674,-1.2045718274862478,-7.610010582545855e-2,-0.8889412164591063,-1.2043735182516015,-8.080289741936658e-2,-0.8887948696425372,-1.3195652705881669,-7.610010582545855e-2,-0.7070765662578083,-1.2045718274862478,-7.610010582545855e-2,-0.8889412164591063,-1.3193480299281206,-8.080289741936658e-2,-0.7069601599053674,-1.4070678435679864,-8.080289741936658e-2,-0.5105212780732769,-1.3195652705881669,-7.610010582545855e-2,-0.7070765662578083,-1.3193480299281206,-8.080289741936658e-2,-0.7069601599053674,-1.4072995279606808,-7.610010582545855e-2,-0.5106053392738854,-1.3195652705881669,-7.610010582545855e-2,-0.7070765662578083,-1.4070678435679864,-8.080289741936658e-2,-0.5105212780732769,-1.4657208694281187,-8.080289741936658e-2,-0.3035361999688265,-1.4072995279606808,-7.610010582545855e-2,-0.5106053392738854,-1.4070678435679864,-8.080289741936658e-2,-0.5105212780732769,-1.465962211486389,-7.610010582545855e-2,-0.30358617950639605,-1.4072995279606808,-7.610010582545855e-2,-0.5106053392738854,-1.4657208694281187,-8.080289741936658e-2,-0.3035361999688265,1.3914200294332852,-0.23682060721097178,-0.504843838916652,1.4501421647832997,-0.23216934038588904,-0.300310005331801,1.4494207828025831,-0.23682060721097178,-0.3001606143053649,1.3921125442286786,-0.23216934038588904,-0.5050951015263726,1.4501421647832997,-0.23216934038588904,-0.300310005331801,1.3914200294332852,-0.23682060721097178,-0.504843838916652,1.3046757361608572,-0.23682060721097178,-0.6990981501000787,1.3921125442286786,-0.23216934038588904,-0.5050951015263726,1.3914200294332852,-0.23682060721097178,-0.504843838916652,1.3053250780068635,-0.23216934038588904,-0.6994460937850446,1.3921125442286786,-0.23216934038588904,-0.5050951015263726,1.3046757361608572,-0.23682060721097178,-0.6990981501000787,1.1909798407196297,-0.23682060721097178,-0.8789107002418804,1.3053250780068635,-0.23216934038588904,-0.6994460937850446,1.3046757361608572,-0.23682060721097178,-0.6990981501000787,1.1915725957060945,-0.23216934038588904,-0.8793481372852403,1.3053250780068635,-0.23216934038588904,-0.6994460937850446,1.1909798407196297,-0.23682060721097178,-0.8789107002418804,1.0526810390017896,-0.23682060721097178,-1.0405669753198683,1.1915725957060945,-0.23216934038588904,-0.8793481372852403,1.1909798407196297,-0.23682060721097178,-0.8789107002418804,1.0532049621730233,-0.23216934038588904,-1.0410848692776693,1.1915725957060945,-0.23216934038588904,-0.8793481372852403,1.0526810390017896,-0.23682060721097178,-1.0405669753198683,0.8926362663982638,-0.23682060721097178,-1.1807275282260488,1.0532049621730233,-0.23216934038588904,-1.0410848692776693,1.0526810390017896,-0.23682060721097178,-1.0405669753198683,0.8930805347056829,-0.23216934038588904,-1.1813151805994,1.0532049621730233,-0.23216934038588904,-1.0410848692776693,0.8926362663982638,-0.23682060721097178,-1.1807275282260488,0.7141516800796439,-0.23682060721097178,-1.2964969640688129,0.8930805347056829,-0.23216934038588904,-1.1813151805994,0.8926362663982638,-0.23682060721097178,-1.1807275282260488,0.7145071159610805,-0.23216934038588904,-1.297142235310284,0.8930805347056829,-0.23216934038588904,-1.1813151805994,0.7141516800796439,-0.23682060721097178,-1.2964969640688129,0.5209143613875964,-0.23682060721097178,-1.3854837523769017,0.7145071159610805,-0.23216934038588904,-1.297142235310284,0.7141516800796439,-0.23682060721097178,-1.2964969640688129,0.5211736223546393,-0.23216934038588904,-1.386173312665673,0.7145071159610805,-0.23216934038588904,-1.297142235310284,0.5209143613875964,-0.23682060721097178,-1.3854837523769017,0.3169161492083707,-0.23682060721097178,-1.445849630622481,0.5211736223546393,-0.23216934038588904,-1.386173312665673,0.5209143613875964,-0.23682060721097178,-1.3854837523769017,0.31707387952530114,-0.23216934038588904,-1.446569235227805,0.5211736223546393,-0.23216934038588904,-1.386173312665673,0.3169161492083707,-0.23682060721097178,-1.445849630622481,0.10637117775586664,-0.23682060721097178,-1.4763475785000821,0.31707387952530114,-0.23216934038588904,-1.446569235227805,0.3169161492083707,-0.23682060721097178,-1.445849630622481,0.10642411907684889,-0.23216934038588904,-1.4770823620446822,0.31707387952530114,-0.23216934038588904,-1.446569235227805,0.10637117775586664,-0.23682060721097178,-1.4763475785000821,-0.10637117775586648,-0.23682060721097178,-1.4763475785000824,0.10642411907684889,-0.23216934038588904,-1.4770823620446822,0.10637117775586664,-0.23682060721097178,-1.4763475785000821,-0.10642411907684872,-0.23216934038588904,-1.4770823620446825,0.10642411907684889,-0.23216934038588904,-1.4770823620446822,-0.10637117775586648,-0.23682060721097178,-1.4763475785000824,-0.3169161492083705,-0.23682060721097178,-1.445849630622481,-0.10642411907684872,-0.23216934038588904,-1.4770823620446825,-0.10637117775586648,-0.23682060721097178,-1.4763475785000824,-0.317073879525301,-0.23216934038588904,-1.446569235227805,-0.10642411907684872,-0.23216934038588904,-1.4770823620446825,-0.3169161492083705,-0.23682060721097178,-1.445849630622481,-0.5209143613875962,-0.23682060721097178,-1.3854837523769017,-0.317073879525301,-0.23216934038588904,-1.446569235227805,-0.3169161492083705,-0.23682060721097178,-1.445849630622481,-0.5211736223546392,-0.23216934038588904,-1.386173312665673,-0.317073879525301,-0.23216934038588904,-1.446569235227805,-0.5209143613875962,-0.23682060721097178,-1.3854837523769017,-0.7141516800796439,-0.23682060721097178,-1.2964969640688129,-0.5211736223546392,-0.23216934038588904,-1.386173312665673,-0.5209143613875962,-0.23682060721097178,-1.3854837523769017,-0.7145071159610805,-0.23216934038588904,-1.297142235310284,-0.5211736223546392,-0.23216934038588904,-1.386173312665673,-0.7141516800796439,-0.23682060721097178,-1.2964969640688129,-0.8926362663982634,-0.23682060721097178,-1.180727528226049,-0.7145071159610805,-0.23216934038588904,-1.297142235310284,-0.7141516800796439,-0.23682060721097178,-1.2964969640688129,-0.8930805347056826,-0.23216934038588904,-1.1813151805994,-0.7145071159610805,-0.23216934038588904,-1.297142235310284,-0.8926362663982634,-0.23682060721097178,-1.180727528226049,-1.0526810390017896,-0.23682060721097178,-1.0405669753198683,-0.8930805347056826,-0.23216934038588904,-1.1813151805994,-0.8926362663982634,-0.23682060721097178,-1.180727528226049,-1.0532049621730233,-0.23216934038588904,-1.0410848692776693,-0.8930805347056826,-0.23216934038588904,-1.1813151805994,-1.0526810390017896,-0.23682060721097178,-1.0405669753198683,-1.1909798407196293,-0.23682060721097178,-0.8789107002418808,-1.0532049621730233,-0.23216934038588904,-1.0410848692776693,-1.0526810390017896,-0.23682060721097178,-1.0405669753198683,-1.191572595706094,-0.23216934038588904,-0.8793481372852406,-1.0532049621730233,-0.23216934038588904,-1.0410848692776693,-1.1909798407196293,-0.23682060721097178,-0.8789107002418808,-1.3046757361608572,-0.23682060721097178,-0.6990981501000788,-1.191572595706094,-0.23216934038588904,-0.8793481372852406,-1.1909798407196293,-0.23682060721097178,-0.8789107002418808,-1.3053250780068635,-0.23216934038588904,-0.6994460937850447,-1.191572595706094,-0.23216934038588904,-0.8793481372852406,-1.3046757361608572,-0.23682060721097178,-0.6990981501000788,-1.3914200294332852,-0.23682060721097178,-0.504843838916652,-1.3053250780068635,-0.23216934038588904,-0.6994460937850447,-1.3046757361608572,-0.23682060721097178,-0.6990981501000788,-1.3921125442286786,-0.23216934038588904,-0.5050951015263725,-1.3053250780068635,-0.23216934038588904,-0.6994460937850447,-1.3914200294332852,-0.23682060721097178,-0.504843838916652,-1.4494207828025831,-0.23682060721097178,-0.30016061430536506,-1.3921125442286786,-0.23216934038588904,-0.5050951015263725,-1.3914200294332852,-0.23682060721097178,-0.504843838916652,-1.4501421647832997,-0.23216934038588904,-0.3003100053318012,-1.3921125442286786,-0.23216934038588904,-0.5050951015263725,-1.4494207828025831,-0.23682060721097178,-0.30016061430536506,1.3605275262827525,-0.3902436608720756,-0.49363522501549106,1.4184340570610876,-0.38569487906849387,-0.29374357189488226,1.417240538769864,-0.3902436608720756,-0.2934964061389272,1.3616732841438288,-0.38569487906849387,-0.49405093614859175,1.4184340570610876,-0.38569487906849387,-0.29374357189488226,1.3605275262827525,-0.3902436608720756,-0.49363522501549106,1.2757091420072653,-0.3902436608720756,-0.6835766746666005,1.3616732841438288,-0.38569487906849387,-0.49405093614859175,1.3605275262827525,-0.3902436608720756,-0.49363522501549106,1.2767834707140835,-0.38569487906849387,-0.6841523435402671,1.3616732841438288,-0.38569487906849387,-0.49405093614859175,1.2757091420072653,-0.3902436608720756,-0.6835766746666005,1.1645375388241788,-0.3902436608720756,-0.8593970012854856,1.2767834707140835,-0.38569487906849387,-0.6841523435402671,1.2757091420072653,-0.3902436608720756,-0.6835766746666005,1.1655182452147888,-0.38569487906849387,-0.8601207359038496,1.2767834707140835,-0.38569487906849387,-0.6841523435402671,1.1645375388241788,-0.3902436608720756,-0.8593970012854856,1.0293092665492154,-0.3902436608720756,-1.0174641610126012,1.1655182452147888,-0.38569487906849387,-0.8601207359038496,1.1645375388241788,-0.3902436608720756,-0.8593970012854856,1.0301760914835478,-0.38569487906849387,-1.018321010681809,1.1655182452147888,-0.38569487906849387,-0.8601207359038496,1.0293092665492154,-0.3902436608720756,-1.0174641610126012,0.8728178304920196,-0.3902436608720756,-1.1545128496142278,1.0301760914835478,-0.38569487906849387,-1.018321010681809,1.0293092665492154,-0.3902436608720756,-1.0174641610126012,0.8735528673591574,-0.38569487906849387,-1.155485113789414,1.0301760914835478,-0.38569487906849387,-1.018321010681809,0.8728178304920196,-0.3902436608720756,-1.1545128496142278,0.69829598405678,-0.3902436608720756,-1.2677119561632815,0.8735528673591574,-0.38569487906849387,-1.155485113789414,0.8728178304920196,-0.3902436608720756,-1.1545128496142278,0.6988840486843858,-0.38569487906849387,-1.2687795501011438,0.8735528673591574,-0.38569487906849387,-1.155485113789414,0.69829598405678,-0.3902436608720756,-1.2677119561632815,0.5093489474867497,-0.3902436608720756,-1.3547230472843153,0.6988840486843858,-0.38569487906849387,-1.2687795501011438,0.69829598405678,-0.3902436608720756,-1.2677119561632815,0.5097778918111681,-0.38569487906849387,-1.3558639169477524,0.6988840486843858,-0.38569487906849387,-1.2687795501011438,0.5093489474867497,-0.3902436608720756,-1.3547230472843153,0.3098799322999064,-0.3902436608720756,-1.4137486738126286,0.5097778918111681,-0.38569487906849387,-1.3558639169477524,0.5093489474867497,-0.3902436608720756,-1.3547230472843153,0.3101408953172387,-0.38569487906849387,-1.4149392514564583,0.5097778918111681,-0.38569487906849387,-1.3558639169477524,0.3098799322999064,-0.3902436608720756,-1.4137486738126286,0.10400950990975451,-0.3902436608720756,-1.443569501962927,0.3101408953172387,-0.38569487906849387,-1.4149392514564583,0.3098799322999064,-0.3902436608720756,-1.4137486738126286,0.10409710072383482,-0.38569487906849387,-1.4447851929893356,0.3101408953172387,-0.38569487906849387,-1.4149392514564583,0.10400950990975451,-0.3902436608720756,-1.443569501962927,-0.10400950990975434,-0.3902436608720756,-1.443569501962927,0.10409710072383482,-0.38569487906849387,-1.4447851929893356,0.10400950990975451,-0.3902436608720756,-1.443569501962927,-0.10409710072383466,-0.38569487906849387,-1.4447851929893358,0.10409710072383482,-0.38569487906849387,-1.4447851929893356,-0.10400950990975434,-0.3902436608720756,-1.443569501962927,-0.3098799322999063,-0.3902436608720756,-1.4137486738126286,-0.10409710072383466,-0.38569487906849387,-1.4447851929893358,-0.10400950990975434,-0.3902436608720756,-1.443569501962927,-0.31014089531723854,-0.38569487906849387,-1.4149392514564583,-0.10409710072383466,-0.38569487906849387,-1.4447851929893358,-0.3098799322999063,-0.3902436608720756,-1.4137486738126286,-0.5093489474867496,-0.3902436608720756,-1.3547230472843153,-0.31014089531723854,-0.38569487906849387,-1.4149392514564583,-0.3098799322999063,-0.3902436608720756,-1.4137486738126286,-0.509777891811168,-0.38569487906849387,-1.3558639169477524,-0.31014089531723854,-0.38569487906849387,-1.4149392514564583,-0.5093489474867496,-0.3902436608720756,-1.3547230472843153,-0.69829598405678,-0.3902436608720756,-1.2677119561632815,-0.509777891811168,-0.38569487906849387,-1.3558639169477524,-0.5093489474867496,-0.3902436608720756,-1.3547230472843153,-0.6988840486843858,-0.38569487906849387,-1.2687795501011438,-0.509777891811168,-0.38569487906849387,-1.3558639169477524,-0.69829598405678,-0.3902436608720756,-1.2677119561632815,-0.8728178304920192,-0.3902436608720756,-1.154512849614228,-0.6988840486843858,-0.38569487906849387,-1.2687795501011438,-0.69829598405678,-0.3902436608720756,-1.2677119561632815,-0.873552867359157,-0.38569487906849387,-1.155485113789414,-0.6988840486843858,-0.38569487906849387,-1.2687795501011438,-0.8728178304920192,-0.3902436608720756,-1.154512849614228,-1.0293092665492154,-0.3902436608720756,-1.0174641610126012,-0.873552867359157,-0.38569487906849387,-1.155485113789414,-0.8728178304920192,-0.3902436608720756,-1.154512849614228,-1.0301760914835478,-0.38569487906849387,-1.018321010681809,-0.873552867359157,-0.38569487906849387,-1.155485113789414,-1.0293092665492154,-0.3902436608720756,-1.0174641610126012,-1.1645375388241785,-0.3902436608720756,-0.8593970012854859,-1.0301760914835478,-0.38569487906849387,-1.018321010681809,-1.0293092665492154,-0.3902436608720756,-1.0174641610126012,-1.1655182452147883,-0.38569487906849387,-0.86012073590385,-1.0301760914835478,-0.38569487906849387,-1.018321010681809,-1.1645375388241785,-0.3902436608720756,-0.8593970012854859,-1.2757091420072653,-0.3902436608720756,-0.6835766746666005,-1.1655182452147883,-0.38569487906849387,-0.86012073590385,-1.1645375388241785,-0.3902436608720756,-0.8593970012854859,-1.2767834707140835,-0.38569487906849387,-0.6841523435402672,-1.1655182452147883,-0.38569487906849387,-0.86012073590385,-1.2757091420072653,-0.3902436608720756,-0.6835766746666005,-1.3605275262827525,-0.3902436608720756,-0.49363522501549095,-1.2767834707140835,-0.38569487906849387,-0.6841523435402672,-1.2757091420072653,-0.3902436608720756,-0.6835766746666005,-1.3616732841438288,-0.38569487906849387,-0.49405093614859164,-1.2767834707140835,-0.38569487906849387,-0.6841523435402672,-1.3605275262827525,-0.3902436608720756,-0.49363522501549095,-1.417240538769864,-0.3902436608720756,-0.29349640613892736,-1.3616732841438288,-0.38569487906849387,-0.49405093614859164,-1.3605275262827525,-0.3902436608720756,-0.49363522501549095,-1.4184340570610876,-0.38569487906849387,-0.29374357189488237,-1.3616732841438288,-0.38569487906849387,-0.49405093614859164,-1.417240538769864,-0.3902436608720756,-0.29349640613892736,1.3147287988455776,-0.5393911233209289,-0.4770182402892486,1.3711852889833047,-0.5349946639441819,-0.2839588224144888,1.369532710817733,-0.5393911233209289,-0.28361658992875993,1.3163152466094454,-0.5349946639441819,-0.4775938453275618,1.3711852889833047,-0.5349946639441819,-0.2839588224144888,1.3147287988455776,-0.5393911233209289,-0.4770182402892486,1.2327656115345413,-0.5393911233209289,-0.6605657901378595,1.3163152466094454,-0.5349946639441819,-0.4775938453275618,1.3147287988455776,-0.5393911233209289,-0.4770182402892486,1.2342531565320416,-0.5349946639441819,-0.6613628770515804,1.3163152466094454,-0.5349946639441819,-0.4775938453275618,1.2327656115345413,-0.5393911233209289,-0.6605657901378595,1.1253363199582236,-0.5393911233209289,-0.8304675689426229,1.2342531565320416,-0.5349946639441819,-0.6613628770515804,1.2327656115345413,-0.5393911233209289,-0.6605657901378595,1.1266942329285368,-0.5349946639441819,-0.8314696717480611,1.2342531565320416,-0.5349946639441819,-0.6613628770515804,1.1253363199582236,-0.5393911233209289,-0.8304675689426229,0.9946601663755165,-0.5393911233209289,-0.9832137964392162,1.1266942329285368,-0.5349946639441819,-0.8314696717480611,1.1253363199582236,-0.5393911233209289,-0.8304675689426229,0.9958603959575721,-0.5349946639441819,-0.9844002139955476,1.1266942329285368,-0.5349946639441819,-0.8314696717480611,0.9946601663755165,-0.5393911233209289,-0.9832137964392162,0.8434366197860315,-0.5393911233209289,-1.115649086624688,0.9958603959575721,-0.5349946639441819,-0.9844002139955476,0.9946601663755165,-0.5393911233209289,-0.9832137964392162,0.8444543719951552,-0.5349946639441819,-1.1169953102719459,0.9958603959575721,-0.5349946639441819,-0.9844002139955476,0.8434366197860315,-0.5393911233209289,-1.115649086624688,0.6747896111047609,-0.5393911233209289,-1.2250376307802435,0.8444543719951552,-0.5349946639441819,-1.1169953102719459,0.8434366197860315,-0.5393911233209289,-1.115649086624688,0.6756038615194155,-0.5349946639441819,-1.2265158506319054,0.8444543719951552,-0.5349946639441819,-1.1169953102719459,0.6747896111047609,-0.5393911233209289,-1.2250376307802435,0.49220299992911876,-0.5393911233209289,-1.3091197129916587,0.6756038615194155,-0.5349946639441819,-1.2265158506319054,0.6747896111047609,-0.5393911233209289,-1.2250376307802435,0.4927969280071316,-0.5349946639441819,-1.3106993924229864,0.6756038615194155,-0.5349946639441819,-1.2265158506319054,0.49220299992911876,-0.5393911233209289,-1.3091197129916587,0.2994486060066197,-0.5393911233209289,-1.3661583906865555,0.4927969280071316,-0.5349946639441819,-1.3106993924229864,0.49220299992911876,-0.5393911233209289,-1.3091197129916587,0.2998099425589259,-0.5349946639441819,-1.3678068971510802,0.4927969280071316,-0.5349946639441819,-1.3106993924229864,0.2994486060066197,-0.5393911233209289,-1.3661583906865555,0.1005082921076819,-0.5393911233209289,-1.394975375875927,0.2998099425589259,-0.5349946639441819,-1.3678068971510802,0.2994486060066197,-0.5393911233209289,-1.3661583906865555,0.10062957275157164,-0.5349946639441819,-1.3966586550188589,0.2998099425589259,-0.5349946639441819,-1.3678068971510802,0.1005082921076819,-0.5393911233209289,-1.394975375875927,-0.10050829210768174,-0.5393911233209289,-1.3949753758759271,0.10062957275157164,-0.5349946639441819,-1.3966586550188589,0.1005082921076819,-0.5393911233209289,-1.394975375875927,-0.1006295727515715,-0.5349946639441819,-1.396658655018859,0.10062957275157164,-0.5349946639441819,-1.3966586550188589,-0.10050829210768174,-0.5393911233209289,-1.3949753758759271,-0.29944860600661954,-0.5393911233209289,-1.3661583906865555,-0.1006295727515715,-0.5349946639441819,-1.396658655018859,-0.10050829210768174,-0.5393911233209289,-1.3949753758759271,-0.29980994255892573,-0.5349946639441819,-1.3678068971510802,-0.1006295727515715,-0.5349946639441819,-1.396658655018859,-0.29944860600661954,-0.5393911233209289,-1.3661583906865555,-0.4922029999291186,-0.5393911233209289,-1.3091197129916587,-0.29980994255892573,-0.5349946639441819,-1.3678068971510802,-0.29944860600661954,-0.5393911233209289,-1.3661583906865555,-0.49279692800713143,-0.5349946639441819,-1.3106993924229864,-0.29980994255892573,-0.5349946639441819,-1.3678068971510802,-0.4922029999291186,-0.5393911233209289,-1.3091197129916587,-0.6747896111047609,-0.5393911233209289,-1.2250376307802435,-0.49279692800713143,-0.5349946639441819,-1.3106993924229864,-0.4922029999291186,-0.5393911233209289,-1.3091197129916587,-0.6756038615194155,-0.5349946639441819,-1.2265158506319054,-0.49279692800713143,-0.5349946639441819,-1.3106993924229864,-0.6747896111047609,-0.5393911233209289,-1.2250376307802435,-0.8434366197860311,-0.5393911233209289,-1.1156490866246882,-0.6756038615194155,-0.5349946639441819,-1.2265158506319054,-0.6747896111047609,-0.5393911233209289,-1.2250376307802435,-0.8444543719951548,-0.5349946639441819,-1.1169953102719459,-0.6756038615194155,-0.5349946639441819,-1.2265158506319054,-0.8434366197860311,-0.5393911233209289,-1.1156490866246882,-0.9946601663755165,-0.5393911233209289,-0.9832137964392162,-0.8444543719951548,-0.5349946639441819,-1.1169953102719459,-0.8434366197860311,-0.5393911233209289,-1.1156490866246882,-0.9958603959575721,-0.5349946639441819,-0.9844002139955476,-0.8444543719951548,-0.5349946639441819,-1.1169953102719459,-0.9946601663755165,-0.5393911233209289,-0.9832137964392162,-1.1253363199582231,-0.5393911233209289,-0.830467568942623,-0.9958603959575721,-0.5349946639441819,-0.9844002139955476,-0.9946601663755165,-0.5393911233209289,-0.9832137964392162,-1.1266942329285365,-0.5349946639441819,-0.8314696717480614,-0.9958603959575721,-0.5349946639441819,-0.9844002139955476,-1.1253363199582231,-0.5393911233209289,-0.830467568942623,-1.2327656115345413,-0.5393911233209289,-0.6605657901378597,-1.1266942329285365,-0.5349946639441819,-0.8314696717480614,-1.1253363199582231,-0.5393911233209289,-0.830467568942623,-1.2342531565320416,-0.5349946639441819,-0.6613628770515805,-1.1266942329285365,-0.5349946639441819,-0.8314696717480614,-1.2327656115345413,-0.5393911233209289,-0.6605657901378597,-1.3147287988455776,-0.5393911233209289,-0.4770182402892485,-1.2342531565320416,-0.5349946639441819,-0.6613628770515805,-1.2327656115345413,-0.5393911233209289,-0.6605657901378597,-1.3163152466094454,-0.5349946639441819,-0.4775938453275617,-1.2342531565320416,-0.5349946639441819,-0.6613628770515805,-1.3147287988455776,-0.5393911233209289,-0.4770182402892485,-1.369532710817733,-0.5393911233209289,-0.2836165899287601,-1.3163152466094454,-0.5349946639441819,-0.4775938453275617,-1.3147287988455776,-0.5393911233209289,-0.4770182402892485,-1.3711852889833047,-0.5349946639441819,-0.28395882241448894,-1.3163152466094454,-0.5349946639441819,-0.4775938453275617,-1.369532710817733,-0.5393911233209289,-0.2836165899287601,1.2545256275635621,-0.6826289037478287,-0.4551749438999128,1.3089135279404522,-0.6784329353268667,-0.2710629606535185,1.3068199958927385,-0.6826289037478287,-0.2706294110087461,1.2565353837765347,-0.6784329353268667,-0.4559041363941858,1.3089135279404522,-0.6784329353268667,-0.2710629606535185,1.2545256275635621,-0.6826289037478287,-0.4551749438999128,1.1763156430490562,-0.6826289037478287,-0.6303176085800898,1.2565353837765347,-0.6784329353268667,-0.4559041363941858,1.2545256275635621,-0.6826289037478287,-0.4551749438999128,1.1782001064829568,-0.6784329353268667,-0.631327380482837,1.2565353837765347,-0.6784329353268667,-0.4559041363941858,1.1763156430490562,-0.6826289037478287,-0.6303176085800898,1.0738056808790413,-0.6826289037478287,-0.7924393601279133,1.1782001064829568,-0.6784329353268667,-0.631327380482837,1.1763156430490562,-0.6826289037478287,-0.6303176085800898,1.0755259228503937,-0.6784329353268667,-0.7937088518723858,1.1782001064829568,-0.6784329353268667,-0.631327380482837,1.0738056808790413,-0.6826289037478287,-0.7924393601279133,0.9491133612729862,-0.6826289037478287,-0.9381911357613287,1.0755259228503937,-0.6784329353268667,-0.7937088518723858,1.0738056808790413,-0.6826289037478287,-0.7924393601279133,0.9506338455362997,-0.6784329353268667,-0.9396941225657628,1.0755259228503937,-0.6784329353268667,-0.7937088518723858,0.9491133612729862,-0.6826289037478287,-0.9381911357613287,0.8048145409732078,-0.6826289037478287,-1.0645620387775072,0.9506338455362997,-0.6784329353268667,-0.9396941225657628,0.9491133612729862,-0.6826289037478287,-0.9381911357613287,0.806103857818136,-0.6784329353268667,-1.0662674723888421,0.9506338455362997,-0.6784329353268667,-0.9396941225657628,0.8048145409732078,-0.6826289037478287,-1.0645620387775072,0.6438901019646738,-0.6826289037478287,-1.168941536758772,0.806103857818136,-0.6784329353268667,-1.0662674723888421,0.8048145409732078,-0.6826289037478287,-1.0645620387775072,0.6449216170684412,-0.6784329353268667,-1.1708141868382014,0.806103857818136,-0.6784329353268667,-1.0662674723888421,0.6438901019646738,-0.6826289037478287,-1.168941536758772,0.4696643733041648,-0.6826289037478287,-1.2491733891725538,0.6449216170684412,-0.6784329353268667,-1.1708141868382014,0.6438901019646738,-0.6826289037478287,-1.168941536758772,0.4704167779354621,-0.6784329353268667,-1.2511745710733535,0.6449216170684412,-0.6784329353268667,-1.1708141868382014,0.4696643733041648,-0.6826289037478287,-1.2491733891725538,0.28573645812227505,-0.6826289037478287,-1.3036001903450978,0.4704167779354621,-0.6784329353268667,-1.2511745710733535,0.4696643733041648,-0.6826289037478287,-1.2491733891725538,0.2861942093306735,-0.6784329353268667,-1.3056885642485203,0.4704167779354621,-0.6784329353268667,-1.2511745710733535,0.28573645812227505,-0.6826289037478287,-1.3036001903450978,9.590588442456523e-2,-0.6826289037478287,-1.3310976076534657,0.2861942093306735,-0.6784329353268667,-1.3056885642485203,0.28573645812227505,-0.6826289037478287,-1.3036001903450978,9.605952612215037e-2,-0.6784329353268667,-1.3332300325543824,0.2861942093306735,-0.6784329353268667,-1.3056885642485203,9.590588442456523e-2,-0.6826289037478287,-1.3310976076534657,-9.590588442456507e-2,-0.6826289037478287,-1.331097607653466,9.605952612215037e-2,-0.6784329353268667,-1.3332300325543824,9.590588442456523e-2,-0.6826289037478287,-1.3310976076534657,-9.605952612215021e-2,-0.6784329353268667,-1.3332300325543827,9.605952612215037e-2,-0.6784329353268667,-1.3332300325543824,-9.590588442456507e-2,-0.6826289037478287,-1.331097607653466,-0.28573645812227494,-0.6826289037478287,-1.3036001903450978,-9.605952612215021e-2,-0.6784329353268667,-1.3332300325543827,-9.590588442456507e-2,-0.6826289037478287,-1.331097607653466,-0.2861942093306734,-0.6784329353268667,-1.3056885642485203,-9.605952612215021e-2,-0.6784329353268667,-1.3332300325543827,-0.28573645812227494,-0.6826289037478287,-1.3036001903450978,-0.4696643733041646,-0.6826289037478287,-1.2491733891725538,-0.2861942093306734,-0.6784329353268667,-1.3056885642485203,-0.28573645812227494,-0.6826289037478287,-1.3036001903450978,-0.47041677793546194,-0.6784329353268667,-1.2511745710733535,-0.2861942093306734,-0.6784329353268667,-1.3056885642485203,-0.4696643733041646,-0.6826289037478287,-1.2491733891725538,-0.6438901019646738,-0.6826289037478287,-1.168941536758772,-0.47041677793546194,-0.6784329353268667,-1.2511745710733535,-0.4696643733041646,-0.6826289037478287,-1.2491733891725538,-0.6449216170684412,-0.6784329353268667,-1.1708141868382014,-0.47041677793546194,-0.6784329353268667,-1.2511745710733535,-0.6438901019646738,-0.6826289037478287,-1.168941536758772,-0.8048145409732076,-0.6826289037478287,-1.0645620387775074,-0.6449216170684412,-0.6784329353268667,-1.1708141868382014,-0.6438901019646738,-0.6826289037478287,-1.168941536758772,-0.8061038578181356,-0.6784329353268667,-1.0662674723888423,-0.6449216170684412,-0.6784329353268667,-1.1708141868382014,-0.8048145409732076,-0.6826289037478287,-1.0645620387775074,-0.9491133612729862,-0.6826289037478287,-0.9381911357613287,-0.8061038578181356,-0.6784329353268667,-1.0662674723888423,-0.8048145409732076,-0.6826289037478287,-1.0645620387775074,-0.9506338455362997,-0.6784329353268667,-0.9396941225657628,-0.8061038578181356,-0.6784329353268667,-1.0662674723888423,-0.9491133612729862,-0.6826289037478287,-0.9381911357613287,-1.073805680879041,-0.6826289037478287,-0.7924393601279135,-0.9506338455362997,-0.6784329353268667,-0.9396941225657628,-0.9491133612729862,-0.6826289037478287,-0.9381911357613287,-1.0755259228503935,-0.6784329353268667,-0.793708851872386,-0.9506338455362997,-0.6784329353268667,-0.9396941225657628,-1.073805680879041,-0.6826289037478287,-0.7924393601279135,-1.1763156430490562,-0.6826289037478287,-0.6303176085800899,-1.0755259228503935,-0.6784329353268667,-0.793708851872386,-1.073805680879041,-0.6826289037478287,-0.7924393601279135,-1.1782001064829568,-0.6784329353268667,-0.631327380482837,-1.0755259228503935,-0.6784329353268667,-0.793708851872386,-1.1763156430490562,-0.6826289037478287,-0.6303176085800899,-1.2545256275635621,-0.6826289037478287,-0.4551749438999127,-1.1782001064829568,-0.6784329353268667,-0.631327380482837,-1.1763156430490562,-0.6826289037478287,-0.6303176085800899,-1.2565353837765347,-0.6784329353268667,-0.4559041363941857,-1.1782001064829568,-0.6784329353268667,-0.631327380482837,-1.2545256275635621,-0.6826289037478287,-0.4551749438999127,-1.3068199958927385,-0.6826289037478287,-0.27062941100874627,-1.2565353837765347,-0.6784329353268667,-0.4559041363941857,-1.2545256275635621,-0.6826289037478287,-0.4551749438999127,-1.3089135279404522,-0.6784329353268667,-0.27106296065351865,-1.2565353837765347,-0.6784329353268667,-0.4559041363941857,-1.3068199958927385,-0.6826289037478287,-0.27062941100874627,1.3082922064603395,0.6805317591095806,-0.268791791835905,1.3699132268474414,0.5371935563684052,-0.28594026282546564,1.3074413173728514,0.6805317591095806,-0.2729005798263596,1.3708047729535964,0.5371935563684052,-0.2816351495177831,1.3699132268474414,0.5371935563684052,-0.28594026282546564,1.3082922064603395,0.6805317591095806,-0.268791791835905,1.3708047729535964,0.5371935563684052,-0.2816351495177831,1.417376080335918,0.3879697486086787,-0.29584712446820893,1.3699132268474414,0.5371935563684052,-0.28594026282546564,1.4182985154950334,0.3879697486086787,-0.2913928535656005,1.417376080335918,0.3879697486086787,-0.29584712446820893,1.3708047729535964,0.5371935563684052,-0.2816351495177831,1.4182985154950334,0.3879697486086787,-0.2913928535656005,1.4493098648832217,0.23449526309530616,-0.30251262310528765,1.417376080335918,0.3879697486086787,-0.29584712446820893,1.4502530827026614,0.23449526309530616,-0.2979579965318783,1.4493098648832217,0.23449526309530616,-0.30251262310528765,1.4182985154950334,0.3879697486086787,-0.2913928535656005,1.4502530827026614,0.23449526309530616,-0.2979579965318783,1.465364707263277,7.84515984081728e-2,-0.3058637301387886,1.4493098648832217,0.23449526309530616,-0.30251262310528765,1.4663183736512309,7.84515984081728e-2,-0.3012586493364336,1.465364707263277,7.84515984081728e-2,-0.3058637301387886,1.4502530827026614,0.23449526309530616,-0.2979579965318783,1.4663183736512309,7.84515984081728e-2,-0.3012586493364336,1.465364707263277,-7.84515984081728e-2,-0.3058637301387886,1.465364707263277,7.84515984081728e-2,-0.3058637301387886,1.4663183736512309,-7.84515984081728e-2,-0.3012586493364336,1.465364707263277,-7.84515984081728e-2,-0.3058637301387886,1.4663183736512309,7.84515984081728e-2,-0.3012586493364336,1.4663183736512309,-7.84515984081728e-2,-0.3012586493364336,1.4493098648832217,-0.23449526309530605,-0.30251262310528765,1.465364707263277,-7.84515984081728e-2,-0.3058637301387886,1.4502530827026614,-0.23449526309530605,-0.2979579965318783,1.4493098648832217,-0.23449526309530605,-0.30251262310528765,1.4663183736512309,-7.84515984081728e-2,-0.3012586493364336,1.4502530827026614,-0.23449526309530605,-0.2979579965318783,1.417376080335918,-0.3879697486086786,-0.29584712446820893,1.4493098648832217,-0.23449526309530605,-0.30251262310528765,1.4182985154950334,-0.3879697486086786,-0.2913928535656005,1.417376080335918,-0.3879697486086786,-0.29584712446820893,1.4502530827026614,-0.23449526309530605,-0.2979579965318783,1.4182985154950334,-0.3879697486086786,-0.2913928535656005,1.3699132268474414,-0.5371935563684052,-0.28594026282546564,1.417376080335918,-0.3879697486086786,-0.29584712446820893,1.3708047729535964,-0.5371935563684052,-0.2816351495177831,1.3699132268474414,-0.5371935563684052,-0.28594026282546564,1.4182985154950334,-0.3879697486086786,-0.2913928535656005,1.3708047729535964,-0.5371935563684052,-0.2816351495177831,1.3074413173728514,-0.6805317591095806,-0.2729005798263596,1.3699132268474414,-0.5371935563684052,-0.28594026282546564,1.3082922064603395,-0.6805317591095806,-0.268791791835905,1.3074413173728514,-0.6805317591095806,-0.2729005798263596,1.3708047729535964,-0.5371935563684052,-0.2816351495177831,1.2562460660949462,0.6805317591095806,-0.45356735581850705,1.3147722715320653,0.5371935563684052,-0.47937246166908504,1.2548149452451507,0.6805317591095806,-0.4575117244755916,1.3162717739229577,0.5371935563684052,-0.47523962394772545,1.3147722715320653,0.5371935563684052,-0.47937246166908504,1.2562460660949462,0.6805317591095806,-0.45356735581850705,1.3162717739229577,0.5371935563684052,-0.47523962394772545,1.360324677678289,0.3879697486086787,-0.49598109385739525,1.3147722715320653,0.5371935563684052,-0.47937246166908504,1.3618761327482927,0.3879697486086787,-0.4917050673066876,1.360324677678289,0.3879697486086787,-0.49598109385739525,1.3162717739229577,0.5371935563684052,-0.47523962394772545,1.3618761327482927,0.3879697486086787,-0.4917050673066876,1.3909730819896295,0.23449526309530616,-0.5071556533906874,1.360324677678289,0.3879697486086787,-0.49598109385739525,1.3925594916723345,0.23449526309530616,-0.5027832870523374,1.3909730819896295,0.23449526309530616,-0.5071556533906874,1.3618761327482927,0.3879697486086787,-0.4917050673066876,1.3925594916723345,0.23449526309530616,-0.5027832870523374,1.406381694134861,7.84515984081728e-2,-0.5127737094562875,1.3909730819896295,0.23449526309530616,-0.5071556533906874,1.407985677393806,7.84515984081728e-2,-0.5083529078908748,1.406381694134861,7.84515984081728e-2,-0.5127737094562875,1.3925594916723345,0.23449526309530616,-0.5027832870523374,1.407985677393806,7.84515984081728e-2,-0.5083529078908748,1.406381694134861,-7.84515984081728e-2,-0.5127737094562875,1.406381694134861,7.84515984081728e-2,-0.5127737094562875,1.407985677393806,-7.84515984081728e-2,-0.5083529078908748,1.406381694134861,-7.84515984081728e-2,-0.5127737094562875,1.407985677393806,7.84515984081728e-2,-0.5083529078908748,1.407985677393806,-7.84515984081728e-2,-0.5083529078908748,1.3909730819896295,-0.23449526309530605,-0.5071556533906874,1.406381694134861,-7.84515984081728e-2,-0.5127737094562875,1.3925594916723345,-0.23449526309530605,-0.5027832870523374,1.3909730819896295,-0.23449526309530605,-0.5071556533906874,1.407985677393806,-7.84515984081728e-2,-0.5083529078908748,1.3925594916723345,-0.23449526309530605,-0.5027832870523374,1.360324677678289,-0.3879697486086786,-0.49598109385739525,1.3909730819896295,-0.23449526309530605,-0.5071556533906874,1.3618761327482927,-0.3879697486086786,-0.4917050673066876,1.360324677678289,-0.3879697486086786,-0.49598109385739525,1.3925594916723345,-0.23449526309530605,-0.5027832870523374,1.3618761327482927,-0.3879697486086786,-0.4917050673066876,1.3147722715320653,-0.5371935563684052,-0.47937246166908504,1.360324677678289,-0.3879697486086786,-0.49598109385739525,1.3162717739229577,-0.5371935563684052,-0.47523962394772545,1.3147722715320653,-0.5371935563684052,-0.47937246166908504,1.3618761327482927,-0.3879697486086786,-0.4917050673066876,1.3162717739229577,-0.5371935563684052,-0.47523962394772545,1.2548149452451507,-0.6805317591095806,-0.4575117244755916,1.3147722715320653,-0.5371935563684052,-0.47937246166908504,1.2562460660949462,-0.6805317591095806,-0.45356735581850705,1.2548149452451507,-0.6805317591095806,-0.4575117244755916,1.3162717739229577,-0.5371935563684052,-0.47523962394772545,1.1782487692382544,0.6805317591095806,-0.6289732606651564,1.2324711428320194,0.5371935563684052,-0.6629019271978336,1.1762669802937589,0.6805317591095806,-0.6326717283977703,1.2345476252345637,0.5371935563684052,-0.6590267399916063,1.2324711428320194,0.5371935563684052,-0.6629019271978336,1.1782487692382544,0.6805317591095806,-0.6289732606651564,1.2345476252345637,0.5371935563684052,-0.6590267399916063,1.2751720936182447,0.3879697486086787,-0.6858692337623709,1.2324711428320194,0.5371935563684052,-0.6629019271978336,1.2773205191031043,0.3879697486086787,-0.6818597844444967,1.2751720936182447,0.3879697486086787,-0.6858692337623709,1.2345476252345637,0.5371935563684052,-0.6590267399916063,1.2773205191031043,0.3879697486086787,-0.6818597844444967,1.3039019920998727,0.23449526309530616,-0.7013220134744433,1.2751720936182447,0.3879697486086787,-0.6858692337623709,1.306098822067848,0.23449526309530616,-0.69722223041068,1.3039019920998727,0.23449526309530616,-0.7013220134744433,1.2773205191031043,0.3879697486086787,-0.6818597844444967,1.306098822067848,0.23449526309530616,-0.69722223041068,1.3183460674970207,7.84515984081728e-2,-0.7090909624458233,1.3039019920998727,0.23449526309530616,-0.7013220134744433,1.3205672330192666,7.84515984081728e-2,-0.7049457637173522,1.3183460674970207,7.84515984081728e-2,-0.7090909624458233,1.306098822067848,0.23449526309530616,-0.69722223041068,1.3205672330192666,7.84515984081728e-2,-0.7049457637173522,1.3183460674970207,-7.84515984081728e-2,-0.7090909624458233,1.3183460674970207,7.84515984081728e-2,-0.7090909624458233,1.3205672330192666,-7.84515984081728e-2,-0.7049457637173522,1.3183460674970207,-7.84515984081728e-2,-0.7090909624458233,1.3205672330192666,7.84515984081728e-2,-0.7049457637173522,1.3205672330192666,-7.84515984081728e-2,-0.7049457637173522,1.3039019920998727,-0.23449526309530605,-0.7013220134744433,1.3183460674970207,-7.84515984081728e-2,-0.7090909624458233,1.306098822067848,-0.23449526309530605,-0.69722223041068,1.3039019920998727,-0.23449526309530605,-0.7013220134744433,1.3205672330192666,-7.84515984081728e-2,-0.7049457637173522,1.306098822067848,-0.23449526309530605,-0.69722223041068,1.2751720936182447,-0.3879697486086786,-0.6858692337623709,1.3039019920998727,-0.23449526309530605,-0.7013220134744433,1.2773205191031043,-0.3879697486086786,-0.6818597844444967,1.2751720936182447,-0.3879697486086786,-0.6858692337623709,1.306098822067848,-0.23449526309530605,-0.69722223041068,1.2773205191031043,-0.3879697486086786,-0.6818597844444967,1.2324711428320194,-0.5371935563684052,-0.6629019271978336,1.2751720936182447,-0.3879697486086786,-0.6858692337623709,1.2345476252345637,-0.5371935563684052,-0.6590267399916063,1.2324711428320194,-0.5371935563684052,-0.6629019271978336,1.2773205191031043,-0.3879697486086786,-0.6818597844444967,1.2345476252345637,-0.5371935563684052,-0.6590267399916063,1.1762669802937589,-0.6805317591095806,-0.6326717283977703,1.2324711428320194,-0.5371935563684052,-0.6629019271978336,1.1782487692382544,-0.6805317591095806,-0.6289732606651564,1.1762669802937589,-0.6805317591095806,-0.6326717283977703,1.2345476252345637,-0.5371935563684052,-0.6590267399916063,1.0759115607818943,0.6805317591095806,-0.7913860235176559,1.1247099929133069,0.5371935563684052,-0.8327373624602264,1.0734200429475405,0.6805317591095806,-0.7947621884826431,1.127320559973453,0.5371935563684052,-0.8291998782304574,1.1247099929133069,0.5371935563684052,-0.8327373624602264,1.0759115607818943,0.6805317591095806,-0.7913860235176559,1.127320559973453,0.5371935563684052,-0.8291998782304574,1.163677384836018,0.3879697486086787,-0.8615888916331994,1.1247099929133069,0.5371935563684052,-0.8327373624602264,1.1663783992029493,0.3879697486086787,-0.8579288455561359,1.163677384836018,0.3879697486086787,-0.8615888916331994,1.127320559973453,0.5371935563684052,-0.8291998782304574,1.1663783992029493,0.3879697486086787,-0.8579288455561359,1.1898952838153178,0.23449526309530616,-0.8810006726103722,1.163677384836018,0.3879697486086787,-0.8615888916331994,1.1926571526104064,0.23449526309530616,-0.8772581649167487,1.1898952838153178,0.23449526309530616,-0.8810006726103722,1.1663783992029493,0.3879697486086787,-0.8579288455561359,1.1926571526104064,0.23449526309530616,-0.8772581649167487,1.2030764410635406,7.84515984081728e-2,-0.8907600258571825,1.1898952838153178,0.23449526309530616,-0.8810006726103722,1.2058689046743092,7.84515984081728e-2,-0.8869760602444603,1.2030764410635406,7.84515984081728e-2,-0.8907600258571825,1.1926571526104064,0.23449526309530616,-0.8772581649167487,1.2058689046743092,7.84515984081728e-2,-0.8869760602444603,1.2030764410635406,-7.84515984081728e-2,-0.8907600258571825,1.2030764410635406,7.84515984081728e-2,-0.8907600258571825,1.2058689046743092,-7.84515984081728e-2,-0.8869760602444603,1.2030764410635406,-7.84515984081728e-2,-0.8907600258571825,1.2058689046743092,7.84515984081728e-2,-0.8869760602444603,1.2058689046743092,-7.84515984081728e-2,-0.8869760602444603,1.1898952838153178,-0.23449526309530605,-0.8810006726103722,1.2030764410635406,-7.84515984081728e-2,-0.8907600258571825,1.1926571526104064,-0.23449526309530605,-0.8772581649167487,1.1898952838153178,-0.23449526309530605,-0.8810006726103722,1.2058689046743092,-7.84515984081728e-2,-0.8869760602444603,1.1926571526104064,-0.23449526309530605,-0.8772581649167487,1.163677384836018,-0.3879697486086786,-0.8615888916331994,1.1898952838153178,-0.23449526309530605,-0.8810006726103722,1.1663783992029493,-0.3879697486086786,-0.8579288455561359,1.163677384836018,-0.3879697486086786,-0.8615888916331994,1.1926571526104064,-0.23449526309530605,-0.8772581649167487,1.1663783992029493,-0.3879697486086786,-0.8579288455561359,1.1247099929133069,-0.5371935563684052,-0.8327373624602264,1.163677384836018,-0.3879697486086786,-0.8615888916331994,1.127320559973453,-0.5371935563684052,-0.8291998782304574,1.1247099929133069,-0.5371935563684052,-0.8327373624602264,1.1663783992029493,-0.3879697486086786,-0.8579288455561359,1.127320559973453,-0.5371935563684052,-0.8291998782304574,1.0734200429475405,-0.6805317591095806,-0.7947621884826431,1.1247099929133069,-0.5371935563684052,-0.8327373624602264,1.0759115607818943,-0.6805317591095806,-0.7913860235176559,1.0734200429475405,-0.6805317591095806,-0.7947621884826431,1.127320559973453,-0.5371935563684052,-0.8291998782304574,0.9513484922506519,0.6805317591095806,-0.9374505699692285,0.9937149194654645,0.5371935563684052,-0.9853703576970493,0.9483987145586341,0.6805317591095806,-0.9404346883578628,0.9968056428676243,0.5371935563684052,-0.9822436527377145,0.9937149194654645,0.5371935563684052,-0.9853703576970493,0.9513484922506519,0.6805317591095806,-0.9374505699692285,0.9968056428676243,0.5371935563684052,-0.9822436527377145,1.0281437757664158,0.3879697486086787,-1.0195101031953029,0.9937149194654645,0.5371935563684052,-0.9853703576970493,1.0313415822663476,0.3879697486086787,-1.0162750684991073,1.0281437757664158,0.3879697486086787,-1.0195101031953029,0.9968056428676243,0.5371935563684052,-0.9822436527377145,1.0313415822663476,0.3879697486086787,-1.0162750684991073,1.0513080737071532,0.23449526309530616,-1.0424798826567439,1.0281437757664158,0.3879697486086787,-1.0195101031953029,1.05457792746766,0.23449526309530616,-1.0391719619407938,1.0513080737071532,0.23449526309530616,-1.0424798826567439,1.0313415822663476,0.3879697486086787,-1.0162750684991073,1.05457792746766,0.23449526309530616,-1.0391719619407938,1.062954020391997,7.84515984081728e-2,-1.0540280343708575,1.0513080737071532,0.23449526309530616,-1.0424798826567439,1.0662600962110178,7.84515984081728e-2,-1.050683469906743,1.062954020391997,7.84515984081728e-2,-1.0540280343708575,1.05457792746766,0.23449526309530616,-1.0391719619407938,1.0662600962110178,7.84515984081728e-2,-1.050683469906743,1.062954020391997,-7.84515984081728e-2,-1.0540280343708575,1.062954020391997,7.84515984081728e-2,-1.0540280343708575,1.0662600962110178,-7.84515984081728e-2,-1.050683469906743,1.062954020391997,-7.84515984081728e-2,-1.0540280343708575,1.0662600962110178,7.84515984081728e-2,-1.050683469906743,1.0662600962110178,-7.84515984081728e-2,-1.050683469906743,1.0513080737071532,-0.23449526309530605,-1.0424798826567439,1.062954020391997,-7.84515984081728e-2,-1.0540280343708575,1.05457792746766,-0.23449526309530605,-1.0391719619407938,1.0513080737071532,-0.23449526309530605,-1.0424798826567439,1.0662600962110178,-7.84515984081728e-2,-1.050683469906743,1.05457792746766,-0.23449526309530605,-1.0391719619407938,1.0281437757664158,-0.3879697486086786,-1.0195101031953029,1.0513080737071532,-0.23449526309530605,-1.0424798826567439,1.0313415822663476,-0.3879697486086786,-1.0162750684991073,1.0281437757664158,-0.3879697486086786,-1.0195101031953029,1.05457792746766,-0.23449526309530605,-1.0391719619407938,1.0313415822663476,-0.3879697486086786,-1.0162750684991073,0.9937149194654645,-0.5371935563684052,-0.9853703576970493,1.0281437757664158,-0.3879697486086786,-1.0195101031953029,0.9968056428676243,-0.5371935563684052,-0.9822436527377145,0.9937149194654645,-0.5371935563684052,-0.9853703576970493,1.0313415822663476,-0.3879697486086786,-1.0162750684991073,0.9968056428676243,-0.5371935563684052,-0.9822436527377145,0.9483987145586341,-0.6805317591095806,-0.9404346883578628,0.9937149194654645,-0.5371935563684052,-0.9853703576970493,0.9513484922506519,-0.6805317591095806,-0.9374505699692285,0.9483987145586341,-0.6805317591095806,-0.9404346883578628,0.9968056428676243,-0.5371935563684052,-0.9822436527377145,0.8071327503566957,0.6805317591095806,-1.0641495421907843,0.8421919796395415,0.5371935563684052,-1.117647866023595,0.803785648434648,0.6805317591095806,-1.066679968975565,0.8456990121416451,0.5371935563684052,-1.1149965308730385,0.8421919796395415,0.5371935563684052,-1.117647866023595,0.8071327503566957,0.6805317591095806,-1.0641495421907843,0.8456990121416451,0.5371935563684052,-1.1149965308730385,0.8713710792755023,0.3879697486086787,-1.1563705791686183,0.8421919796395415,0.5371935563684052,-1.117647866023595,0.8749996185756749,0.3879697486086787,-1.1536273842350235,0.8713710792755023,0.3879697486086787,-1.1563705791686183,0.8456990121416451,0.5371935563684052,-1.1149965308730385,0.8749996185756749,0.3879697486086787,-1.1536273842350235,0.8910032550207994,0.23449526309530616,-1.1824238542621694,0.8713710792755023,0.3879697486086787,-1.1563705791686183,0.8947135460831475,0.23449526309530616,-1.179618854563279,0.8910032550207994,0.23449526309530616,-1.1824238542621694,0.8749996185756749,0.3879697486086787,-1.1536273842350235,0.8947135460831475,0.23449526309530616,-1.179618854563279,0.9008734126496704,7.84515984081728e-2,-1.1955222461703252,0.8910032550207994,0.23449526309530616,-1.1824238542621694,0.9046248047488868,7.84515984081728e-2,-1.1926861738698245,0.9008734126496704,7.84515984081728e-2,-1.1955222461703252,0.8947135460831475,0.23449526309530616,-1.179618854563279,0.9046248047488868,7.84515984081728e-2,-1.1926861738698245,0.9008734126496704,-7.84515984081728e-2,-1.1955222461703252,0.9008734126496704,7.84515984081728e-2,-1.1955222461703252,0.9046248047488868,-7.84515984081728e-2,-1.1926861738698245,0.9008734126496704,-7.84515984081728e-2,-1.1955222461703252,0.9046248047488868,7.84515984081728e-2,-1.1926861738698245,0.9046248047488868,-7.84515984081728e-2,-1.1926861738698245,0.8910032550207994,-0.23449526309530605,-1.1824238542621694,0.9008734126496704,-7.84515984081728e-2,-1.1955222461703252,0.8947135460831475,-0.23449526309530605,-1.179618854563279,0.8910032550207994,-0.23449526309530605,-1.1824238542621694,0.9046248047488868,-7.84515984081728e-2,-1.1926861738698245,0.8947135460831475,-0.23449526309530605,-1.179618854563279,0.8713710792755023,-0.3879697486086786,-1.1563705791686183,0.8910032550207994,-0.23449526309530605,-1.1824238542621694,0.8749996185756749,-0.3879697486086786,-1.1536273842350235,0.8713710792755023,-0.3879697486086786,-1.1563705791686183,0.8947135460831475,-0.23449526309530605,-1.179618854563279,0.8749996185756749,-0.3879697486086786,-1.1536273842350235,0.8421919796395415,-0.5371935563684052,-1.117647866023595,0.8713710792755023,-0.3879697486086786,-1.1563705791686183,0.8456990121416451,-0.5371935563684052,-1.1149965308730385,0.8421919796395415,-0.5371935563684052,-1.117647866023595,0.8749996185756749,-0.3879697486086786,-1.1536273842350235,0.8456990121416451,-0.5371935563684052,-1.1149965308730385,0.803785648434648,-0.6805317591095806,-1.066679968975565,0.8421919796395415,-0.5371935563684052,-1.117647866023595,0.8071327503566957,-0.6805317591095806,-1.0641495421907843,0.803785648434648,-0.6805317591095806,-1.066679968975565,0.8456990121416451,-0.5371935563684052,-1.1149965308730385,0.6462435008760691,0.6805317591095806,-1.168865630608866,0.6732712891266994,0.5371935563684052,-1.226837338131644,0.6425682181570462,0.6805317591095806,-1.1708900929881074,0.6771221834974771,0.5371935563684052,-1.224716143280505,0.6732712891266994,0.5371935563684052,-1.226837338131644,0.6462435008760691,0.6805317591095806,-1.168865630608866,0.6771221834974771,0.5371935563684052,-1.224716143280505,0.6965978589616058,0.3879697486086787,-1.2693430966663923,0.6732712891266994,0.5371935563684052,-1.226837338131644,0.7005821737795604,0.3879697486086787,-1.2671484095980332,0.6965978589616058,0.3879697486086787,-1.2693430966663923,0.6771221834974771,0.5371935563684052,-1.224716143280505,0.7005821737795604,0.3879697486086787,-1.2671484095980332,0.712292356881255,0.23449526309530616,-1.2979416666069437,0.6965978589616058,0.3879697486086787,-1.2693430966663923,0.7163664391594698,0.23449526309530616,-1.2956975327721534,0.712292356881255,0.23449526309530616,-1.2979416666069437,0.7005821737795604,0.3879697486086787,-1.2671484095980332,0.7163664391594698,0.23449526309530616,-1.2956975327721534,0.7201828306821547,7.84515984081728e-2,-1.3123197160363942,0.712292356881255,0.23449526309530616,-1.2979416666069437,0.7243020439226331,7.84515984081728e-2,-1.3100507226350815,0.7201828306821547,7.84515984081728e-2,-1.3123197160363942,0.7163664391594698,0.23449526309530616,-1.2956975327721534,0.7243020439226331,7.84515984081728e-2,-1.3100507226350815,0.7201828306821547,-7.84515984081728e-2,-1.3123197160363942,0.7201828306821547,7.84515984081728e-2,-1.3123197160363942,0.7243020439226331,-7.84515984081728e-2,-1.3100507226350815,0.7201828306821547,-7.84515984081728e-2,-1.3123197160363942,0.7243020439226331,7.84515984081728e-2,-1.3100507226350815,0.7243020439226331,-7.84515984081728e-2,-1.3100507226350815,0.712292356881255,-0.23449526309530605,-1.2979416666069437,0.7201828306821547,-7.84515984081728e-2,-1.3123197160363942,0.7163664391594698,-0.23449526309530605,-1.2956975327721534,0.712292356881255,-0.23449526309530605,-1.2979416666069437,0.7243020439226331,-7.84515984081728e-2,-1.3100507226350815,0.7163664391594698,-0.23449526309530605,-1.2956975327721534,0.6965978589616058,-0.3879697486086786,-1.2693430966663923,0.712292356881255,-0.23449526309530605,-1.2979416666069437,0.7005821737795604,-0.3879697486086786,-1.2671484095980332,0.6965978589616058,-0.3879697486086786,-1.2693430966663923,0.7163664391594698,-0.23449526309530605,-1.2956975327721534,0.7005821737795604,-0.3879697486086786,-1.2671484095980332,0.6732712891266994,-0.5371935563684052,-1.226837338131644,0.6965978589616058,-0.3879697486086786,-1.2693430966663923,0.6771221834974771,-0.5371935563684052,-1.224716143280505,0.6732712891266994,-0.5371935563684052,-1.226837338131644,0.7005821737795604,-0.3879697486086786,-1.2671484095980332,0.6771221834974771,-0.5371935563684052,-1.224716143280505,0.6425682181570462,-0.6805317591095806,-1.1708900929881074,0.6732712891266994,-0.5371935563684052,-1.226837338131644,0.6462435008760691,-0.6805317591095806,-1.168865630608866,0.6425682181570462,-0.6805317591095806,-1.1708900929881074,0.6771221834974771,-0.5371935563684052,-1.224716143280505,0.4720043459307816,0.6805317591095806,-1.2494356415060661,0.4904423611619884,0.5371935563684052,-1.3106831704779454,0.46807680530884527,0.6805317591095806,-1.250912318739841,0.49455756677426205,0.5371935563684052,-1.3091359349366996,0.4904423611619884,0.5371935563684052,-1.3106831704779454,0.4720043459307816,0.6805317591095806,-1.2494356415060661,0.49455756677426205,0.5371935563684052,-1.3091359349366996,0.5074345278745785,0.3879697486086787,-1.3560939031222075,0.4904423611619884,0.5371935563684052,-1.3106831704779454,0.5116923114233393,0.3879697486086787,-1.35449306110986,0.5074345278745785,0.3879697486086787,-1.3560939031222075,0.49455756677426205,0.5371935563684052,-1.3091359349366996,0.5116923114233393,0.3879697486086787,-1.35449306110986,0.5188671357122733,0.23449526309530616,-1.386646987182969,0.5074345278745785,0.3879697486086787,-1.3560939031222075,0.5232208480299624,0.23449526309530616,-1.3850100778596055,0.5188671357122733,0.23449526309530616,-1.386646987182969,0.5116923114233393,0.3879697486086787,-1.35449306110986,0.5232208480299624,0.23449526309530616,-1.3850100778596055,0.5246149266311756,7.84515984081728e-2,-1.4020076766776168,0.5188671357122733,0.23449526309530616,-1.386646987182969,0.5290168675345677,7.84515984081728e-2,-1.400352634364329,0.5246149266311756,7.84515984081728e-2,-1.4020076766776168,0.5232208480299624,0.23449526309530616,-1.3850100778596055,0.5290168675345677,7.84515984081728e-2,-1.400352634364329,0.5246149266311756,-7.84515984081728e-2,-1.4020076766776168,0.5246149266311756,7.84515984081728e-2,-1.4020076766776168,0.5290168675345677,-7.84515984081728e-2,-1.400352634364329,0.5246149266311756,-7.84515984081728e-2,-1.4020076766776168,0.5290168675345677,7.84515984081728e-2,-1.400352634364329,0.5290168675345677,-7.84515984081728e-2,-1.400352634364329,0.5188671357122733,-0.23449526309530605,-1.386646987182969,0.5246149266311756,-7.84515984081728e-2,-1.4020076766776168,0.5232208480299624,-0.23449526309530605,-1.3850100778596055,0.5188671357122733,-0.23449526309530605,-1.386646987182969,0.5290168675345677,-7.84515984081728e-2,-1.400352634364329,0.5232208480299624,-0.23449526309530605,-1.3850100778596055,0.5074345278745785,-0.3879697486086786,-1.3560939031222075,0.5188671357122733,-0.23449526309530605,-1.386646987182969,0.5116923114233393,-0.3879697486086786,-1.35449306110986,0.5074345278745785,-0.3879697486086786,-1.3560939031222075,0.5232208480299624,-0.23449526309530605,-1.3850100778596055,0.5116923114233393,-0.3879697486086786,-1.35449306110986,0.4904423611619884,-0.5371935563684052,-1.3106831704779454,0.5074345278745785,-0.3879697486086786,-1.3560939031222075,0.49455756677426205,-0.5371935563684052,-1.3091359349366996,0.4904423611619884,-0.5371935563684052,-1.3106831704779454,0.5116923114233393,-0.3879697486086786,-1.35449306110986,0.49455756677426205,-0.5371935563684052,-1.3091359349366996,0.46807680530884527,-0.6805317591095806,-1.250912318739841,0.4904423611619884,-0.5371935563684052,-1.3106831704779454,0.4720043459307816,-0.6805317591095806,-1.2494356415060661,0.46807680530884527,-0.6805317591095806,-1.250912318739841,0.49455756677426205,-0.5371935563684052,-1.3091359349366996,0.28801466600761416,0.6805317591095806,-1.304195183631554,0.2974820212008674,0.5371935563684052,-1.3674533008693615,0.28391600144533424,0.6805317591095806,-1.3050935709620646,0.3017765273646779,0.5371935563684052,-1.3665119869682743,0.2974820212008674,0.5371935563684052,-1.3674533008693615,0.28801466600761416,0.6805317591095806,-1.304195183631554,0.3017765273646779,0.5371935563684052,-1.3665119869682743,0.3077887656800089,0.3879697486086787,-1.4148309262543337,0.2974820212008674,0.5371935563684052,-1.3674533008693615,0.312232061937136,0.3879697486086787,-1.4138569990147531,0.3077887656800089,0.3879697486086787,-1.4148309262543337,0.3017765273646779,0.5371935563684052,-1.3665119869682743,0.312232061937136,0.3879697486086787,-1.4138569990147531,0.3147233120334202,0.23449526309530616,-1.446707367938858,0.3077887656800089,0.3879697486086787,-1.4148309262543337,0.3192667167002513,0.23449526309530616,-1.445711497911428,0.3147233120334202,0.23449526309530616,-1.446707367938858,0.312232061937136,0.3879697486086787,-1.4138569990147531,0.3192667167002513,0.23449526309530616,-1.445711497911428,0.3182096839201062,7.84515984081728e-2,-1.4627333809572642,0.3147233120334202,0.23449526309530616,-1.446707367938858,0.32280341850434296,7.84515984081728e-2,-1.4617264791024047,0.3182096839201062,7.84515984081728e-2,-1.4627333809572642,0.3192667167002513,0.23449526309530616,-1.445711497911428,0.32280341850434296,7.84515984081728e-2,-1.4617264791024047,0.3182096839201062,-7.84515984081728e-2,-1.4627333809572642,0.3182096839201062,7.84515984081728e-2,-1.4627333809572642,0.32280341850434296,-7.84515984081728e-2,-1.4617264791024047,0.3182096839201062,-7.84515984081728e-2,-1.4627333809572642,0.32280341850434296,7.84515984081728e-2,-1.4617264791024047,0.32280341850434296,-7.84515984081728e-2,-1.4617264791024047,0.3147233120334202,-0.23449526309530605,-1.446707367938858,0.3182096839201062,-7.84515984081728e-2,-1.4627333809572642,0.3192667167002513,-0.23449526309530605,-1.445711497911428,0.3147233120334202,-0.23449526309530605,-1.446707367938858,0.32280341850434296,-7.84515984081728e-2,-1.4617264791024047,0.3192667167002513,-0.23449526309530605,-1.445711497911428,0.3077887656800089,-0.3879697486086786,-1.4148309262543337,0.3147233120334202,-0.23449526309530605,-1.446707367938858,0.312232061937136,-0.3879697486086786,-1.4138569990147531,0.3077887656800089,-0.3879697486086786,-1.4148309262543337,0.3192667167002513,-0.23449526309530605,-1.445711497911428,0.312232061937136,-0.3879697486086786,-1.4138569990147531,0.2974820212008674,-0.5371935563684052,-1.3674533008693615,0.3077887656800089,-0.3879697486086786,-1.4148309262543337,0.3017765273646779,-0.5371935563684052,-1.3665119869682743,0.2974820212008674,-0.5371935563684052,-1.3674533008693615,0.312232061937136,-0.3879697486086786,-1.4138569990147531,0.3017765273646779,-0.5371935563684052,-1.3665119869682743,0.28391600144533424,-0.6805317591095806,-1.3050935709620646,0.2974820212008674,-0.5371935563684052,-1.3674533008693615,0.28801466600761416,-0.6805317591095806,-1.304195183631554,0.28391600144533424,-0.6805317591095806,-1.3050935709620646,0.3017765273646779,-0.5371935563684052,-1.3665119869682743,9.807526502972766e-2,0.6805317591095806,-1.3320130506990422,9.837638638558688e-2,0.5371935563684052,-1.395974988886971,9.389014551698761e-2,0.6805317591095806,-1.3323145895088062,0.10276147847366635,0.5371935563684052,-1.395659042007815,9.837638638558688e-2,0.5371935563684052,-1.395974988886971,9.807526502972766e-2,0.6805317591095806,-1.3320130506990422,0.10276147847366635,0.5371935563684052,-1.395659042007815,0.10178479497836329,0.3879697486086787,-1.4443407941603434,9.837638638558688e-2,0.5371935563684052,-1.395974988886971,0.1063218156552257,0.3879697486086787,-1.4440139007919195,0.10178479497836329,0.3879697486086787,-1.4443407941603434,0.10276147847366635,0.5371935563684052,-1.395659042007815,0.1063218156552257,0.3879697486086787,-1.4440139007919195,0.10407802805751909,0.23449526309530616,-1.4768820994451528,0.10178479497836329,0.3879697486086787,-1.4443407941603434,0.10871726877519611,0.23449526309530616,-1.4765478410996122,0.10407802805751909,0.23449526309530616,-1.4768820994451528,0.1063218156552257,0.3879697486086787,-1.4440139007919195,0.10871726877519611,0.23449526309530616,-1.4765478410996122,0.1052309604815491,7.84515984081728e-2,-1.4932423753909925,0.10407802805751909,0.23449526309530616,-1.4768820994451528,0.10992159274791433,7.84515984081728e-2,-1.492904414272717,0.1052309604815491,7.84515984081728e-2,-1.4932423753909925,0.10871726877519611,0.23449526309530616,-1.4765478410996122,0.10992159274791433,7.84515984081728e-2,-1.492904414272717,0.1052309604815491,-7.84515984081728e-2,-1.4932423753909925,0.1052309604815491,7.84515984081728e-2,-1.4932423753909925,0.10992159274791433,-7.84515984081728e-2,-1.492904414272717,0.1052309604815491,-7.84515984081728e-2,-1.4932423753909925,0.10992159274791433,7.84515984081728e-2,-1.492904414272717,0.10992159274791433,-7.84515984081728e-2,-1.492904414272717,0.10407802805751909,-0.23449526309530605,-1.4768820994451528,0.1052309604815491,-7.84515984081728e-2,-1.4932423753909925,0.10871726877519611,-0.23449526309530605,-1.4765478410996122,0.10407802805751909,-0.23449526309530605,-1.4768820994451528,0.10992159274791433,-7.84515984081728e-2,-1.492904414272717,0.10871726877519611,-0.23449526309530605,-1.4765478410996122,0.10178479497836329,-0.3879697486086786,-1.4443407941603434,0.10407802805751909,-0.23449526309530605,-1.4768820994451528,0.1063218156552257,-0.3879697486086786,-1.4440139007919195,0.10178479497836329,-0.3879697486086786,-1.4443407941603434,0.10871726877519611,-0.23449526309530605,-1.4765478410996122,0.1063218156552257,-0.3879697486086786,-1.4440139007919195,9.837638638558688e-2,-0.5371935563684052,-1.395974988886971,0.10178479497836329,-0.3879697486086786,-1.4443407941603434,0.10276147847366635,-0.5371935563684052,-1.395659042007815,9.837638638558688e-2,-0.5371935563684052,-1.395974988886971,0.1063218156552257,-0.3879697486086786,-1.4440139007919195,0.10276147847366635,-0.5371935563684052,-1.395659042007815,9.389014551698761e-2,-0.6805317591095806,-1.3323145895088062,9.837638638558688e-2,-0.5371935563684052,-1.395974988886971,9.807526502972766e-2,-0.6805317591095806,-1.3320130506990422,9.389014551698761e-2,-0.6805317591095806,-1.3323145895088062,0.10276147847366635,-0.5371935563684052,-1.395659042007815,-9.389014551698807e-2,0.6805317591095806,-1.3323145895088062,-0.10276147847366618,0.5371935563684052,-1.395659042007815,-9.80752650297275e-2,0.6805317591095806,-1.3320130506990422,-9.837638638558734e-2,0.5371935563684052,-1.395974988886971,-0.10276147847366618,0.5371935563684052,-1.395659042007815,-9.389014551698807e-2,0.6805317591095806,-1.3323145895088062,-9.837638638558734e-2,0.5371935563684052,-1.395974988886971,-0.10632181565522554,0.3879697486086787,-1.4440139007919195,-0.10276147847366618,0.5371935563684052,-1.395659042007815,-0.10178479497836379,0.3879697486086787,-1.4443407941603434,-0.10632181565522554,0.3879697486086787,-1.4440139007919195,-9.837638638558734e-2,0.5371935563684052,-1.395974988886971,-0.10178479497836379,0.3879697486086787,-1.4443407941603434,-0.10871726877519594,0.23449526309530616,-1.4765478410996122,-0.10632181565522554,0.3879697486086787,-1.4440139007919195,-0.10407802805751959,0.23449526309530616,-1.4768820994451528,-0.10871726877519594,0.23449526309530616,-1.4765478410996122,-0.10178479497836379,0.3879697486086787,-1.4443407941603434,-0.10407802805751959,0.23449526309530616,-1.4768820994451528,-0.10992159274791416,7.84515984081728e-2,-1.492904414272717,-0.10871726877519594,0.23449526309530616,-1.4765478410996122,-0.1052309604815496,7.84515984081728e-2,-1.4932423753909925,-0.10992159274791416,7.84515984081728e-2,-1.492904414272717,-0.10407802805751959,0.23449526309530616,-1.4768820994451528,-0.1052309604815496,7.84515984081728e-2,-1.4932423753909925,-0.10992159274791416,-7.84515984081728e-2,-1.492904414272717,-0.10992159274791416,7.84515984081728e-2,-1.492904414272717,-0.1052309604815496,-7.84515984081728e-2,-1.4932423753909925,-0.10992159274791416,-7.84515984081728e-2,-1.492904414272717,-0.1052309604815496,7.84515984081728e-2,-1.4932423753909925,-0.1052309604815496,-7.84515984081728e-2,-1.4932423753909925,-0.10871726877519594,-0.23449526309530605,-1.4765478410996122,-0.10992159274791416,-7.84515984081728e-2,-1.492904414272717,-0.10407802805751959,-0.23449526309530605,-1.4768820994451528,-0.10871726877519594,-0.23449526309530605,-1.4765478410996122,-0.1052309604815496,-7.84515984081728e-2,-1.4932423753909925,-0.10407802805751959,-0.23449526309530605,-1.4768820994451528,-0.10632181565522554,-0.3879697486086786,-1.4440139007919195,-0.10871726877519594,-0.23449526309530605,-1.4765478410996122,-0.10178479497836379,-0.3879697486086786,-1.4443407941603434,-0.10632181565522554,-0.3879697486086786,-1.4440139007919195,-0.10407802805751959,-0.23449526309530605,-1.4768820994451528,-0.10178479497836379,-0.3879697486086786,-1.4443407941603434,-0.10276147847366618,-0.5371935563684052,-1.395659042007815,-0.10632181565522554,-0.3879697486086786,-1.4440139007919195,-9.837638638558734e-2,-0.5371935563684052,-1.395974988886971,-0.10276147847366618,-0.5371935563684052,-1.395659042007815,-0.10178479497836379,-0.3879697486086786,-1.4443407941603434,-9.837638638558734e-2,-0.5371935563684052,-1.395974988886971,-9.80752650297275e-2,-0.6805317591095806,-1.3320130506990422,-0.10276147847366618,-0.5371935563684052,-1.395659042007815,-9.389014551698807e-2,-0.6805317591095806,-1.3323145895088062,-9.80752650297275e-2,-0.6805317591095806,-1.3320130506990422,-9.837638638558734e-2,-0.5371935563684052,-1.395974988886971,-0.28391600144533374,0.6805317591095806,-1.3050935709620646,-0.3017765273646781,0.5371935563684052,-1.3665119869682743,-0.2880146660076143,0.6805317591095806,-1.304195183631554,-0.297482021200867,0.5371935563684052,-1.3674533008693615,-0.3017765273646781,0.5371935563684052,-1.3665119869682743,-0.28391600144533374,0.6805317591095806,-1.3050935709620646,-0.297482021200867,0.5371935563684052,-1.3674533008693615,-0.31223206193713615,0.3879697486086787,-1.4138569990147531,-0.3017765273646781,0.5371935563684052,-1.3665119869682743,-0.30778876568000846,0.3879697486086787,-1.4148309262543337,-0.31223206193713615,0.3879697486086787,-1.4138569990147531,-0.297482021200867,0.5371935563684052,-1.3674533008693615,-0.30778876568000846,0.3879697486086787,-1.4148309262543337,-0.3192667167002515,0.23449526309530616,-1.445711497911428,-0.31223206193713615,0.3879697486086787,-1.4138569990147531,-0.3147233120334197,0.23449526309530616,-1.4467073679388582,-0.3192667167002515,0.23449526309530616,-1.445711497911428,-0.30778876568000846,0.3879697486086787,-1.4148309262543337,-0.3147233120334197,0.23449526309530616,-1.4467073679388582,-0.32280341850434313,7.84515984081728e-2,-1.4617264791024047,-0.3192667167002515,0.23449526309530616,-1.445711497911428,-0.3182096839201057,7.84515984081728e-2,-1.4627333809572645,-0.32280341850434313,7.84515984081728e-2,-1.4617264791024047,-0.3147233120334197,0.23449526309530616,-1.4467073679388582,-0.3182096839201057,7.84515984081728e-2,-1.4627333809572645,-0.32280341850434313,-7.84515984081728e-2,-1.4617264791024047,-0.32280341850434313,7.84515984081728e-2,-1.4617264791024047,-0.3182096839201057,-7.84515984081728e-2,-1.4627333809572645,-0.32280341850434313,-7.84515984081728e-2,-1.4617264791024047,-0.3182096839201057,7.84515984081728e-2,-1.4627333809572645,-0.3182096839201057,-7.84515984081728e-2,-1.4627333809572645,-0.3192667167002515,-0.23449526309530605,-1.445711497911428,-0.32280341850434313,-7.84515984081728e-2,-1.4617264791024047,-0.3147233120334197,-0.23449526309530605,-1.4467073679388582,-0.3192667167002515,-0.23449526309530605,-1.445711497911428,-0.3182096839201057,-7.84515984081728e-2,-1.4627333809572645,-0.3147233120334197,-0.23449526309530605,-1.4467073679388582,-0.31223206193713615,-0.3879697486086786,-1.4138569990147531,-0.3192667167002515,-0.23449526309530605,-1.445711497911428,-0.30778876568000846,-0.3879697486086786,-1.4148309262543337,-0.31223206193713615,-0.3879697486086786,-1.4138569990147531,-0.3147233120334197,-0.23449526309530605,-1.4467073679388582,-0.30778876568000846,-0.3879697486086786,-1.4148309262543337,-0.3017765273646781,-0.5371935563684052,-1.3665119869682743,-0.31223206193713615,-0.3879697486086786,-1.4138569990147531,-0.297482021200867,-0.5371935563684052,-1.3674533008693615,-0.3017765273646781,-0.5371935563684052,-1.3665119869682743,-0.30778876568000846,-0.3879697486086786,-1.4148309262543337,-0.297482021200867,-0.5371935563684052,-1.3674533008693615,-0.2880146660076143,-0.6805317591095806,-1.304195183631554,-0.3017765273646781,-0.5371935563684052,-1.3665119869682743,-0.28391600144533374,-0.6805317591095806,-1.3050935709620646,-0.2880146660076143,-0.6805317591095806,-1.304195183631554,-0.297482021200867,-0.5371935563684052,-1.3674533008693615,-0.4680768053088452,0.6805317591095806,-1.250912318739841,-0.4945575667742619,0.5371935563684052,-1.3091359349366996,-0.47200434593078144,0.6805317591095806,-1.2494356415060661,-0.49044236116198825,0.5371935563684052,-1.3106831704779456,-0.4945575667742619,0.5371935563684052,-1.3091359349366996,-0.4680768053088452,0.6805317591095806,-1.250912318739841,-0.49044236116198825,0.5371935563684052,-1.3106831704779456,-0.5116923114233392,0.3879697486086787,-1.35449306110986,-0.4945575667742619,0.5371935563684052,-1.3091359349366996,-0.5074345278745784,0.3879697486086787,-1.3560939031222075,-0.5116923114233392,0.3879697486086787,-1.35449306110986,-0.49044236116198825,0.5371935563684052,-1.3106831704779456,-0.5074345278745784,0.3879697486086787,-1.3560939031222075,-0.5232208480299622,0.23449526309530616,-1.3850100778596055,-0.5116923114233392,0.3879697486086787,-1.35449306110986,-0.518867135712273,0.23449526309530616,-1.3866469871829692,-0.5232208480299622,0.23449526309530616,-1.3850100778596055,-0.5074345278745784,0.3879697486086787,-1.3560939031222075,-0.518867135712273,0.23449526309530616,-1.3866469871829692,-0.5290168675345674,7.84515984081728e-2,-1.400352634364329,-0.5232208480299622,0.23449526309530616,-1.3850100778596055,-0.5246149266311755,7.84515984081728e-2,-1.402007676677617,-0.5290168675345674,7.84515984081728e-2,-1.400352634364329,-0.518867135712273,0.23449526309530616,-1.3866469871829692,-0.5246149266311755,7.84515984081728e-2,-1.402007676677617,-0.5290168675345674,-7.84515984081728e-2,-1.400352634364329,-0.5290168675345674,7.84515984081728e-2,-1.400352634364329,-0.5246149266311755,-7.84515984081728e-2,-1.402007676677617,-0.5290168675345674,-7.84515984081728e-2,-1.400352634364329,-0.5246149266311755,7.84515984081728e-2,-1.402007676677617,-0.5246149266311755,-7.84515984081728e-2,-1.402007676677617,-0.5232208480299622,-0.23449526309530605,-1.3850100778596055,-0.5290168675345674,-7.84515984081728e-2,-1.400352634364329,-0.518867135712273,-0.23449526309530605,-1.3866469871829692,-0.5232208480299622,-0.23449526309530605,-1.3850100778596055,-0.5246149266311755,-7.84515984081728e-2,-1.402007676677617,-0.518867135712273,-0.23449526309530605,-1.3866469871829692,-0.5116923114233392,-0.3879697486086786,-1.35449306110986,-0.5232208480299622,-0.23449526309530605,-1.3850100778596055,-0.5074345278745784,-0.3879697486086786,-1.3560939031222075,-0.5116923114233392,-0.3879697486086786,-1.35449306110986,-0.518867135712273,-0.23449526309530605,-1.3866469871829692,-0.5074345278745784,-0.3879697486086786,-1.3560939031222075,-0.4945575667742619,-0.5371935563684052,-1.3091359349366996,-0.5116923114233392,-0.3879697486086786,-1.35449306110986,-0.49044236116198825,-0.5371935563684052,-1.3106831704779456,-0.4945575667742619,-0.5371935563684052,-1.3091359349366996,-0.5074345278745784,-0.3879697486086786,-1.3560939031222075,-0.49044236116198825,-0.5371935563684052,-1.3106831704779456,-0.47200434593078144,-0.6805317591095806,-1.2494356415060661,-0.4945575667742619,-0.5371935563684052,-1.3091359349366996,-0.4680768053088452,-0.6805317591095806,-1.250912318739841,-0.47200434593078144,-0.6805317591095806,-1.2494356415060661,-0.49044236116198825,-0.5371935563684052,-1.3106831704779456,-0.6425682181570462,0.6805317591095806,-1.1708900929881074,-0.6771221834974769,0.5371935563684052,-1.2247161432805052,-0.6462435008760687,0.6805317591095806,-1.1688656306088663,-0.6732712891266994,0.5371935563684052,-1.226837338131644,-0.6771221834974769,0.5371935563684052,-1.2247161432805052,-0.6425682181570462,0.6805317591095806,-1.1708900929881074,-0.6732712891266994,0.5371935563684052,-1.226837338131644,-0.7005821737795601,0.3879697486086787,-1.2671484095980334,-0.6771221834974769,0.5371935563684052,-1.2247161432805052,-0.6965978589616058,0.3879697486086787,-1.2693430966663923,-0.7005821737795601,0.3879697486086787,-1.2671484095980334,-0.6732712891266994,0.5371935563684052,-1.226837338131644,-0.6965978589616058,0.3879697486086787,-1.2693430966663923,-0.7163664391594695,0.23449526309530616,-1.2956975327721534,-0.7005821737795601,0.3879697486086787,-1.2671484095980334,-0.712292356881255,0.23449526309530616,-1.2979416666069437,-0.7163664391594695,0.23449526309530616,-1.2956975327721534,-0.6965978589616058,0.3879697486086787,-1.2693430966663923,-0.712292356881255,0.23449526309530616,-1.2979416666069437,-0.7243020439226328,7.84515984081728e-2,-1.3100507226350817,-0.7163664391594695,0.23449526309530616,-1.2956975327721534,-0.7201828306821547,7.84515984081728e-2,-1.3123197160363942,-0.7243020439226328,7.84515984081728e-2,-1.3100507226350817,-0.712292356881255,0.23449526309530616,-1.2979416666069437,-0.7201828306821547,7.84515984081728e-2,-1.3123197160363942,-0.7243020439226328,-7.84515984081728e-2,-1.3100507226350817,-0.7243020439226328,7.84515984081728e-2,-1.3100507226350817,-0.7201828306821547,-7.84515984081728e-2,-1.3123197160363942,-0.7243020439226328,-7.84515984081728e-2,-1.3100507226350817,-0.7201828306821547,7.84515984081728e-2,-1.3123197160363942,-0.7201828306821547,-7.84515984081728e-2,-1.3123197160363942,-0.7163664391594695,-0.23449526309530605,-1.2956975327721534,-0.7243020439226328,-7.84515984081728e-2,-1.3100507226350817,-0.712292356881255,-0.23449526309530605,-1.2979416666069437,-0.7163664391594695,-0.23449526309530605,-1.2956975327721534,-0.7201828306821547,-7.84515984081728e-2,-1.3123197160363942,-0.712292356881255,-0.23449526309530605,-1.2979416666069437,-0.7005821737795601,-0.3879697486086786,-1.2671484095980334,-0.7163664391594695,-0.23449526309530605,-1.2956975327721534,-0.6965978589616058,-0.3879697486086786,-1.2693430966663923,-0.7005821737795601,-0.3879697486086786,-1.2671484095980334,-0.712292356881255,-0.23449526309530605,-1.2979416666069437,-0.6965978589616058,-0.3879697486086786,-1.2693430966663923,-0.6771221834974769,-0.5371935563684052,-1.2247161432805052,-0.7005821737795601,-0.3879697486086786,-1.2671484095980334,-0.6732712891266994,-0.5371935563684052,-1.226837338131644,-0.6771221834974769,-0.5371935563684052,-1.2247161432805052,-0.6965978589616058,-0.3879697486086786,-1.2693430966663923,-0.6732712891266994,-0.5371935563684052,-1.226837338131644,-0.6462435008760687,-0.6805317591095806,-1.1688656306088663,-0.6771221834974769,-0.5371935563684052,-1.2247161432805052,-0.6425682181570462,-0.6805317591095806,-1.1708900929881074,-0.6462435008760687,-0.6805317591095806,-1.1688656306088663,-0.6732712891266994,-0.5371935563684052,-1.226837338131644,-0.8037856484346475,0.6805317591095806,-1.0666799689755657,-0.8456990121416448,0.5371935563684052,-1.114996530873039,-0.8071327503566955,0.6805317591095806,-1.0641495421907845,-0.8421919796395408,0.5371935563684052,-1.1176478660235956,-0.8456990121416448,0.5371935563684052,-1.114996530873039,-0.8037856484346475,0.6805317591095806,-1.0666799689755657,-0.8421919796395408,0.5371935563684052,-1.1176478660235956,-0.8749996185756745,0.3879697486086787,-1.1536273842350238,-0.8456990121416448,0.5371935563684052,-1.114996530873039,-0.8713710792755016,0.3879697486086787,-1.156370579168619,-0.8749996185756745,0.3879697486086787,-1.1536273842350238,-0.8421919796395408,0.5371935563684052,-1.1176478660235956,-0.8713710792755016,0.3879697486086787,-1.156370579168619,-0.8947135460831471,0.23449526309530616,-1.1796188545632795,-0.8749996185756745,0.3879697486086787,-1.1536273842350238,-0.8910032550207987,0.23449526309530616,-1.1824238542621701,-0.8947135460831471,0.23449526309530616,-1.1796188545632795,-0.8713710792755016,0.3879697486086787,-1.156370579168619,-0.8910032550207987,0.23449526309530616,-1.1824238542621701,-0.9046248047488865,7.84515984081728e-2,-1.1926861738698247,-0.8947135460831471,0.23449526309530616,-1.1796188545632795,-0.9008734126496697,7.84515984081728e-2,-1.1955222461703259,-0.9046248047488865,7.84515984081728e-2,-1.1926861738698247,-0.8910032550207987,0.23449526309530616,-1.1824238542621701,-0.9008734126496697,7.84515984081728e-2,-1.1955222461703259,-0.9046248047488865,-7.84515984081728e-2,-1.1926861738698247,-0.9046248047488865,7.84515984081728e-2,-1.1926861738698247,-0.9008734126496697,-7.84515984081728e-2,-1.1955222461703259,-0.9046248047488865,-7.84515984081728e-2,-1.1926861738698247,-0.9008734126496697,7.84515984081728e-2,-1.1955222461703259,-0.9008734126496697,-7.84515984081728e-2,-1.1955222461703259,-0.8947135460831471,-0.23449526309530605,-1.1796188545632795,-0.9046248047488865,-7.84515984081728e-2,-1.1926861738698247,-0.8910032550207987,-0.23449526309530605,-1.1824238542621701,-0.8947135460831471,-0.23449526309530605,-1.1796188545632795,-0.9008734126496697,-7.84515984081728e-2,-1.1955222461703259,-0.8910032550207987,-0.23449526309530605,-1.1824238542621701,-0.8749996185756745,-0.3879697486086786,-1.1536273842350238,-0.8947135460831471,-0.23449526309530605,-1.1796188545632795,-0.8713710792755016,-0.3879697486086786,-1.156370579168619,-0.8749996185756745,-0.3879697486086786,-1.1536273842350238,-0.8910032550207987,-0.23449526309530605,-1.1824238542621701,-0.8713710792755016,-0.3879697486086786,-1.156370579168619,-0.8456990121416448,-0.5371935563684052,-1.114996530873039,-0.8749996185756745,-0.3879697486086786,-1.1536273842350238,-0.8421919796395408,-0.5371935563684052,-1.1176478660235956,-0.8456990121416448,-0.5371935563684052,-1.114996530873039,-0.8713710792755016,-0.3879697486086786,-1.156370579168619,-0.8421919796395408,-0.5371935563684052,-1.1176478660235956,-0.8071327503566955,-0.6805317591095806,-1.0641495421907845,-0.8456990121416448,-0.5371935563684052,-1.114996530873039,-0.8037856484346475,-0.6805317591095806,-1.0666799689755657,-0.8071327503566955,-0.6805317591095806,-1.0641495421907845,-0.8421919796395408,-0.5371935563684052,-1.1176478660235956,-0.9483987145586341,0.6805317591095806,-0.9404346883578628,-0.9968056428676236,0.5371935563684052,-0.9822436527377151,-0.9513484922506512,0.6805317591095806,-0.9374505699692292,-0.9937149194654645,0.5371935563684052,-0.9853703576970493,-0.9968056428676236,0.5371935563684052,-0.9822436527377151,-0.9483987145586341,0.6805317591095806,-0.9404346883578628,-0.9937149194654645,0.5371935563684052,-0.9853703576970493,-1.031341582266347,0.3879697486086787,-1.016275068499108,-0.9968056428676236,0.5371935563684052,-0.9822436527377151,-1.0281437757664158,0.3879697486086787,-1.0195101031953029,-1.031341582266347,0.3879697486086787,-1.016275068499108,-0.9937149194654645,0.5371935563684052,-0.9853703576970493,-1.0281437757664158,0.3879697486086787,-1.0195101031953029,-1.0545779274676597,0.23449526309530616,-1.0391719619407944,-1.031341582266347,0.3879697486086787,-1.016275068499108,-1.0513080737071532,0.23449526309530616,-1.0424798826567439,-1.0545779274676597,0.23449526309530616,-1.0391719619407944,-1.0281437757664158,0.3879697486086787,-1.0195101031953029,-1.0513080737071532,0.23449526309530616,-1.0424798826567439,-1.0662600962110171,7.84515984081728e-2,-1.0506834699067438,-1.0545779274676597,0.23449526309530616,-1.0391719619407944,-1.062954020391997,7.84515984081728e-2,-1.0540280343708575,-1.0662600962110171,7.84515984081728e-2,-1.0506834699067438,-1.0513080737071532,0.23449526309530616,-1.0424798826567439,-1.062954020391997,7.84515984081728e-2,-1.0540280343708575,-1.0662600962110171,-7.84515984081728e-2,-1.0506834699067438,-1.0662600962110171,7.84515984081728e-2,-1.0506834699067438,-1.062954020391997,-7.84515984081728e-2,-1.0540280343708575,-1.0662600962110171,-7.84515984081728e-2,-1.0506834699067438,-1.062954020391997,7.84515984081728e-2,-1.0540280343708575,-1.062954020391997,-7.84515984081728e-2,-1.0540280343708575,-1.0545779274676597,-0.23449526309530605,-1.0391719619407944,-1.0662600962110171,-7.84515984081728e-2,-1.0506834699067438,-1.0513080737071532,-0.23449526309530605,-1.0424798826567439,-1.0545779274676597,-0.23449526309530605,-1.0391719619407944,-1.062954020391997,-7.84515984081728e-2,-1.0540280343708575,-1.0513080737071532,-0.23449526309530605,-1.0424798826567439,-1.031341582266347,-0.3879697486086786,-1.016275068499108,-1.0545779274676597,-0.23449526309530605,-1.0391719619407944,-1.0281437757664158,-0.3879697486086786,-1.0195101031953029,-1.031341582266347,-0.3879697486086786,-1.016275068499108,-1.0513080737071532,-0.23449526309530605,-1.0424798826567439,-1.0281437757664158,-0.3879697486086786,-1.0195101031953029,-0.9968056428676236,-0.5371935563684052,-0.9822436527377151,-1.031341582266347,-0.3879697486086786,-1.016275068499108,-0.9937149194654645,-0.5371935563684052,-0.9853703576970493,-0.9968056428676236,-0.5371935563684052,-0.9822436527377151,-1.0281437757664158,-0.3879697486086786,-1.0195101031953029,-0.9937149194654645,-0.5371935563684052,-0.9853703576970493,-0.9513484922506512,-0.6805317591095806,-0.9374505699692292,-0.9968056428676236,-0.5371935563684052,-0.9822436527377151,-0.9483987145586341,-0.6805317591095806,-0.9404346883578628,-0.9513484922506512,-0.6805317591095806,-0.9374505699692292,-0.9937149194654645,-0.5371935563684052,-0.9853703576970493,-1.0734200429475405,0.6805317591095806,-0.7947621884826434,-1.1273205599734533,0.5371935563684052,-0.8291998782304573,-1.0759115607818945,0.6805317591095806,-0.7913860235176559,-1.1247099929133066,0.5371935563684052,-0.8327373624602267,-1.1273205599734533,0.5371935563684052,-0.8291998782304573,-1.0734200429475405,0.6805317591095806,-0.7947621884826434,-1.1247099929133066,0.5371935563684052,-0.8327373624602267,-1.1663783992029495,0.3879697486086787,-0.8579288455561358,-1.1273205599734533,0.5371935563684052,-0.8291998782304573,-1.1636773848360178,0.3879697486086787,-0.8615888916331996,-1.1663783992029495,0.3879697486086787,-0.8579288455561358,-1.1247099929133066,0.5371935563684052,-0.8327373624602267,-1.1636773848360178,0.3879697486086787,-0.8615888916331996,-1.1926571526104066,0.23449526309530616,-0.8772581649167485,-1.1663783992029495,0.3879697486086787,-0.8579288455561358,-1.1898952838153176,0.23449526309530616,-0.8810006726103724,-1.1926571526104066,0.23449526309530616,-0.8772581649167485,-1.1636773848360178,0.3879697486086787,-0.8615888916331996,-1.1898952838153176,0.23449526309530616,-0.8810006726103724,-1.2058689046743094,7.84515984081728e-2,-0.8869760602444602,-1.1926571526104066,0.23449526309530616,-0.8772581649167485,-1.2030764410635404,7.84515984081728e-2,-0.8907600258571826,-1.2058689046743094,7.84515984081728e-2,-0.8869760602444602,-1.1898952838153176,0.23449526309530616,-0.8810006726103724,-1.2030764410635404,7.84515984081728e-2,-0.8907600258571826,-1.2058689046743094,-7.84515984081728e-2,-0.8869760602444602,-1.2058689046743094,7.84515984081728e-2,-0.8869760602444602,-1.2030764410635404,-7.84515984081728e-2,-0.8907600258571826,-1.2058689046743094,-7.84515984081728e-2,-0.8869760602444602,-1.2030764410635404,7.84515984081728e-2,-0.8907600258571826,-1.2030764410635404,-7.84515984081728e-2,-0.8907600258571826,-1.1926571526104066,-0.23449526309530605,-0.8772581649167485,-1.2058689046743094,-7.84515984081728e-2,-0.8869760602444602,-1.1898952838153176,-0.23449526309530605,-0.8810006726103724,-1.1926571526104066,-0.23449526309530605,-0.8772581649167485,-1.2030764410635404,-7.84515984081728e-2,-0.8907600258571826,-1.1898952838153176,-0.23449526309530605,-0.8810006726103724,-1.1663783992029495,-0.3879697486086786,-0.8579288455561358,-1.1926571526104066,-0.23449526309530605,-0.8772581649167485,-1.1636773848360178,-0.3879697486086786,-0.8615888916331996,-1.1663783992029495,-0.3879697486086786,-0.8579288455561358,-1.1898952838153176,-0.23449526309530605,-0.8810006726103724,-1.1636773848360178,-0.3879697486086786,-0.8615888916331996,-1.1273205599734533,-0.5371935563684052,-0.8291998782304573,-1.1663783992029495,-0.3879697486086786,-0.8579288455561358,-1.1247099929133066,-0.5371935563684052,-0.8327373624602267,-1.1273205599734533,-0.5371935563684052,-0.8291998782304573,-1.1636773848360178,-0.3879697486086786,-0.8615888916331996,-1.1247099929133066,-0.5371935563684052,-0.8327373624602267,-1.0759115607818945,-0.6805317591095806,-0.7913860235176559,-1.1273205599734533,-0.5371935563684052,-0.8291998782304573,-1.0734200429475405,-0.6805317591095806,-0.7947621884826434,-1.0759115607818945,-0.6805317591095806,-0.7913860235176559,-1.1247099929133066,-0.5371935563684052,-0.8327373624602267,-1.1762669802937589,0.6805317591095806,-0.6326717283977703,-1.234547625234564,0.5371935563684052,-0.659026739991606,-1.1782487692382544,0.6805317591095806,-0.6289732606651561,-1.2324711428320194,0.5371935563684052,-0.6629019271978336,-1.234547625234564,0.5371935563684052,-0.659026739991606,-1.1762669802937589,0.6805317591095806,-0.6326717283977703,-1.2324711428320194,0.5371935563684052,-0.6629019271978336,-1.2773205191031043,0.3879697486086787,-0.6818597844444964,-1.234547625234564,0.5371935563684052,-0.659026739991606,-1.2751720936182447,0.3879697486086787,-0.6858692337623709,-1.2773205191031043,0.3879697486086787,-0.6818597844444964,-1.2324711428320194,0.5371935563684052,-0.6629019271978336,-1.2751720936182447,0.3879697486086787,-0.6858692337623709,-1.3060988220678482,0.23449526309530616,-0.6972222304106797,-1.2773205191031043,0.3879697486086787,-0.6818597844444964,-1.3039019920998727,0.23449526309530616,-0.7013220134744433,-1.3060988220678482,0.23449526309530616,-0.6972222304106797,-1.2751720936182447,0.3879697486086787,-0.6858692337623709,-1.3039019920998727,0.23449526309530616,-0.7013220134744433,-1.3205672330192668,7.84515984081728e-2,-0.7049457637173518,-1.3060988220678482,0.23449526309530616,-0.6972222304106797,-1.3183460674970207,7.84515984081728e-2,-0.7090909624458233,-1.3205672330192668,7.84515984081728e-2,-0.7049457637173518,-1.3039019920998727,0.23449526309530616,-0.7013220134744433,-1.3183460674970207,7.84515984081728e-2,-0.7090909624458233,-1.3205672330192668,-7.84515984081728e-2,-0.7049457637173518,-1.3205672330192668,7.84515984081728e-2,-0.7049457637173518,-1.3183460674970207,-7.84515984081728e-2,-0.7090909624458233,-1.3205672330192668,-7.84515984081728e-2,-0.7049457637173518,-1.3183460674970207,7.84515984081728e-2,-0.7090909624458233,-1.3183460674970207,-7.84515984081728e-2,-0.7090909624458233,-1.3060988220678482,-0.23449526309530605,-0.6972222304106797,-1.3205672330192668,-7.84515984081728e-2,-0.7049457637173518,-1.3039019920998727,-0.23449526309530605,-0.7013220134744433,-1.3060988220678482,-0.23449526309530605,-0.6972222304106797,-1.3183460674970207,-7.84515984081728e-2,-0.7090909624458233,-1.3039019920998727,-0.23449526309530605,-0.7013220134744433,-1.2773205191031043,-0.3879697486086786,-0.6818597844444964,-1.3060988220678482,-0.23449526309530605,-0.6972222304106797,-1.2751720936182447,-0.3879697486086786,-0.6858692337623709,-1.2773205191031043,-0.3879697486086786,-0.6818597844444964,-1.3039019920998727,-0.23449526309530605,-0.7013220134744433,-1.2751720936182447,-0.3879697486086786,-0.6858692337623709,-1.234547625234564,-0.5371935563684052,-0.659026739991606,-1.2773205191031043,-0.3879697486086786,-0.6818597844444964,-1.2324711428320194,-0.5371935563684052,-0.6629019271978336,-1.234547625234564,-0.5371935563684052,-0.659026739991606,-1.2751720936182447,-0.3879697486086786,-0.6858692337623709,-1.2324711428320194,-0.5371935563684052,-0.6629019271978336,-1.1782487692382544,-0.6805317591095806,-0.6289732606651561,-1.234547625234564,-0.5371935563684052,-0.659026739991606,-1.1762669802937589,-0.6805317591095806,-0.6326717283977703,-1.1782487692382544,-0.6805317591095806,-0.6289732606651561,-1.2324711428320194,-0.5371935563684052,-0.6629019271978336,-1.254814945245151,0.6805317591095806,-0.45751172447559085,-1.3162717739229577,0.5371935563684052,-0.47523962394772545,-1.2562460660949462,0.6805317591095806,-0.45356735581850705,-1.3147722715320656,0.5371935563684052,-0.4793724616690843,-1.3162717739229577,0.5371935563684052,-0.47523962394772545,-1.254814945245151,0.6805317591095806,-0.45751172447559085,-1.3147722715320656,0.5371935563684052,-0.4793724616690843,-1.3618761327482927,0.3879697486086787,-0.4917050673066876,-1.3162717739229577,0.5371935563684052,-0.47523962394772545,-1.3603246776782891,0.3879697486086787,-0.4959810938573944,-1.3618761327482927,0.3879697486086787,-0.4917050673066876,-1.3147722715320656,0.5371935563684052,-0.4793724616690843,-1.3603246776782891,0.3879697486086787,-0.4959810938573944,-1.3925594916723345,0.23449526309530616,-0.5027832870523374,-1.3618761327482927,0.3879697486086787,-0.4917050673066876,-1.3909730819896298,0.23449526309530616,-0.5071556533906865,-1.3925594916723345,0.23449526309530616,-0.5027832870523374,-1.3603246776782891,0.3879697486086787,-0.4959810938573944,-1.3909730819896298,0.23449526309530616,-0.5071556533906865,-1.407985677393806,7.84515984081728e-2,-0.5083529078908748,-1.3925594916723345,0.23449526309530616,-0.5027832870523374,-1.4063816941348612,7.84515984081728e-2,-0.5127737094562868,-1.407985677393806,7.84515984081728e-2,-0.5083529078908748,-1.3909730819896298,0.23449526309530616,-0.5071556533906865,-1.4063816941348612,7.84515984081728e-2,-0.5127737094562868,-1.407985677393806,-7.84515984081728e-2,-0.5083529078908748,-1.407985677393806,7.84515984081728e-2,-0.5083529078908748,-1.4063816941348612,-7.84515984081728e-2,-0.5127737094562868,-1.407985677393806,-7.84515984081728e-2,-0.5083529078908748,-1.4063816941348612,7.84515984081728e-2,-0.5127737094562868,-1.4063816941348612,-7.84515984081728e-2,-0.5127737094562868,-1.3925594916723345,-0.23449526309530605,-0.5027832870523374,-1.407985677393806,-7.84515984081728e-2,-0.5083529078908748,-1.3909730819896298,-0.23449526309530605,-0.5071556533906865,-1.3925594916723345,-0.23449526309530605,-0.5027832870523374,-1.4063816941348612,-7.84515984081728e-2,-0.5127737094562868,-1.3909730819896298,-0.23449526309530605,-0.5071556533906865,-1.3618761327482927,-0.3879697486086786,-0.4917050673066876,-1.3925594916723345,-0.23449526309530605,-0.5027832870523374,-1.3603246776782891,-0.3879697486086786,-0.4959810938573944,-1.3618761327482927,-0.3879697486086786,-0.4917050673066876,-1.3909730819896298,-0.23449526309530605,-0.5071556533906865,-1.3603246776782891,-0.3879697486086786,-0.4959810938573944,-1.3162717739229577,-0.5371935563684052,-0.47523962394772545,-1.3618761327482927,-0.3879697486086786,-0.4917050673066876,-1.3147722715320656,-0.5371935563684052,-0.4793724616690843,-1.3162717739229577,-0.5371935563684052,-0.47523962394772545,-1.3603246776782891,-0.3879697486086786,-0.4959810938573944,-1.3147722715320656,-0.5371935563684052,-0.4793724616690843,-1.2562460660949462,-0.6805317591095806,-0.45356735581850705,-1.3162717739229577,-0.5371935563684052,-0.47523962394772545,-1.254814945245151,-0.6805317591095806,-0.45751172447559085,-1.2562460660949462,-0.6805317591095806,-0.45356735581850705,-1.3147722715320656,-0.5371935563684052,-0.4793724616690843,-1.3074413173728512,0.6805317591095806,-0.27290057982636023,-1.3708047729535966,0.5371935563684052,-0.2816351495177828,-1.3082922064603395,0.6805317591095806,-0.2687917918359047,-1.3699132268474412,0.5371935563684052,-0.28594026282546625,-1.3708047729535966,0.5371935563684052,-0.2816351495177828,-1.3074413173728512,0.6805317591095806,-0.27290057982636023,-1.3699132268474412,0.5371935563684052,-0.28594026282546625,-1.4182985154950336,0.3879697486086787,-0.29139285356560013,-1.3708047729535966,0.5371935563684052,-0.2816351495177828,-1.417376080335918,0.3879697486086787,-0.2958471244682096,-1.4182985154950336,0.3879697486086787,-0.29139285356560013,-1.3699132268474412,0.5371935563684052,-0.28594026282546625,-1.417376080335918,0.3879697486086787,-0.2958471244682096,-1.4502530827026616,0.23449526309530616,-0.297957996531878,-1.4182985154950336,0.3879697486086787,-0.29139285356560013,-1.4493098648832214,0.23449526309530616,-0.3025126231052883,-1.4502530827026616,0.23449526309530616,-0.297957996531878,-1.417376080335918,0.3879697486086787,-0.2958471244682096,-1.4493098648832214,0.23449526309530616,-0.3025126231052883,-1.4663183736512309,7.84515984081728e-2,-0.3012586493364332,-1.4502530827026616,0.23449526309530616,-0.297957996531878,-1.4653647072632767,7.84515984081728e-2,-0.30586373013878926,-1.4663183736512309,7.84515984081728e-2,-0.3012586493364332,-1.4493098648832214,0.23449526309530616,-0.3025126231052883,-1.4653647072632767,7.84515984081728e-2,-0.30586373013878926,-1.4663183736512309,-7.84515984081728e-2,-0.3012586493364332,-1.4663183736512309,7.84515984081728e-2,-0.3012586493364332,-1.4653647072632767,-7.84515984081728e-2,-0.30586373013878926,-1.4663183736512309,-7.84515984081728e-2,-0.3012586493364332,-1.4653647072632767,7.84515984081728e-2,-0.30586373013878926,-1.4653647072632767,-7.84515984081728e-2,-0.30586373013878926,-1.4502530827026616,-0.23449526309530605,-0.297957996531878,-1.4663183736512309,-7.84515984081728e-2,-0.3012586493364332,-1.4493098648832214,-0.23449526309530605,-0.3025126231052883,-1.4502530827026616,-0.23449526309530605,-0.297957996531878,-1.4653647072632767,-7.84515984081728e-2,-0.30586373013878926,-1.4493098648832214,-0.23449526309530605,-0.3025126231052883,-1.4182985154950336,-0.3879697486086786,-0.29139285356560013,-1.4502530827026616,-0.23449526309530605,-0.297957996531878,-1.417376080335918,-0.3879697486086786,-0.2958471244682096,-1.4182985154950336,-0.3879697486086786,-0.29139285356560013,-1.4493098648832214,-0.23449526309530605,-0.3025126231052883,-1.417376080335918,-0.3879697486086786,-0.2958471244682096,-1.3708047729535966,-0.5371935563684052,-0.2816351495177828,-1.4182985154950336,-0.3879697486086786,-0.29139285356560013,-1.3699132268474412,-0.5371935563684052,-0.28594026282546625,-1.3708047729535966,-0.5371935563684052,-0.2816351495177828,-1.417376080335918,-0.3879697486086786,-0.2958471244682096,-1.3699132268474412,-0.5371935563684052,-0.28594026282546625,-1.3082922064603395,-0.6805317591095806,-0.2687917918359047,-1.3708047729535966,-0.5371935563684052,-0.2816351495177828,-1.3074413173728512,-0.6805317591095806,-0.27290057982636023,-1.3082922064603395,-0.6805317591095806,-0.2687917918359047,-1.3699132268474412,-0.5371935563684052,-0.28594026282546625],[1.2563696343765676,0.6809857496093201,-0.45584399814577214,1.371274873764085,0.5375519243179504,-0.2839773745307207,1.3087408693470959,0.6809857496093201,-0.27102720477774944,1.3164012465204369,0.5375519243179504,-0.47762504836064523,1.371274873764085,0.5375519243179504,-0.2839773745307207,1.2563696343765676,0.6809857496093201,-0.45584399814577214,1.3164012465204369,0.5375519243179504,-0.47762504836064523,1.4187849037063995,0.3882285676537812,-0.2938162287422472,1.371274873764085,0.5375519243179504,-0.2839773745307207,1.3620100911327575,0.3882285676537812,-0.4941731385962121,1.4187849037063995,0.3882285676537812,-0.2938162287422472,1.3164012465204369,0.5375519243179504,-0.47762504836064523,1.3620100911327575,0.3882285676537812,-0.4941731385962121,1.450750429343878,0.23465169756034637,-0.3004359708666751,1.4187849037063995,0.3882285676537812,-0.2938162287422472,1.3926964681676928,0.23465169756034637,-0.5053069645129413,1.450750429343878,0.23465169756034637,-0.3004359708666751,1.3620100911327575,0.3882285676537812,-0.4941731385962121,1.3926964681676928,0.23465169756034637,-0.5053069645129413,1.4668212296884202,7.850393436441573e-2,-0.3037640736240189,1.450750429343878,0.23465169756034637,-0.3004359708666751,1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,1.4668212296884202,7.850393436441573e-2,-0.3037640736240189,1.3926964681676928,0.23465169756034637,-0.5053069645129413,1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,1.4668212296884202,-7.850393436441573e-2,-0.3037640736240189,1.4668212296884202,7.850393436441573e-2,-0.3037640736240189,1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,1.4668212296884202,-7.850393436441573e-2,-0.3037640736240189,1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,1.450750429343878,-0.23465169756034626,-0.3004359708666751,1.4668212296884202,-7.850393436441573e-2,-0.3037640736240189,1.3926964681676928,-0.23465169756034626,-0.5053069645129413,1.450750429343878,-0.23465169756034626,-0.3004359708666751,1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,1.3926964681676928,-0.23465169756034626,-0.5053069645129413,1.4187849037063995,-0.3882285676537811,-0.2938162287422472,1.450750429343878,-0.23465169756034626,-0.3004359708666751,1.3620100911327575,-0.3882285676537811,-0.4941731385962121,1.4187849037063995,-0.3882285676537811,-0.2938162287422472,1.3926964681676928,-0.23465169756034626,-0.5053069645129413,1.3620100911327575,-0.3882285676537811,-0.4941731385962121,1.371274873764085,-0.5375519243179504,-0.2839773745307207,1.4187849037063995,-0.3882285676537811,-0.2938162287422472,1.3164012465204369,-0.5375519243179504,-0.47762504836064523,1.371274873764085,-0.5375519243179504,-0.2839773745307207,1.3620100911327575,-0.3882285676537811,-0.4941731385962121,1.3164012465204369,-0.5375519243179504,-0.47762504836064523,1.3087408693470959,-0.6809857496093201,-0.27102720477774944,1.371274873764085,-0.5375519243179504,-0.2839773745307207,1.2563696343765676,-0.6809857496093201,-0.45584399814577214,1.3087408693470959,-0.6809857496093201,-0.27102720477774944,1.3164012465204369,-0.5375519243179504,-0.47762504836064523,1.1780446902780395,0.6809857496093201,-0.6312441021797751,1.3164012465204369,0.5375519243179504,-0.47762504836064523,1.2563696343765676,0.6809857496093201,-0.45584399814577214,1.2343337950127369,0.5375519243179504,-0.6614060864185659,1.3164012465204369,0.5375519243179504,-0.47762504836064523,1.1780446902780395,0.6809857496093201,-0.6312441021797751,1.2343337950127369,0.5375519243179504,-0.6614060864185659,1.3620100911327575,0.3882285676537812,-0.4941731385962121,1.3164012465204369,0.5375519243179504,-0.47762504836064523,1.277099280388322,0.3882285676537812,-0.6843215671664332,1.3620100911327575,0.3882285676537812,-0.4941731385962121,1.2343337950127369,0.5375519243179504,-0.6614060864185659,1.277099280388322,0.3882285676537812,-0.6843215671664332,1.3926964681676928,0.23465169756034637,-0.5053069645129413,1.3620100911327575,0.3882285676537812,-0.4941731385962121,1.3058725987977675,0.23465169756034637,-0.6997394776209307,1.3926964681676928,0.23465169756034637,-0.5053069645129413,1.277099280388322,0.3882285676537812,-0.6843215671664332,1.3058725987977675,0.23465169756034637,-0.6997394776209307,1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,1.3926964681676928,0.23465169756034637,-0.5053069645129413,1.3203385037433746,7.850393436441573e-2,-0.7074908959287137,1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,1.3058725987977675,0.23465169756034637,-0.6997394776209307,1.3203385037433746,7.850393436441573e-2,-0.7074908959287137,1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,1.3203385037433746,-7.850393436441573e-2,-0.7074908959287137,1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,1.3203385037433746,7.850393436441573e-2,-0.7074908959287137,1.3203385037433746,-7.850393436441573e-2,-0.7074908959287137,1.3926964681676928,-0.23465169756034626,-0.5053069645129413,1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,1.3058725987977675,-0.23465169756034626,-0.6997394776209307,1.3926964681676928,-0.23465169756034626,-0.5053069645129413,1.3203385037433746,-7.850393436441573e-2,-0.7074908959287137,1.3058725987977675,-0.23465169756034626,-0.6997394776209307,1.3620100911327575,-0.3882285676537811,-0.4941731385962121,1.3926964681676928,-0.23465169756034626,-0.5053069645129413,1.277099280388322,-0.3882285676537811,-0.6843215671664332,1.3620100911327575,-0.3882285676537811,-0.4941731385962121,1.3058725987977675,-0.23465169756034626,-0.6997394776209307,1.277099280388322,-0.3882285676537811,-0.6843215671664332,1.3164012465204369,-0.5375519243179504,-0.47762504836064523,1.3620100911327575,-0.3882285676537811,-0.4941731385962121,1.2343337950127369,-0.5375519243179504,-0.6614060864185659,1.3164012465204369,-0.5375519243179504,-0.47762504836064523,1.277099280388322,-0.3882285676537811,-0.6843215671664332,1.2343337950127369,-0.5375519243179504,-0.6614060864185659,1.2563696343765676,-0.6809857496093201,-0.45584399814577214,1.3164012465204369,-0.5375519243179504,-0.47762504836064523,1.1780446902780395,-0.6809857496093201,-0.6312441021797751,1.2563696343765676,-0.6809857496093201,-0.45584399814577214,1.2343337950127369,-0.5375519243179504,-0.6614060864185659,1.0753840503821261,0.6809857496093201,-0.7936041538530183,1.2343337950127369,0.5375519243179504,-0.6614060864185659,1.1780446902780395,0.6809857496093201,-0.6312441021797751,1.1267678441732563,0.5375519243179504,-0.8315239948427351,1.2343337950127369,0.5375519243179504,-0.6614060864185659,1.0753840503821261,0.6809857496093201,-0.7936041538530183,1.1267678441732563,0.5375519243179504,-0.8315239948427351,1.277099280388322,0.3882285676537812,-0.6843215671664332,1.2343337950127369,0.5375519243179504,-0.6614060864185659,1.1658065336722943,0.3882285676537812,-0.8603334849373723,1.277099280388322,0.3882285676537812,-0.6843215671664332,1.1267678441732563,0.5375519243179504,-0.8315239948427351,1.1658065336722943,0.3882285676537812,-0.8603334849373723,1.3058725987977675,0.23465169756034637,-0.6997394776209307,1.277099280388322,0.3882285676537812,-0.6843215671664332,1.1920724028277174,0.23465169756034637,-0.8797169813346795,1.3058725987977675,0.23465169756034637,-0.6997394776209307,1.1658065336722943,0.3882285676537812,-0.8603334849373723,1.1920724028277174,0.23465169756034637,-0.8797169813346795,1.3203385037433746,7.850393436441573e-2,-0.7074908959287137,1.3058725987977675,0.23465169756034637,-0.6997394776209307,1.205277677280572,7.850393436441573e-2,-0.8894621143918703,1.3203385037433746,7.850393436441573e-2,-0.7074908959287137,1.1920724028277174,0.23465169756034637,-0.8797169813346795,1.205277677280572,7.850393436441573e-2,-0.8894621143918703,1.3203385037433746,-7.850393436441573e-2,-0.7074908959287137,1.3203385037433746,7.850393436441573e-2,-0.7074908959287137,1.205277677280572,-7.850393436441573e-2,-0.8894621143918703,1.3203385037433746,-7.850393436441573e-2,-0.7074908959287137,1.205277677280572,7.850393436441573e-2,-0.8894621143918703,1.205277677280572,-7.850393436441573e-2,-0.8894621143918703,1.3058725987977675,-0.23465169756034626,-0.6997394776209307,1.3203385037433746,-7.850393436441573e-2,-0.7074908959287137,1.1920724028277174,-0.23465169756034626,-0.8797169813346795,1.3058725987977675,-0.23465169756034626,-0.6997394776209307,1.205277677280572,-7.850393436441573e-2,-0.8894621143918703,1.1920724028277174,-0.23465169756034626,-0.8797169813346795,1.277099280388322,-0.3882285676537811,-0.6843215671664332,1.3058725987977675,-0.23465169756034626,-0.6997394776209307,1.1658065336722943,-0.3882285676537811,-0.8603334849373723,1.277099280388322,-0.3882285676537811,-0.6843215671664332,1.1920724028277174,-0.23465169756034626,-0.8797169813346795,1.1658065336722943,-0.3882285676537811,-0.8603334849373723,1.2343337950127369,-0.5375519243179504,-0.6614060864185659,1.277099280388322,-0.3882285676537811,-0.6843215671664332,1.1267678441732563,-0.5375519243179504,-0.8315239948427351,1.2343337950127369,-0.5375519243179504,-0.6614060864185659,1.1658065336722943,-0.3882285676537811,-0.8603334849373723,1.1267678441732563,-0.5375519243179504,-0.8315239948427351,1.1780446902780395,-0.6809857496093201,-0.6312441021797751,1.2343337950127369,-0.5375519243179504,-0.6614060864185659,1.0753840503821261,-0.6809857496093201,-0.7936041538530183,1.1780446902780395,-0.6809857496093201,-0.6312441021797751,1.1267678441732563,-0.5375519243179504,-0.8315239948427351,0.9505084475638106,0.6809857496093201,-0.9395701676505419,1.1267678441732563,0.5375519243179504,-0.8315239948427351,1.0753840503821261,0.6809857496093201,-0.7936041538530183,0.9959254593271811,0.5375519243179504,-0.984464528627625,1.1267678441732563,0.5375519243179504,-0.8315239948427351,0.9505084475638106,0.6809857496093201,-0.9395701676505419,0.9959254593271811,0.5375519243179504,-0.984464528627625,1.1658065336722943,0.3882285676537812,-0.8603334849373723,1.1267678441732563,0.5375519243179504,-0.8315239948427351,1.0304309033473622,0.3882285676537812,-1.0185728902166171,1.1658065336722943,0.3882285676537812,-0.8603334849373723,0.9959254593271811,0.5375519243179504,-0.984464528627625,1.0304309033473622,0.3882285676537812,-1.0185728902166171,1.1920724028277174,0.23465169756034637,-0.8797169813346795,1.1658065336722943,0.3882285676537812,-0.8603334849373723,1.0536467307589406,0.23465169756034637,-1.0415215540703153,1.1920724028277174,0.23465169756034637,-0.8797169813346795,1.0304309033473622,0.3882285676537812,-1.0185728902166171,1.0536467307589406,0.23465169756034637,-1.0415215540703153,1.205277677280572,7.850393436441573e-2,-0.8894621143918703,1.1920724028277174,0.23465169756034637,-0.8797169813346795,1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,1.205277677280572,7.850393436441573e-2,-0.8894621143918703,1.0536467307589406,0.23465169756034637,-1.0415215540703153,1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,1.205277677280572,-7.850393436441573e-2,-0.8894621143918703,1.205277677280572,7.850393436441573e-2,-0.8894621143918703,1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,1.205277677280572,-7.850393436441573e-2,-0.8894621143918703,1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,1.1920724028277174,-0.23465169756034626,-0.8797169813346795,1.205277677280572,-7.850393436441573e-2,-0.8894621143918703,1.0536467307589406,-0.23465169756034626,-1.0415215540703153,1.1920724028277174,-0.23465169756034626,-0.8797169813346795,1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,1.0536467307589406,-0.23465169756034626,-1.0415215540703153,1.1658065336722943,-0.3882285676537811,-0.8603334849373723,1.1920724028277174,-0.23465169756034626,-0.8797169813346795,1.0304309033473622,-0.3882285676537811,-1.0185728902166171,1.1658065336722943,-0.3882285676537811,-0.8603334849373723,1.0536467307589406,-0.23465169756034626,-1.0415215540703153,1.0304309033473622,-0.3882285676537811,-1.0185728902166171,1.1267678441732563,-0.5375519243179504,-0.8315239948427351,1.1658065336722943,-0.3882285676537811,-0.8603334849373723,0.9959254593271811,-0.5375519243179504,-0.984464528627625,1.1267678441732563,-0.5375519243179504,-0.8315239948427351,1.0304309033473622,-0.3882285676537811,-1.0185728902166171,0.9959254593271811,-0.5375519243179504,-0.984464528627625,1.0753840503821261,-0.6809857496093201,-0.7936041538530183,1.1267678441732563,-0.5375519243179504,-0.8315239948427351,0.9505084475638106,-0.6809857496093201,-0.9395701676505419,1.0753840503821261,-0.6809857496093201,-0.7936041538530183,0.9959254593271811,-0.5375519243179504,-0.984464528627625,0.8059975247753344,0.6809857496093201,-1.0661268212014303,0.9959254593271811,0.5375519243179504,-0.984464528627625,0.9505084475638106,0.6809857496093201,-0.9395701676505419,0.8445095434299729,0.5375519243179504,-1.1170682878489422,0.9959254593271811,0.5375519243179504,-0.984464528627625,0.8059975247753344,0.6809857496093201,-1.0661268212014303,0.8445095434299729,0.5375519243179504,-1.1170682878489422,1.0304309033473622,0.3882285676537812,-1.0185728902166171,0.9959254593271811,0.5375519243179504,-0.984464528627625,0.8737689388018092,0.3882285676537812,-1.1557709205732136,1.0304309033473622,0.3882285676537812,-1.0185728902166171,0.8445095434299729,0.5375519243179504,-1.1170682878489422,0.8737689388018092,0.3882285676537812,-1.1557709205732136,1.0536467307589406,0.23465169756034637,-1.0415215540703153,1.0304309033473622,0.3882285676537812,-1.0185728902166171,0.8934551388322273,0.23465169756034637,-1.181810685231071,1.0536467307589406,0.23465169756034637,-1.0415215540703153,0.8737689388018092,0.3882285676537812,-1.1557709205732136,0.8934551388322273,0.23465169756034637,-1.181810685231071,1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,1.0536467307589406,0.23465169756034637,-1.0415215540703153,0.9033524573940916,7.850393436441573e-2,-1.1949022847117514,1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,0.8934551388322273,0.23465169756034637,-1.181810685231071,0.9033524573940916,7.850393436441573e-2,-1.1949022847117514,1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,0.9033524573940916,-7.850393436441573e-2,-1.1949022847117514,1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,0.9033524573940916,7.850393436441573e-2,-1.1949022847117514,0.9033524573940916,-7.850393436441573e-2,-1.1949022847117514,1.0536467307589406,-0.23465169756034626,-1.0415215540703153,1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,0.8934551388322273,-0.23465169756034626,-1.181810685231071,1.0536467307589406,-0.23465169756034626,-1.0415215540703153,0.9033524573940916,-7.850393436441573e-2,-1.1949022847117514,0.8934551388322273,-0.23465169756034626,-1.181810685231071,1.0304309033473622,-0.3882285676537811,-1.0185728902166171,1.0536467307589406,-0.23465169756034626,-1.0415215540703153,0.8737689388018092,-0.3882285676537811,-1.1557709205732136,1.0304309033473622,-0.3882285676537811,-1.0185728902166171,0.8934551388322273,-0.23465169756034626,-1.181810685231071,0.8737689388018092,-0.3882285676537811,-1.1557709205732136,0.9959254593271811,-0.5375519243179504,-0.984464528627625,1.0304309033473622,-0.3882285676537811,-1.0185728902166171,0.8445095434299729,-0.5375519243179504,-1.1170682878489422,0.9959254593271811,-0.5375519243179504,-0.984464528627625,0.8737689388018092,-0.3882285676537811,-1.1557709205732136,0.8445095434299729,-0.5375519243179504,-1.1170682878489422,0.9505084475638106,-0.6809857496093201,-0.9395701676505419,0.9959254593271811,-0.5375519243179504,-0.984464528627625,0.8059975247753344,-0.6809857496093201,-1.0661268212014303,0.9505084475638106,-0.6809857496093201,-0.9395701676505419,0.8445095434299729,-0.5375519243179504,-1.1170682878489422,0.644836545551606,0.6809857496093201,-1.1706597449088716,0.8445095434299729,0.5375519243179504,-1.1170682878489422,0.8059975247753344,0.6809857496093201,-1.0661268212014303,0.6756480013043992,0.5375519243179504,-1.226595983604805,0.8445095434299729,0.5375519243179504,-1.1170682878489422,0.644836545551606,0.6809857496093201,-1.1706597449088716,0.6756480013043992,0.5375519243179504,-1.226595983604805,0.8737689388018092,0.3882285676537812,-1.1557709205732136,0.8445095434299729,0.5375519243179504,-1.1170682878489422,0.6990569161664674,0.3882285676537812,-1.269093380022653,0.8737689388018092,0.3882285676537812,-1.1557709205732136,0.6756480013043992,0.5375519243179504,-1.226595983604805,0.6990569161664674,0.3882285676537812,-1.269093380022653,0.8934551388322273,0.23465169756034637,-1.181810685231071,0.8737689388018092,0.3882285676537812,-1.1557709205732136,0.7148068171679516,0.23465169756034637,-1.2976863237941094,0.8934551388322273,0.23465169756034637,-1.181810685231071,0.6990569161664674,0.3882285676537812,-1.269093380022653,0.7148068171679516,0.23465169756034637,-1.2976863237941094,0.9033524573940916,7.850393436441573e-2,-1.1949022847117514,0.8934551388322273,0.23465169756034637,-1.181810685231071,0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,0.9033524573940916,7.850393436441573e-2,-1.1949022847117514,0.7148068171679516,0.23465169756034637,-1.2976863237941094,0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,0.9033524573940916,-7.850393436441573e-2,-1.1949022847117514,0.9033524573940916,7.850393436441573e-2,-1.1949022847117514,0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,0.9033524573940916,-7.850393436441573e-2,-1.1949022847117514,0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,0.8934551388322273,-0.23465169756034626,-1.181810685231071,0.9033524573940916,-7.850393436441573e-2,-1.1949022847117514,0.7148068171679516,-0.23465169756034626,-1.2976863237941094,0.8934551388322273,-0.23465169756034626,-1.181810685231071,0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,0.7148068171679516,-0.23465169756034626,-1.2976863237941094,0.8737689388018092,-0.3882285676537811,-1.1557709205732136,0.8934551388322273,-0.23465169756034626,-1.181810685231071,0.6990569161664674,-0.3882285676537811,-1.269093380022653,0.8737689388018092,-0.3882285676537811,-1.1557709205732136,0.7148068171679516,-0.23465169756034626,-1.2976863237941094,0.6990569161664674,-0.3882285676537811,-1.269093380022653,0.8445095434299729,-0.5375519243179504,-1.1170682878489422,0.8737689388018092,-0.3882285676537811,-1.1557709205732136,0.6756480013043992,-0.5375519243179504,-1.226595983604805,0.8445095434299729,-0.5375519243179504,-1.1170682878489422,0.6990569161664674,-0.3882285676537811,-1.269093380022653,0.6756480013043992,-0.5375519243179504,-1.226595983604805,0.8059975247753344,-0.6809857496093201,-1.0661268212014303,0.8445095434299729,-0.5375519243179504,-1.1170682878489422,0.644836545551606,-0.6809857496093201,-1.1706597449088716,0.8059975247753344,-0.6809857496093201,-1.0661268212014303,0.6756480013043992,-0.5375519243179504,-1.226595983604805,0.4703547253266107,0.6809857496093201,-1.2510095288174112,0.6756480013043992,0.5375519243179504,-1.226595983604805,0.644836545551606,0.6809857496093201,-1.1706597449088716,0.4928291243157702,0.5375519243179504,-1.3107850254287388,0.6756480013043992,0.5375519243179504,-1.226595983604805,0.4703547253266107,0.6809857496093201,-1.2510095288174112,0.4928291243157702,0.5375519243179504,-1.3107850254287388,0.6990569161664674,0.3882285676537812,-1.269093380022653,0.6756480013043992,0.5375519243179504,-1.226595983604805,0.5099039842878016,0.3882285676537812,-1.3561992870020685,0.6990569161664674,0.3882285676537812,-1.269093380022653,0.4928291243157702,0.5375519243179504,-1.3107850254287388,0.5099039842878016,0.3882285676537812,-1.3561992870020685,0.7148068171679516,0.23465169756034637,-1.2976863237941094,0.6990569161664674,0.3882285676537812,-1.269093380022653,0.5213922295037072,0.23465169756034637,-1.3867547453840574,0.7148068171679516,0.23465169756034637,-1.2976863237941094,0.5099039842878016,0.3882285676537812,-1.3561992870020685,0.5213922295037072,0.23465169756034637,-1.3867547453840574,0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,0.7148068171679516,0.23465169756034637,-1.2976863237941094,0.5271679923448317,7.850393436441573e-2,-1.4021166285785314,0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,0.5213922295037072,0.23465169756034637,-1.3867547453840574,0.5271679923448317,7.850393436441573e-2,-1.4021166285785314,0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,0.5271679923448317,-7.850393436441573e-2,-1.4021166285785314,0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,0.5271679923448317,7.850393436441573e-2,-1.4021166285785314,0.5271679923448317,-7.850393436441573e-2,-1.4021166285785314,0.7148068171679516,-0.23465169756034626,-1.2976863237941094,0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,0.5213922295037072,-0.23465169756034626,-1.3867547453840574,0.7148068171679516,-0.23465169756034626,-1.2976863237941094,0.5271679923448317,-7.850393436441573e-2,-1.4021166285785314,0.5213922295037072,-0.23465169756034626,-1.3867547453840574,0.6990569161664674,-0.3882285676537811,-1.269093380022653,0.7148068171679516,-0.23465169756034626,-1.2976863237941094,0.5099039842878016,-0.3882285676537811,-1.3561992870020685,0.6990569161664674,-0.3882285676537811,-1.269093380022653,0.5213922295037072,-0.23465169756034626,-1.3867547453840574,0.5099039842878016,-0.3882285676537811,-1.3561992870020685,0.6756480013043992,-0.5375519243179504,-1.226595983604805,0.6990569161664674,-0.3882285676537811,-1.269093380022653,0.4928291243157702,-0.5375519243179504,-1.3107850254287388,0.6756480013043992,-0.5375519243179504,-1.226595983604805,0.5099039842878016,-0.3882285676537811,-1.3561992870020685,0.4928291243157702,-0.5375519243179504,-1.3107850254287388,0.644836545551606,-0.6809857496093201,-1.1706597449088716,0.6756480013043992,-0.5375519243179504,-1.226595983604805,0.4703547253266107,-0.6809857496093201,-1.2510095288174112,0.644836545551606,-0.6809857496093201,-1.1706597449088716,0.4928291243157702,-0.5375519243179504,-1.3107850254287388,0.2861564574940893,0.6809857496093201,-1.3055163310596565,0.4928291243157702,0.5375519243179504,-1.3107850254287388,0.4703547253266107,0.6809857496093201,-1.2510095288174112,0.2998295302894797,0.5375519243179504,-1.367896261208597,0.4928291243157702,0.5375519243179504,-1.3107850254287388,0.2861564574940893,0.6809857496093201,-1.3055163310596565,0.2998295302894797,0.5375519243179504,-1.367896261208597,0.5099039842878016,0.3882285676537812,-1.3561992870020685,0.4928291243157702,0.5375519243179504,-1.3107850254287388,0.31021760800765563,0.3882285676537812,-1.4152892336690415,0.5099039842878016,0.3882285676537812,-1.3561992870020685,0.2998295302894797,0.5375519243179504,-1.367896261208597,0.31021760800765563,0.3882285676537812,-1.4152892336690415,0.5213922295037072,0.23465169756034637,-1.3867547453840574,0.5099039842878016,0.3882285676537812,-1.3561992870020685,0.31720687669529174,0.23465169756034637,-1.447176001117095,0.5213922295037072,0.23465169756034637,-1.3867547453840574,0.31021760800765563,0.3882285676537812,-1.4152892336690415,0.31720687669529174,0.23465169756034637,-1.447176001117095,0.5271679923448317,7.850393436441573e-2,-1.4021166285785314,0.5213922295037072,0.23465169756034637,-1.3867547453840574,0.32072076046204784,7.850393436441573e-2,-1.463207205455862,0.5271679923448317,7.850393436441573e-2,-1.4021166285785314,0.31720687669529174,0.23465169756034637,-1.447176001117095,0.32072076046204784,7.850393436441573e-2,-1.463207205455862,0.5271679923448317,-7.850393436441573e-2,-1.4021166285785314,0.5271679923448317,7.850393436441573e-2,-1.4021166285785314,0.32072076046204784,-7.850393436441573e-2,-1.463207205455862,0.5271679923448317,-7.850393436441573e-2,-1.4021166285785314,0.32072076046204784,7.850393436441573e-2,-1.463207205455862,0.32072076046204784,-7.850393436441573e-2,-1.463207205455862,0.5213922295037072,-0.23465169756034626,-1.3867547453840574,0.5271679923448317,-7.850393436441573e-2,-1.4021166285785314,0.31720687669529174,-0.23465169756034626,-1.447176001117095,0.5213922295037072,-0.23465169756034626,-1.3867547453840574,0.32072076046204784,-7.850393436441573e-2,-1.463207205455862,0.31720687669529174,-0.23465169756034626,-1.447176001117095,0.5099039842878016,-0.3882285676537811,-1.3561992870020685,0.5213922295037072,-0.23465169756034626,-1.3867547453840574,0.31021760800765563,-0.3882285676537811,-1.4152892336690415,0.5099039842878016,-0.3882285676537811,-1.3561992870020685,0.31720687669529174,-0.23465169756034626,-1.447176001117095,0.31021760800765563,-0.3882285676537811,-1.4152892336690415,0.4928291243157702,-0.5375519243179504,-1.3107850254287388,0.5099039842878016,-0.3882285676537811,-1.3561992870020685,0.2998295302894797,-0.5375519243179504,-1.367896261208597,0.4928291243157702,-0.5375519243179504,-1.3107850254287388,0.31021760800765563,-0.3882285676537811,-1.4152892336690415,0.2998295302894797,-0.5375519243179504,-1.367896261208597,0.4703547253266107,-0.6809857496093201,-1.2510095288174112,0.4928291243157702,-0.5375519243179504,-1.3107850254287388,0.2861564574940893,-0.6809857496093201,-1.3055163310596565,0.4703547253266107,-0.6809857496093201,-1.2510095288174112,0.2998295302894797,-0.5375519243179504,-1.367896261208597,9.604685492401183e-2,0.6809857496093201,-1.3330541663744344,0.2998295302894797,0.5375519243179504,-1.367896261208597,0.2861564574940893,0.6809857496093201,-1.3055163310596565,0.10063614726654595,0.5375519243179504,-1.3967499040720979,0.2998295302894797,0.5375519243179504,-1.367896261208597,9.604685492401183e-2,0.6809857496093201,-1.3330541663744344,0.10063614726654595,0.5375519243179504,-1.3967499040720979,0.31021760800765563,0.3882285676537812,-1.4152892336690415,0.2998295302894797,0.5375519243179504,-1.367896261208597,0.10412284892016008,0.3882285676537812,-1.4451425575320391,0.31021760800765563,0.3882285676537812,-1.4152892336690415,0.10063614726654595,0.5375519243179504,-1.3967499040720979,0.10412284892016008,0.3882285676537812,-1.4451425575320391,0.31720687669529174,0.23465169756034637,-1.447176001117095,0.31021760800765563,0.3882285676537812,-1.4152892336690415,0.10646875885189802,0.23465169756034637,-1.4777019267161373,0.31720687669529174,0.23465169756034637,-1.447176001117095,0.10412284892016008,0.3882285676537812,-1.4451425575320391,0.10646875885189802,0.23465169756034637,-1.4777019267161373,0.32072076046204784,7.850393436441573e-2,-1.463207205455862,0.31720687669529174,0.23465169756034637,-1.447176001117095,0.1076481747816344,7.850393436441573e-2,-1.4940712843621249,0.32072076046204784,7.850393436441573e-2,-1.463207205455862,0.10646875885189802,0.23465169756034637,-1.4777019267161373,0.1076481747816344,7.850393436441573e-2,-1.4940712843621249,0.32072076046204784,-7.850393436441573e-2,-1.463207205455862,0.32072076046204784,7.850393436441573e-2,-1.463207205455862,0.1076481747816344,-7.850393436441573e-2,-1.4940712843621249,0.32072076046204784,-7.850393436441573e-2,-1.463207205455862,0.1076481747816344,7.850393436441573e-2,-1.4940712843621249,0.1076481747816344,-7.850393436441573e-2,-1.4940712843621249,0.31720687669529174,-0.23465169756034626,-1.447176001117095,0.32072076046204784,-7.850393436441573e-2,-1.463207205455862,0.10646875885189802,-0.23465169756034626,-1.4777019267161373,0.31720687669529174,-0.23465169756034626,-1.447176001117095,0.1076481747816344,-7.850393436441573e-2,-1.4940712843621249,0.10646875885189802,-0.23465169756034626,-1.4777019267161373,0.31021760800765563,-0.3882285676537811,-1.4152892336690415,0.31720687669529174,-0.23465169756034626,-1.447176001117095,0.10412284892016008,-0.3882285676537811,-1.4451425575320391,0.31021760800765563,-0.3882285676537811,-1.4152892336690415,0.10646875885189802,-0.23465169756034626,-1.4777019267161373,0.10412284892016008,-0.3882285676537811,-1.4451425575320391,0.2998295302894797,-0.5375519243179504,-1.367896261208597,0.31021760800765563,-0.3882285676537811,-1.4152892336690415,0.10063614726654595,-0.5375519243179504,-1.3967499040720979,0.2998295302894797,-0.5375519243179504,-1.367896261208597,0.10412284892016008,-0.3882285676537811,-1.4451425575320391,0.10063614726654595,-0.5375519243179504,-1.3967499040720979,0.2861564574940893,-0.6809857496093201,-1.3055163310596565,0.2998295302894797,-0.5375519243179504,-1.367896261208597,9.604685492401183e-2,-0.6809857496093201,-1.3330541663744344,0.2861564574940893,-0.6809857496093201,-1.3055163310596565,0.10063614726654595,-0.5375519243179504,-1.3967499040720979,-9.604685492401169e-2,0.6809857496093201,-1.3330541663744344,0.10063614726654595,0.5375519243179504,-1.3967499040720979,9.604685492401183e-2,0.6809857496093201,-1.3330541663744344,-0.1006361472665458,0.5375519243179504,-1.3967499040720979,0.10063614726654595,0.5375519243179504,-1.3967499040720979,-9.604685492401169e-2,0.6809857496093201,-1.3330541663744344,-0.1006361472665458,0.5375519243179504,-1.3967499040720979,0.10412284892016008,0.3882285676537812,-1.4451425575320391,0.10063614726654595,0.5375519243179504,-1.3967499040720979,-0.10412284892015994,0.3882285676537812,-1.4451425575320394,0.10412284892016008,0.3882285676537812,-1.4451425575320391,-0.1006361472665458,0.5375519243179504,-1.3967499040720979,-0.10412284892015994,0.3882285676537812,-1.4451425575320394,0.10646875885189802,0.23465169756034637,-1.4777019267161373,0.10412284892016008,0.3882285676537812,-1.4451425575320391,-0.10646875885189785,0.23465169756034637,-1.4777019267161373,0.10646875885189802,0.23465169756034637,-1.4777019267161373,-0.10412284892015994,0.3882285676537812,-1.4451425575320394,-0.10646875885189785,0.23465169756034637,-1.4777019267161373,0.1076481747816344,7.850393436441573e-2,-1.4940712843621249,0.10646875885189802,0.23465169756034637,-1.4777019267161373,-0.10764817478163423,7.850393436441573e-2,-1.494071284362125,0.1076481747816344,7.850393436441573e-2,-1.4940712843621249,-0.10646875885189785,0.23465169756034637,-1.4777019267161373,-0.10764817478163423,7.850393436441573e-2,-1.494071284362125,0.1076481747816344,-7.850393436441573e-2,-1.4940712843621249,0.1076481747816344,7.850393436441573e-2,-1.4940712843621249,-0.10764817478163423,-7.850393436441573e-2,-1.494071284362125,0.1076481747816344,-7.850393436441573e-2,-1.4940712843621249,-0.10764817478163423,7.850393436441573e-2,-1.494071284362125,-0.10764817478163423,-7.850393436441573e-2,-1.494071284362125,0.10646875885189802,-0.23465169756034626,-1.4777019267161373,0.1076481747816344,-7.850393436441573e-2,-1.4940712843621249,-0.10646875885189785,-0.23465169756034626,-1.4777019267161373,0.10646875885189802,-0.23465169756034626,-1.4777019267161373,-0.10764817478163423,-7.850393436441573e-2,-1.494071284362125,-0.10646875885189785,-0.23465169756034626,-1.4777019267161373,0.10412284892016008,-0.3882285676537811,-1.4451425575320391,0.10646875885189802,-0.23465169756034626,-1.4777019267161373,-0.10412284892015994,-0.3882285676537811,-1.4451425575320394,0.10412284892016008,-0.3882285676537811,-1.4451425575320391,-0.10646875885189785,-0.23465169756034626,-1.4777019267161373,-0.10412284892015994,-0.3882285676537811,-1.4451425575320394,0.10063614726654595,-0.5375519243179504,-1.3967499040720979,0.10412284892016008,-0.3882285676537811,-1.4451425575320391,-0.1006361472665458,-0.5375519243179504,-1.3967499040720979,0.10063614726654595,-0.5375519243179504,-1.3967499040720979,-0.10412284892015994,-0.3882285676537811,-1.4451425575320394,-0.1006361472665458,-0.5375519243179504,-1.3967499040720979,9.604685492401183e-2,-0.6809857496093201,-1.3330541663744344,0.10063614726654595,-0.5375519243179504,-1.3967499040720979,-9.604685492401169e-2,-0.6809857496093201,-1.3330541663744344,9.604685492401183e-2,-0.6809857496093201,-1.3330541663744344,-0.1006361472665458,-0.5375519243179504,-1.3967499040720979,-0.28615645749408913,0.6809857496093201,-1.3055163310596565,-0.1006361472665458,0.5375519243179504,-1.3967499040720979,-9.604685492401169e-2,0.6809857496093201,-1.3330541663744344,-0.2998295302894795,0.5375519243179504,-1.367896261208597,-0.1006361472665458,0.5375519243179504,-1.3967499040720979,-0.28615645749408913,0.6809857496093201,-1.3055163310596565,-0.2998295302894795,0.5375519243179504,-1.367896261208597,-0.10412284892015994,0.3882285676537812,-1.4451425575320394,-0.1006361472665458,0.5375519243179504,-1.3967499040720979,-0.31021760800765547,0.3882285676537812,-1.4152892336690415,-0.10412284892015994,0.3882285676537812,-1.4451425575320394,-0.2998295302894795,0.5375519243179504,-1.367896261208597,-0.31021760800765547,0.3882285676537812,-1.4152892336690415,-0.10646875885189785,0.23465169756034637,-1.4777019267161373,-0.10412284892015994,0.3882285676537812,-1.4451425575320394,-0.31720687669529163,0.23465169756034637,-1.447176001117095,-0.10646875885189785,0.23465169756034637,-1.4777019267161373,-0.31021760800765547,0.3882285676537812,-1.4152892336690415,-0.31720687669529163,0.23465169756034637,-1.447176001117095,-0.10764817478163423,7.850393436441573e-2,-1.494071284362125,-0.10646875885189785,0.23465169756034637,-1.4777019267161373,-0.32072076046204767,7.850393436441573e-2,-1.463207205455862,-0.10764817478163423,7.850393436441573e-2,-1.494071284362125,-0.31720687669529163,0.23465169756034637,-1.447176001117095,-0.32072076046204767,7.850393436441573e-2,-1.463207205455862,-0.10764817478163423,-7.850393436441573e-2,-1.494071284362125,-0.10764817478163423,7.850393436441573e-2,-1.494071284362125,-0.32072076046204767,-7.850393436441573e-2,-1.463207205455862,-0.10764817478163423,-7.850393436441573e-2,-1.494071284362125,-0.32072076046204767,7.850393436441573e-2,-1.463207205455862,-0.32072076046204767,-7.850393436441573e-2,-1.463207205455862,-0.10646875885189785,-0.23465169756034626,-1.4777019267161373,-0.10764817478163423,-7.850393436441573e-2,-1.494071284362125,-0.31720687669529163,-0.23465169756034626,-1.447176001117095,-0.10646875885189785,-0.23465169756034626,-1.4777019267161373,-0.32072076046204767,-7.850393436441573e-2,-1.463207205455862,-0.31720687669529163,-0.23465169756034626,-1.447176001117095,-0.10412284892015994,-0.3882285676537811,-1.4451425575320394,-0.10646875885189785,-0.23465169756034626,-1.4777019267161373,-0.31021760800765547,-0.3882285676537811,-1.4152892336690415,-0.10412284892015994,-0.3882285676537811,-1.4451425575320394,-0.31720687669529163,-0.23465169756034626,-1.447176001117095,-0.31021760800765547,-0.3882285676537811,-1.4152892336690415,-0.1006361472665458,-0.5375519243179504,-1.3967499040720979,-0.10412284892015994,-0.3882285676537811,-1.4451425575320394,-0.2998295302894795,-0.5375519243179504,-1.367896261208597,-0.1006361472665458,-0.5375519243179504,-1.3967499040720979,-0.31021760800765547,-0.3882285676537811,-1.4152892336690415,-0.2998295302894795,-0.5375519243179504,-1.367896261208597,-9.604685492401169e-2,-0.6809857496093201,-1.3330541663744344,-0.1006361472665458,-0.5375519243179504,-1.3967499040720979,-0.28615645749408913,-0.6809857496093201,-1.3055163310596565,-9.604685492401169e-2,-0.6809857496093201,-1.3330541663744344,-0.2998295302894795,-0.5375519243179504,-1.367896261208597,-0.4703547253266106,0.6809857496093201,-1.2510095288174112,-0.2998295302894795,0.5375519243179504,-1.367896261208597,-0.28615645749408913,0.6809857496093201,-1.3055163310596565,-0.49282912431577,0.5375519243179504,-1.3107850254287388,-0.2998295302894795,0.5375519243179504,-1.367896261208597,-0.4703547253266106,0.6809857496093201,-1.2510095288174112,-0.49282912431577,0.5375519243179504,-1.3107850254287388,-0.31021760800765547,0.3882285676537812,-1.4152892336690415,-0.2998295302894795,0.5375519243179504,-1.367896261208597,-0.5099039842878014,0.3882285676537812,-1.3561992870020685,-0.31021760800765547,0.3882285676537812,-1.4152892336690415,-0.49282912431577,0.5375519243179504,-1.3107850254287388,-0.5099039842878014,0.3882285676537812,-1.3561992870020685,-0.31720687669529163,0.23465169756034637,-1.447176001117095,-0.31021760800765547,0.3882285676537812,-1.4152892336690415,-0.521392229503707,0.23465169756034637,-1.3867547453840574,-0.31720687669529163,0.23465169756034637,-1.447176001117095,-0.5099039842878014,0.3882285676537812,-1.3561992870020685,-0.521392229503707,0.23465169756034637,-1.3867547453840574,-0.32072076046204767,7.850393436441573e-2,-1.463207205455862,-0.31720687669529163,0.23465169756034637,-1.447176001117095,-0.5271679923448315,7.850393436441573e-2,-1.4021166285785314,-0.32072076046204767,7.850393436441573e-2,-1.463207205455862,-0.521392229503707,0.23465169756034637,-1.3867547453840574,-0.5271679923448315,7.850393436441573e-2,-1.4021166285785314,-0.32072076046204767,-7.850393436441573e-2,-1.463207205455862,-0.32072076046204767,7.850393436441573e-2,-1.463207205455862,-0.5271679923448315,-7.850393436441573e-2,-1.4021166285785314,-0.32072076046204767,-7.850393436441573e-2,-1.463207205455862,-0.5271679923448315,7.850393436441573e-2,-1.4021166285785314,-0.5271679923448315,-7.850393436441573e-2,-1.4021166285785314,-0.31720687669529163,-0.23465169756034626,-1.447176001117095,-0.32072076046204767,-7.850393436441573e-2,-1.463207205455862,-0.521392229503707,-0.23465169756034626,-1.3867547453840574,-0.31720687669529163,-0.23465169756034626,-1.447176001117095,-0.5271679923448315,-7.850393436441573e-2,-1.4021166285785314,-0.521392229503707,-0.23465169756034626,-1.3867547453840574,-0.31021760800765547,-0.3882285676537811,-1.4152892336690415,-0.31720687669529163,-0.23465169756034626,-1.447176001117095,-0.5099039842878014,-0.3882285676537811,-1.3561992870020685,-0.31021760800765547,-0.3882285676537811,-1.4152892336690415,-0.521392229503707,-0.23465169756034626,-1.3867547453840574,-0.5099039842878014,-0.3882285676537811,-1.3561992870020685,-0.2998295302894795,-0.5375519243179504,-1.367896261208597,-0.31021760800765547,-0.3882285676537811,-1.4152892336690415,-0.49282912431577,-0.5375519243179504,-1.3107850254287388,-0.2998295302894795,-0.5375519243179504,-1.367896261208597,-0.5099039842878014,-0.3882285676537811,-1.3561992870020685,-0.49282912431577,-0.5375519243179504,-1.3107850254287388,-0.28615645749408913,-0.6809857496093201,-1.3055163310596565,-0.2998295302894795,-0.5375519243179504,-1.367896261208597,-0.4703547253266106,-0.6809857496093201,-1.2510095288174112,-0.28615645749408913,-0.6809857496093201,-1.3055163310596565,-0.49282912431577,-0.5375519243179504,-1.3107850254287388,-0.644836545551606,0.6809857496093201,-1.1706597449088716,-0.49282912431577,0.5375519243179504,-1.3107850254287388,-0.4703547253266106,0.6809857496093201,-1.2510095288174112,-0.6756480013043992,0.5375519243179504,-1.226595983604805,-0.49282912431577,0.5375519243179504,-1.3107850254287388,-0.644836545551606,0.6809857496093201,-1.1706597449088716,-0.6756480013043992,0.5375519243179504,-1.226595983604805,-0.5099039842878014,0.3882285676537812,-1.3561992870020685,-0.49282912431577,0.5375519243179504,-1.3107850254287388,-0.6990569161664674,0.3882285676537812,-1.269093380022653,-0.5099039842878014,0.3882285676537812,-1.3561992870020685,-0.6756480013043992,0.5375519243179504,-1.226595983604805,-0.6990569161664674,0.3882285676537812,-1.269093380022653,-0.521392229503707,0.23465169756034637,-1.3867547453840574,-0.5099039842878014,0.3882285676537812,-1.3561992870020685,-0.7148068171679516,0.23465169756034637,-1.2976863237941094,-0.521392229503707,0.23465169756034637,-1.3867547453840574,-0.6990569161664674,0.3882285676537812,-1.269093380022653,-0.7148068171679516,0.23465169756034637,-1.2976863237941094,-0.5271679923448315,7.850393436441573e-2,-1.4021166285785314,-0.521392229503707,0.23465169756034637,-1.3867547453840574,-0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,-0.5271679923448315,7.850393436441573e-2,-1.4021166285785314,-0.7148068171679516,0.23465169756034637,-1.2976863237941094,-0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,-0.5271679923448315,-7.850393436441573e-2,-1.4021166285785314,-0.5271679923448315,7.850393436441573e-2,-1.4021166285785314,-0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,-0.5271679923448315,-7.850393436441573e-2,-1.4021166285785314,-0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,-0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,-0.521392229503707,-0.23465169756034626,-1.3867547453840574,-0.5271679923448315,-7.850393436441573e-2,-1.4021166285785314,-0.7148068171679516,-0.23465169756034626,-1.2976863237941094,-0.521392229503707,-0.23465169756034626,-1.3867547453840574,-0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,-0.7148068171679516,-0.23465169756034626,-1.2976863237941094,-0.5099039842878014,-0.3882285676537811,-1.3561992870020685,-0.521392229503707,-0.23465169756034626,-1.3867547453840574,-0.6990569161664674,-0.3882285676537811,-1.269093380022653,-0.5099039842878014,-0.3882285676537811,-1.3561992870020685,-0.7148068171679516,-0.23465169756034626,-1.2976863237941094,-0.6990569161664674,-0.3882285676537811,-1.269093380022653,-0.49282912431577,-0.5375519243179504,-1.3107850254287388,-0.5099039842878014,-0.3882285676537811,-1.3561992870020685,-0.6756480013043992,-0.5375519243179504,-1.226595983604805,-0.49282912431577,-0.5375519243179504,-1.3107850254287388,-0.6990569161664674,-0.3882285676537811,-1.269093380022653,-0.6756480013043992,-0.5375519243179504,-1.226595983604805,-0.4703547253266106,-0.6809857496093201,-1.2510095288174112,-0.49282912431577,-0.5375519243179504,-1.3107850254287388,-0.644836545551606,-0.6809857496093201,-1.1706597449088716,-0.4703547253266106,-0.6809857496093201,-1.2510095288174112,-0.6756480013043992,-0.5375519243179504,-1.226595983604805,-0.8059975247753342,0.6809857496093201,-1.0661268212014305,-0.6756480013043992,0.5375519243179504,-1.226595983604805,-0.644836545551606,0.6809857496093201,-1.1706597449088716,-0.8445095434299725,0.5375519243179504,-1.1170682878489424,-0.6756480013043992,0.5375519243179504,-1.226595983604805,-0.8059975247753342,0.6809857496093201,-1.0661268212014305,-0.8445095434299725,0.5375519243179504,-1.1170682878489424,-0.6990569161664674,0.3882285676537812,-1.269093380022653,-0.6756480013043992,0.5375519243179504,-1.226595983604805,-0.8737689388018088,0.3882285676537812,-1.1557709205732138,-0.6990569161664674,0.3882285676537812,-1.269093380022653,-0.8445095434299725,0.5375519243179504,-1.1170682878489424,-0.8737689388018088,0.3882285676537812,-1.1557709205732138,-0.7148068171679516,0.23465169756034637,-1.2976863237941094,-0.6990569161664674,0.3882285676537812,-1.269093380022653,-0.8934551388322269,0.23465169756034637,-1.1818106852310712,-0.7148068171679516,0.23465169756034637,-1.2976863237941094,-0.8737689388018088,0.3882285676537812,-1.1557709205732138,-0.8934551388322269,0.23465169756034637,-1.1818106852310712,-0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,-0.7148068171679516,0.23465169756034637,-1.2976863237941094,-0.9033524573940913,7.850393436441573e-2,-1.1949022847117516,-0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,-0.8934551388322269,0.23465169756034637,-1.1818106852310712,-0.9033524573940913,7.850393436441573e-2,-1.1949022847117516,-0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,-0.7227251450976004,7.850393436441573e-2,-1.3120615446437562,-0.9033524573940913,-7.850393436441573e-2,-1.1949022847117516,-0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,-0.9033524573940913,7.850393436441573e-2,-1.1949022847117516,-0.9033524573940913,-7.850393436441573e-2,-1.1949022847117516,-0.7148068171679516,-0.23465169756034626,-1.2976863237941094,-0.7227251450976004,-7.850393436441573e-2,-1.3120615446437562,-0.8934551388322269,-0.23465169756034626,-1.1818106852310712,-0.7148068171679516,-0.23465169756034626,-1.2976863237941094,-0.9033524573940913,-7.850393436441573e-2,-1.1949022847117516,-0.8934551388322269,-0.23465169756034626,-1.1818106852310712,-0.6990569161664674,-0.3882285676537811,-1.269093380022653,-0.7148068171679516,-0.23465169756034626,-1.2976863237941094,-0.8737689388018088,-0.3882285676537811,-1.1557709205732138,-0.6990569161664674,-0.3882285676537811,-1.269093380022653,-0.8934551388322269,-0.23465169756034626,-1.1818106852310712,-0.8737689388018088,-0.3882285676537811,-1.1557709205732138,-0.6756480013043992,-0.5375519243179504,-1.226595983604805,-0.6990569161664674,-0.3882285676537811,-1.269093380022653,-0.8445095434299725,-0.5375519243179504,-1.1170682878489424,-0.6756480013043992,-0.5375519243179504,-1.226595983604805,-0.8737689388018088,-0.3882285676537811,-1.1557709205732138,-0.8445095434299725,-0.5375519243179504,-1.1170682878489424,-0.644836545551606,-0.6809857496093201,-1.1706597449088716,-0.6756480013043992,-0.5375519243179504,-1.226595983604805,-0.8059975247753342,-0.6809857496093201,-1.0661268212014305,-0.644836545551606,-0.6809857496093201,-1.1706597449088716,-0.8445095434299725,-0.5375519243179504,-1.1170682878489424,-0.9505084475638106,0.6809857496093201,-0.9395701676505419,-0.8445095434299725,0.5375519243179504,-1.1170682878489424,-0.8059975247753342,0.6809857496093201,-1.0661268212014305,-0.9959254593271811,0.5375519243179504,-0.984464528627625,-0.8445095434299725,0.5375519243179504,-1.1170682878489424,-0.9505084475638106,0.6809857496093201,-0.9395701676505419,-0.9959254593271811,0.5375519243179504,-0.984464528627625,-0.8737689388018088,0.3882285676537812,-1.1557709205732138,-0.8445095434299725,0.5375519243179504,-1.1170682878489424,-1.0304309033473622,0.3882285676537812,-1.0185728902166171,-0.8737689388018088,0.3882285676537812,-1.1557709205732138,-0.9959254593271811,0.5375519243179504,-0.984464528627625,-1.0304309033473622,0.3882285676537812,-1.0185728902166171,-0.8934551388322269,0.23465169756034637,-1.1818106852310712,-0.8737689388018088,0.3882285676537812,-1.1557709205732138,-1.0536467307589406,0.23465169756034637,-1.0415215540703153,-0.8934551388322269,0.23465169756034637,-1.1818106852310712,-1.0304309033473622,0.3882285676537812,-1.0185728902166171,-1.0536467307589406,0.23465169756034637,-1.0415215540703153,-0.9033524573940913,7.850393436441573e-2,-1.1949022847117516,-0.8934551388322269,0.23465169756034637,-1.1818106852310712,-1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,-0.9033524573940913,7.850393436441573e-2,-1.1949022847117516,-1.0536467307589406,0.23465169756034637,-1.0415215540703153,-1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,-0.9033524573940913,-7.850393436441573e-2,-1.1949022847117516,-0.9033524573940913,7.850393436441573e-2,-1.1949022847117516,-1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,-0.9033524573940913,-7.850393436441573e-2,-1.1949022847117516,-1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,-1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,-0.8934551388322269,-0.23465169756034626,-1.1818106852310712,-0.9033524573940913,-7.850393436441573e-2,-1.1949022847117516,-1.0536467307589406,-0.23465169756034626,-1.0415215540703153,-0.8934551388322269,-0.23465169756034626,-1.1818106852310712,-1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,-1.0536467307589406,-0.23465169756034626,-1.0415215540703153,-0.8737689388018088,-0.3882285676537811,-1.1557709205732138,-0.8934551388322269,-0.23465169756034626,-1.1818106852310712,-1.0304309033473622,-0.3882285676537811,-1.0185728902166171,-0.8737689388018088,-0.3882285676537811,-1.1557709205732138,-1.0536467307589406,-0.23465169756034626,-1.0415215540703153,-1.0304309033473622,-0.3882285676537811,-1.0185728902166171,-0.8445095434299725,-0.5375519243179504,-1.1170682878489424,-0.8737689388018088,-0.3882285676537811,-1.1557709205732138,-0.9959254593271811,-0.5375519243179504,-0.984464528627625,-0.8445095434299725,-0.5375519243179504,-1.1170682878489424,-1.0304309033473622,-0.3882285676537811,-1.0185728902166171,-0.9959254593271811,-0.5375519243179504,-0.984464528627625,-0.8059975247753342,-0.6809857496093201,-1.0661268212014305,-0.8445095434299725,-0.5375519243179504,-1.1170682878489424,-0.9505084475638106,-0.6809857496093201,-0.9395701676505419,-0.8059975247753342,-0.6809857496093201,-1.0661268212014305,-0.9959254593271811,-0.5375519243179504,-0.984464528627625,-1.0753840503821261,0.6809857496093201,-0.7936041538530185,-0.9959254593271811,0.5375519243179504,-0.984464528627625,-0.9505084475638106,0.6809857496093201,-0.9395701676505419,-1.1267678441732563,0.5375519243179504,-0.8315239948427353,-0.9959254593271811,0.5375519243179504,-0.984464528627625,-1.0753840503821261,0.6809857496093201,-0.7936041538530185,-1.1267678441732563,0.5375519243179504,-0.8315239948427353,-1.0304309033473622,0.3882285676537812,-1.0185728902166171,-0.9959254593271811,0.5375519243179504,-0.984464528627625,-1.1658065336722938,0.3882285676537812,-0.8603334849373725,-1.0304309033473622,0.3882285676537812,-1.0185728902166171,-1.1267678441732563,0.5375519243179504,-0.8315239948427353,-1.1658065336722938,0.3882285676537812,-0.8603334849373725,-1.0536467307589406,0.23465169756034637,-1.0415215540703153,-1.0304309033473622,0.3882285676537812,-1.0185728902166171,-1.192072402827717,0.23465169756034637,-0.8797169813346798,-1.0536467307589406,0.23465169756034637,-1.0415215540703153,-1.1658065336722938,0.3882285676537812,-0.8603334849373725,-1.192072402827717,0.23465169756034637,-0.8797169813346798,-1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,-1.0536467307589406,0.23465169756034637,-1.0415215540703153,-1.2052776772805718,7.850393436441573e-2,-0.8894621143918706,-1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,-1.192072402827717,0.23465169756034637,-0.8797169813346798,-1.2052776772805718,7.850393436441573e-2,-0.8894621143918706,-1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,-1.0653185840985702,7.850393436441573e-2,-1.0530590898252221,-1.2052776772805718,-7.850393436441573e-2,-0.8894621143918706,-1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,-1.2052776772805718,7.850393436441573e-2,-0.8894621143918706,-1.2052776772805718,-7.850393436441573e-2,-0.8894621143918706,-1.0536467307589406,-0.23465169756034626,-1.0415215540703153,-1.0653185840985702,-7.850393436441573e-2,-1.0530590898252221,-1.192072402827717,-0.23465169756034626,-0.8797169813346798,-1.0536467307589406,-0.23465169756034626,-1.0415215540703153,-1.2052776772805718,-7.850393436441573e-2,-0.8894621143918706,-1.192072402827717,-0.23465169756034626,-0.8797169813346798,-1.0304309033473622,-0.3882285676537811,-1.0185728902166171,-1.0536467307589406,-0.23465169756034626,-1.0415215540703153,-1.1658065336722938,-0.3882285676537811,-0.8603334849373725,-1.0304309033473622,-0.3882285676537811,-1.0185728902166171,-1.192072402827717,-0.23465169756034626,-0.8797169813346798,-1.1658065336722938,-0.3882285676537811,-0.8603334849373725,-0.9959254593271811,-0.5375519243179504,-0.984464528627625,-1.0304309033473622,-0.3882285676537811,-1.0185728902166171,-1.1267678441732563,-0.5375519243179504,-0.8315239948427353,-0.9959254593271811,-0.5375519243179504,-0.984464528627625,-1.1658065336722938,-0.3882285676537811,-0.8603334849373725,-1.1267678441732563,-0.5375519243179504,-0.8315239948427353,-0.9505084475638106,-0.6809857496093201,-0.9395701676505419,-0.9959254593271811,-0.5375519243179504,-0.984464528627625,-1.0753840503821261,-0.6809857496093201,-0.7936041538530185,-0.9505084475638106,-0.6809857496093201,-0.9395701676505419,-1.1267678441732563,-0.5375519243179504,-0.8315239948427353,-1.1780446902780395,0.6809857496093201,-0.6312441021797752,-1.1267678441732563,0.5375519243179504,-0.8315239948427353,-1.0753840503821261,0.6809857496093201,-0.7936041538530185,-1.2343337950127369,0.5375519243179504,-0.661406086418566,-1.1267678441732563,0.5375519243179504,-0.8315239948427353,-1.1780446902780395,0.6809857496093201,-0.6312441021797752,-1.2343337950127369,0.5375519243179504,-0.661406086418566,-1.1658065336722938,0.3882285676537812,-0.8603334849373725,-1.1267678441732563,0.5375519243179504,-0.8315239948427353,-1.277099280388322,0.3882285676537812,-0.6843215671664333,-1.1658065336722938,0.3882285676537812,-0.8603334849373725,-1.2343337950127369,0.5375519243179504,-0.661406086418566,-1.277099280388322,0.3882285676537812,-0.6843215671664333,-1.192072402827717,0.23465169756034637,-0.8797169813346798,-1.1658065336722938,0.3882285676537812,-0.8603334849373725,-1.3058725987977675,0.23465169756034637,-0.6997394776209308,-1.192072402827717,0.23465169756034637,-0.8797169813346798,-1.277099280388322,0.3882285676537812,-0.6843215671664333,-1.3058725987977675,0.23465169756034637,-0.6997394776209308,-1.2052776772805718,7.850393436441573e-2,-0.8894621143918706,-1.192072402827717,0.23465169756034637,-0.8797169813346798,-1.3203385037433746,7.850393436441573e-2,-0.7074908959287138,-1.2052776772805718,7.850393436441573e-2,-0.8894621143918706,-1.3058725987977675,0.23465169756034637,-0.6997394776209308,-1.3203385037433746,7.850393436441573e-2,-0.7074908959287138,-1.2052776772805718,-7.850393436441573e-2,-0.8894621143918706,-1.2052776772805718,7.850393436441573e-2,-0.8894621143918706,-1.3203385037433746,-7.850393436441573e-2,-0.7074908959287138,-1.2052776772805718,-7.850393436441573e-2,-0.8894621143918706,-1.3203385037433746,7.850393436441573e-2,-0.7074908959287138,-1.3203385037433746,-7.850393436441573e-2,-0.7074908959287138,-1.192072402827717,-0.23465169756034626,-0.8797169813346798,-1.2052776772805718,-7.850393436441573e-2,-0.8894621143918706,-1.3058725987977675,-0.23465169756034626,-0.6997394776209308,-1.192072402827717,-0.23465169756034626,-0.8797169813346798,-1.3203385037433746,-7.850393436441573e-2,-0.7074908959287138,-1.3058725987977675,-0.23465169756034626,-0.6997394776209308,-1.1658065336722938,-0.3882285676537811,-0.8603334849373725,-1.192072402827717,-0.23465169756034626,-0.8797169813346798,-1.277099280388322,-0.3882285676537811,-0.6843215671664333,-1.1658065336722938,-0.3882285676537811,-0.8603334849373725,-1.3058725987977675,-0.23465169756034626,-0.6997394776209308,-1.277099280388322,-0.3882285676537811,-0.6843215671664333,-1.1267678441732563,-0.5375519243179504,-0.8315239948427353,-1.1658065336722938,-0.3882285676537811,-0.8603334849373725,-1.2343337950127369,-0.5375519243179504,-0.661406086418566,-1.1267678441732563,-0.5375519243179504,-0.8315239948427353,-1.277099280388322,-0.3882285676537811,-0.6843215671664333,-1.2343337950127369,-0.5375519243179504,-0.661406086418566,-1.0753840503821261,-0.6809857496093201,-0.7936041538530185,-1.1267678441732563,-0.5375519243179504,-0.8315239948427353,-1.1780446902780395,-0.6809857496093201,-0.6312441021797752,-1.0753840503821261,-0.6809857496093201,-0.7936041538530185,-1.2343337950127369,-0.5375519243179504,-0.661406086418566,-1.2563696343765676,0.6809857496093201,-0.45584399814577214,-1.2343337950127369,0.5375519243179504,-0.661406086418566,-1.1780446902780395,0.6809857496093201,-0.6312441021797752,-1.3164012465204369,0.5375519243179504,-0.4776250483606452,-1.2343337950127369,0.5375519243179504,-0.661406086418566,-1.2563696343765676,0.6809857496093201,-0.45584399814577214,-1.3164012465204369,0.5375519243179504,-0.4776250483606452,-1.277099280388322,0.3882285676537812,-0.6843215671664333,-1.2343337950127369,0.5375519243179504,-0.661406086418566,-1.3620100911327575,0.3882285676537812,-0.49417313859621204,-1.277099280388322,0.3882285676537812,-0.6843215671664333,-1.3164012465204369,0.5375519243179504,-0.4776250483606452,-1.3620100911327575,0.3882285676537812,-0.49417313859621204,-1.3058725987977675,0.23465169756034637,-0.6997394776209308,-1.277099280388322,0.3882285676537812,-0.6843215671664333,-1.3926964681676928,0.23465169756034637,-0.5053069645129412,-1.3058725987977675,0.23465169756034637,-0.6997394776209308,-1.3620100911327575,0.3882285676537812,-0.49417313859621204,-1.3926964681676928,0.23465169756034637,-0.5053069645129412,-1.3203385037433746,7.850393436441573e-2,-0.7074908959287138,-1.3058725987977675,0.23465169756034637,-0.6997394776209308,-1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,-1.3203385037433746,7.850393436441573e-2,-0.7074908959287138,-1.3926964681676928,0.23465169756034637,-0.5053069645129412,-1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,-1.3203385037433746,-7.850393436441573e-2,-0.7074908959287138,-1.3203385037433746,7.850393436441573e-2,-0.7074908959287138,-1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,-1.3203385037433746,-7.850393436441573e-2,-0.7074908959287138,-1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,-1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,-1.3058725987977675,-0.23465169756034626,-0.6997394776209308,-1.3203385037433746,-7.850393436441573e-2,-0.7074908959287138,-1.3926964681676928,-0.23465169756034626,-0.5053069645129412,-1.3058725987977675,-0.23465169756034626,-0.6997394776209308,-1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,-1.3926964681676928,-0.23465169756034626,-0.5053069645129412,-1.277099280388322,-0.3882285676537811,-0.6843215671664333,-1.3058725987977675,-0.23465169756034626,-0.6997394776209308,-1.3620100911327575,-0.3882285676537811,-0.49417313859621204,-1.277099280388322,-0.3882285676537811,-0.6843215671664333,-1.3926964681676928,-0.23465169756034626,-0.5053069645129412,-1.3620100911327575,-0.3882285676537811,-0.49417313859621204,-1.2343337950127369,-0.5375519243179504,-0.661406086418566,-1.277099280388322,-0.3882285676537811,-0.6843215671664333,-1.3164012465204369,-0.5375519243179504,-0.4776250483606452,-1.2343337950127369,-0.5375519243179504,-0.661406086418566,-1.3620100911327575,-0.3882285676537811,-0.49417313859621204,-1.3164012465204369,-0.5375519243179504,-0.4776250483606452,-1.1780446902780395,-0.6809857496093201,-0.6312441021797752,-1.2343337950127369,-0.5375519243179504,-0.661406086418566,-1.2563696343765676,-0.6809857496093201,-0.45584399814577214,-1.1780446902780395,-0.6809857496093201,-0.6312441021797752,-1.3164012465204369,-0.5375519243179504,-0.4776250483606452,-1.3087408693470959,0.6809857496093201,-0.27102720477774955,-1.3164012465204369,0.5375519243179504,-0.4776250483606452,-1.2563696343765676,0.6809857496093201,-0.45584399814577214,-1.371274873764085,0.5375519243179504,-0.2839773745307208,-1.3164012465204369,0.5375519243179504,-0.4776250483606452,-1.3087408693470959,0.6809857496093201,-0.27102720477774955,-1.371274873764085,0.5375519243179504,-0.2839773745307208,-1.3620100911327575,0.3882285676537812,-0.49417313859621204,-1.3164012465204369,0.5375519243179504,-0.4776250483606452,-1.4187849037063995,0.3882285676537812,-0.29381622874224744,-1.3620100911327575,0.3882285676537812,-0.49417313859621204,-1.371274873764085,0.5375519243179504,-0.2839773745307208,-1.4187849037063995,0.3882285676537812,-0.29381622874224744,-1.3926964681676928,0.23465169756034637,-0.5053069645129412,-1.3620100911327575,0.3882285676537812,-0.49417313859621204,-1.450750429343878,0.23465169756034637,-0.3004359708666753,-1.3926964681676928,0.23465169756034637,-0.5053069645129412,-1.4187849037063995,0.3882285676537812,-0.29381622874224744,-1.450750429343878,0.23465169756034637,-0.3004359708666753,-1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,-1.3926964681676928,0.23465169756034637,-0.5053069645129412,-1.4668212296884202,7.850393436441573e-2,-0.3037640736240191,-1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,-1.450750429343878,0.23465169756034637,-0.3004359708666753,-1.4668212296884202,7.850393436441573e-2,-0.3037640736240191,-1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,-1.4081241712569101,7.850393436441573e-2,-0.5109045415841862,-1.4668212296884202,-7.850393436441573e-2,-0.3037640736240191,-1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,-1.4668212296884202,7.850393436441573e-2,-0.3037640736240191,-1.4668212296884202,-7.850393436441573e-2,-0.3037640736240191,-1.3926964681676928,-0.23465169756034626,-0.5053069645129412,-1.4081241712569101,-7.850393436441573e-2,-0.5109045415841862,-1.450750429343878,-0.23465169756034626,-0.3004359708666753,-1.3926964681676928,-0.23465169756034626,-0.5053069645129412,-1.4668212296884202,-7.850393436441573e-2,-0.3037640736240191,-1.450750429343878,-0.23465169756034626,-0.3004359708666753,-1.3620100911327575,-0.3882285676537811,-0.49417313859621204,-1.3926964681676928,-0.23465169756034626,-0.5053069645129412,-1.4187849037063995,-0.3882285676537811,-0.29381622874224744,-1.3620100911327575,-0.3882285676537811,-0.49417313859621204,-1.450750429343878,-0.23465169756034626,-0.3004359708666753,-1.4187849037063995,-0.3882285676537811,-0.29381622874224744,-1.3164012465204369,-0.5375519243179504,-0.4776250483606452,-1.3620100911327575,-0.3882285676537811,-0.49417313859621204,-1.371274873764085,-0.5375519243179504,-0.2839773745307208,-1.3164012465204369,-0.5375519243179504,-0.4776250483606452,-1.4187849037063995,-0.3882285676537811,-0.29381622874224744,-1.371274873764085,-0.5375519243179504,-0.2839773745307208,-1.2563696343765676,-0.6809857496093201,-0.45584399814577214,-1.3164012465204369,-0.5375519243179504,-0.4776250483606452,-1.3087408693470959,-0.6809857496093201,-0.27102720477774955,-1.2563696343765676,-0.6809857496093201,-0.45584399814577214,-1.371274873764085,-0.5375519243179504,-0.2839773745307208],[[1.0226629275469556,0.303478973012677,-0.7430081086773284,1.1078014609022506,0.1831716015188576,-0.6551519576555168,1.0880475531649874,0.30347897301267707,-0.6434695291860019,1.0412297531037233,0.1831716015188576,-0.756497697054377,1.1078014609022506,0.1831716015188576,-0.6551519576555168,1.0226629275469556,0.303478973012677,-0.7430081086773284,1.0412297531037233,0.1831716015188576,-0.756497697054377,1.1177224444470264,6.123838592253551e-2,-0.6610192109681584,1.1078014609022506,0.1831716015188576,-0.6551519576555168,1.05055454965929,6.123838592253551e-2,-0.763272558316994,1.1177224444470264,6.123838592253551e-2,-0.6610192109681584,1.0412297531037233,0.1831716015188576,-0.756497697054377,1.05055454965929,6.123838592253551e-2,-0.763272558316994,1.1177224444470264,-6.123838592253551e-2,-0.6610192109681584,1.1177224444470264,6.123838592253551e-2,-0.6610192109681584,1.05055454965929,-6.123838592253551e-2,-0.763272558316994,1.1177224444470264,-6.123838592253551e-2,-0.6610192109681584,1.05055454965929,6.123838592253551e-2,-0.763272558316994,1.05055454965929,-6.123838592253551e-2,-0.763272558316994,1.1078014609022506,-0.1831716015188576,-0.6551519576555168,1.1177224444470264,-6.123838592253551e-2,-0.6610192109681584,1.0412297531037233,-0.1831716015188576,-0.756497697054377,1.1078014609022506,-0.1831716015188576,-0.6551519576555168,1.05055454965929,-6.123838592253551e-2,-0.763272558316994,1.0412297531037233,-0.1831716015188576,-0.756497697054377,1.0880475531649874,-0.3034789730126769,-0.643469529186002,1.1078014609022506,-0.1831716015188576,-0.6551519576555168,1.0226629275469559,-0.30347897301267684,-0.7430081086773285,1.0880475531649874,-0.3034789730126769,-0.643469529186002,1.0412297531037233,-0.1831716015188576,-0.756497697054377,0.9482010733857819,0.30347897301267707,-0.8359516955956379,1.0412297531037233,0.1831716015188576,-0.756497697054377,1.0226629275469556,0.303478973012677,-0.7430081086773284,0.9654160163039952,0.1831716015188576,-0.8511287093388057,1.0412297531037233,0.1831716015188576,-0.756497697054377,0.9482010733857819,0.30347897301267707,-0.8359516955956379,0.9654160163039952,0.1831716015188576,-0.8511287093388057,1.05055454965929,6.123838592253551e-2,-0.763272558316994,1.0412297531037233,0.1831716015188576,-0.756497697054377,0.9740618583159874,6.123838592253552e-2,-0.8587510444032128,1.05055454965929,6.123838592253551e-2,-0.763272558316994,0.9654160163039952,0.1831716015188576,-0.8511287093388057,0.9740618583159874,6.123838592253552e-2,-0.8587510444032128,1.05055454965929,-6.123838592253551e-2,-0.763272558316994,1.05055454965929,6.123838592253551e-2,-0.763272558316994,0.9740618583159874,-6.123838592253552e-2,-0.8587510444032128,1.05055454965929,-6.123838592253551e-2,-0.763272558316994,0.9740618583159874,6.123838592253552e-2,-0.8587510444032128,0.9740618583159874,-6.123838592253552e-2,-0.8587510444032128,1.0412297531037233,-0.1831716015188576,-0.756497697054377,1.05055454965929,-6.123838592253551e-2,-0.763272558316994,0.9654160163039952,-0.1831716015188576,-0.8511287093388057,1.0412297531037233,-0.1831716015188576,-0.756497697054377,0.9740618583159874,-6.123838592253552e-2,-0.8587510444032128,0.9654160163039952,-0.1831716015188576,-0.8511287093388057,1.0226629275469559,-0.30347897301267684,-0.7430081086773285,1.0412297531037233,-0.1831716015188576,-0.756497697054377,0.948201073385782,-0.3034789730126769,-0.835951695595638,1.0226629275469559,-0.30347897301267684,-0.7430081086773285,0.9654160163039952,-0.1831716015188576,-0.8511287093388057,0.8653229193704408,0.30347897301267707,-0.9214753160836098,0.9654160163039952,0.1831716015188576,-0.8511287093388057,0.9482010733857819,0.30347897301267707,-0.8359516955956379,0.8810331785980458,0.1831716015188576,-0.9382050429444734,0.9654160163039952,0.1831716015188576,-0.8511287093388057,0.8653229193704408,0.30347897301267707,-0.9214753160836098,0.8810331785980458,0.1831716015188576,-0.9382050429444734,0.9740618583159874,6.123838592253552e-2,-0.8587510444032128,0.9654160163039952,0.1831716015188576,-0.8511287093388057,0.8889233249606927,6.123838592253552e-2,-0.9466071954250245,0.9740618583159874,6.123838592253552e-2,-0.8587510444032128,0.8810331785980458,0.1831716015188576,-0.9382050429444734,0.8889233249606927,6.123838592253552e-2,-0.9466071954250245,0.9740618583159874,-6.123838592253552e-2,-0.8587510444032128,0.9740618583159874,6.123838592253552e-2,-0.8587510444032128,0.8889233249606927,-6.123838592253552e-2,-0.9466071954250245,0.9740618583159874,-6.123838592253552e-2,-0.8587510444032128,0.8889233249606927,6.123838592253552e-2,-0.9466071954250245,0.8889233249606927,-6.123838592253552e-2,-0.9466071954250245,0.9654160163039952,-0.1831716015188576,-0.8511287093388057,0.9740618583159874,-6.123838592253552e-2,-0.8587510444032128,0.8810331785980458,-0.1831716015188576,-0.9382050429444734,0.9654160163039952,-0.1831716015188576,-0.8511287093388057,0.8889233249606927,-6.123838592253552e-2,-0.9466071954250245,0.8810331785980458,-0.1831716015188576,-0.9382050429444734,0.948201073385782,-0.3034789730126769,-0.835951695595638,0.9654160163039952,-0.1831716015188576,-0.8511287093388057,0.865322919370441,-0.3034789730126769,-0.9214753160836098,0.948201073385782,-0.3034789730126769,-0.835951695595638,0.8810331785980458,-0.1831716015188576,-0.9382050429444734,0.774764097863235,0.30347897301267707,-0.9988198564312478,0.8810331785980458,0.1831716015188576,-0.9382050429444734,0.8653229193704408,0.30347897301267707,-0.9214753160836098,0.7888302280271381,0.1831716015188576,-1.0169538021698286,0.8810331785980458,0.1831716015188576,-0.9382050429444734,0.774764097863235,0.30347897301267707,-0.9988198564312478,0.7888302280271381,0.1831716015188576,-1.0169538021698286,0.8889233249606927,6.123838592253552e-2,-0.9466071954250245,0.8810331785980458,0.1831716015188576,-0.9382050429444734,0.7958946452427506,6.123838592253551e-2,-1.0260611939662854,0.8889233249606927,6.123838592253552e-2,-0.9466071954250245,0.7888302280271381,0.1831716015188576,-1.0169538021698286,0.7958946452427506,6.123838592253551e-2,-1.0260611939662854,0.8889233249606927,-6.123838592253552e-2,-0.9466071954250245,0.8889233249606927,6.123838592253552e-2,-0.9466071954250245,0.7958946452427506,-6.123838592253551e-2,-1.0260611939662854,0.8889233249606927,-6.123838592253552e-2,-0.9466071954250245,0.7958946452427506,6.123838592253551e-2,-1.0260611939662854,0.7958946452427506,-6.123838592253551e-2,-1.0260611939662854,0.8810331785980458,-0.1831716015188576,-0.9382050429444734,0.8889233249606927,-6.123838592253552e-2,-0.9466071954250245,0.7888302280271381,-0.1831716015188576,-1.0169538021698286,0.8810331785980458,-0.1831716015188576,-0.9382050429444734,0.7958946452427506,-6.123838592253551e-2,-1.0260611939662854,0.7888302280271381,-0.1831716015188576,-1.0169538021698286,0.865322919370441,-0.3034789730126769,-0.9214753160836098,0.8810331785980458,-0.1831716015188576,-0.9382050429444734,0.7747640978632351,-0.3034789730126769,-0.9988198564312479,0.865322919370441,-0.3034789730126769,-0.9214753160836098,0.7888302280271381,-0.1831716015188576,-1.0169538021698286,0.6773284153748707,0.30347897301267707,-1.0672988010229083,0.7888302280271381,0.1831716015188576,-1.0169538021698286,0.774764097863235,0.30347897301267707,-0.9988198564312478,0.6896255645079409,0.1831716015188576,-1.0866760074530593,0.7888302280271381,0.1831716015188576,-1.0169538021698286,0.6773284153748707,0.30347897301267707,-1.0672988010229083,0.6896255645079409,0.1831716015188576,-1.0866760074530593,0.7958946452427506,6.123838592253551e-2,-1.0260611939662854,0.7888302280271381,0.1831716015188576,-1.0169538021698286,0.6958015483091968,6.123838592253551e-2,-1.0964078007110891,0.7958946452427506,6.123838592253551e-2,-1.0260611939662854,0.6896255645079409,0.1831716015188576,-1.0866760074530593,0.6958015483091968,6.123838592253551e-2,-1.0964078007110891,0.7958946452427506,-6.123838592253551e-2,-1.0260611939662854,0.7958946452427506,6.123838592253551e-2,-1.0260611939662854,0.6958015483091968,-6.123838592253551e-2,-1.0964078007110891,0.7958946452427506,-6.123838592253551e-2,-1.0260611939662854,0.6958015483091968,6.123838592253551e-2,-1.0964078007110891,0.6958015483091968,-6.123838592253551e-2,-1.0964078007110891,0.7888302280271381,-0.1831716015188576,-1.0169538021698286,0.7958946452427506,-6.123838592253551e-2,-1.0260611939662854,0.6896255645079409,-0.1831716015188576,-1.0866760074530593,0.7888302280271381,-0.1831716015188576,-1.0169538021698286,0.6958015483091968,-6.123838592253551e-2,-1.0964078007110891,0.6896255645079409,-0.1831716015188576,-1.0866760074530593,0.7747640978632351,-0.3034789730126769,-0.9988198564312479,0.7888302280271381,-0.1831716015188576,-1.0169538021698286,0.6773284153748707,-0.3034789730126769,-1.0672988010229085,0.7747640978632351,-0.3034789730126769,-0.9988198564312479,0.6896255645079409,-0.1831716015188576,-1.0866760074530593],[0.502027061643441,0.303478973012677,-1.1601160900172116,0.6200317949183421,0.1831716015188576,-1.1278336480560225,0.6089756162588454,0.30347897301267707,-1.107722533733665,0.5111415495428594,0.1831716015188576,-1.1811784288277285,0.6200317949183421,0.1831716015188576,-1.1278336480560225,0.502027061643441,0.303478973012677,-1.1601160900172116,0.5111415495428594,0.1831716015188576,-1.1811784288277285,0.625584527471712,6.123838592253551e-2,-1.13793403107456,0.6200317949183421,0.1831716015188576,-1.1278336480560225,0.5157191088628622,6.123838592253551e-2,-1.1917565442838138,0.625584527471712,6.123838592253551e-2,-1.13793403107456,0.5111415495428594,0.1831716015188576,-1.1811784288277285,0.5157191088628622,6.123838592253551e-2,-1.1917565442838138,0.625584527471712,-6.123838592253551e-2,-1.13793403107456,0.625584527471712,6.123838592253551e-2,-1.13793403107456,0.5157191088628622,-6.123838592253551e-2,-1.1917565442838138,0.625584527471712,-6.123838592253551e-2,-1.13793403107456,0.5157191088628622,6.123838592253551e-2,-1.1917565442838138,0.5157191088628622,-6.123838592253551e-2,-1.1917565442838138,0.6200317949183421,-0.1831716015188576,-1.1278336480560225,0.625584527471712,-6.123838592253551e-2,-1.13793403107456,0.5111415495428594,-0.1831716015188576,-1.1811784288277285,0.6200317949183421,-0.1831716015188576,-1.1278336480560225,0.5157191088628622,-6.123838592253551e-2,-1.1917565442838138,0.5111415495428594,-0.1831716015188576,-1.1811784288277285,0.6089756162588456,-0.3034789730126769,-1.107722533733665,0.6200317949183421,-0.1831716015188576,-1.1278336480560225,0.502027061643441,-0.30347897301267684,-1.1601160900172116,0.6089756162588456,-0.3034789730126769,-1.107722533733665,0.5111415495428594,-0.1831716015188576,-1.1811784288277285,0.3906224792884663,0.30347897301267707,-1.202212373756693,0.5111415495428594,0.1831716015188576,-1.1811784288277285,0.502027061643441,0.303478973012677,-1.1601160900172116,0.3977143755879614,0.1831716015188576,-1.2240389862450032,0.5111415495428594,0.1831716015188576,-1.1811784288277285,0.3906224792884663,0.30347897301267707,-1.202212373756693,0.3977143755879614,0.1831716015188576,-1.2240389862450032,0.5157191088628622,6.123838592253551e-2,-1.1917565442838138,0.5111415495428594,0.1831716015188576,-1.1811784288277285,0.40127613093400993,6.123838592253552e-2,-1.2350009420369827,0.5157191088628622,6.123838592253551e-2,-1.1917565442838138,0.3977143755879614,0.1831716015188576,-1.2240389862450032,0.40127613093400993,6.123838592253552e-2,-1.2350009420369827,0.5157191088628622,-6.123838592253551e-2,-1.1917565442838138,0.5157191088628622,6.123838592253551e-2,-1.1917565442838138,0.40127613093400993,-6.123838592253552e-2,-1.2350009420369827,0.5157191088628622,-6.123838592253551e-2,-1.1917565442838138,0.40127613093400993,6.123838592253552e-2,-1.2350009420369827,0.40127613093400993,-6.123838592253552e-2,-1.2350009420369827,0.5111415495428594,-0.1831716015188576,-1.1811784288277285,0.5157191088628622,-6.123838592253551e-2,-1.1917565442838138,0.3977143755879614,-0.1831716015188576,-1.2240389862450032,0.5111415495428594,-0.1831716015188576,-1.1811784288277285,0.40127613093400993,-6.123838592253552e-2,-1.2350009420369827,0.3977143755879614,-0.1831716015188576,-1.2240389862450032,0.502027061643441,-0.30347897301267684,-1.1601160900172116,0.5111415495428594,-0.1831716015188576,-1.1811784288277285,0.3906224792884663,-0.3034789730126769,-1.2022123737566932,0.502027061643441,-0.30347897301267684,-1.1601160900172116,0.3977143755879614,-0.1831716015188576,-1.2240389862450032,0.27575070415366215,0.30347897301267707,-1.2336377353574794,0.3977143755879614,0.1831716015188576,-1.2240389862450032,0.3906224792884663,0.30347897301267707,-1.202212373756693,0.2807570606796171,0.1831716015188576,-1.2560348869659468,0.3977143755879614,0.1831716015188576,-1.2240389862450032,0.27575070415366215,0.30347897301267707,-1.2336377353574794,0.2807570606796171,0.1831716015188576,-1.2560348869659468,0.40127613093400993,6.123838592253552e-2,-1.2350009420369827,0.3977143755879614,0.1831716015188576,-1.2240389862450032,0.28327139765910936,6.123838592253552e-2,-1.267283383998172,0.40127613093400993,6.123838592253552e-2,-1.2350009420369827,0.2807570606796171,0.1831716015188576,-1.2560348869659468,0.28327139765910936,6.123838592253552e-2,-1.267283383998172,0.40127613093400993,-6.123838592253552e-2,-1.2350009420369827,0.40127613093400993,6.123838592253552e-2,-1.2350009420369827,0.28327139765910936,-6.123838592253552e-2,-1.267283383998172,0.40127613093400993,-6.123838592253552e-2,-1.2350009420369827,0.28327139765910936,6.123838592253552e-2,-1.267283383998172,0.28327139765910936,-6.123838592253552e-2,-1.267283383998172,0.3977143755879614,-0.1831716015188576,-1.2240389862450032,0.40127613093400993,-6.123838592253552e-2,-1.2350009420369827,0.2807570606796171,-0.1831716015188576,-1.2560348869659468,0.3977143755879614,-0.1831716015188576,-1.2240389862450032,0.28327139765910936,-6.123838592253552e-2,-1.267283383998172,0.2807570606796171,-0.1831716015188576,-1.2560348869659468,0.3906224792884663,-0.3034789730126769,-1.2022123737566932,0.3977143755879614,-0.1831716015188576,-1.2240389862450032,0.27575070415366215,-0.3034789730126769,-1.2336377353574794,0.3906224792884663,-0.3034789730126769,-1.2022123737566932,0.2807570606796171,-0.1831716015188576,-1.2560348869659468,0.158431346247338,0.30347897301267707,-1.2541132410852804,0.2807570606796171,0.1831716015188576,-1.2560348869659468,0.27575070415366215,0.30347897301267707,-1.2336377353574794,0.16130772622480954,0.1831716015188576,-1.2768821331106481,0.2807570606796171,0.1831716015188576,-1.2560348869659468,0.158431346247338,0.30347897301267707,-1.2541132410852804,0.16130772622480954,0.1831716015188576,-1.2768821331106481,0.28327139765910936,6.123838592253552e-2,-1.267283383998172,0.2807570606796171,0.1831716015188576,-1.2560348869659468,0.16275232740471574,6.123838592253551e-2,-1.2883173289271364,0.28327139765910936,6.123838592253552e-2,-1.267283383998172,0.16130772622480954,0.1831716015188576,-1.2768821331106481,0.16275232740471574,6.123838592253551e-2,-1.2883173289271364,0.28327139765910936,-6.123838592253552e-2,-1.267283383998172,0.28327139765910936,6.123838592253552e-2,-1.267283383998172,0.16275232740471574,-6.123838592253551e-2,-1.2883173289271364,0.28327139765910936,-6.123838592253552e-2,-1.267283383998172,0.16275232740471574,6.123838592253551e-2,-1.2883173289271364,0.16275232740471574,-6.123838592253551e-2,-1.2883173289271364,0.2807570606796171,-0.1831716015188576,-1.2560348869659468,0.28327139765910936,-6.123838592253552e-2,-1.267283383998172,0.16130772622480954,-0.1831716015188576,-1.2768821331106481,0.2807570606796171,-0.1831716015188576,-1.2560348869659468,0.16275232740471574,-6.123838592253551e-2,-1.2883173289271364,0.16130772622480954,-0.1831716015188576,-1.2768821331106481,0.27575070415366215,-0.3034789730126769,-1.2336377353574794,0.2807570606796171,-0.1831716015188576,-1.2560348869659468,0.158431346247338,-0.3034789730126769,-1.2541132410852807,0.27575070415366215,-0.3034789730126769,-1.2336377353574794,0.16130772622480954,-0.1831716015188576,-1.2768821331106481,3.970574049576024e-2,0.30347897301267707,-1.2634571489017163,0.16130772622480954,0.1831716015188576,-1.2768821331106481,0.158431346247338,0.30347897301267707,-1.2541132410852804,4.042661297243777e-2,0.1831716015188576,-1.2863956830464698,0.16130772622480954,0.1831716015188576,-1.2768821331106481,3.970574049576024e-2,0.30347897301267707,-1.2634571489017163,4.042661297243777e-2,0.1831716015188576,-1.2863956830464698,0.16275232740471574,6.123838592253551e-2,-1.2883173289271364,0.16130772622480954,0.1831716015188576,-1.2768821331106481,4.0788655970416775e-2,6.123838592253551e-2,-1.2979160780396126,0.16275232740471574,6.123838592253551e-2,-1.2883173289271364,4.042661297243777e-2,0.1831716015188576,-1.2863956830464698,4.0788655970416775e-2,6.123838592253551e-2,-1.2979160780396126,0.16275232740471574,-6.123838592253551e-2,-1.2883173289271364,0.16275232740471574,6.123838592253551e-2,-1.2883173289271364,4.0788655970416775e-2,-6.123838592253551e-2,-1.2979160780396126,0.16275232740471574,-6.123838592253551e-2,-1.2883173289271364,4.0788655970416775e-2,6.123838592253551e-2,-1.2979160780396126,4.0788655970416775e-2,-6.123838592253551e-2,-1.2979160780396126,0.16130772622480954,-0.1831716015188576,-1.2768821331106481,0.16275232740471574,-6.123838592253551e-2,-1.2883173289271364,4.042661297243777e-2,-0.1831716015188576,-1.2863956830464698,0.16130772622480954,-0.1831716015188576,-1.2768821331106481,4.0788655970416775e-2,-6.123838592253551e-2,-1.2979160780396126,4.042661297243777e-2,-0.1831716015188576,-1.2863956830464698,0.158431346247338,-0.3034789730126769,-1.2541132410852807,0.16130772622480954,-0.1831716015188576,-1.2768821331106481,3.970574049576024e-2,-0.3034789730126769,-1.2634571489017166,0.158431346247338,-0.3034789730126769,-1.2541132410852807,4.042661297243777e-2,-0.1831716015188576,-1.2863956830464698],[-0.15843134624733782,0.303478973012677,-1.2541132410852802,-4.0426612972436764e-2,0.1831716015188576,-1.28639568304647,-3.970574049575925e-2,0.30347897301267707,-1.2634571489017166,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-4.0426612972436764e-2,0.1831716015188576,-1.28639568304647,-0.15843134624733782,0.303478973012677,-1.2541132410852802,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-4.0788655970415755e-2,6.123838592253551e-2,-1.2979160780396126,-4.0426612972436764e-2,0.1831716015188576,-1.28639568304647,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-4.0788655970415755e-2,6.123838592253551e-2,-1.2979160780396126,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-4.0788655970415755e-2,-6.123838592253551e-2,-1.2979160780396126,-4.0788655970415755e-2,6.123838592253551e-2,-1.2979160780396126,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-4.0788655970415755e-2,-6.123838592253551e-2,-1.2979160780396126,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-4.0426612972436764e-2,-0.1831716015188576,-1.28639568304647,-4.0788655970415755e-2,-6.123838592253551e-2,-1.2979160780396126,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-4.0426612972436764e-2,-0.1831716015188576,-1.28639568304647,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-3.970574049575925e-2,-0.3034789730126769,-1.2634571489017166,-4.0426612972436764e-2,-0.1831716015188576,-1.28639568304647,-0.15843134624733785,-0.30347897301267684,-1.2541132410852804,-3.970574049575925e-2,-0.3034789730126769,-1.2634571489017166,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-0.27575070415366115,0.30347897301267707,-1.2336377353574794,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-0.15843134624733782,0.303478973012677,-1.2541132410852802,-0.2807570606796161,0.1831716015188576,-1.256034886965947,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-0.27575070415366115,0.30347897301267707,-1.2336377353574794,-0.2807570606796161,0.1831716015188576,-1.256034886965947,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-0.1613077262248091,0.1831716015188576,-1.2768821331106481,-0.2832713976591084,6.123838592253552e-2,-1.267283383998172,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-0.2807570606796161,0.1831716015188576,-1.256034886965947,-0.2832713976591084,6.123838592253552e-2,-1.267283383998172,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-0.1627523274047156,6.123838592253551e-2,-1.2883173289271364,-0.2832713976591084,-6.123838592253552e-2,-1.267283383998172,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-0.2832713976591084,6.123838592253552e-2,-1.267283383998172,-0.2832713976591084,-6.123838592253552e-2,-1.267283383998172,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-0.1627523274047156,-6.123838592253551e-2,-1.2883173289271364,-0.2807570606796161,-0.1831716015188576,-1.256034886965947,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-0.2832713976591084,-6.123838592253552e-2,-1.267283383998172,-0.2807570606796161,-0.1831716015188576,-1.256034886965947,-0.15843134624733785,-0.30347897301267684,-1.2541132410852804,-0.1613077262248091,-0.1831716015188576,-1.2768821331106481,-0.2757507041536612,-0.3034789730126769,-1.2336377353574797,-0.15843134624733785,-0.30347897301267684,-1.2541132410852804,-0.2807570606796161,-0.1831716015188576,-1.256034886965947,-0.3906224792884654,0.30347897301267707,-1.2022123737566934,-0.2807570606796161,0.1831716015188576,-1.256034886965947,-0.27575070415366115,0.30347897301267707,-1.2336377353574794,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.2807570606796161,0.1831716015188576,-1.256034886965947,-0.3906224792884654,0.30347897301267707,-1.2022123737566934,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.2832713976591084,6.123838592253552e-2,-1.267283383998172,-0.2807570606796161,0.1831716015188576,-1.256034886965947,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.2832713976591084,6.123838592253552e-2,-1.267283383998172,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.2832713976591084,-6.123838592253552e-2,-1.267283383998172,-0.2832713976591084,6.123838592253552e-2,-1.267283383998172,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.2832713976591084,-6.123838592253552e-2,-1.267283383998172,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.2807570606796161,-0.1831716015188576,-1.256034886965947,-0.2832713976591084,-6.123838592253552e-2,-1.267283383998172,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.2807570606796161,-0.1831716015188576,-1.256034886965947,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.2757507041536612,-0.3034789730126769,-1.2336377353574797,-0.2807570606796161,-0.1831716015188576,-1.256034886965947,-0.39062247928846544,-0.3034789730126769,-1.2022123737566934,-0.2757507041536612,-0.3034789730126769,-1.2336377353574797,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.5020270616434406,0.30347897301267707,-1.160116090017212,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.3906224792884654,0.30347897301267707,-1.2022123737566934,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.5020270616434406,0.30347897301267707,-1.160116090017212,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.39771437558796047,0.1831716015188576,-1.2240389862450034,-0.5157191088628618,6.123838592253551e-2,-1.191756544283814,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.5157191088628618,6.123838592253551e-2,-1.191756544283814,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.401276130934009,6.123838592253552e-2,-1.235000942036983,-0.5157191088628618,-6.123838592253551e-2,-1.191756544283814,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.5157191088628618,6.123838592253551e-2,-1.191756544283814,-0.5157191088628618,-6.123838592253551e-2,-1.191756544283814,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.401276130934009,-6.123838592253552e-2,-1.235000942036983,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.5157191088628618,-6.123838592253551e-2,-1.191756544283814,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.39062247928846544,-0.3034789730126769,-1.2022123737566934,-0.39771437558796047,-0.1831716015188576,-1.2240389862450034,-0.5020270616434407,-0.3034789730126769,-1.160116090017212,-0.39062247928846544,-0.3034789730126769,-1.2022123737566934,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.6089756162588442,0.30347897301267707,-1.1077225337336654,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.5020270616434406,0.30347897301267707,-1.160116090017212,-0.6200317949183408,0.1831716015188576,-1.127833648056023,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.6089756162588442,0.30347897301267707,-1.1077225337336654,-0.6200317949183408,0.1831716015188576,-1.127833648056023,-0.5157191088628618,6.123838592253551e-2,-1.191756544283814,-0.5111415495428587,0.1831716015188576,-1.181178428827729,-0.6255845274717108,6.123838592253551e-2,-1.1379340310745607,-0.5157191088628618,6.123838592253551e-2,-1.191756544283814,-0.6200317949183408,0.1831716015188576,-1.127833648056023,-0.6255845274717108,6.123838592253551e-2,-1.1379340310745607,-0.5157191088628618,-6.123838592253551e-2,-1.191756544283814,-0.5157191088628618,6.123838592253551e-2,-1.191756544283814,-0.6255845274717108,-6.123838592253551e-2,-1.1379340310745607,-0.5157191088628618,-6.123838592253551e-2,-1.191756544283814,-0.6255845274717108,6.123838592253551e-2,-1.1379340310745607,-0.6255845274717108,-6.123838592253551e-2,-1.1379340310745607,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.5157191088628618,-6.123838592253551e-2,-1.191756544283814,-0.6200317949183408,-0.1831716015188576,-1.127833648056023,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.6255845274717108,-6.123838592253551e-2,-1.1379340310745607,-0.6200317949183408,-0.1831716015188576,-1.127833648056023,-0.5020270616434407,-0.3034789730126769,-1.160116090017212,-0.5111415495428587,-0.1831716015188576,-1.181178428827729,-0.6089756162588442,-0.3034789730126769,-1.1077225337336656,-0.5020270616434407,-0.3034789730126769,-1.160116090017212,-0.6200317949183408,-0.1831716015188576,-1.127833648056023],[-0.7747640978632345,0.303478973012677,-0.9988198564312477,-0.6896255645079399,0.1831716015188576,-1.0866760074530597,-0.6773284153748697,0.30347897301267707,-1.0672988010229087,-0.7888302280271374,0.1831716015188576,-1.0169538021698288,-0.6896255645079399,0.1831716015188576,-1.0866760074530597,-0.7747640978632345,0.303478973012677,-0.9988198564312477,-0.7888302280271374,0.1831716015188576,-1.0169538021698288,-0.6958015483091958,6.123838592253551e-2,-1.0964078007110898,-0.6896255645079399,0.1831716015188576,-1.0866760074530597,-0.7958946452427502,6.123838592253551e-2,-1.0260611939662856,-0.6958015483091958,6.123838592253551e-2,-1.0964078007110898,-0.7888302280271374,0.1831716015188576,-1.0169538021698288,-0.7958946452427502,6.123838592253551e-2,-1.0260611939662856,-0.6958015483091958,-6.123838592253551e-2,-1.0964078007110898,-0.6958015483091958,6.123838592253551e-2,-1.0964078007110898,-0.7958946452427502,-6.123838592253551e-2,-1.0260611939662856,-0.6958015483091958,-6.123838592253551e-2,-1.0964078007110898,-0.7958946452427502,6.123838592253551e-2,-1.0260611939662856,-0.7958946452427502,-6.123838592253551e-2,-1.0260611939662856,-0.6896255645079399,-0.1831716015188576,-1.0866760074530597,-0.6958015483091958,-6.123838592253551e-2,-1.0964078007110898,-0.7888302280271374,-0.1831716015188576,-1.0169538021698288,-0.6896255645079399,-0.1831716015188576,-1.0866760074530597,-0.7958946452427502,-6.123838592253551e-2,-1.0260611939662856,-0.7888302280271374,-0.1831716015188576,-1.0169538021698288,-0.6773284153748698,-0.3034789730126769,-1.067298801022909,-0.6896255645079399,-0.1831716015188576,-1.0866760074530597,-0.7747640978632346,-0.30347897301267684,-0.9988198564312479,-0.6773284153748698,-0.3034789730126769,-1.067298801022909,-0.7888302280271374,-0.1831716015188576,-1.0169538021698288,-0.8653229193704406,0.30347897301267707,-0.9214753160836099,-0.7888302280271374,0.1831716015188576,-1.0169538021698288,-0.7747640978632345,0.303478973012677,-0.9988198564312477,-0.8810331785980454,0.1831716015188576,-0.9382050429444736,-0.7888302280271374,0.1831716015188576,-1.0169538021698288,-0.8653229193704406,0.30347897301267707,-0.9214753160836099,-0.8810331785980454,0.1831716015188576,-0.9382050429444736,-0.7958946452427502,6.123838592253551e-2,-1.0260611939662856,-0.7888302280271374,0.1831716015188576,-1.0169538021698288,-0.8889233249606925,6.123838592253552e-2,-0.9466071954250247,-0.7958946452427502,6.123838592253551e-2,-1.0260611939662856,-0.8810331785980454,0.1831716015188576,-0.9382050429444736,-0.8889233249606925,6.123838592253552e-2,-0.9466071954250247,-0.7958946452427502,-6.123838592253551e-2,-1.0260611939662856,-0.7958946452427502,6.123838592253551e-2,-1.0260611939662856,-0.8889233249606925,-6.123838592253552e-2,-0.9466071954250247,-0.7958946452427502,-6.123838592253551e-2,-1.0260611939662856,-0.8889233249606925,6.123838592253552e-2,-0.9466071954250247,-0.8889233249606925,-6.123838592253552e-2,-0.9466071954250247,-0.7888302280271374,-0.1831716015188576,-1.0169538021698288,-0.7958946452427502,-6.123838592253551e-2,-1.0260611939662856,-0.8810331785980454,-0.1831716015188576,-0.9382050429444736,-0.7888302280271374,-0.1831716015188576,-1.0169538021698288,-0.8889233249606925,-6.123838592253552e-2,-0.9466071954250247,-0.8810331785980454,-0.1831716015188576,-0.9382050429444736,-0.7747640978632346,-0.30347897301267684,-0.9988198564312479,-0.7888302280271374,-0.1831716015188576,-1.0169538021698288,-0.8653229193704407,-0.3034789730126769,-0.92147531608361,-0.7747640978632346,-0.30347897301267684,-0.9988198564312479,-0.8810331785980454,-0.1831716015188576,-0.9382050429444736,-0.948201073385781,0.30347897301267707,-0.8359516955956388,-0.8810331785980454,0.1831716015188576,-0.9382050429444736,-0.8653229193704406,0.30347897301267707,-0.9214753160836099,-0.9654160163039943,0.1831716015188576,-0.8511287093388066,-0.8810331785980454,0.1831716015188576,-0.9382050429444736,-0.948201073385781,0.30347897301267707,-0.8359516955956388,-0.9654160163039943,0.1831716015188576,-0.8511287093388066,-0.8889233249606925,6.123838592253552e-2,-0.9466071954250247,-0.8810331785980454,0.1831716015188576,-0.9382050429444736,-0.9740618583159865,6.123838592253552e-2,-0.8587510444032136,-0.8889233249606925,6.123838592253552e-2,-0.9466071954250247,-0.9654160163039943,0.1831716015188576,-0.8511287093388066,-0.9740618583159865,6.123838592253552e-2,-0.8587510444032136,-0.8889233249606925,-6.123838592253552e-2,-0.9466071954250247,-0.8889233249606925,6.123838592253552e-2,-0.9466071954250247,-0.9740618583159865,-6.123838592253552e-2,-0.8587510444032136,-0.8889233249606925,-6.123838592253552e-2,-0.9466071954250247,-0.9740618583159865,6.123838592253552e-2,-0.8587510444032136,-0.9740618583159865,-6.123838592253552e-2,-0.8587510444032136,-0.8810331785980454,-0.1831716015188576,-0.9382050429444736,-0.8889233249606925,-6.123838592253552e-2,-0.9466071954250247,-0.9654160163039943,-0.1831716015188576,-0.8511287093388066,-0.8810331785980454,-0.1831716015188576,-0.9382050429444736,-0.9740618583159865,-6.123838592253552e-2,-0.8587510444032136,-0.9654160163039943,-0.1831716015188576,-0.8511287093388066,-0.8653229193704407,-0.3034789730126769,-0.92147531608361,-0.8810331785980454,-0.1831716015188576,-0.9382050429444736,-0.9482010733857811,-0.3034789730126769,-0.8359516955956389,-0.8653229193704407,-0.3034789730126769,-0.92147531608361,-0.9654160163039943,-0.1831716015188576,-0.8511287093388066,-1.0226629275469556,0.30347897301267707,-0.7430081086773286,-0.9654160163039943,0.1831716015188576,-0.8511287093388066,-0.948201073385781,0.30347897301267707,-0.8359516955956388,-1.0412297531037231,0.1831716015188576,-0.7564976970543775,-0.9654160163039943,0.1831716015188576,-0.8511287093388066,-1.0226629275469556,0.30347897301267707,-0.7430081086773286,-1.0412297531037231,0.1831716015188576,-0.7564976970543775,-0.9740618583159865,6.123838592253552e-2,-0.8587510444032136,-0.9654160163039943,0.1831716015188576,-0.8511287093388066,-1.0505545496592898,6.123838592253551e-2,-0.7632725583169943,-0.9740618583159865,6.123838592253552e-2,-0.8587510444032136,-1.0412297531037231,0.1831716015188576,-0.7564976970543775,-1.0505545496592898,6.123838592253551e-2,-0.7632725583169943,-0.9740618583159865,-6.123838592253552e-2,-0.8587510444032136,-0.9740618583159865,6.123838592253552e-2,-0.8587510444032136,-1.0505545496592898,-6.123838592253551e-2,-0.7632725583169943,-0.9740618583159865,-6.123838592253552e-2,-0.8587510444032136,-1.0505545496592898,6.123838592253551e-2,-0.7632725583169943,-1.0505545496592898,-6.123838592253551e-2,-0.7632725583169943,-0.9654160163039943,-0.1831716015188576,-0.8511287093388066,-0.9740618583159865,-6.123838592253552e-2,-0.8587510444032136,-1.0412297531037231,-0.1831716015188576,-0.7564976970543775,-0.9654160163039943,-0.1831716015188576,-0.8511287093388066,-1.0505545496592898,-6.123838592253551e-2,-0.7632725583169943,-1.0412297531037231,-0.1831716015188576,-0.7564976970543775,-0.9482010733857811,-0.3034789730126769,-0.8359516955956389,-0.9654160163039943,-0.1831716015188576,-0.8511287093388066,-1.0226629275469559,-0.3034789730126769,-0.7430081086773287,-0.9482010733857811,-0.3034789730126769,-0.8359516955956389,-1.0412297531037231,-0.1831716015188576,-0.7564976970543775,-1.088047553164987,0.30347897301267707,-0.6434695291860026,-1.0412297531037231,0.1831716015188576,-0.7564976970543775,-1.0226629275469556,0.30347897301267707,-0.7430081086773286,-1.1078014609022502,0.1831716015188576,-0.6551519576555175,-1.0412297531037231,0.1831716015188576,-0.7564976970543775,-1.088047553164987,0.30347897301267707,-0.6434695291860026,-1.1078014609022502,0.1831716015188576,-0.6551519576555175,-1.0505545496592898,6.123838592253551e-2,-0.7632725583169943,-1.0412297531037231,0.1831716015188576,-0.7564976970543775,-1.117722444447026,6.123838592253551e-2,-0.6610192109681592,-1.0505545496592898,6.123838592253551e-2,-0.7632725583169943,-1.1078014609022502,0.1831716015188576,-0.6551519576555175,-1.117722444447026,6.123838592253551e-2,-0.6610192109681592,-1.0505545496592898,-6.123838592253551e-2,-0.7632725583169943,-1.0505545496592898,6.123838592253551e-2,-0.7632725583169943,-1.117722444447026,-6.123838592253551e-2,-0.6610192109681592,-1.0505545496592898,-6.123838592253551e-2,-0.7632725583169943,-1.117722444447026,6.123838592253551e-2,-0.6610192109681592,-1.117722444447026,-6.123838592253551e-2,-0.6610192109681592,-1.0412297531037231,-0.1831716015188576,-0.7564976970543775,-1.0505545496592898,-6.123838592253551e-2,-0.7632725583169943,-1.1078014609022502,-0.1831716015188576,-0.6551519576555175,-1.0412297531037231,-0.1831716015188576,-0.7564976970543775,-1.117722444447026,-6.123838592253551e-2,-0.6610192109681592,-1.1078014609022502,-0.1831716015188576,-0.6551519576555175,-1.0226629275469559,-0.3034789730126769,-0.7430081086773287,-1.0412297531037231,-0.1831716015188576,-0.7564976970543775,-1.088047553164987,-0.3034789730126769,-0.6434695291860026,-1.0226629275469559,-0.3034789730126769,-0.7430081086773287,-1.1078014609022502,-0.1831716015188576,-0.6551519576555175]]]
      case 5:
        return [[1.2904122258874484,0.6784329353268667,-0.34866321793956584,1.3319147297205405,0.6826289037478287,8.379697203060224e-2,1.3340484636550194,0.6784329353268667,8.393121519109781e-2,1.2883482856102721,0.6826289037478287,-0.34810555113802855,1.3319147297205405,0.6826289037478287,8.379697203060224e-2,1.2904122258874484,0.6784329353268667,-0.34866321793956584,1.1102460001099754,0.6784329353268667,-0.7443679006402496,1.2883482856102721,0.6826289037478287,-0.34810555113802855,1.2904122258874484,0.6784329353268667,-0.34866321793956584,1.1084702253682064,0.6826289037478287,-0.7431773269147784,1.2883482856102721,0.6826289037478287,-0.34810555113802855,1.1102460001099754,0.6784329353268667,-0.7443679006402496,0.8126119837752565,0.6784329353268667,-1.0613159360381876,1.1084702253682064,0.6826289037478287,-0.7431773269147784,1.1102460001099754,0.6784329353268667,-0.7443679006402496,0.8113122575564691,0.6826289037478287,-1.0596184221249947,1.1084702253682064,0.6826289037478287,-0.7431773269147784,0.8126119837752565,0.6784329353268667,-1.0613159360381876,0.42900086479394856,0.6784329353268667,-1.2659731475311065,0.8113122575564691,0.6826289037478287,-1.0596184221249947,0.8126119837752565,0.6784329353268667,-1.0613159360381876,0.42831470253817583,0.6826289037478287,-1.2639482961567976,0.8113122575564691,0.6826289037478287,-1.0596184221249947,0.42900086479394856,0.6784329353268667,-1.2659731475311065,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,0.42831470253817583,0.6826289037478287,-1.2639482961567976,0.42900086479394856,0.6784329353268667,-1.2659731475311065,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,0.42831470253817583,0.6826289037478287,-1.2639482961567976,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,-0.4290008647939484,0.6784329353268667,-1.2659731475311065,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,8.184841820295902e-17,0.6784329353268667,-1.3366861083529566,-0.4283147025381757,0.6826289037478287,-1.2639482961567976,8.171750635667977e-17,0.6826289037478287,-1.3345481556572014,-0.4290008647939484,0.6784329353268667,-1.2659731475311065,-0.8126119837752559,0.6784329353268667,-1.061315936038188,-0.4283147025381757,0.6826289037478287,-1.2639482961567976,-0.4290008647939484,0.6784329353268667,-1.2659731475311065,-0.8113122575564685,0.6826289037478287,-1.0596184221249951,-0.4283147025381757,0.6826289037478287,-1.2639482961567976,-0.8126119837752559,0.6784329353268667,-1.061315936038188,-1.1102460001099754,0.6784329353268667,-0.7443679006402497,-0.8113122575564685,0.6826289037478287,-1.0596184221249951,-0.8126119837752559,0.6784329353268667,-1.061315936038188,-1.1084702253682064,0.6826289037478287,-0.7431773269147786,-0.8113122575564685,0.6826289037478287,-1.0596184221249951,-1.1102460001099754,0.6784329353268667,-0.7443679006402497,-1.2904122258874484,0.6784329353268667,-0.34866321793956573,-1.1084702253682064,0.6826289037478287,-0.7431773269147786,-1.1102460001099754,0.6784329353268667,-0.7443679006402497,-1.2883482856102721,0.6826289037478287,-0.34810555113802844,-1.1084702253682064,0.6826289037478287,-0.7431773269147786,-1.2904122258874484,0.6784329353268667,-0.34866321793956573,-1.3340484636550194,0.6784329353268667,8.393121519109815e-2,-1.2883482856102721,0.6826289037478287,-0.34810555113802844,-1.2904122258874484,0.6784329353268667,-0.34866321793956573,-1.3319147297205405,0.6826289037478287,8.379697203060259e-2,-1.2883482856102721,0.6826289037478287,-0.34810555113802844,-1.3340484636550194,0.6784329353268667,8.393121519109815e-2,1.3902802511054366,0.41594602267436875,-0.3756470811913471,1.435982354502078,0.42046828134539266,9.034435201561032e-2,1.4372935995407192,0.41594602267436875,9.042684856090442e-2,1.3890118964128633,0.42046828134539266,-0.3753043778135198,1.435982354502078,0.42046828134539266,9.034435201561032e-2,1.3902802511054366,0.41594602267436875,-0.3756470811913471,1.1961705390385335,0.41594602267436875,-0.8019762763060001,1.3890118964128633,0.42046828134539266,-0.3753043778135198,1.3902802511054366,0.41594602267436875,-0.3756470811913471,1.1950792709182385,0.42046828134539266,-0.8012446321841935,1.3890118964128633,0.42046828134539266,-0.3753043778135198,1.1961705390385335,0.41594602267436875,-0.8019762763060001,0.8755019289106531,0.41594602267436875,-1.143453662679472,1.1950792709182385,0.42046828134539266,-0.8012446321841935,1.1961705390385335,0.41594602267436875,-0.8019762763060001,0.8747032072291736,0.42046828134539266,-1.1424104882420545,1.1950792709182385,0.42046828134539266,-0.8012446321841935,0.8755019289106531,0.41594602267436875,-1.143453662679472,0.4622022467432836,0.41594602267436875,-1.36394977522152,0.8747032072291736,0.42046828134539266,-1.1424104882420545,0.8755019289106531,0.41594602267436875,-1.143453662679472,0.4617805789621952,0.42046828134539266,-1.3627054418603428,0.8747032072291736,0.42046828134539266,-1.1424104882420545,0.4622022467432836,0.41594602267436875,-1.36394977522152,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,0.4617805789621952,0.42046828134539266,-1.3627054418603428,0.4622022467432836,0.41594602267436875,-1.36394977522152,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,0.4617805789621952,0.42046828134539266,-1.3627054418603428,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,-0.46220224674328353,0.41594602267436875,-1.36394977522152,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,8.818285903447244e-17,0.41594602267436875,-1.440135377741056,-0.46178057896219504,0.42046828134539266,-1.3627054418603428,8.810240968408284e-17,0.42046828134539266,-1.4388215401440347,-0.46220224674328353,0.41594602267436875,-1.36394977522152,-0.8755019289106527,0.41594602267436875,-1.1434536626794722,-0.46178057896219504,0.42046828134539266,-1.3627054418603428,-0.46220224674328353,0.41594602267436875,-1.36394977522152,-0.8747032072291732,0.42046828134539266,-1.142410488242055,-0.46178057896219504,0.42046828134539266,-1.3627054418603428,-0.8755019289106527,0.41594602267436875,-1.1434536626794722,-1.1961705390385335,0.41594602267436875,-0.8019762763060003,-0.8747032072291732,0.42046828134539266,-1.142410488242055,-0.8755019289106527,0.41594602267436875,-1.1434536626794722,-1.1950792709182385,0.42046828134539266,-0.8012446321841937,-0.8747032072291732,0.42046828134539266,-1.142410488242055,-1.1961705390385335,0.41594602267436875,-0.8019762763060003,-1.3902802511054366,0.41594602267436875,-0.37564708119134693,-1.1950792709182385,0.42046828134539266,-0.8012446321841937,-1.1961705390385335,0.41594602267436875,-0.8019762763060003,-1.3890118964128633,0.42046828134539266,-0.3753043778135196,-1.1950792709182385,0.42046828134539266,-0.8012446321841937,-1.3902802511054366,0.41594602267436875,-0.37564708119134693,-1.4372935995407192,0.41594602267436875,9.04268485609048e-2,-1.3890118964128633,0.42046828134539266,-0.3753043778135196,-1.3902802511054366,0.41594602267436875,-0.37564708119134693,-1.435982354502078,0.42046828134539266,9.03443520156107e-2,-1.3890118964128633,0.42046828134539266,-0.3753043778135196,-1.4372935995407192,0.41594602267436875,9.04268485609048e-2,1.440896905314052,0.13872401480181032,-0.3893234593158438,1.4891795884769703,0.14341236045493638,9.36912382899552e-2,1.4896218931105683,0.13872401480181032,9.371906573880276e-2,1.4404690683033952,0.14341236045493638,-0.3892078597997382,1.4891795884769703,0.14341236045493638,9.36912382899552e-2,1.440896905314052,0.13872401480181032,-0.3893234593158438,1.239720140279654,0.13872401480181032,-0.8311742425642544,1.4404690683033952,0.14341236045493638,-0.3892078597997382,1.440896905314052,0.13872401480181032,-0.3893234593158438,1.2393520374980378,0.14341236045493638,-0.8309274469038833,1.4404690683033952,0.14341236045493638,-0.3892078597997382,1.239720140279654,0.13872401480181032,-0.8311742425642544,0.9073767817393622,0.13872401480181032,-1.185083973259948,1.2393520374980378,0.14341236045493638,-0.8309274469038833,1.239720140279654,0.13872401480181032,-0.8311742425642544,0.9071073597089538,0.14341236045493638,-1.1847320932728187,1.2393520374980378,0.14341236045493638,-0.8309274469038833,0.9073767817393622,0.13872401480181032,-1.185083973259948,0.47902988367421756,0.13872401480181032,-1.413607802137613,0.9071073597089538,0.14341236045493638,-1.1847320932728187,0.9073767817393622,0.13872401480181032,-1.185083973259948,0.47888764815917784,0.14341236045493638,-1.4131880679192403,0.9071073597089538,0.14341236045493638,-1.1847320932728187,0.47902988367421756,0.13872401480181032,-1.413607802137613,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,0.47888764815917784,0.14341236045493638,-1.4131880679192403,0.47902988367421756,0.13872401480181032,-1.413607802137613,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,0.47888764815917784,0.14341236045493638,-1.4131880679192403,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,-0.4790298836742174,0.13872401480181032,-1.413607802137613,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,9.139337812177583e-17,0.13872401480181032,-1.4925671334038102,-0.47888764815917767,0.14341236045493638,-1.4131880679192403,9.136624122562092e-17,0.14341236045493638,-1.4921239542574014,-0.4790298836742174,0.13872401480181032,-1.413607802137613,-0.9073767817393618,0.13872401480181032,-1.1850839732599485,-0.47888764815917767,0.14341236045493638,-1.4131880679192403,-0.4790298836742174,0.13872401480181032,-1.413607802137613,-0.9071073597089533,0.14341236045493638,-1.184732093272819,-0.47888764815917767,0.14341236045493638,-1.4131880679192403,-0.9073767817393618,0.13872401480181032,-1.1850839732599485,-1.239720140279654,0.13872401480181032,-0.8311742425642547,-0.9071073597089533,0.14341236045493638,-1.184732093272819,-0.9073767817393618,0.13872401480181032,-1.1850839732599485,-1.2393520374980378,0.14341236045493638,-0.8309274469038835,-0.9071073597089533,0.14341236045493638,-1.184732093272819,-1.239720140279654,0.13872401480181032,-0.8311742425642547,-1.440896905314052,0.13872401480181032,-0.3893234593158436,-1.2393520374980378,0.14341236045493638,-0.8309274469038835,-1.239720140279654,0.13872401480181032,-0.8311742425642547,-1.4404690683033952,0.14341236045493638,-0.38920785979973804,-1.2393520374980378,0.14341236045493638,-0.8309274469038835,-1.440896905314052,0.13872401480181032,-0.3893234593158436,-1.4896218931105683,0.13872401480181032,9.371906573880316e-2,-1.4404690683033952,0.14341236045493638,-0.38920785979973804,-1.440896905314052,0.13872401480181032,-0.3893234593158436,-1.4891795884769703,0.14341236045493638,9.36912382899556e-2,-1.4404690683033952,0.14341236045493638,-0.38920785979973804,-1.4896218931105683,0.13872401480181032,9.371906573880316e-2,1.4404690683033952,-0.14341236045493638,-0.3892078597997382,1.4896218931105683,-0.13872401480181032,9.371906573880276e-2,1.4891795884769703,-0.14341236045493638,9.36912382899552e-2,1.440896905314052,-0.13872401480181032,-0.3893234593158438,1.4896218931105683,-0.13872401480181032,9.371906573880276e-2,1.4404690683033952,-0.14341236045493638,-0.3892078597997382,1.2393520374980378,-0.14341236045493638,-0.8309274469038833,1.440896905314052,-0.13872401480181032,-0.3893234593158438,1.4404690683033952,-0.14341236045493638,-0.3892078597997382,1.239720140279654,-0.13872401480181032,-0.8311742425642544,1.440896905314052,-0.13872401480181032,-0.3893234593158438,1.2393520374980378,-0.14341236045493638,-0.8309274469038833,0.9071073597089538,-0.14341236045493638,-1.1847320932728187,1.239720140279654,-0.13872401480181032,-0.8311742425642544,1.2393520374980378,-0.14341236045493638,-0.8309274469038833,0.9073767817393622,-0.13872401480181032,-1.185083973259948,1.239720140279654,-0.13872401480181032,-0.8311742425642544,0.9071073597089538,-0.14341236045493638,-1.1847320932728187,0.47888764815917784,-0.14341236045493638,-1.4131880679192403,0.9073767817393622,-0.13872401480181032,-1.185083973259948,0.9071073597089538,-0.14341236045493638,-1.1847320932728187,0.47902988367421756,-0.13872401480181032,-1.413607802137613,0.9073767817393622,-0.13872401480181032,-1.185083973259948,0.47888764815917784,-0.14341236045493638,-1.4131880679192403,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,0.47902988367421756,-0.13872401480181032,-1.413607802137613,0.47888764815917784,-0.14341236045493638,-1.4131880679192403,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,0.47902988367421756,-0.13872401480181032,-1.413607802137613,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,-0.47888764815917767,-0.14341236045493638,-1.4131880679192403,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,9.136624122562092e-17,-0.14341236045493638,-1.4921239542574014,-0.4790298836742174,-0.13872401480181032,-1.413607802137613,9.139337812177583e-17,-0.13872401480181032,-1.4925671334038102,-0.47888764815917767,-0.14341236045493638,-1.4131880679192403,-0.9071073597089533,-0.14341236045493638,-1.184732093272819,-0.4790298836742174,-0.13872401480181032,-1.413607802137613,-0.47888764815917767,-0.14341236045493638,-1.4131880679192403,-0.9073767817393618,-0.13872401480181032,-1.1850839732599485,-0.4790298836742174,-0.13872401480181032,-1.413607802137613,-0.9071073597089533,-0.14341236045493638,-1.184732093272819,-1.2393520374980378,-0.14341236045493638,-0.8309274469038835,-0.9073767817393618,-0.13872401480181032,-1.1850839732599485,-0.9071073597089533,-0.14341236045493638,-1.184732093272819,-1.239720140279654,-0.13872401480181032,-0.8311742425642547,-0.9073767817393618,-0.13872401480181032,-1.1850839732599485,-1.2393520374980378,-0.14341236045493638,-0.8309274469038835,-1.4404690683033952,-0.14341236045493638,-0.38920785979973804,-1.239720140279654,-0.13872401480181032,-0.8311742425642547,-1.2393520374980378,-0.14341236045493638,-0.8309274469038835,-1.440896905314052,-0.13872401480181032,-0.3893234593158436,-1.239720140279654,-0.13872401480181032,-0.8311742425642547,-1.4404690683033952,-0.14341236045493638,-0.38920785979973804,-1.4891795884769703,-0.14341236045493638,9.36912382899556e-2,-1.440896905314052,-0.13872401480181032,-0.3893234593158436,-1.4404690683033952,-0.14341236045493638,-0.38920785979973804,-1.4896218931105683,-0.13872401480181032,9.371906573880316e-2,-1.440896905314052,-0.13872401480181032,-0.3893234593158436,-1.4891795884769703,-0.14341236045493638,9.36912382899556e-2,1.3890118964128633,-0.42046828134539266,-0.3753043778135198,1.4372935995407192,-0.41594602267436875,9.042684856090442e-2,1.435982354502078,-0.42046828134539266,9.034435201561032e-2,1.3902802511054366,-0.41594602267436875,-0.3756470811913471,1.4372935995407192,-0.41594602267436875,9.042684856090442e-2,1.3890118964128633,-0.42046828134539266,-0.3753043778135198,1.1950792709182385,-0.42046828134539266,-0.8012446321841935,1.3902802511054366,-0.41594602267436875,-0.3756470811913471,1.3890118964128633,-0.42046828134539266,-0.3753043778135198,1.1961705390385335,-0.41594602267436875,-0.8019762763060001,1.3902802511054366,-0.41594602267436875,-0.3756470811913471,1.1950792709182385,-0.42046828134539266,-0.8012446321841935,0.8747032072291736,-0.42046828134539266,-1.1424104882420545,1.1961705390385335,-0.41594602267436875,-0.8019762763060001,1.1950792709182385,-0.42046828134539266,-0.8012446321841935,0.8755019289106531,-0.41594602267436875,-1.143453662679472,1.1961705390385335,-0.41594602267436875,-0.8019762763060001,0.8747032072291736,-0.42046828134539266,-1.1424104882420545,0.4617805789621952,-0.42046828134539266,-1.3627054418603428,0.8755019289106531,-0.41594602267436875,-1.143453662679472,0.8747032072291736,-0.42046828134539266,-1.1424104882420545,0.4622022467432836,-0.41594602267436875,-1.36394977522152,0.8755019289106531,-0.41594602267436875,-1.143453662679472,0.4617805789621952,-0.42046828134539266,-1.3627054418603428,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,0.4622022467432836,-0.41594602267436875,-1.36394977522152,0.4617805789621952,-0.42046828134539266,-1.3627054418603428,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,0.4622022467432836,-0.41594602267436875,-1.36394977522152,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,-0.46178057896219504,-0.42046828134539266,-1.3627054418603428,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,8.810240968408284e-17,-0.42046828134539266,-1.4388215401440347,-0.46220224674328353,-0.41594602267436875,-1.36394977522152,8.818285903447244e-17,-0.41594602267436875,-1.440135377741056,-0.46178057896219504,-0.42046828134539266,-1.3627054418603428,-0.8747032072291732,-0.42046828134539266,-1.142410488242055,-0.46220224674328353,-0.41594602267436875,-1.36394977522152,-0.46178057896219504,-0.42046828134539266,-1.3627054418603428,-0.8755019289106527,-0.41594602267436875,-1.1434536626794722,-0.46220224674328353,-0.41594602267436875,-1.36394977522152,-0.8747032072291732,-0.42046828134539266,-1.142410488242055,-1.1950792709182385,-0.42046828134539266,-0.8012446321841937,-0.8755019289106527,-0.41594602267436875,-1.1434536626794722,-0.8747032072291732,-0.42046828134539266,-1.142410488242055,-1.1961705390385335,-0.41594602267436875,-0.8019762763060003,-0.8755019289106527,-0.41594602267436875,-1.1434536626794722,-1.1950792709182385,-0.42046828134539266,-0.8012446321841937,-1.3890118964128633,-0.42046828134539266,-0.3753043778135196,-1.1961705390385335,-0.41594602267436875,-0.8019762763060003,-1.1950792709182385,-0.42046828134539266,-0.8012446321841937,-1.3902802511054366,-0.41594602267436875,-0.37564708119134693,-1.1961705390385335,-0.41594602267436875,-0.8019762763060003,-1.3890118964128633,-0.42046828134539266,-0.3753043778135196,-1.435982354502078,-0.42046828134539266,9.03443520156107e-2,-1.3902802511054366,-0.41594602267436875,-0.37564708119134693,-1.3890118964128633,-0.42046828134539266,-0.3753043778135196,-1.4372935995407192,-0.41594602267436875,9.04268485609048e-2,-1.3902802511054366,-0.41594602267436875,-0.37564708119134693,-1.435982354502078,-0.42046828134539266,9.03443520156107e-2,1.2883482856102721,-0.6826289037478287,-0.34810555113802855,1.3340484636550194,-0.6784329353268667,8.393121519109781e-2,1.3319147297205405,-0.6826289037478287,8.379697203060224e-2,1.2904122258874484,-0.6784329353268667,-0.34866321793956584,1.3340484636550194,-0.6784329353268667,8.393121519109781e-2,1.2883482856102721,-0.6826289037478287,-0.34810555113802855,1.1084702253682064,-0.6826289037478287,-0.7431773269147784,1.2904122258874484,-0.6784329353268667,-0.34866321793956584,1.2883482856102721,-0.6826289037478287,-0.34810555113802855,1.1102460001099754,-0.6784329353268667,-0.7443679006402496,1.2904122258874484,-0.6784329353268667,-0.34866321793956584,1.1084702253682064,-0.6826289037478287,-0.7431773269147784,0.8113122575564691,-0.6826289037478287,-1.0596184221249947,1.1102460001099754,-0.6784329353268667,-0.7443679006402496,1.1084702253682064,-0.6826289037478287,-0.7431773269147784,0.8126119837752565,-0.6784329353268667,-1.0613159360381876,1.1102460001099754,-0.6784329353268667,-0.7443679006402496,0.8113122575564691,-0.6826289037478287,-1.0596184221249947,0.42831470253817583,-0.6826289037478287,-1.2639482961567976,0.8126119837752565,-0.6784329353268667,-1.0613159360381876,0.8113122575564691,-0.6826289037478287,-1.0596184221249947,0.42900086479394856,-0.6784329353268667,-1.2659731475311065,0.8126119837752565,-0.6784329353268667,-1.0613159360381876,0.42831470253817583,-0.6826289037478287,-1.2639482961567976,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,0.42900086479394856,-0.6784329353268667,-1.2659731475311065,0.42831470253817583,-0.6826289037478287,-1.2639482961567976,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,0.42900086479394856,-0.6784329353268667,-1.2659731475311065,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,-0.4283147025381757,-0.6826289037478287,-1.2639482961567976,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,8.171750635667977e-17,-0.6826289037478287,-1.3345481556572014,-0.4290008647939484,-0.6784329353268667,-1.2659731475311065,8.184841820295902e-17,-0.6784329353268667,-1.3366861083529566,-0.4283147025381757,-0.6826289037478287,-1.2639482961567976,-0.8113122575564685,-0.6826289037478287,-1.0596184221249951,-0.4290008647939484,-0.6784329353268667,-1.2659731475311065,-0.4283147025381757,-0.6826289037478287,-1.2639482961567976,-0.8126119837752559,-0.6784329353268667,-1.061315936038188,-0.4290008647939484,-0.6784329353268667,-1.2659731475311065,-0.8113122575564685,-0.6826289037478287,-1.0596184221249951,-1.1084702253682064,-0.6826289037478287,-0.7431773269147786,-0.8126119837752559,-0.6784329353268667,-1.061315936038188,-0.8113122575564685,-0.6826289037478287,-1.0596184221249951,-1.1102460001099754,-0.6784329353268667,-0.7443679006402497,-0.8126119837752559,-0.6784329353268667,-1.061315936038188,-1.1084702253682064,-0.6826289037478287,-0.7431773269147786,-1.2883482856102721,-0.6826289037478287,-0.34810555113802844,-1.1102460001099754,-0.6784329353268667,-0.7443679006402497,-1.1084702253682064,-0.6826289037478287,-0.7431773269147786,-1.2904122258874484,-0.6784329353268667,-0.34866321793956573,-1.1102460001099754,-0.6784329353268667,-0.7443679006402497,-1.2883482856102721,-0.6826289037478287,-0.34810555113802844,-1.3319147297205405,-0.6826289037478287,8.379697203060259e-2,-1.2904122258874484,-0.6784329353268667,-0.34866321793956573,-1.2883482856102721,-0.6826289037478287,-0.34810555113802844,-1.3340484636550194,-0.6784329353268667,8.393121519109815e-2,-1.2904122258874484,-0.6784329353268667,-0.34866321793956573,-1.3319147297205405,-0.6826289037478287,8.379697203060259e-2,1.3328498631692396,0.6805317591095806,8.595793792873038e-2,1.436779954507098,0.41820766795280473,8.812893277498343e-2,1.3331133302063203,0.6805317591095806,8.177024929296971e-2,1.4364959995356987,0.41820766795280473,9.264226780153134e-2,1.436779954507098,0.41820766795280473,8.812893277498343e-2,1.3328498631692396,0.6805317591095806,8.595793792873038e-2,1.4364959995356987,0.41820766795280473,9.264226780153134e-2,1.4895479326234156,0.14106836166445297,9.13656048774141e-2,1.436779954507098,0.41820766795280473,8.812893277498343e-2,1.4892535489641228,0.14106836166445297,9.60446991513439e-2,1.4895479326234156,0.14106836166445297,9.13656048774141e-2,1.4364959995356987,0.41820766795280473,9.264226780153134e-2,1.4892535489641228,0.14106836166445297,9.60446991513439e-2,1.4895479326234156,-0.14106836166445297,9.13656048774141e-2,1.4895479326234156,0.14106836166445297,9.13656048774141e-2,1.4892535489641228,-0.14106836166445297,9.60446991513439e-2,1.4895479326234156,-0.14106836166445297,9.13656048774141e-2,1.4892535489641228,0.14106836166445297,9.60446991513439e-2,1.4892535489641228,-0.14106836166445297,9.60446991513439e-2,1.436779954507098,-0.41820766795280473,8.812893277498343e-2,1.4895479326234156,-0.14106836166445297,9.13656048774141e-2,1.4364959995356987,-0.41820766795280473,9.264226780153134e-2,1.436779954507098,-0.41820766795280473,8.812893277498343e-2,1.4892535489641228,-0.14106836166445297,9.60446991513439e-2,1.4364959995356987,-0.41820766795280473,9.264226780153134e-2,1.3331133302063203,-0.6805317591095806,8.177024929296971e-2,1.436779954507098,-0.41820766795280473,8.812893277498343e-2,1.3328498631692396,-0.6805317591095806,8.595793792873038e-2,1.3331133302063203,-0.6805317591095806,8.177024929296971e-2,1.4364959995356987,-0.41820766795280473,9.264226780153134e-2,1.2899274971104944,0.6805317591095806,-0.3463590291034367,1.3890562773773587,0.41820766795280473,-0.3776585822459639,1.288833014387226,0.6805317591095806,-0.35040973997415775,1.390235870140941,0.41820766795280473,-0.3732928767589028,1.3890562773773587,0.41820766795280473,-0.3776585822459639,1.2899274971104944,0.6805317591095806,-0.3463590291034367,1.390235870140941,0.41820766795280473,-0.3732928767589028,1.440071529237639,0.14106836166445297,-0.39152868096280685,1.3890562773773587,0.41820766795280473,-0.3776585822459639,1.441294444379808,0.14106836166445297,-0.38700263815277514,1.440071529237639,0.14106836166445297,-0.39152868096280685,1.390235870140941,0.41820766795280473,-0.3732928767589028,1.441294444379808,0.14106836166445297,-0.38700263815277514,1.440071529237639,-0.14106836166445297,-0.39152868096280685,1.440071529237639,0.14106836166445297,-0.39152868096280685,1.441294444379808,-0.14106836166445297,-0.38700263815277514,1.440071529237639,-0.14106836166445297,-0.39152868096280685,1.441294444379808,0.14106836166445297,-0.38700263815277514,1.441294444379808,-0.14106836166445297,-0.38700263815277514,1.3890562773773587,-0.41820766795280473,-0.3776585822459639,1.440071529237639,-0.14106836166445297,-0.39152868096280685,1.390235870140941,-0.41820766795280473,-0.3732928767589028,1.3890562773773587,-0.41820766795280473,-0.3776585822459639,1.441294444379808,-0.14106836166445297,-0.38700263815277514,1.390235870140941,-0.41820766795280473,-0.3732928767589028,1.288833014387226,-0.6805317591095806,-0.35040973997415775,1.3890562773773587,-0.41820766795280473,-0.3776585822459639,1.2899274971104944,-0.6805317591095806,-0.3463590291034367,1.288833014387226,-0.6805317591095806,-0.35040973997415775,1.390235870140941,-0.41820766795280473,-0.3732928767589028,1.110526428989685,0.6805317591095806,-0.7420300366957111,1.1943657371857133,0.41820766795280473,-0.8034885389987244,1.1081897964884968,0.6805317591095806,-0.7455151908593173,1.1968840727710588,0.41820766795280473,-0.7997323694914693,1.1943657371857133,0.41820766795280473,-0.8034885389987244,1.110526428989685,0.6805317591095806,-0.7420300366957111,1.1968840727710588,0.41820766795280473,-0.7997323694914693,1.238230676200899,0.14106836166445297,-0.8329979050708184,1.1943657371857133,0.41820766795280473,-0.8034885389987244,1.240841501576793,0.14106836166445297,-0.8291037843973196,1.238230676200899,0.14106836166445297,-0.8329979050708184,1.1968840727710588,0.41820766795280473,-0.7997323694914693,1.240841501576793,0.14106836166445297,-0.8291037843973196,1.238230676200899,-0.14106836166445297,-0.8329979050708184,1.238230676200899,0.14106836166445297,-0.8329979050708184,1.240841501576793,-0.14106836166445297,-0.8291037843973196,1.238230676200899,-0.14106836166445297,-0.8329979050708184,1.240841501576793,0.14106836166445297,-0.8291037843973196,1.240841501576793,-0.14106836166445297,-0.8291037843973196,1.1943657371857133,-0.41820766795280473,-0.8034885389987244,1.238230676200899,-0.14106836166445297,-0.8329979050708184,1.1968840727710588,-0.41820766795280473,-0.7997323694914693,1.1943657371857133,-0.41820766795280473,-0.8034885389987244,1.240841501576793,-0.14106836166445297,-0.8291037843973196,1.1968840727710588,-0.41820766795280473,-0.7997323694914693,1.1081897964884968,-0.6805317591095806,-0.7455151908593173,1.1943657371857133,-0.41820766795280473,-0.8034885389987244,1.110526428989685,-0.6805317591095806,-0.7420300366957111,1.1081897964884968,-0.6805317591095806,-0.7455151908593173,1.1968840727710588,-0.41820766795280473,-0.7997323694914693,0.8136278999854994,0.6805317591095806,-1.059191750915955,0.8733072530874157,0.41820766795280473,-1.144306684490827,0.8102963413462261,0.6805317591095806,-1.0617426072472276,0.8768978830524109,0.41820766795280473,-1.1415574664306993,0.8733072530874157,0.41820766795280473,-1.144306684490827,0.8136278999854994,0.6805317591095806,-1.059191750915955,0.8768978830524109,0.41820766795280473,-1.1415574664306993,0.9053808200070956,0.14106836166445297,-1.1863331269506834,0.8733072530874157,0.41820766795280473,-1.144306684490827,0.9091033214412202,0.14106836166445297,-1.1834829395820832,0.9053808200070956,0.14106836166445297,-1.1863331269506834,0.8768978830524109,0.41820766795280473,-1.1415574664306993,0.9091033214412202,0.14106836166445297,-1.1834829395820832,0.9053808200070956,-0.14106836166445297,-1.1863331269506834,0.9053808200070956,0.14106836166445297,-1.1863331269506834,0.9091033214412202,-0.14106836166445297,-1.1834829395820832,0.9053808200070956,-0.14106836166445297,-1.1863331269506834,0.9091033214412202,0.14106836166445297,-1.1834829395820832,0.9091033214412202,-0.14106836166445297,-1.1834829395820832,0.8733072530874157,-0.41820766795280473,-1.144306684490827,0.9053808200070956,-0.14106836166445297,-1.1863331269506834,0.8768978830524109,-0.41820766795280473,-1.1415574664306993,0.8733072530874157,-0.41820766795280473,-1.144306684490827,0.9091033214412202,-0.14106836166445297,-1.1834829395820832,0.8768978830524109,-0.41820766795280473,-1.1415574664306993,0.8102963413462261,-0.6805317591095806,-1.0617426072472276,0.8733072530874157,-0.41820766795280473,-1.144306684490827,0.8136278999854994,-0.6805317591095806,-1.059191750915955,0.8102963413462261,-0.6805317591095806,-1.0617426072472276,0.8768978830524109,-0.41820766795280473,-1.1415574664306993,0.43064478095571473,0.6805317591095806,-1.2642873872181217,0.45984990109170226,0.41820766795280473,-1.3640533035521114,0.42667078637641004,0.6805317591095806,-1.2656340564697823,0.46413292461377686,0.41820766795280473,-1.3626019135297511,0.45984990109170226,0.41820766795280473,-1.3640533035521114,0.43064478095571473,0.6805317591095806,-1.2642873872181217,0.46413292461377686,0.41820766795280473,-1.3626019135297511,0.47673860380604527,0.14106836166445297,-1.4141502823173957,0.45984990109170226,0.41820766795280473,-1.3640533035521114,0.48117892802735046,0.14106836166445297,-1.4126455877394575,0.47673860380604527,0.14106836166445297,-1.4141502823173957,0.46413292461377686,0.41820766795280473,-1.3626019135297511,0.48117892802735046,0.14106836166445297,-1.4126455877394575,0.47673860380604527,-0.14106836166445297,-1.4141502823173957,0.47673860380604527,0.14106836166445297,-1.4141502823173957,0.48117892802735046,-0.14106836166445297,-1.4126455877394575,0.47673860380604527,-0.14106836166445297,-1.4141502823173957,0.48117892802735046,0.14106836166445297,-1.4126455877394575,0.48117892802735046,-0.14106836166445297,-1.4126455877394575,0.45984990109170226,-0.41820766795280473,-1.3640533035521114,0.47673860380604527,-0.14106836166445297,-1.4141502823173957,0.46413292461377686,-0.41820766795280473,-1.3626019135297511,0.45984990109170226,-0.41820766795280473,-1.3640533035521114,0.48117892802735046,-0.14106836166445297,-1.4126455877394575,0.46413292461377686,-0.41820766795280473,-1.3626019135297511,0.42667078637641004,-0.6805317591095806,-1.2656340564697823,0.45984990109170226,-0.41820766795280473,-1.3640533035521114,0.43064478095571473,-0.6805317591095806,-1.2642873872181217,0.42667078637641004,-0.6805317591095806,-1.2656340564697823,0.46413292461377686,-0.41820766795280473,-1.3626019135297511,2.0979842104809974e-3,0.6805317591095806,-1.3356171320050791,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,-2.097984210480834e-3,0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,2.0979842104809974e-3,0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,-2.2611293355117427e-3,0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,0.14106836166445297,-1.492345543830606,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,2.344172826562981e-3,0.14106836166445297,-1.492345543830606,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,-2.3441728265627985e-3,-0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,2.344172826562981e-3,-0.14106836166445297,-1.492345543830606,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-2.097984210480834e-3,-0.6805317591095806,-1.3356171320050791,-2.2611293355117427e-3,-0.41820766795280473,-1.4394784589425453,2.0979842104809974e-3,-0.6805317591095806,-1.3356171320050791,-2.097984210480834e-3,-0.6805317591095806,-1.3356171320050791,2.2611293355119183e-3,-0.41820766795280473,-1.4394784589425453,-0.42667078637640954,0.6805317591095806,-1.2656340564697823,-0.4641329246137767,0.41820766795280473,-1.3626019135297511,-0.43064478095571457,0.6805317591095806,-1.2642873872181217,-0.45984990109170176,0.41820766795280473,-1.3640533035521116,-0.4641329246137767,0.41820766795280473,-1.3626019135297511,-0.42667078637640954,0.6805317591095806,-1.2656340564697823,-0.45984990109170176,0.41820766795280473,-1.3640533035521116,-0.4811789280273503,0.14106836166445297,-1.4126455877394575,-0.4641329246137767,0.41820766795280473,-1.3626019135297511,-0.47673860380604477,0.14106836166445297,-1.414150282317396,-0.4811789280273503,0.14106836166445297,-1.4126455877394575,-0.45984990109170176,0.41820766795280473,-1.3640533035521116,-0.47673860380604477,0.14106836166445297,-1.414150282317396,-0.4811789280273503,-0.14106836166445297,-1.4126455877394575,-0.4811789280273503,0.14106836166445297,-1.4126455877394575,-0.47673860380604477,-0.14106836166445297,-1.414150282317396,-0.4811789280273503,-0.14106836166445297,-1.4126455877394575,-0.47673860380604477,0.14106836166445297,-1.414150282317396,-0.47673860380604477,-0.14106836166445297,-1.414150282317396,-0.4641329246137767,-0.41820766795280473,-1.3626019135297511,-0.4811789280273503,-0.14106836166445297,-1.4126455877394575,-0.45984990109170176,-0.41820766795280473,-1.3640533035521116,-0.4641329246137767,-0.41820766795280473,-1.3626019135297511,-0.47673860380604477,-0.14106836166445297,-1.414150282317396,-0.45984990109170176,-0.41820766795280473,-1.3640533035521116,-0.43064478095571457,-0.6805317591095806,-1.2642873872181217,-0.4641329246137767,-0.41820766795280473,-1.3626019135297511,-0.42667078637640954,-0.6805317591095806,-1.2656340564697823,-0.43064478095571457,-0.6805317591095806,-1.2642873872181217,-0.45984990109170176,-0.41820766795280473,-1.3640533035521116,-0.8102963413462259,0.6805317591095806,-1.0617426072472278,-0.8768978830524103,0.41820766795280473,-1.1415574664306998,-0.8136278999854991,0.6805317591095806,-1.0591917509159552,-0.8733072530874156,0.41820766795280473,-1.1443066844908272,-0.8768978830524103,0.41820766795280473,-1.1415574664306998,-0.8102963413462259,0.6805317591095806,-1.0617426072472278,-0.8733072530874156,0.41820766795280473,-1.1443066844908272,-0.9091033214412196,0.14106836166445297,-1.1834829395820834,-0.8768978830524103,0.41820766795280473,-1.1415574664306998,-0.9053808200070955,0.14106836166445297,-1.1863331269506836,-0.9091033214412196,0.14106836166445297,-1.1834829395820834,-0.8733072530874156,0.41820766795280473,-1.1443066844908272,-0.9053808200070955,0.14106836166445297,-1.1863331269506836,-0.9091033214412196,-0.14106836166445297,-1.1834829395820834,-0.9091033214412196,0.14106836166445297,-1.1834829395820834,-0.9053808200070955,-0.14106836166445297,-1.1863331269506836,-0.9091033214412196,-0.14106836166445297,-1.1834829395820834,-0.9053808200070955,0.14106836166445297,-1.1863331269506836,-0.9053808200070955,-0.14106836166445297,-1.1863331269506836,-0.8768978830524103,-0.41820766795280473,-1.1415574664306998,-0.9091033214412196,-0.14106836166445297,-1.1834829395820834,-0.8733072530874156,-0.41820766795280473,-1.1443066844908272,-0.8768978830524103,-0.41820766795280473,-1.1415574664306998,-0.9053808200070955,-0.14106836166445297,-1.1863331269506836,-0.8733072530874156,-0.41820766795280473,-1.1443066844908272,-0.8136278999854991,-0.6805317591095806,-1.0591917509159552,-0.8768978830524103,-0.41820766795280473,-1.1415574664306998,-0.8102963413462259,-0.6805317591095806,-1.0617426072472278,-0.8136278999854991,-0.6805317591095806,-1.0591917509159552,-0.8733072530874156,-0.41820766795280473,-1.1443066844908272,-1.1081897964884964,0.6805317591095806,-0.7455151908593177,-1.196884072771059,0.41820766795280473,-0.7997323694914692,-1.1105264289896852,0.6805317591095806,-0.7420300366957108,-1.1943657371857128,0.41820766795280473,-0.803488538998725,-1.196884072771059,0.41820766795280473,-0.7997323694914692,-1.1081897964884964,0.6805317591095806,-0.7455151908593177,-1.1943657371857128,0.41820766795280473,-0.803488538998725,-1.2408415015767933,0.14106836166445297,-0.8291037843973192,-1.196884072771059,0.41820766795280473,-0.7997323694914692,-1.2382306762008986,0.14106836166445297,-0.832997905070819,-1.2408415015767933,0.14106836166445297,-0.8291037843973192,-1.1943657371857128,0.41820766795280473,-0.803488538998725,-1.2382306762008986,0.14106836166445297,-0.832997905070819,-1.2408415015767933,-0.14106836166445297,-0.8291037843973192,-1.2408415015767933,0.14106836166445297,-0.8291037843973192,-1.2382306762008986,-0.14106836166445297,-0.832997905070819,-1.2408415015767933,-0.14106836166445297,-0.8291037843973192,-1.2382306762008986,0.14106836166445297,-0.832997905070819,-1.2382306762008986,-0.14106836166445297,-0.832997905070819,-1.196884072771059,-0.41820766795280473,-0.7997323694914692,-1.2408415015767933,-0.14106836166445297,-0.8291037843973192,-1.1943657371857128,-0.41820766795280473,-0.803488538998725,-1.196884072771059,-0.41820766795280473,-0.7997323694914692,-1.2382306762008986,-0.14106836166445297,-0.832997905070819,-1.1943657371857128,-0.41820766795280473,-0.803488538998725,-1.1105264289896852,-0.6805317591095806,-0.7420300366957108,-1.196884072771059,-0.41820766795280473,-0.7997323694914692,-1.1081897964884964,-0.6805317591095806,-0.7455151908593177,-1.1105264289896852,-0.6805317591095806,-0.7420300366957108,-1.1943657371857128,-0.41820766795280473,-0.803488538998725,-1.288833014387226,0.6805317591095806,-0.3504097399741575,-1.3902358701409407,0.41820766795280473,-0.3732928767589035,-1.2899274971104944,0.6805317591095806,-0.3463590291034373,-1.3890562773773587,0.41820766795280473,-0.37765858224596366,-1.3902358701409407,0.41820766795280473,-0.3732928767589035,-1.288833014387226,0.6805317591095806,-0.3504097399741575,-1.3890562773773587,0.41820766795280473,-0.37765858224596366,-1.441294444379808,0.14106836166445297,-0.3870026381527758,-1.3902358701409407,0.41820766795280473,-0.3732928767589035,-1.440071529237639,0.14106836166445297,-0.3915286809628065,-1.441294444379808,0.14106836166445297,-0.3870026381527758,-1.3890562773773587,0.41820766795280473,-0.37765858224596366,-1.440071529237639,0.14106836166445297,-0.3915286809628065,-1.441294444379808,-0.14106836166445297,-0.3870026381527758,-1.441294444379808,0.14106836166445297,-0.3870026381527758,-1.440071529237639,-0.14106836166445297,-0.3915286809628065,-1.441294444379808,-0.14106836166445297,-0.3870026381527758,-1.440071529237639,0.14106836166445297,-0.3915286809628065,-1.440071529237639,-0.14106836166445297,-0.3915286809628065,-1.3902358701409407,-0.41820766795280473,-0.3732928767589035,-1.441294444379808,-0.14106836166445297,-0.3870026381527758,-1.3890562773773587,-0.41820766795280473,-0.37765858224596366,-1.3902358701409407,-0.41820766795280473,-0.3732928767589035,-1.440071529237639,-0.14106836166445297,-0.3915286809628065,-1.3890562773773587,-0.41820766795280473,-0.37765858224596366,-1.2899274971104944,-0.6805317591095806,-0.3463590291034373,-1.3902358701409407,-0.41820766795280473,-0.3732928767589035,-1.288833014387226,-0.6805317591095806,-0.3504097399741575,-1.2899274971104944,-0.6805317591095806,-0.3463590291034373,-1.3890562773773587,-0.41820766795280473,-0.37765858224596366,-1.3331133302063203,0.6805317591095806,8.177024929296936e-2,-1.436495999535699,0.41820766795280473,9.264226780153055e-2,-1.3328498631692396,0.6805317591095806,8.595793792872963e-2,-1.4367799545070983,0.41820766795280473,8.812893277498306e-2,-1.436495999535699,0.41820766795280473,9.264226780153055e-2,-1.3331133302063203,0.6805317591095806,8.177024929296936e-2,-1.4367799545070983,0.41820766795280473,8.812893277498306e-2,-1.489253548964123,0.14106836166445297,9.604469915134307e-2,-1.436495999535699,0.41820766795280473,9.264226780153055e-2,-1.4895479326234158,0.14106836166445297,9.136560487741371e-2,-1.489253548964123,0.14106836166445297,9.604469915134307e-2,-1.4367799545070983,0.41820766795280473,8.812893277498306e-2,-1.4895479326234158,0.14106836166445297,9.136560487741371e-2,-1.489253548964123,-0.14106836166445297,9.604469915134307e-2,-1.489253548964123,0.14106836166445297,9.604469915134307e-2,-1.4895479326234158,-0.14106836166445297,9.136560487741371e-2,-1.489253548964123,-0.14106836166445297,9.604469915134307e-2,-1.4895479326234158,0.14106836166445297,9.136560487741371e-2,-1.4895479326234158,-0.14106836166445297,9.136560487741371e-2,-1.436495999535699,-0.41820766795280473,9.264226780153055e-2,-1.489253548964123,-0.14106836166445297,9.604469915134307e-2,-1.4367799545070983,-0.41820766795280473,8.812893277498306e-2,-1.436495999535699,-0.41820766795280473,9.264226780153055e-2,-1.4895479326234158,-0.14106836166445297,9.136560487741371e-2,-1.4367799545070983,-0.41820766795280473,8.812893277498306e-2,-1.3328498631692396,-0.6805317591095806,8.595793792872963e-2,-1.436495999535699,-0.41820766795280473,9.264226780153055e-2,-1.3331133302063203,-0.6805317591095806,8.177024929296936e-2,-1.3328498631692396,-0.6805317591095806,8.595793792872963e-2,-1.4367799545070983,-0.41820766795280473,8.812893277498306e-2],[1.2902420077981585,0.6809857496093201,-0.34861722582512933,1.4375981481697295,0.4184866590588439,9.044600913656117e-2,1.3338724895159437,0.6809857496093201,8.39201438366926e-2,1.3905748380600782,0.4184866590588439,-0.3757266771789763,1.4375981481697295,0.4184866590588439,9.044600913656117e-2,1.2902420077981585,0.6809857496093201,-0.34861722582512933,1.3905748380600782,0.4184866590588439,-0.3757266771789763,1.4903961757206519,0.14116246997777146,9.376777947157647e-2,1.4375981481697295,0.4184866590588439,9.044600913656117e-2,1.4416458614228929,0.14116246997777146,-0.38952582367799504,1.4903961757206519,0.14116246997777146,9.376777947157647e-2,1.3905748380600782,0.4184866590588439,-0.3757266771789763,1.4416458614228929,0.14116246997777146,-0.38952582367799504,1.4903961757206519,-0.14116246997777146,9.376777947157647e-2,1.4903961757206519,0.14116246997777146,9.376777947157647e-2,1.4416458614228929,-0.14116246997777146,-0.38952582367799504,1.4903961757206519,-0.14116246997777146,9.376777947157647e-2,1.4416458614228929,0.14116246997777146,-0.38952582367799504,1.4416458614228929,-0.14116246997777146,-0.38952582367799504,1.4375981481697295,-0.4184866590588439,9.044600913656117e-2,1.4903961757206519,-0.14116246997777146,9.376777947157647e-2,1.3905748380600782,-0.4184866590588439,-0.3757266771789763,1.4375981481697295,-0.4184866590588439,9.044600913656117e-2,1.4416458614228929,-0.14116246997777146,-0.38952582367799504,1.3905748380600782,-0.4184866590588439,-0.3757266771789763,1.3338724895159437,-0.6809857496093201,8.39201438366926e-2,1.4375981481697295,-0.4184866590588439,9.044600913656117e-2,1.2902420077981585,-0.6809857496093201,-0.34861722582512933,1.3338724895159437,-0.6809857496093201,8.39201438366926e-2,1.3905748380600782,-0.4184866590588439,-0.3757266771789763,1.1100995477213593,0.6809857496093201,-0.7442697111785925,1.3905748380600782,0.4184866590588439,-0.3757266771789763,1.2902420077981585,0.6809857496093201,-0.34861722582512933,1.1964239960203522,0.4184866590588439,-0.8021462073315929,1.3905748380600782,0.4184866590588439,-0.3757266771789763,1.1100995477213593,0.6809857496093201,-0.7442697111785925,1.1964239960203522,0.4184866590588439,-0.8021462073315929,1.4416458614228929,0.14116246997777146,-0.38952582367799504,1.3905748380600782,0.4184866590588439,-0.3757266771789763,1.2403645277919675,0.14116246997777146,-0.8316062741857977,1.4416458614228929,0.14116246997777146,-0.38952582367799504,1.1964239960203522,0.4184866590588439,-0.8021462073315929,1.2403645277919675,0.14116246997777146,-0.8316062741857977,1.4416458614228929,-0.14116246997777146,-0.38952582367799504,1.4416458614228929,0.14116246997777146,-0.38952582367799504,1.2403645277919675,-0.14116246997777146,-0.8316062741857977,1.4416458614228929,-0.14116246997777146,-0.38952582367799504,1.2403645277919675,0.14116246997777146,-0.8316062741857977,1.2403645277919675,-0.14116246997777146,-0.8316062741857977,1.3905748380600782,-0.4184866590588439,-0.3757266771789763,1.4416458614228929,-0.14116246997777146,-0.38952582367799504,1.1964239960203522,-0.4184866590588439,-0.8021462073315929,1.3905748380600782,-0.4184866590588439,-0.3757266771789763,1.2403645277919675,-0.14116246997777146,-0.8316062741857977,1.1964239960203522,-0.4184866590588439,-0.8021462073315929,1.2902420077981585,-0.6809857496093201,-0.34861722582512933,1.3905748380600782,-0.4184866590588439,-0.3757266771789763,1.1100995477213593,-0.6809857496093201,-0.7442697111785925,1.2902420077981585,-0.6809857496093201,-0.34861722582512933,1.1964239960203522,-0.4184866590588439,-0.8021462073315929,0.8125047922465052,0.6809857496093201,-1.0611759380072165,1.1964239960203522,0.4184866590588439,-0.8021462073315929,1.1100995477213593,0.6809857496093201,-0.7442697111785925,0.8756874393117506,0.4184866590588439,-1.1436959494644527,1.1964239960203522,0.4184866590588439,-0.8021462073315929,0.8125047922465052,0.6809857496093201,-1.0611759380072165,0.8756874393117506,0.4184866590588439,-1.1436959494644527,1.2403645277919675,0.14116246997777146,-0.8316062741857977,1.1964239960203522,0.4184866590588439,-0.8021462073315929,0.907848422271865,0.14116246997777146,-1.1856999617307324,1.2403645277919675,0.14116246997777146,-0.8316062741857977,0.8756874393117506,0.4184866590588439,-1.1436959494644527,0.907848422271865,0.14116246997777146,-1.1856999617307324,1.2403645277919675,-0.14116246997777146,-0.8316062741857977,1.2403645277919675,0.14116246997777146,-0.8316062741857977,0.907848422271865,-0.14116246997777146,-1.1856999617307324,1.2403645277919675,-0.14116246997777146,-0.8316062741857977,0.907848422271865,0.14116246997777146,-1.1856999617307324,0.907848422271865,-0.14116246997777146,-1.1856999617307324,1.1964239960203522,-0.4184866590588439,-0.8021462073315929,1.2403645277919675,-0.14116246997777146,-0.8316062741857977,0.8756874393117506,-0.4184866590588439,-1.1436959494644527,1.1964239960203522,-0.4184866590588439,-0.8021462073315929,0.907848422271865,-0.14116246997777146,-1.1856999617307324,0.8756874393117506,-0.4184866590588439,-1.1436959494644527,1.1100995477213593,-0.6809857496093201,-0.7442697111785925,1.1964239960203522,-0.4184866590588439,-0.8021462073315929,0.8125047922465052,-0.6809857496093201,-1.0611759380072165,1.1100995477213593,-0.6809857496093201,-0.7442697111785925,0.8756874393117506,-0.4184866590588439,-1.1436959494644527,0.42894427535218416,0.6809857496093201,-1.2658061531970928,0.8756874393117506,0.4184866590588439,-1.1436959494644527,0.8125047922465052,0.6809857496093201,-1.0611759380072165,0.4623001829343417,0.4184866590588439,-1.3642387830026828,0.8756874393117506,0.4184866590588439,-1.1436959494644527,0.42894427535218416,0.6809857496093201,-1.2658061531970928,0.4623001829343417,0.4184866590588439,-1.3642387830026828,0.907848422271865,0.14116246997777146,-1.1856999617307324,0.8756874393117506,0.4184866590588439,-1.1436959494644527,0.47927887605970465,0.14116246997777146,-1.4143425737892221,0.907848422271865,0.14116246997777146,-1.1856999617307324,0.4623001829343417,0.4184866590588439,-1.3642387830026828,0.47927887605970465,0.14116246997777146,-1.4143425737892221,0.907848422271865,-0.14116246997777146,-1.1856999617307324,0.907848422271865,0.14116246997777146,-1.1856999617307324,0.47927887605970465,-0.14116246997777146,-1.4143425737892221,0.907848422271865,-0.14116246997777146,-1.1856999617307324,0.47927887605970465,0.14116246997777146,-1.4143425737892221,0.47927887605970465,-0.14116246997777146,-1.4143425737892221,0.8756874393117506,-0.4184866590588439,-1.1436959494644527,0.907848422271865,-0.14116246997777146,-1.1856999617307324,0.4623001829343417,-0.4184866590588439,-1.3642387830026828,0.8756874393117506,-0.4184866590588439,-1.1436959494644527,0.47927887605970465,-0.14116246997777146,-1.4143425737892221,0.4623001829343417,-0.4184866590588439,-1.3642387830026828,0.8125047922465052,-0.6809857496093201,-1.0611759380072165,0.8756874393117506,-0.4184866590588439,-1.1436959494644527,0.42894427535218416,-0.6809857496093201,-1.2658061531970928,0.8125047922465052,-0.6809857496093201,-1.0611759380072165,0.4623001829343417,-0.4184866590588439,-1.3642387830026828,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,0.4623001829343417,0.4184866590588439,-1.3642387830026828,0.42894427535218416,0.6809857496093201,-1.2658061531970928,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,0.4623001829343417,0.4184866590588439,-1.3642387830026828,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,0.47927887605970465,0.14116246997777146,-1.4143425737892221,0.4623001829343417,0.4184866590588439,-1.3642387830026828,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,0.47927887605970465,0.14116246997777146,-1.4143425737892221,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,0.47927887605970465,-0.14116246997777146,-1.4143425737892221,0.47927887605970465,0.14116246997777146,-1.4143425737892221,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,0.47927887605970465,-0.14116246997777146,-1.4143425737892221,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,0.4623001829343417,-0.4184866590588439,-1.3642387830026828,0.47927887605970465,-0.14116246997777146,-1.4143425737892221,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,0.4623001829343417,-0.4184866590588439,-1.3642387830026828,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,0.42894427535218416,-0.6809857496093201,-1.2658061531970928,0.4623001829343417,-0.4184866590588439,-1.3642387830026828,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,0.42894427535218416,-0.6809857496093201,-1.2658061531970928,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.42894427535218405,0.6809857496093201,-1.2658061531970928,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,8.183762159000201e-17,0.6809857496093201,-1.336509786282552,-0.46230018293434155,0.4184866590588439,-1.3642387830026828,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,-0.42894427535218405,0.6809857496093201,-1.2658061531970928,-0.46230018293434155,0.4184866590588439,-1.3642387830026828,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,8.820154413042621e-17,0.4184866590588439,-1.4404405285154147,-0.47927887605970443,0.14116246997777146,-1.4143425737892221,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,-0.46230018293434155,0.4184866590588439,-1.3642387830026828,-0.47927887605970443,0.14116246997777146,-1.4143425737892221,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,9.144088299780094e-17,0.14116246997777146,-1.49334294690462,-0.47927887605970443,-0.14116246997777146,-1.4143425737892221,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,-0.47927887605970443,0.14116246997777146,-1.4143425737892221,-0.47927887605970443,-0.14116246997777146,-1.4143425737892221,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,9.144088299780094e-17,-0.14116246997777146,-1.49334294690462,-0.46230018293434155,-0.4184866590588439,-1.3642387830026828,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.47927887605970443,-0.14116246997777146,-1.4143425737892221,-0.46230018293434155,-0.4184866590588439,-1.3642387830026828,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,8.820154413042621e-17,-0.4184866590588439,-1.4404405285154147,-0.42894427535218405,-0.6809857496093201,-1.2658061531970928,8.183762159000201e-17,-0.6809857496093201,-1.336509786282552,-0.46230018293434155,-0.4184866590588439,-1.3642387830026828,-0.8125047922465047,0.6809857496093201,-1.061175938007217,-0.46230018293434155,0.4184866590588439,-1.3642387830026828,-0.42894427535218405,0.6809857496093201,-1.2658061531970928,-0.8756874393117503,0.4184866590588439,-1.143695949464453,-0.46230018293434155,0.4184866590588439,-1.3642387830026828,-0.8125047922465047,0.6809857496093201,-1.061175938007217,-0.8756874393117503,0.4184866590588439,-1.143695949464453,-0.47927887605970443,0.14116246997777146,-1.4143425737892221,-0.46230018293434155,0.4184866590588439,-1.3642387830026828,-0.9078484222718645,0.14116246997777146,-1.1856999617307329,-0.47927887605970443,0.14116246997777146,-1.4143425737892221,-0.8756874393117503,0.4184866590588439,-1.143695949464453,-0.9078484222718645,0.14116246997777146,-1.1856999617307329,-0.47927887605970443,-0.14116246997777146,-1.4143425737892221,-0.47927887605970443,0.14116246997777146,-1.4143425737892221,-0.9078484222718645,-0.14116246997777146,-1.1856999617307329,-0.47927887605970443,-0.14116246997777146,-1.4143425737892221,-0.9078484222718645,0.14116246997777146,-1.1856999617307329,-0.9078484222718645,-0.14116246997777146,-1.1856999617307329,-0.46230018293434155,-0.4184866590588439,-1.3642387830026828,-0.47927887605970443,-0.14116246997777146,-1.4143425737892221,-0.8756874393117503,-0.4184866590588439,-1.143695949464453,-0.46230018293434155,-0.4184866590588439,-1.3642387830026828,-0.9078484222718645,-0.14116246997777146,-1.1856999617307329,-0.8756874393117503,-0.4184866590588439,-1.143695949464453,-0.42894427535218405,-0.6809857496093201,-1.2658061531970928,-0.46230018293434155,-0.4184866590588439,-1.3642387830026828,-0.8125047922465047,-0.6809857496093201,-1.061175938007217,-0.42894427535218405,-0.6809857496093201,-1.2658061531970928,-0.8756874393117503,-0.4184866590588439,-1.143695949464453,-1.1100995477213593,0.6809857496093201,-0.7442697111785928,-0.8756874393117503,0.4184866590588439,-1.143695949464453,-0.8125047922465047,0.6809857496093201,-1.061175938007217,-1.1964239960203522,0.4184866590588439,-0.802146207331593,-0.8756874393117503,0.4184866590588439,-1.143695949464453,-1.1100995477213593,0.6809857496093201,-0.7442697111785928,-1.1964239960203522,0.4184866590588439,-0.802146207331593,-0.9078484222718645,0.14116246997777146,-1.1856999617307329,-0.8756874393117503,0.4184866590588439,-1.143695949464453,-1.2403645277919675,0.14116246997777146,-0.8316062741857979,-0.9078484222718645,0.14116246997777146,-1.1856999617307329,-1.1964239960203522,0.4184866590588439,-0.802146207331593,-1.2403645277919675,0.14116246997777146,-0.8316062741857979,-0.9078484222718645,-0.14116246997777146,-1.1856999617307329,-0.9078484222718645,0.14116246997777146,-1.1856999617307329,-1.2403645277919675,-0.14116246997777146,-0.8316062741857979,-0.9078484222718645,-0.14116246997777146,-1.1856999617307329,-1.2403645277919675,0.14116246997777146,-0.8316062741857979,-1.2403645277919675,-0.14116246997777146,-0.8316062741857979,-0.8756874393117503,-0.4184866590588439,-1.143695949464453,-0.9078484222718645,-0.14116246997777146,-1.1856999617307329,-1.1964239960203522,-0.4184866590588439,-0.802146207331593,-0.8756874393117503,-0.4184866590588439,-1.143695949464453,-1.2403645277919675,-0.14116246997777146,-0.8316062741857979,-1.1964239960203522,-0.4184866590588439,-0.802146207331593,-0.8125047922465047,-0.6809857496093201,-1.061175938007217,-0.8756874393117503,-0.4184866590588439,-1.143695949464453,-1.1100995477213593,-0.6809857496093201,-0.7442697111785928,-0.8125047922465047,-0.6809857496093201,-1.061175938007217,-1.1964239960203522,-0.4184866590588439,-0.802146207331593,-1.2902420077981585,0.6809857496093201,-0.3486172258251292,-1.1964239960203522,0.4184866590588439,-0.802146207331593,-1.1100995477213593,0.6809857496093201,-0.7442697111785928,-1.3905748380600782,0.4184866590588439,-0.3757266771789761,-1.1964239960203522,0.4184866590588439,-0.802146207331593,-1.2902420077981585,0.6809857496093201,-0.3486172258251292,-1.3905748380600782,0.4184866590588439,-0.3757266771789761,-1.2403645277919675,0.14116246997777146,-0.8316062741857979,-1.1964239960203522,0.4184866590588439,-0.802146207331593,-1.4416458614228929,0.14116246997777146,-0.3895258236779949,-1.2403645277919675,0.14116246997777146,-0.8316062741857979,-1.3905748380600782,0.4184866590588439,-0.3757266771789761,-1.4416458614228929,0.14116246997777146,-0.3895258236779949,-1.2403645277919675,-0.14116246997777146,-0.8316062741857979,-1.2403645277919675,0.14116246997777146,-0.8316062741857979,-1.4416458614228929,-0.14116246997777146,-0.3895258236779949,-1.2403645277919675,-0.14116246997777146,-0.8316062741857979,-1.4416458614228929,0.14116246997777146,-0.3895258236779949,-1.4416458614228929,-0.14116246997777146,-0.3895258236779949,-1.1964239960203522,-0.4184866590588439,-0.802146207331593,-1.2403645277919675,-0.14116246997777146,-0.8316062741857979,-1.3905748380600782,-0.4184866590588439,-0.3757266771789761,-1.1964239960203522,-0.4184866590588439,-0.802146207331593,-1.4416458614228929,-0.14116246997777146,-0.3895258236779949,-1.3905748380600782,-0.4184866590588439,-0.3757266771789761,-1.1100995477213593,-0.6809857496093201,-0.7442697111785928,-1.1964239960203522,-0.4184866590588439,-0.802146207331593,-1.2902420077981585,-0.6809857496093201,-0.3486172258251292,-1.1100995477213593,-0.6809857496093201,-0.7442697111785928,-1.3905748380600782,-0.4184866590588439,-0.3757266771789761,-1.3338724895159437,0.6809857496093201,8.392014383669297e-2,-1.3905748380600782,0.4184866590588439,-0.3757266771789761,-1.2902420077981585,0.6809857496093201,-0.3486172258251292,-1.4375981481697295,0.4184866590588439,9.044600913656156e-2,-1.3905748380600782,0.4184866590588439,-0.3757266771789761,-1.3338724895159437,0.6809857496093201,8.392014383669297e-2,-1.4375981481697295,0.4184866590588439,9.044600913656156e-2,-1.4416458614228929,0.14116246997777146,-0.3895258236779949,-1.3905748380600782,0.4184866590588439,-0.3757266771789761,-1.4903961757206519,0.14116246997777146,9.376777947157687e-2,-1.4416458614228929,0.14116246997777146,-0.3895258236779949,-1.4375981481697295,0.4184866590588439,9.044600913656156e-2,-1.4903961757206519,0.14116246997777146,9.376777947157687e-2,-1.4416458614228929,-0.14116246997777146,-0.3895258236779949,-1.4416458614228929,0.14116246997777146,-0.3895258236779949,-1.4903961757206519,-0.14116246997777146,9.376777947157687e-2,-1.4416458614228929,-0.14116246997777146,-0.3895258236779949,-1.4903961757206519,0.14116246997777146,9.376777947157687e-2,-1.4903961757206519,-0.14116246997777146,9.376777947157687e-2,-1.3905748380600782,-0.4184866590588439,-0.3757266771789761,-1.4416458614228929,-0.14116246997777146,-0.3895258236779949,-1.4375981481697295,-0.4184866590588439,9.044600913656156e-2,-1.3905748380600782,-0.4184866590588439,-0.3757266771789761,-1.4903961757206519,-0.14116246997777146,9.376777947157687e-2,-1.4375981481697295,-0.4184866590588439,9.044600913656156e-2,-1.2902420077981585,-0.6809857496093201,-0.3486172258251292,-1.3905748380600782,-0.4184866590588439,-0.3757266771789761,-1.3338724895159437,-0.6809857496093201,8.392014383669297e-2,-1.2902420077981585,-0.6809857496093201,-0.3486172258251292,-1.4375981481697295,-0.4184866590588439,9.044600913656156e-2],[[1.1824769291328754,0.303478973012677,-0.44682079742068315,1.2414150625489981,0.1831716015188576,-0.3396127307294057,1.2192786062662366,0.30347897301267707,-0.3335568815668989,1.203945237288697,0.1831716015188576,-0.4549329950738682,1.2414150625489981,0.1831716015188576,-0.3396127307294057,1.1824769291328754,0.303478973012677,-0.44682079742068315,1.203945237288697,0.1831716015188576,-0.4549329950738682,1.2525326308520357,6.123838592253551e-2,-0.3426541532514691,1.2414150625489981,0.1831716015188576,-0.3396127307294057,1.2147272422864375,6.123838592253551e-2,-0.45900717525632384,1.2525326308520357,6.123838592253551e-2,-0.3426541532514691,1.203945237288697,0.1831716015188576,-0.4549329950738682,1.2147272422864375,6.123838592253551e-2,-0.45900717525632384,1.2525326308520357,-6.123838592253551e-2,-0.3426541532514691,1.2525326308520357,6.123838592253551e-2,-0.3426541532514691,1.2147272422864375,-6.123838592253551e-2,-0.45900717525632384,1.2525326308520357,-6.123838592253551e-2,-0.3426541532514691,1.2147272422864375,6.123838592253551e-2,-0.45900717525632384,1.2147272422864375,-6.123838592253551e-2,-0.45900717525632384,1.2414150625489981,-0.1831716015188576,-0.3396127307294057,1.2525326308520357,-6.123838592253551e-2,-0.3426541532514691,1.203945237288697,-0.1831716015188576,-0.4549329950738682,1.2414150625489981,-0.1831716015188576,-0.3396127307294057,1.2147272422864375,-6.123838592253551e-2,-0.45900717525632384,1.203945237288697,-0.1831716015188576,-0.4549329950738682,1.2192786062662369,-0.3034789730126769,-0.3335568815668989,1.2414150625489981,-0.1831716015188576,-0.3396127307294057,1.1824769291328756,-0.30347897301267684,-0.44682079742068326,1.2192786062662369,-0.3034789730126769,-0.3335568815668989,1.203945237288697,-0.1831716015188576,-0.4549329950738682,1.1351795030644487,0.30347897301267707,-0.5561187002444015,1.203945237288697,0.1831716015188576,-0.4549329950738682,1.1824769291328754,0.303478973012677,-0.44682079742068315,1.1557891088703147,0.1831716015188576,-0.5662152419476012,1.203945237288697,0.1831716015188576,-0.4549329950738682,1.1351795030644487,0.30347897301267707,-0.5561187002444015,1.1557891088703147,0.1831716015188576,-0.5662152419476012,1.2147272422864375,6.123838592253551e-2,-0.45900717525632384,1.203945237288697,0.1831716015188576,-0.4549329950738682,1.1661398487230992,6.123838592253552e-2,-0.571286017078723,1.2147272422864375,6.123838592253551e-2,-0.45900717525632384,1.1557891088703147,0.1831716015188576,-0.5662152419476012,1.1661398487230992,6.123838592253552e-2,-0.571286017078723,1.2147272422864375,-6.123838592253551e-2,-0.45900717525632384,1.2147272422864375,6.123838592253551e-2,-0.45900717525632384,1.1661398487230992,-6.123838592253552e-2,-0.571286017078723,1.2147272422864375,-6.123838592253551e-2,-0.45900717525632384,1.1661398487230992,6.123838592253552e-2,-0.571286017078723,1.1661398487230992,-6.123838592253552e-2,-0.571286017078723,1.203945237288697,-0.1831716015188576,-0.4549329950738682,1.2147272422864375,-6.123838592253551e-2,-0.45900717525632384,1.1557891088703147,-0.1831716015188576,-0.5662152419476012,1.203945237288697,-0.1831716015188576,-0.4549329950738682,1.1661398487230992,-6.123838592253552e-2,-0.571286017078723,1.1557891088703147,-0.1831716015188576,-0.5662152419476012,1.1824769291328756,-0.30347897301267684,-0.44682079742068326,1.203945237288697,-0.1831716015188576,-0.4549329950738682,1.1351795030644487,-0.3034789730126769,-0.5561187002444015,1.1824769291328756,-0.30347897301267684,-0.44682079742068326,1.1557891088703147,-0.1831716015188576,-0.5662152419476012,1.0778061433631057,0.30347897301267707,-0.6604804541149717,1.1557891088703147,0.1831716015188576,-0.5662152419476012,1.1351795030644487,0.30347897301267707,-0.5561187002444015,1.0973741144988507,0.1831716015188576,-0.6724717222492559,1.1557891088703147,0.1831716015188576,-0.5662152419476012,1.0778061433631057,0.30347897301267707,-0.6604804541149717,1.0973741144988507,0.1831716015188576,-0.6724717222492559,1.1661398487230992,6.123838592253552e-2,-0.571286017078723,1.1557891088703147,0.1831716015188576,-0.5662152419476012,1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,1.1661398487230992,6.123838592253552e-2,-0.571286017078723,1.0973741144988507,0.1831716015188576,-0.6724717222492559,1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,1.1661398487230992,-6.123838592253552e-2,-0.571286017078723,1.1661398487230992,6.123838592253552e-2,-0.571286017078723,1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,1.1661398487230992,-6.123838592253552e-2,-0.571286017078723,1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,1.1557891088703147,-0.1831716015188576,-0.5662152419476012,1.1661398487230992,-6.123838592253552e-2,-0.571286017078723,1.0973741144988507,-0.1831716015188576,-0.6724717222492559,1.1557891088703147,-0.1831716015188576,-0.5662152419476012,1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,1.0973741144988507,-0.1831716015188576,-0.6724717222492559,1.1351795030644487,-0.3034789730126769,-0.5561187002444015,1.1557891088703147,-0.1831716015188576,-0.5662152419476012,1.0778061433631057,-0.3034789730126769,-0.6604804541149719,1.1351795030644487,-0.3034789730126769,-0.5561187002444015,1.0973741144988507,-0.1831716015188576,-0.6724717222492559,1.0108661000312356,0.30347897301267707,-0.7589797367168709,1.0973741144988507,0.1831716015188576,-0.6724717222492559,1.0778061433631057,0.30347897301267707,-0.6604804541149717,1.0292187497997671,0.1831716015188576,-0.772759295937371,1.0973741144988507,0.1831716015188576,-0.6724717222492559,1.0108661000312356,0.30347897301267707,-0.7589797367168709,1.0292187497997671,0.1831716015188576,-0.772759295937371,1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,1.0973741144988507,0.1831716015188576,-0.6724717222492559,1.038435981082728,6.123838592253551e-2,-0.7796797889405338,1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,1.0292187497997671,0.1831716015188576,-0.772759295937371,1.038435981082728,6.123838592253551e-2,-0.7796797889405338,1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,1.038435981082728,-6.123838592253551e-2,-0.7796797889405338,1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,1.038435981082728,6.123838592253551e-2,-0.7796797889405338,1.038435981082728,-6.123838592253551e-2,-0.7796797889405338,1.0973741144988507,-0.1831716015188576,-0.6724717222492559,1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,1.0292187497997671,-0.1831716015188576,-0.772759295937371,1.0973741144988507,-0.1831716015188576,-0.6724717222492559,1.038435981082728,-6.123838592253551e-2,-0.7796797889405338,1.0292187497997671,-0.1831716015188576,-0.772759295937371,1.0778061433631057,-0.3034789730126769,-0.6604804541149719,1.0973741144988507,-0.1831716015188576,-0.6724717222492559,1.0108661000312358,-0.3034789730126769,-0.7589797367168709,1.0778061433631057,-0.3034789730126769,-0.6604804541149719,1.0292187497997671,-0.1831716015188576,-0.772759295937371,0.9349535376323963,0.30347897301267707,-0.85074226144458,1.0292187497997671,0.1831716015188576,-0.772759295937371,1.0108661000312356,0.30347897301267707,-0.7589797367168709,0.9519279666151136,0.1831716015188576,-0.8661878034081479,1.0292187497997671,0.1831716015188576,-0.772759295937371,0.9349535376323963,0.30347897301267707,-0.85074226144458,0.9519279666151136,0.1831716015188576,-0.8661878034081479,1.038435981082728,6.123838592253551e-2,-0.7796797889405338,1.0292187497997671,0.1831716015188576,-0.772759295937371,0.960453015575519,6.123838592253551e-2,-0.8739450011079042,1.038435981082728,6.123838592253551e-2,-0.7796797889405338,0.9519279666151136,0.1831716015188576,-0.8661878034081479,0.960453015575519,6.123838592253551e-2,-0.8739450011079042,1.038435981082728,-6.123838592253551e-2,-0.7796797889405338,1.038435981082728,6.123838592253551e-2,-0.7796797889405338,0.960453015575519,-6.123838592253551e-2,-0.8739450011079042,1.038435981082728,-6.123838592253551e-2,-0.7796797889405338,0.960453015575519,6.123838592253551e-2,-0.8739450011079042,0.960453015575519,-6.123838592253551e-2,-0.8739450011079042,1.0292187497997671,-0.1831716015188576,-0.772759295937371,1.038435981082728,-6.123838592253551e-2,-0.7796797889405338,0.9519279666151136,-0.1831716015188576,-0.8661878034081479,1.0292187497997671,-0.1831716015188576,-0.772759295937371,0.960453015575519,-6.123838592253551e-2,-0.8739450011079042,0.9519279666151136,-0.1831716015188576,-0.8661878034081479,1.0108661000312358,-0.3034789730126769,-0.7589797367168709,1.0292187497997671,-0.1831716015188576,-0.772759295937371,0.9349535376323964,-0.3034789730126769,-0.8507422614445802,1.0108661000312358,-0.3034789730126769,-0.7589797367168709,0.9519279666151136,-0.1831716015188576,-0.8661878034081479],[0.7903572975614958,0.303478973012677,-0.9865271689773475,0.8956611720343062,0.1831716015188576,-0.9242506312181505,0.8796900718140512,0.30347897301267707,-0.9077697342889112,0.8047065280616588,0.1831716015188576,-1.0044379364062113,0.8956611720343062,0.1831716015188576,-0.9242506312181505,0.7903572975614958,0.303478973012677,-0.9865271689773475,0.8047065280616588,0.1831716015188576,-1.0044379364062113,0.9036823202842912,6.123838592253551e-2,-0.9325278141134464,0.8956611720343062,0.1831716015188576,-0.9242506312181505,0.8119131264504813,6.123838592253551e-2,-1.013433241603515,0.9036823202842912,6.123838592253551e-2,-0.9325278141134464,0.8047065280616588,0.1831716015188576,-1.0044379364062113,0.8119131264504813,6.123838592253551e-2,-1.013433241603515,0.9036823202842912,-6.123838592253551e-2,-0.9325278141134464,0.9036823202842912,6.123838592253551e-2,-0.9325278141134464,0.8119131264504813,-6.123838592253551e-2,-1.013433241603515,0.9036823202842912,-6.123838592253551e-2,-0.9325278141134464,0.8119131264504813,6.123838592253551e-2,-1.013433241603515,0.8119131264504813,-6.123838592253551e-2,-1.013433241603515,0.8956611720343062,-0.1831716015188576,-0.9242506312181505,0.9036823202842912,-6.123838592253551e-2,-0.9325278141134464,0.8047065280616588,-0.1831716015188576,-1.0044379364062113,0.8956611720343062,-0.1831716015188576,-0.9242506312181505,0.8119131264504813,-6.123838592253551e-2,-1.013433241603515,0.8047065280616588,-0.1831716015188576,-1.0044379364062113,0.8796900718140512,-0.3034789730126769,-0.9077697342889113,0.8956611720343062,-0.1831716015188576,-0.9242506312181505,0.7903572975614958,-0.30347897301267684,-0.9865271689773475,0.8796900718140512,-0.3034789730126769,-0.9077697342889113,0.8047065280616588,-0.1831716015188576,-1.0044379364062113,0.6940092559833568,0.30347897301267707,-1.0565281186738942,0.8047065280616588,0.1831716015188576,-1.0044379364062113,0.7903572975614958,0.303478973012677,-0.9865271689773475,0.706609251977671,0.1831716015188576,-1.0757097793627113,0.8047065280616588,0.1831716015188576,-1.0044379364062113,0.6940092559833568,0.30347897301267707,-1.0565281186738942,0.706609251977671,0.1831716015188576,-1.0757097793627113,0.8119131264504813,6.123838592253551e-2,-1.013433241603515,0.8047065280616588,0.1831716015188576,-1.0044379364062113,0.712937334227849,6.123838592253552e-2,-1.0853433638962795,0.8119131264504813,6.123838592253551e-2,-1.013433241603515,0.706609251977671,0.1831716015188576,-1.0757097793627113,0.712937334227849,6.123838592253552e-2,-1.0853433638962795,0.8119131264504813,-6.123838592253551e-2,-1.013433241603515,0.8119131264504813,6.123838592253551e-2,-1.013433241603515,0.712937334227849,-6.123838592253552e-2,-1.0853433638962795,0.8119131264504813,-6.123838592253551e-2,-1.013433241603515,0.712937334227849,6.123838592253552e-2,-1.0853433638962795,0.712937334227849,-6.123838592253552e-2,-1.0853433638962795,0.8047065280616588,-0.1831716015188576,-1.0044379364062113,0.8119131264504813,-6.123838592253551e-2,-1.013433241603515,0.706609251977671,-0.1831716015188576,-1.0757097793627113,0.8047065280616588,-0.1831716015188576,-1.0044379364062113,0.712937334227849,-6.123838592253552e-2,-1.0853433638962795,0.706609251977671,-0.1831716015188576,-1.0757097793627113,0.7903572975614958,-0.30347897301267684,-0.9865271689773475,0.8047065280616588,-0.1831716015188576,-1.0044379364062113,0.6940092559833569,-0.3034789730126769,-1.0565281186738944,0.7903572975614958,-0.30347897301267684,-0.9865271689773475,0.706609251977671,-0.1831716015188576,-1.0757097793627113,0.5915011391175296,0.30347897301267707,-1.1171512499934089,0.706609251977671,0.1831716015188576,-1.0757097793627113,0.6940092559833568,0.30347897301267707,-1.0565281186738942,0.6022400621495471,0.1831716015188576,-1.1374335461639624,0.706609251977671,0.1831716015188576,-1.0757097793627113,0.5915011391175296,0.30347897301267707,-1.1171512499934089,0.6022400621495471,0.1831716015188576,-1.1374335461639624,0.712937334227849,6.123838592253552e-2,-1.0853433638962795,0.706609251977671,0.1831716015188576,-1.0757097793627113,0.6076334597550389,6.123838592253552e-2,-1.1476199016554762,0.712937334227849,6.123838592253552e-2,-1.0853433638962795,0.6022400621495471,0.1831716015188576,-1.1374335461639624,0.6076334597550389,6.123838592253552e-2,-1.1476199016554762,0.712937334227849,-6.123838592253552e-2,-1.0853433638962795,0.712937334227849,6.123838592253552e-2,-1.0853433638962795,0.6076334597550389,-6.123838592253552e-2,-1.1476199016554762,0.712937334227849,-6.123838592253552e-2,-1.0853433638962795,0.6076334597550389,6.123838592253552e-2,-1.1476199016554762,0.6076334597550389,-6.123838592253552e-2,-1.1476199016554762,0.706609251977671,-0.1831716015188576,-1.0757097793627113,0.712937334227849,-6.123838592253552e-2,-1.0853433638962795,0.6022400621495471,-0.1831716015188576,-1.1374335461639624,0.706609251977671,-0.1831716015188576,-1.0757097793627113,0.6076334597550389,-6.123838592253552e-2,-1.1476199016554762,0.6022400621495471,-0.1831716015188576,-1.1374335461639624,0.6940092559833569,-0.3034789730126769,-1.0565281186738944,0.706609251977671,-0.1831716015188576,-1.0757097793627113,0.5915011391175296,-0.3034789730126769,-1.1171512499934089,0.6940092559833569,-0.3034789730126769,-1.0565281186738944,0.6022400621495471,-0.1831716015188576,-1.1374335461639624,0.4837428162662571,0.30347897301267707,-1.1678584677305557,0.6022400621495471,0.1831716015188576,-1.1374335461639624,0.5915011391175296,0.30347897301267707,-1.1171512499934089,0.492525346894897,0.1831716015188576,-1.1890613722861736,0.6022400621495471,0.1831716015188576,-1.1374335461639624,0.4837428162662571,0.30347897301267707,-1.1678584677305557,0.492525346894897,0.1831716015188576,-1.1890613722861736,0.6076334597550389,6.123838592253552e-2,-1.1476199016554762,0.6022400621495471,0.1831716015188576,-1.1374335461639624,0.49693618767673614,6.123838592253551e-2,-1.1997100839231596,0.6076334597550389,6.123838592253552e-2,-1.1476199016554762,0.492525346894897,0.1831716015188576,-1.1890613722861736,0.49693618767673614,6.123838592253551e-2,-1.1997100839231596,0.6076334597550389,-6.123838592253552e-2,-1.1476199016554762,0.6076334597550389,6.123838592253552e-2,-1.1476199016554762,0.49693618767673614,-6.123838592253551e-2,-1.1997100839231596,0.6076334597550389,-6.123838592253552e-2,-1.1476199016554762,0.49693618767673614,6.123838592253551e-2,-1.1997100839231596,0.49693618767673614,-6.123838592253551e-2,-1.1997100839231596,0.6022400621495471,-0.1831716015188576,-1.1374335461639624,0.6076334597550389,-6.123838592253552e-2,-1.1476199016554762,0.492525346894897,-0.1831716015188576,-1.1890613722861736,0.6022400621495471,-0.1831716015188576,-1.1374335461639624,0.49693618767673614,-6.123838592253551e-2,-1.1997100839231596,0.492525346894897,-0.1831716015188576,-1.1890613722861736,0.5915011391175296,-0.3034789730126769,-1.1171512499934089,0.6022400621495471,-0.1831716015188576,-1.1374335461639624,0.48374281626625715,-0.3034789730126769,-1.1678584677305557,0.5915011391175296,-0.3034789730126769,-1.1171512499934089,0.492525346894897,-0.1831716015188576,-1.1890613722861736,0.3716907579317954,0.30347897301267707,-1.2081996910309398,0.492525346894897,0.1831716015188576,-1.1890613722861736,0.4837428162662571,0.30347897301267707,-1.1678584677305557,0.3784389417934477,0.1831716015188576,-1.230135005489752,0.492525346894897,0.1831716015188576,-1.1890613722861736,0.3716907579317954,0.30347897301267707,-1.2081996910309398,0.3784389417934477,0.1831716015188576,-1.230135005489752,0.49693618767673614,6.123838592253551e-2,-1.1997100839231596,0.492525346894897,0.1831716015188576,-1.1890613722861736,0.3818280748165955,6.123838592253551e-2,-1.2411515545538567,0.49693618767673614,6.123838592253551e-2,-1.1997100839231596,0.3784389417934477,0.1831716015188576,-1.230135005489752,0.3818280748165955,6.123838592253551e-2,-1.2411515545538567,0.49693618767673614,-6.123838592253551e-2,-1.1997100839231596,0.49693618767673614,6.123838592253551e-2,-1.1997100839231596,0.3818280748165955,-6.123838592253551e-2,-1.2411515545538567,0.49693618767673614,-6.123838592253551e-2,-1.1997100839231596,0.3818280748165955,6.123838592253551e-2,-1.2411515545538567,0.3818280748165955,-6.123838592253551e-2,-1.2411515545538567,0.492525346894897,-0.1831716015188576,-1.1890613722861736,0.49693618767673614,-6.123838592253551e-2,-1.1997100839231596,0.3784389417934477,-0.1831716015188576,-1.230135005489752,0.492525346894897,-0.1831716015188576,-1.1890613722861736,0.3818280748165955,-6.123838592253551e-2,-1.2411515545538567,0.3784389417934477,-0.1831716015188576,-1.230135005489752,0.48374281626625715,-0.3034789730126769,-1.1678584677305557,0.492525346894897,-0.1831716015188576,-1.1890613722861736,0.3716907579317955,-0.3034789730126769,-1.2081996910309398,0.48374281626625715,-0.3034789730126769,-1.1678584677305557,0.3784389417934477,-0.1831716015188576,-1.230135005489752],[0.17811055558800623,0.303478973012677,-1.251469992819365,0.300451362902075,0.1831716015188576,-1.2514699928193653,0.29509382483070545,0.30347897301267707,-1.2291542407224392,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.300451362902075,0.1831716015188576,-1.2514699928193653,0.17811055558800623,0.303478973012677,-1.251469992819365,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.3031420734062201,6.123838592253551e-2,-1.2626776086636606,0.300451362902075,0.1831716015188576,-1.2514699928193653,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.3031420734062201,6.123838592253551e-2,-1.2626776086636606,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.3031420734062201,-6.123838592253551e-2,-1.2626776086636606,0.3031420734062201,6.123838592253551e-2,-1.2626776086636606,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.3031420734062201,-6.123838592253551e-2,-1.2626776086636606,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.300451362902075,-0.1831716015188576,-1.2514699928193653,0.3031420734062201,-6.123838592253551e-2,-1.2626776086636606,0.18134421892549893,-0.1831716015188576,-1.274190895690013,0.300451362902075,-0.1831716015188576,-1.2514699928193653,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.18134421892549893,-0.1831716015188576,-1.274190895690013,0.2950938248307055,-0.3034789730126769,-1.2291542407224392,0.300451362902075,-0.1831716015188576,-1.2514699928193653,0.17811055558800623,-0.30347897301267684,-1.2514699928193653,0.2950938248307055,-0.3034789730126769,-1.2291542407224392,0.18134421892549893,-0.1831716015188576,-1.274190895690013,5.9546364444778024e-2,0.30347897301267707,-1.2626776086636606,0.18134421892549893,0.1831716015188576,-1.274190895690013,0.17811055558800623,0.303478973012677,-1.251469992819365,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,0.18134421892549893,0.1831716015188576,-1.274190895690013,5.9546364444778024e-2,0.30347897301267707,-1.2626776086636606,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,0.18134421892549893,0.1831716015188576,-1.274190895690013,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,0.18296825813775525,6.123838592253551e-2,-1.2856019899656497,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,0.18134421892549893,-0.1831716015188576,-1.274190895690013,0.18296825813775525,-6.123838592253551e-2,-1.2856019899656497,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,0.18134421892549893,-0.1831716015188576,-1.274190895690013,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,0.17811055558800623,-0.30347897301267684,-1.2514699928193653,0.18134421892549893,-0.1831716015188576,-1.274190895690013,5.954636444477803e-2,-0.3034789730126769,-1.2626776086636609,0.17811055558800623,-0.30347897301267684,-1.2514699928193653,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,-5.954636444477759e-2,0.30347897301267707,-1.2626776086636606,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,5.9546364444778024e-2,0.30347897301267707,-1.2626776086636606,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,-5.954636444477759e-2,0.30347897301267707,-1.2626776086636606,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,6.062745082368673e-2,0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,6.117040365703434e-2,-6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,5.954636444477803e-2,-0.3034789730126769,-1.2626776086636609,6.062745082368673e-2,-0.1831716015188576,-1.2856019899656497,-5.95463644447776e-2,-0.3034789730126769,-1.2626776086636609,5.954636444477803e-2,-0.3034789730126769,-1.2626776086636609,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-0.17811055558800643,0.30347897301267707,-1.2514699928193653,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-5.954636444477759e-2,0.30347897301267707,-1.2626776086636606,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-0.17811055558800643,0.30347897301267707,-1.2514699928193653,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-6.062745082368628e-2,0.1831716015188576,-1.2856019899656497,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-6.117040365703389e-2,6.123838592253552e-2,-1.2971152769920025,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-6.117040365703389e-2,-6.123838592253552e-2,-1.2971152769920025,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-5.95463644447776e-2,-0.3034789730126769,-1.2626776086636609,-6.062745082368628e-2,-0.1831716015188576,-1.2856019899656497,-0.17811055558800643,-0.3034789730126769,-1.2514699928193656,-5.95463644447776e-2,-0.3034789730126769,-1.2626776086636609,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.2950938248307048,0.30347897301267707,-1.2291542407224392,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.17811055558800643,0.30347897301267707,-1.2514699928193653,-0.30045136290207436,0.1831716015188576,-1.2514699928193656,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.2950938248307048,0.30347897301267707,-1.2291542407224392,-0.30045136290207436,0.1831716015188576,-1.2514699928193656,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.1813442189254988,0.1831716015188576,-1.274190895690013,-0.3031420734062194,6.123838592253551e-2,-1.2626776086636609,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.30045136290207436,0.1831716015188576,-1.2514699928193656,-0.3031420734062194,6.123838592253551e-2,-1.2626776086636609,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.18296825813775539,6.123838592253551e-2,-1.2856019899656497,-0.3031420734062194,-6.123838592253551e-2,-1.2626776086636609,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.3031420734062194,6.123838592253551e-2,-1.2626776086636609,-0.3031420734062194,-6.123838592253551e-2,-1.2626776086636609,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.18296825813775539,-6.123838592253551e-2,-1.2856019899656497,-0.30045136290207436,-0.1831716015188576,-1.2514699928193656,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.3031420734062194,-6.123838592253551e-2,-1.2626776086636609,-0.30045136290207436,-0.1831716015188576,-1.2514699928193656,-0.17811055558800643,-0.3034789730126769,-1.2514699928193656,-0.1813442189254988,-0.1831716015188576,-1.274190895690013,-0.29509382483070484,-0.3034789730126769,-1.2291542407224392,-0.17811055558800643,-0.3034789730126769,-1.2514699928193656,-0.30045136290207436,-0.1831716015188576,-1.2514699928193656],[-0.4837428162662574,0.303478973012677,-1.167858467730555,-0.37843894179344706,0.1831716015188576,-1.2301350054897524,-0.37169075793179485,0.30347897301267707,-1.20819969103094,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.37843894179344706,0.1831716015188576,-1.2301350054897524,-0.4837428162662574,0.303478973012677,-1.167858467730555,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.38182807481659486,6.123838592253551e-2,-1.241151554553857,-0.37843894179344706,0.1831716015188576,-1.2301350054897524,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.38182807481659486,6.123838592253551e-2,-1.241151554553857,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.38182807481659486,-6.123838592253551e-2,-1.241151554553857,-0.38182807481659486,6.123838592253551e-2,-1.241151554553857,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.38182807481659486,-6.123838592253551e-2,-1.241151554553857,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.37843894179344706,-0.1831716015188576,-1.2301350054897524,-0.38182807481659486,-6.123838592253551e-2,-1.241151554553857,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.37843894179344706,-0.1831716015188576,-1.2301350054897524,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.37169075793179485,-0.3034789730126769,-1.20819969103094,-0.37843894179344706,-0.1831716015188576,-1.2301350054897524,-0.48374281626625754,-0.30347897301267684,-1.1678584677305552,-0.37169075793179485,-0.3034789730126769,-1.20819969103094,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.5915011391175291,0.30347897301267707,-1.1171512499934089,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.4837428162662574,0.303478973012677,-1.167858467730555,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.5915011391175291,0.30347897301267707,-1.1171512499934089,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.4925253468948969,0.1831716015188576,-1.1890613722861736,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.4969361876767366,6.123838592253551e-2,-1.1997100839231594,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.4969361876767366,-6.123838592253551e-2,-1.1997100839231594,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.48374281626625754,-0.30347897301267684,-1.1678584677305552,-0.4925253468948969,-0.1831716015188576,-1.1890613722861736,-0.5915011391175292,-0.3034789730126769,-1.117151249993409,-0.48374281626625754,-0.30347897301267684,-1.1678584677305552,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.6940092559833565,0.30347897301267707,-1.0565281186738944,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.5915011391175291,0.30347897301267707,-1.1171512499934089,-0.7066092519776707,0.1831716015188576,-1.0757097793627115,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.6940092559833565,0.30347897301267707,-1.0565281186738944,-0.7066092519776707,0.1831716015188576,-1.0757097793627115,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.6022400621495467,0.1831716015188576,-1.1374335461639624,-0.7129373342278488,6.123838592253552e-2,-1.0853433638962795,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.7066092519776707,0.1831716015188576,-1.0757097793627115,-0.7129373342278488,6.123838592253552e-2,-1.0853433638962795,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.6076334597550385,6.123838592253552e-2,-1.1476199016554762,-0.7129373342278488,-6.123838592253552e-2,-1.0853433638962795,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.7129373342278488,6.123838592253552e-2,-1.0853433638962795,-0.7129373342278488,-6.123838592253552e-2,-1.0853433638962795,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.6076334597550385,-6.123838592253552e-2,-1.1476199016554762,-0.7066092519776707,-0.1831716015188576,-1.0757097793627115,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.7129373342278488,-6.123838592253552e-2,-1.0853433638962795,-0.7066092519776707,-0.1831716015188576,-1.0757097793627115,-0.5915011391175292,-0.3034789730126769,-1.117151249993409,-0.6022400621495467,-0.1831716015188576,-1.1374335461639624,-0.6940092559833566,-0.3034789730126769,-1.0565281186738944,-0.5915011391175292,-0.3034789730126769,-1.117151249993409,-0.7066092519776707,-0.1831716015188576,-1.0757097793627115,-0.790357297561496,0.30347897301267707,-0.9865271689773474,-0.7066092519776707,0.1831716015188576,-1.0757097793627115,-0.6940092559833565,0.30347897301267707,-1.0565281186738944,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.7066092519776707,0.1831716015188576,-1.0757097793627115,-0.790357297561496,0.30347897301267707,-0.9865271689773474,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.7129373342278488,6.123838592253552e-2,-1.0853433638962795,-0.7066092519776707,0.1831716015188576,-1.0757097793627115,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.7129373342278488,6.123838592253552e-2,-1.0853433638962795,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.7129373342278488,-6.123838592253552e-2,-1.0853433638962795,-0.7129373342278488,6.123838592253552e-2,-1.0853433638962795,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.7129373342278488,-6.123838592253552e-2,-1.0853433638962795,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.7066092519776707,-0.1831716015188576,-1.0757097793627115,-0.7129373342278488,-6.123838592253552e-2,-1.0853433638962795,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.7066092519776707,-0.1831716015188576,-1.0757097793627115,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.6940092559833566,-0.3034789730126769,-1.0565281186738944,-0.7066092519776707,-0.1831716015188576,-1.0757097793627115,-0.790357297561496,-0.3034789730126769,-0.9865271689773474,-0.6940092559833566,-0.3034789730126769,-1.0565281186738944,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.8796900718140508,0.30347897301267707,-0.9077697342889117,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.790357297561496,0.30347897301267707,-0.9865271689773474,-0.8956611720343057,0.1831716015188576,-0.9242506312181511,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.8796900718140508,0.30347897301267707,-0.9077697342889117,-0.8956611720343057,0.1831716015188576,-0.9242506312181511,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.8047065280616584,0.1831716015188576,-1.0044379364062113,-0.9036823202842907,6.123838592253551e-2,-0.932527814113447,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.8956611720343057,0.1831716015188576,-0.9242506312181511,-0.9036823202842907,6.123838592253551e-2,-0.932527814113447,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.8119131264504814,6.123838592253551e-2,-1.0134332416035146,-0.9036823202842907,-6.123838592253551e-2,-0.932527814113447,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.9036823202842907,6.123838592253551e-2,-0.932527814113447,-0.9036823202842907,-6.123838592253551e-2,-0.932527814113447,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.8119131264504814,-6.123838592253551e-2,-1.0134332416035146,-0.8956611720343057,-0.1831716015188576,-0.9242506312181511,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.9036823202842907,-6.123838592253551e-2,-0.932527814113447,-0.8956611720343057,-0.1831716015188576,-0.9242506312181511,-0.790357297561496,-0.3034789730126769,-0.9865271689773474,-0.8047065280616584,-0.1831716015188576,-1.0044379364062113,-0.8796900718140508,-0.3034789730126769,-0.9077697342889117,-0.790357297561496,-0.3034789730126769,-0.9865271689773474,-0.8956611720343057,-0.1831716015188576,-0.9242506312181511],[-1.0108661000312358,0.303478973012677,-0.7589797367168702,-0.9519279666151136,0.1831716015188576,-0.8661878034081479,-0.9349535376323963,0.30347897301267707,-0.85074226144458,-1.0292187497997674,0.1831716015188576,-0.7727592959373708,-0.9519279666151136,0.1831716015188576,-0.8661878034081479,-1.0108661000312358,0.303478973012677,-0.7589797367168702,-1.0292187497997674,0.1831716015188576,-0.7727592959373708,-0.960453015575519,6.123838592253551e-2,-0.8739450011079042,-0.9519279666151136,0.1831716015188576,-0.8661878034081479,-1.0384359810827284,6.123838592253551e-2,-0.7796797889405332,-0.960453015575519,6.123838592253551e-2,-0.8739450011079042,-1.0292187497997674,0.1831716015188576,-0.7727592959373708,-1.0384359810827284,6.123838592253551e-2,-0.7796797889405332,-0.960453015575519,-6.123838592253551e-2,-0.8739450011079042,-0.960453015575519,6.123838592253551e-2,-0.8739450011079042,-1.0384359810827284,-6.123838592253551e-2,-0.7796797889405332,-0.960453015575519,-6.123838592253551e-2,-0.8739450011079042,-1.0384359810827284,6.123838592253551e-2,-0.7796797889405332,-1.0384359810827284,-6.123838592253551e-2,-0.7796797889405332,-0.9519279666151136,-0.1831716015188576,-0.8661878034081479,-0.960453015575519,-6.123838592253551e-2,-0.8739450011079042,-1.0292187497997674,-0.1831716015188576,-0.7727592959373708,-0.9519279666151136,-0.1831716015188576,-0.8661878034081479,-1.0384359810827284,-6.123838592253551e-2,-0.7796797889405332,-1.0292187497997674,-0.1831716015188576,-0.7727592959373708,-0.9349535376323964,-0.3034789730126769,-0.8507422614445802,-0.9519279666151136,-0.1831716015188576,-0.8661878034081479,-1.010866100031236,-0.30347897301267684,-0.7589797367168702,-0.9349535376323964,-0.3034789730126769,-0.8507422614445802,-1.0292187497997674,-0.1831716015188576,-0.7727592959373708,-1.0778061433631057,0.30347897301267707,-0.6604804541149717,-1.0292187497997674,0.1831716015188576,-0.7727592959373708,-1.0108661000312358,0.303478973012677,-0.7589797367168702,-1.0973741144988507,0.1831716015188576,-0.6724717222492559,-1.0292187497997674,0.1831716015188576,-0.7727592959373708,-1.0778061433631057,0.30347897301267707,-0.6604804541149717,-1.0973741144988507,0.1831716015188576,-0.6724717222492559,-1.0384359810827284,6.123838592253551e-2,-0.7796797889405332,-1.0292187497997674,0.1831716015188576,-0.7727592959373708,-1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,-1.0384359810827284,6.123838592253551e-2,-0.7796797889405332,-1.0973741144988507,0.1831716015188576,-0.6724717222492559,-1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,-1.0384359810827284,-6.123838592253551e-2,-0.7796797889405332,-1.0384359810827284,6.123838592253551e-2,-0.7796797889405332,-1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,-1.0384359810827284,-6.123838592253551e-2,-0.7796797889405332,-1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,-1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,-1.0292187497997674,-0.1831716015188576,-0.7727592959373708,-1.0384359810827284,-6.123838592253551e-2,-0.7796797889405332,-1.0973741144988507,-0.1831716015188576,-0.6724717222492559,-1.0292187497997674,-0.1831716015188576,-0.7727592959373708,-1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,-1.0973741144988507,-0.1831716015188576,-0.6724717222492559,-1.010866100031236,-0.30347897301267684,-0.7589797367168702,-1.0292187497997674,-0.1831716015188576,-0.7727592959373708,-1.0778061433631057,-0.3034789730126769,-0.6604804541149719,-1.010866100031236,-0.30347897301267684,-0.7589797367168702,-1.0973741144988507,-0.1831716015188576,-0.6724717222492559,-1.1351795030644487,0.30347897301267707,-0.5561187002444012,-1.0973741144988507,0.1831716015188576,-0.6724717222492559,-1.0778061433631057,0.30347897301267707,-0.6604804541149717,-1.155789108870315,0.1831716015188576,-0.566215241947601,-1.0973741144988507,0.1831716015188576,-0.6724717222492559,-1.1351795030644487,0.30347897301267707,-0.5561187002444012,-1.155789108870315,0.1831716015188576,-0.566215241947601,-1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,-1.0973741144988507,0.1831716015188576,-0.6724717222492559,-1.1661398487230994,6.123838592253552e-2,-0.5712860170787227,-1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,-1.155789108870315,0.1831716015188576,-0.566215241947601,-1.1661398487230994,6.123838592253552e-2,-0.5712860170787227,-1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,-1.1072017153069766,6.123838592253552e-2,-0.6784940837700001,-1.1661398487230994,-6.123838592253552e-2,-0.5712860170787227,-1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,-1.1661398487230994,6.123838592253552e-2,-0.5712860170787227,-1.1661398487230994,-6.123838592253552e-2,-0.5712860170787227,-1.0973741144988507,-0.1831716015188576,-0.6724717222492559,-1.1072017153069766,-6.123838592253552e-2,-0.6784940837700001,-1.155789108870315,-0.1831716015188576,-0.566215241947601,-1.0973741144988507,-0.1831716015188576,-0.6724717222492559,-1.1661398487230994,-6.123838592253552e-2,-0.5712860170787227,-1.155789108870315,-0.1831716015188576,-0.566215241947601,-1.0778061433631057,-0.3034789730126769,-0.6604804541149719,-1.0973741144988507,-0.1831716015188576,-0.6724717222492559,-1.1351795030644487,-0.3034789730126769,-0.5561187002444012,-1.0778061433631057,-0.3034789730126769,-0.6604804541149719,-1.155789108870315,-0.1831716015188576,-0.566215241947601,-1.1824769291328756,0.30347897301267707,-0.4468207974206829,-1.155789108870315,0.1831716015188576,-0.566215241947601,-1.1351795030644487,0.30347897301267707,-0.5561187002444012,-1.2039452372886972,0.1831716015188576,-0.4549329950738679,-1.155789108870315,0.1831716015188576,-0.566215241947601,-1.1824769291328756,0.30347897301267707,-0.4468207974206829,-1.2039452372886972,0.1831716015188576,-0.4549329950738679,-1.1661398487230994,6.123838592253552e-2,-0.5712860170787227,-1.155789108870315,0.1831716015188576,-0.566215241947601,-1.2147272422864375,6.123838592253551e-2,-0.45900717525632356,-1.1661398487230994,6.123838592253552e-2,-0.5712860170787227,-1.2039452372886972,0.1831716015188576,-0.4549329950738679,-1.2147272422864375,6.123838592253551e-2,-0.45900717525632356,-1.1661398487230994,-6.123838592253552e-2,-0.5712860170787227,-1.1661398487230994,6.123838592253552e-2,-0.5712860170787227,-1.2147272422864375,-6.123838592253551e-2,-0.45900717525632356,-1.1661398487230994,-6.123838592253552e-2,-0.5712860170787227,-1.2147272422864375,6.123838592253551e-2,-0.45900717525632356,-1.2147272422864375,-6.123838592253551e-2,-0.45900717525632356,-1.155789108870315,-0.1831716015188576,-0.566215241947601,-1.1661398487230994,-6.123838592253552e-2,-0.5712860170787227,-1.2039452372886972,-0.1831716015188576,-0.4549329950738679,-1.155789108870315,-0.1831716015188576,-0.566215241947601,-1.2147272422864375,-6.123838592253551e-2,-0.45900717525632356,-1.2039452372886972,-0.1831716015188576,-0.4549329950738679,-1.1351795030644487,-0.3034789730126769,-0.5561187002444012,-1.155789108870315,-0.1831716015188576,-0.566215241947601,-1.1824769291328758,-0.3034789730126769,-0.44682079742068304,-1.1351795030644487,-0.3034789730126769,-0.5561187002444012,-1.2039452372886972,-0.1831716015188576,-0.4549329950738679,-1.2192786062662366,0.30347897301267707,-0.3335568815668993,-1.2039452372886972,0.1831716015188576,-0.4549329950738679,-1.1824769291328756,0.30347897301267707,-0.4468207974206829,-1.241415062548998,0.1831716015188576,-0.3396127307294062,-1.2039452372886972,0.1831716015188576,-0.4549329950738679,-1.2192786062662366,0.30347897301267707,-0.3335568815668993,-1.241415062548998,0.1831716015188576,-0.3396127307294062,-1.2147272422864375,6.123838592253551e-2,-0.45900717525632356,-1.2039452372886972,0.1831716015188576,-0.4549329950738679,-1.2525326308520355,6.123838592253551e-2,-0.3426541532514696,-1.2147272422864375,6.123838592253551e-2,-0.45900717525632356,-1.241415062548998,0.1831716015188576,-0.3396127307294062,-1.2525326308520355,6.123838592253551e-2,-0.3426541532514696,-1.2147272422864375,-6.123838592253551e-2,-0.45900717525632356,-1.2147272422864375,6.123838592253551e-2,-0.45900717525632356,-1.2525326308520355,-6.123838592253551e-2,-0.3426541532514696,-1.2147272422864375,-6.123838592253551e-2,-0.45900717525632356,-1.2525326308520355,6.123838592253551e-2,-0.3426541532514696,-1.2525326308520355,-6.123838592253551e-2,-0.3426541532514696,-1.2039452372886972,-0.1831716015188576,-0.4549329950738679,-1.2147272422864375,-6.123838592253551e-2,-0.45900717525632356,-1.241415062548998,-0.1831716015188576,-0.3396127307294062,-1.2039452372886972,-0.1831716015188576,-0.4549329950738679,-1.2525326308520355,-6.123838592253551e-2,-0.3426541532514696,-1.241415062548998,-0.1831716015188576,-0.3396127307294062,-1.1824769291328758,-0.3034789730126769,-0.44682079742068304,-1.2039452372886972,-0.1831716015188576,-0.4549329950738679,-1.2192786062662366,-0.3034789730126769,-0.3335568815668994,-1.1824769291328758,-0.3034789730126769,-0.44682079742068304,-1.241415062548998,-0.1831716015188576,-0.3396127307294062]]];
      default:
        console.error("no geometry generated for " + n)
    }
  }
}
